using System;
using System.ComponentModel;
using System.Globalization;

namespace GeoFramework.Gps
{
    /// <summary>
    /// Indicates the type of GPS corrective signals currently in use.
    /// </summary>
    public enum DifferentialGpsType
    {
        /// <summary>
        /// No differential information is available.
        /// </summary>
        None,
        /// <summary>
        /// A DGPS ground station is currently being used to improve GPS precision.
        /// </summary>
        GroundStation,
        /// <summary>
        /// A WAAS geostationary satellite is being used to improve GPS precision. (United States only)
        /// </summary>
        Waas,
        /// <summary>
        /// An EGNOS geostationary satellite is being used to improve GPS precision. (Europe only)
        /// </summary>
        Egnos,
        /// <summary>
        /// An MSAS geostationary satellite is being used to improve GPS precision. (Asia only)
        /// </summary>
        Msas
    }

	/// <summary>
	/// Represents additional information about a GPS fix when DGPS ground stations are
	/// involved.
	/// </summary>
#if !PocketPC || DesignTime
	[Serializable()]
    [TypeConverter(typeof(ExpandableObjectConverter))]
#endif
	public sealed class DifferentialGps : IFormattable
	{
        private DifferentialGpsType pType;
		private DateTime pLastUpdateUtc;
		private bool pIsActive;
		private int pStationId;
        private SatelliteCollection pSatellites;

		/// <summary>Occurs when a ground stations reports updated GPS correction 
		/// information.</summary>
		public event DifferentialGpsEventHandler LastUpdateChanged;
		/// <summary>
		/// Occurs when a fix using both satellites and ground stations has been
		/// acquired.
		/// </summary>
		public event DifferentialGpsEventHandler DifferentialFixObtained;
		/// <summary>
		/// Occurs when the active fix using both satellites and ground stations has been
		/// lost.
		/// </summary>
		public event DifferentialGpsEventHandler DifferentialFixLost;
		/// <summary>
		/// Occurs when the unique ID for the current differential GPS station has changed.
		/// </summary>
		public event DifferentialGpsEventHandler StationIdChanged;

		internal DifferentialGps()
		{}

        internal void SetType(DifferentialGpsType type)
        {
            pType = type;
        }

		internal void SetAge(DateTime lastUpdateUtc, TimeSpan value)
		{
            System.DateTime NewLastUpdateUtc = lastUpdateUtc;
			// Adjust the date by the age
			NewLastUpdateUtc = NewLastUpdateUtc.AddMilliseconds(-value.TotalMilliseconds);

			if (pLastUpdateUtc.Equals(NewLastUpdateUtc))
			{
				return;
			}
			pLastUpdateUtc = NewLastUpdateUtc;

            if (LastUpdateChanged != null)
#if PocketPC && !Framework20
				LastUpdateChanged(this, new DifferentialGpsEventArgs(this));
#else
			DynamicEvent.BeginInvoke(LastUpdateChanged, this, new DifferentialGpsEventArgs(this));
#endif
		}

        internal void SetSatellites(SatelliteCollection satellites)
        {
            pSatellites = satellites;
        }

		internal void SetIsActive(bool value)
		{
			if (pIsActive == value)
			{
				return;
			}
			pIsActive = value;
			if (pIsActive)
			{
                if (DifferentialFixObtained != null)
#if PocketPC && !Framework20
					DifferentialFixObtained(this, new DifferentialGpsEventArgs(this));
#else
				DynamicEvent.BeginInvoke(DifferentialFixObtained, this, new DifferentialGpsEventArgs(this));
#endif

			}
			else
			{
                if (DifferentialFixLost != null)
#if PocketPC && !Framework20
					DifferentialFixLost(this, new DifferentialGpsEventArgs(this));
#else
				DynamicEvent.BeginInvoke(DifferentialFixLost, this, new DifferentialGpsEventArgs(this));
#endif
			}
		}

		internal void SetStationId(int value)
		{
			if (pStationId == value)
			{
				return;
			}
			pStationId = value;
            if (StationIdChanged != null)
#if PocketPC && !Framework20
				StationIdChanged(this, new DifferentialGpsEventArgs(this));
#else
			DynamicEvent.BeginInvoke(StationIdChanged, this, new DifferentialGpsEventArgs(this));
#endif

		}

		/// <summary>
		/// Returns the date/time (in UTC) when the last DGPS correction information was
		/// received.
		/// </summary>
		/// <value>A <strong>DateTime</strong> object.</value>
		/// <remarks>
		/// This property is useful for determining when correction information was received
		/// by a DGPS ground station. Note that the value is given in UTC and may have to be
		/// converted to the local time zone using <strong>DateTime.ToLocalTime</strong>.
		/// </remarks>
		public System.DateTime LastUpdateUtc
		{
			get
			{
				return pLastUpdateUtc;
			}
		}

        /// <summary>
        /// Returns the type of technology being used to improve GPS precision.
        /// </summary>
        /// <remarks>Several technologies exist which can improve the quality of GPS data.
        /// For example, in the United States, the WAAS system provides a geostationary 
        /// satellite which transmits corrective signals.  This property is used to identify
        /// which specific technology is currently in use.</remarks>
        public DifferentialGpsType Type
        {
            get
            {
                return pType;
            }
        }

        /// <summary>
        /// Returns a collection of satellites used to send corrective signals to the GPS receiver.
        /// </summary>
        /// <remarks>This collection will return any satellite which has been identified as a WAAS, EGNOS or MSAS 
        /// satellite.  This collection will typically contain one to three satellites since a augmentation satellite
        /// typically covers a large surface area.</remarks>
        public SatelliteCollection Satellites
        {
            get
            {
                return pSatellites;
            }
        }

		/// <summary>
		/// Indicates the age (in seconds) since the last correction information was
		/// received.
		/// </summary>
		/// <value>A <see cref="System.TimeSpan">TimeSpan</see> object.</value>
		/// <remarks>
		///	 <para>This property returns the number of seconds elapsed since the DGPS correction
		///	 (to the current time). This property is most useful for determining the reliability
		///	 of correction information.</para>
		/// </remarks>
		/// <seealso cref="DifferentialGps.LastUpdateUtc">LastUpdateUtc Property</seealso>
		public TimeSpan Age
		{
			get
			{
				if(this.IsActive)
					return System.DateTime.Now.ToUniversalTime().Subtract(pLastUpdateUtc);
				else
					return TimeSpan.Zero;
			}
		}

		/// <summary>Indicates if a differential GPS fix is currently active.</summary>
		/// <value>
		/// A <strong>Boolean</strong> value. Returns <strong>True</strong> if DGPS
		/// correction information is being used to calculate current positional
		/// information.
		/// </value>
		/// <remarks>
		/// This property is most useful for determining whether DGPS information in this
		/// class is valid. If a DGPS fix is lost, the
		/// <see cref="DifferentialGps.LastUpdateUtc">LastUpdateUtc</see> and
		/// <see cref="DifferentialGps.Age">Age</see> properties are not cleared out.
		/// </remarks>
		public bool IsActive
		{
			get
			{
				return pIsActive;
			}
		}

		/// <summary>
		/// Returns the unique identifier of the DGPS ground station involved in the current
		/// fix.
		/// </summary>
		/// <value>An <strong>Integer</strong> value between 0 and 1,024.</value>
		/// <remarks>
		/// 	<para>Each DGPS ground station transmits a uniquely-identifying code, the same
		///  way GPS satellite have their own unique ID's. This property is most useful as a
		///  starting point to learn more information about a specific ground station.</para>
		/// </remarks>
		public int StationId
		{
			get
			{
				return pStationId;
			}
		}

        #region IFormattable Members

        public override string ToString()
        {
            return ToString("G", CultureInfo.InvariantCulture);
        }

        public string ToString(string format, CultureInfo culture)
        {
            if (!pIsActive)
            {
                return "(No differential GPS in effect.)";
            }
            else
            {
                switch (pType)
                {
                    case DifferentialGpsType.Waas:
                    case DifferentialGpsType.Msas:
                    case DifferentialGpsType.Egnos:
                        // Get a satellite
                        if (pSatellites == null)
                            return "Augmented GPS satellite fix (unknown satellite)";
                        else if (pSatellites.Count == 1)
                            return "GPS satellite fix enhanced by " + pSatellites[0].Name + " (Signal is " + pSatellites[0].SignalToNoiseRatio.Rating.ToString() + ")";
                        else
                            return "Augmented GPS satellite fix (undocumented satellite)";
                    case DifferentialGpsType.GroundStation:
                        return "GPS satellite fix enhanced by DGPS ground station #"  + pStationId.ToString();
                    default:
                        return "Unidentified DGPS corrections are in effect.";
                }
            }
        }

        string IFormattable.ToString(string format, IFormatProvider formatProvider)
        {
            return ToString(format, (CultureInfo)formatProvider);
        }

        #endregion
    }

	/// <summary>
	/// Represents information about DGPS/WAAS when a DGPA/WAAS event is raised.
	/// </summary>
	/// <remarks>This object is used primarily by the <see cref="DifferentialGps">DifferentialGps</see>
	/// class to provide notification when new DGPS/WAAS corrective information has been processed
	/// by the GPS device.</remarks>
	/// <example>This example demonstrates how to use this class when raising an event.
	/// <code lang="VB">
	/// ' Declare a new event
	/// Dim MyDifferentialGpsEvent As EventHandler
	/// ' Create an DifferentialGps of 90
	/// Dim MyDifferentialGps As New DifferentialGps(90)
	/// 
	/// Sub Main()
	///   ' Raise our custom event
	///   RaiseEvent MyDifferentialGpsEvent(Me, New DifferentialGpsEventArgs(MyDifferentialGps))
	/// End Sub
	/// </code>
	/// <code lang="C#">
	/// // Declare a new event
	/// EventHandler MyDifferentialGpsEvent;
	/// // Create an DifferentialGps of 90
	/// DifferentialGps MyDifferentialGps = new DifferentialGps(90);
	/// 
	/// void Main()
	/// {
	///   // Raise our custom event
	///   MyDifferentialGpsEvent(this, New DifferentialGpsEventArgs(MyDifferentialGps));
	/// }
	/// </code>
	/// </example>
	public sealed class DifferentialGpsEventArgs : EventArgs
	{
		private DifferentialGps oDifferentialGps;

		/// <summary>
		/// Creates a new instance.
		/// </summary>
		/// <param name="differentialGps">Differential GPS information provided by a DGPS/WAAS ground station.</param>

		public DifferentialGpsEventArgs(DifferentialGps differentialGps)
		{
			oDifferentialGps = differentialGps;
		}

		/// <summary>
		/// Represents differential GPS information used during a DGPS/WAAS-related event.
		/// </summary>
		/// <value></value>
		public DifferentialGps DifferentialGps
		{
			get
			{
				return oDifferentialGps;
			}
		}
	}

	public delegate void DifferentialGpsEventHandler(object sender, DifferentialGpsEventArgs e);
}