using System;
using System.Security.Permissions;
using System.Globalization;
using GeoFramework;
using System.Xml;
using System.Xml.Serialization;
#if !PocketPC || DesignTime
using System.Runtime.Serialization;
using System.ComponentModel;
#endif

namespace GeoFramework.Gps
{

	/// <summary>Represents a confidence level in positional measurements.</summary>
	/// <remarks>
	/// 	<para>Dilution of precision (or DOP for short) is an important concept in GPS
	/// programming because it describes how much you can actually trust the device when it
	/// reports your position. Any applications which perform recommendations based on the
	/// current location -- such as in-car driving directions -- must analyze DOP values
	/// frequently to minimize incorrect suggestions.</para>
	/// 	<para>DOP values can be measured in three ways: <see cref="DilutionOfPrecision.CurrentMean">mean</see>,
	/// <see cref="DilutionOfPrecision.CurrentHorizontal">horizontal</see>, and
	/// <see cref="DilutionOfPrecision.CurrentVertical">vertical</see>. Mean dilution of precision (also known as MDOP or PDOP) reports the
	/// overall confidence in any measurement of latitude, longitude or altitude.
	/// Horizontal DOP (or HDOP) reports the confidence level of latitude/longitude
	/// measurements. Finally, vertical DOP reports the confidence level of altitude
	/// measurements.</para>
	/// 	<para>A dilution of precision measurement can be between one and fifty, where
	/// one represents ideal accuracy, and fifty means that measurements can be grossly
	/// inaccurate by a factor of fifty. A value of six or less is recommended for
	/// applications which make recommendations based on the current position, and a
	/// value of three or less is required for any application which requires high
	/// precision, such as for land surveys.</para>
	/// 	<para>The exact value of this property depends on the distribution of
	/// <see cref="SatelliteCollection.FixedCount">fixed satellites</see> in the sky. If a fixed satellite is in each quadrant of
	/// the sky (further apart), DOP values are smaller because there are fewer
	/// possible trilateration solutions. On the other hand, if all fixed satellites
	/// are closer together and almost directly overhead, DOP values are larger because
	/// there are many trilateration solutions. (Trilateration is the process of
	/// determining where multiple circles intersect.)</para>
	/// 	<para>
	/// 		<img src="GoodDOP.jpg"/>
	/// 		<img src="PoorDOP.jpg"/>
	/// 	</para>
	/// 	<para>In order to simplify the process of programming with DOP, the
	/// <see cref="DilutionOfPrecision.Rating">Rating</see> property was developed. This property 
	/// returns DOP values in the form of an in-English description. When used frequently, this 
	/// property can help developers to maximize precision and intelligence during automated
	/// business decisions.</para>
	/// </remarks>
	/// <seealso cref="DilutionOfPrecision.Rating">Rating Property</seealso>
	/// <example>
	///     This example demonstrates how DOP values are typically used in applications to make
	///     business decisions. 
	///     <code lang="VB">
	/// ' How accurate is the position
	/// Select Case Receiver.MeanDilutionOfPrecision.Rating 
	///  Case DilutionOfPrecisionRating.Ideal
	///      ' Used when the utmost precision is absolutely necessary
	///  Case DilutionOfPrecisionRating.Excellent
	///      ' Minimum level for high-precision applications
	///  Case DilutionOfPrecisionRating.Good
	///      ' Minimum level for en-route navigation
	///  Case DilutionOfPrecisionRating.Moderate
	///      ' Middle of the road, OK to show as approximate location
	///  Case DilutionOfPrecisionRating.Fair
	///      ' The fix solution should not be used for decision-making
	///  Case DilutionOfPrecisionRating.Poor
	///      ' Either no fix is available or the fix is unstable
	/// End Select
	///     </code>
	/// 	<code lang="CS">
	/// // How accurate is the position
	/// switch(Receiver.MeanDilutionOfPrecision.Rating);
	/// {
	///  case DilutionOfPrecisionRating.Ideal:
	///      // Used when the utmost precision is absolutely necessary
	///  case DilutionOfPrecisionRating.Excellent:
	///      // Minimum level for high-precision applications
	///  case DilutionOfPrecisionRating.Good:
	///      // Minimum level for en-route navigation
	///  case DilutionOfPrecisionRating.Moderate:
	///      // Middle of the road, OK to show as approximate location
	///  case DilutionOfPrecisionRating.Fair:
	///      // The fix solution should not be used for decision-making
	///  case DilutionOfPrecisionRating.Poor:
	///      // Either no fix is available or the fix is unstable
	/// }
	///     </code>
	/// </example>
#if !PocketPC || DesignTime
#if Framework30
    [TypeConverter("GeoFramework.Design.DilutionOfPrecisionConverter, GeoFramework.Gps.Design, Culture=neutral, Version=2.0.0.0, PublicKeyToken=3ed3cdf4fdda3400")]
#elif Framework20
    [TypeConverter("GeoFramework.Design.DilutionOfPrecisionConverter, GeoFramework.Gps.Design, Culture=neutral, Version=2.0.0.0, PublicKeyToken=3ed3cdf4fdda3400")]
#else
	[TypeConverter(typeof(Design.DilutionOfPrecisionConverter))]
#endif
	[Serializable()]
    [ImmutableObject(true)]
#endif
    public struct DilutionOfPrecision : IComparable, IFormattable, IXmlSerializable
#if !PocketPC || DesignTime
, ISerializable
#endif
	{	
		#region  Private Variables 
		private double pValue;
		#endregion

		/// <summary>Represents the worst possible DOP value of fifty.</summary>
		public static readonly DilutionOfPrecision Maximum = new DilutionOfPrecision(50.0);
		/// <summary>Represents the best possible DOP value of one.</summary>
		public static readonly DilutionOfPrecision Minimum = new DilutionOfPrecision(1.0);
		/// <summary>Represents a DOP reading signifying nearly-ideal precision.</summary>
		public static readonly DilutionOfPrecision Excellent = new DilutionOfPrecision(3.0);
		/// <summary>Represents a DOP reading signifying inaccurate positional measurements.</summary>
		public static readonly DilutionOfPrecision Fair = new DilutionOfPrecision(20.0);
		/// <summary>
		/// Represents a FOP reading signifying grossly inaccurate positional
		/// measurements.
		/// </summary>
		public static readonly DilutionOfPrecision Poor = new DilutionOfPrecision(50.0);
		/// <summary>
		/// Represents a value of 1, where the GPS device is making the most accurate
		/// measurements possible.
		/// </summary>
		public static readonly DilutionOfPrecision Ideal = new DilutionOfPrecision(1.0);
		/// <summary>Represents a FOP reading signifying inaccurate positional measurements.</summary>
		public static readonly DilutionOfPrecision Moderate = new DilutionOfPrecision(8.0);
		/// <summary>
		/// Represents a DOP reading signifying fairly accurate positional
		/// measurements.
		/// </summary>
		public static readonly DilutionOfPrecision Good = new DilutionOfPrecision(6.0);

		#region  Constructors

		/// <summary>
		/// Creates a new instance of a mean, horizontal or vertical DOP reading with the
		/// specified value.
		/// </summary>
		/// <remarks>
		/// This constructor is typically used for special cases when a custom DOP reading
		/// must be created, such as for loading previously-saved DOP data. For most applications,
		/// creating a new <strong>DilutionOfPrecision</strong> object is not necessary.
		/// </remarks>
		/// <exception cref="ArgumentOutOfRangeException" caption="ArgumentOutOfRangeException">Dilution of precision values 
		/// must be 1, 50, or any integer in between.</exception>
		/// <example>
		///     This example creates a new horizontal DOP measurement of ten using this
		///     constructor. 
		///     <code lang="VB">
		/// Dim MyDOP As New DilutionOfPrecision(DilutionOfPrecisionType.Horizontal, 10.0)
		///     </code>
		/// 	<code lang="CS">
		/// DilutionOfPrecision MyDOP = new DilutionOfPrecision(DilutionOfPrecisionType.Horizontal, 10.0);
		///     </code>
		/// </example>
		/// <seealso cref="DilutionOfPrecision.Type">Type Property</seealso>
		/// <seealso cref="DilutionOfPrecision.Value">Value Property</seealso>
        public DilutionOfPrecision(double value)
        {
            pValue = value;
            //// Zero is invalid.  Use 50
            //if (value <= 0 || value > 50)
            //    pValue = 50;
            //else
            //    pValue = value;
        }

		public DilutionOfPrecision(string value) 
            : this(value, CultureInfo.CurrentCulture)
		{}

		public DilutionOfPrecision(string value, CultureInfo culture)
		{
            try
            {
                // Remove any reference to "dB"
                value = value.ToUpper(culture);
                // If there's a parenthesis, remove it
                if (value.IndexOf("(") != -1 && value.IndexOf(")") != -1)
                    value = value.Substring(0, value.IndexOf("(") - 1).Trim();
                // Convert rating values
                switch (value)
                {
                    case "EXCELLENT":
                        pValue = DilutionOfPrecision.Excellent.Value;
                        break;
                    case "FAIR":
                        pValue = DilutionOfPrecision.Fair.Value;
                        break;
                    case "GOOD":
                        pValue = DilutionOfPrecision.Good.Value;
                        break;
                    case "IDEAL":
                        pValue = DilutionOfPrecision.Ideal.Value;
                        break;
                    case "MODERATE":
                        pValue = DilutionOfPrecision.Moderate.Value;
                        break;
                    case "POOR":
                        pValue = DilutionOfPrecision.Poor.Value;
                        break;
                    case "MAXIMUM":
                        pValue = DilutionOfPrecision.Maximum.Value;
                        break;
                    case "MINIMUM":
                        pValue = DilutionOfPrecision.Minimum.Value;
                        break;
                    case "UNKNOWN":
                        // "Unknown" is zero
                        pValue = 0.0; 
                        break;
                    default:
                        // Try to parse it as a double
                        pValue = double.Parse(value, NumberStyles.Any, culture);
                        break;
                }
            }
            catch (Exception ex)
            {
                throw ex;
            }
		}

#if !PocketPC || DesignTime
        /// <summary>
        /// Creates a new instance by de-serializing previously-serialized information.
        /// </summary>
        /// <param name="info">A <strong>SerializationInfo</strong> object containing serialized data.</param>
        /// <param name="context">A <strong>StreamingContext</strong> object describing the context of serialization.</param>
        /// <remarks>This constructor will de-serialize a previously-serialized object.</remarks>
        private DilutionOfPrecision(SerializationInfo info, StreamingContext context)
        {
            pValue = info.GetDouble("Value");
        }
#endif

		#endregion

		#region  Public Members 

		/// <summary>
		/// Returns the amount of positional error in the form of a measurable distance.
		/// </summary>
		/// <remarks>This property calculates positional error by multiplying the current 
		/// DOP value by the maximum accuracy of the GPS device.  If the maximum accuracy
		/// cannot be determined (which is nearly always the case), a default value of 6 
		/// meters is used.   Therefore, if the DOP value is 3, this property returns
		/// 6 * 3 = 18 meters of error.</remarks>
		//[CLSCompliant(false)]
		public Distance GetEstimatedError()
		{
			return GetEstimatedError(new Distance(6, DistanceUnit.Meters));
		}

        /// <summary>
        /// Returns the amount of positional error in the form of a measurable distance.
        /// </summary>
        /// <remarks>This property calculates positional error by multiplying the current 
        /// DOP value by the maximum accuracy of the GPS device.  If the maximum accuracy
        /// cannot be determined (which is nearly always the case), a default value of 6 
        /// meters is used.   Therefore, if the DOP value is 3, this property returns
        /// 6 * 3 = 18 meters of error.</remarks>
        //[CLSCompliant(false)]
		public Distance GetEstimatedError(Distance maximumDeviceAccuracy)
		{
			return new Distance(Value * maximumDeviceAccuracy.ToMeters().Value, DistanceUnit.Meters);
		}

		#region Overloaded Operators
		//		public DilutionOfPrecision Add(DilutionOfPrecision value)
		//		{
		//			return this + value;
		//		}
		//
		//		public DilutionOfPrecision Subtract(DilutionOfPrecision value)
		//		{
		//			return this - value;
		//		}
		//
		//		public DilutionOfPrecision Multiply(DilutionOfPrecision value)
		//		{
		//			return this + value;
		//		}
		//
		//		public DilutionOfPrecision Divide(DilutionOfPrecision value)
		//		{
		//			return this / value;
		//		}
		//
		//		public static DilutionOfPrecision operator +(DilutionOfPrecision left, DilutionOfPrecision right) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value + right.Value);
		//		}
		//
		//		public static DilutionOfPrecision operator +(DilutionOfPrecision left, double right) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value + right);
		//		}
		//
		//		public static DilutionOfPrecision operator ++(DilutionOfPrecision left) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value + 1.0);
		//		}
		//
		//		public static DilutionOfPrecision operator --(DilutionOfPrecision left) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value - 1.0);
		//		}
		//
		//		public static DilutionOfPrecision operator -(DilutionOfPrecision left, DilutionOfPrecision right) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value - right.Value);
		//		}
		//
		//		public static DilutionOfPrecision operator -(DilutionOfPrecision left, double right) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value - right);
		//		}
		//
		//		public static DilutionOfPrecision operator *(DilutionOfPrecision left, DilutionOfPrecision right) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value * right.Value);
		//		}
		//
		//		public static DilutionOfPrecision operator *(DilutionOfPrecision left, double right) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value * right);
		//		}
		//
		//		public static DilutionOfPrecision operator /(DilutionOfPrecision left, DilutionOfPrecision right) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value / right.Value);
		//		}
		//
		//		public static DilutionOfPrecision operator /(DilutionOfPrecision left, double right) 
		//		{
		//			return new DilutionOfPrecision(left.Type, left.Value / right);
		//		}
		//
		//		public static bool operator <(DilutionOfPrecision left, DilutionOfPrecision right) 
		//		{
		//			return left.Value < right.Value;
		//		}
		//
		//		public static bool operator <=(DilutionOfPrecision left, DilutionOfPrecision right) 
		//		{
		//			return left.Value <= right.Value;
		//		}
		//
		////		public static bool operator ==(DilutionOfPrecision left, DilutionOfPrecision right) 
		////		{
		////			return left.Value == right.Value;
		////		}
		////		
		////		public static bool operator !=(DilutionOfPrecision left, DilutionOfPrecision right) 
		////		{
		////			return left.Value != right.Value;
		////		}
		//
		//		public static bool operator >=(DilutionOfPrecision left, DilutionOfPrecision right) 
		//		{
		//			return left.Value >= right.Value;
		//		}
		//		
		//		public static bool operator >(DilutionOfPrecision left, DilutionOfPrecision right) 
		//		{
		//			return left.Value > right.Value;
		//		}

		#endregion

		/// <summary>
		/// Compares the current instance to another DilutionOfPrecision object.
		/// </summary>
		/// <param name="value"></param>
		/// <returns></returns>
		public bool Equals(DilutionOfPrecision value)
		{
			return (pValue == value.Value);
		}

		/// <summary>
		/// Compares the current instance with the specified in-English rating.
		/// </summary>
		/// <param name="value"></param>
		/// <returns></returns>
		public bool Equals(DilutionOfPrecisionRating value)
		{
			return (Rating == value);
		}

		/// <summary>
		/// Compares the current instance with the specified object.
		/// </summary>
		/// <param name="value"></param>
		/// <returns></returns>
		public override bool Equals(object obj)
		{
			if (obj is DilutionOfPrecision)
                return this.Equals((DilutionOfPrecision)obj);
            else
                return false;
		}

		/// <summary>
		/// Returns a unique code for this object used for hash tables.
		/// </summary>
		/// <returns></returns>
		public override int GetHashCode()
		{
			return Value.GetHashCode();
		}

		/// <summary>
		/// Compares the current instance to the specified DOP value.
		/// </summary>
		/// <param name="value"></param>
		/// <returns></returns>
		public bool Equals(double value)
		{
			return (pValue == value);
		}

		/// <summary>Returns an interpretation of the current value.</summary>
		/// <value>
		/// Read only. A value from the <see cref="DilutionOfPrecisionRating">DilutionOfPrecisionRating</see> 
		/// enumeration.
		/// </value>
		/// <remarks>
		/// 	<para>This property is provided to simplify the process of learing about DOP and
		/// enforcing minimum precision in applications.  This property can also create more readable
		/// code. The possible values of this property are as follows:</para>
		/// 	<table width="90%" cellpadding="2" cellspacing="1" border="1">
		/// 		<tr>
		/// 			<td><strong>Rating</strong></td>
		/// 			<td><strong>Range</strong></td>
		/// 			<td><strong>Description</strong></td>
		/// 		</tr>
		/// 		<tr>
		/// 			<td><strong>Ideal</strong></td>
		/// 			<td>1</td>
		/// 			<td>This is the highest possible confidence level to be used for applications demanding 
		/// the highest possible precision at all times.  The receiver probably has an almost
		/// completely unobstructed view of the sky (no nearby buildings or other tall objects).
		/// This level of accuracy is unrealistic for most applications.</td>
		/// 		</tr>
		/// 		<tr>
		/// 			<td><strong>Excellent</strong></td>
		/// 			<td>2-3</td>
		/// 			<td>At this confidence level, positional measurements are considered accurate enough 
		/// to meet all but the most sensitive surveying applications.  The receiver has multiple
		/// satellite acquired from various positions in the sky.</td>
		/// 		</tr>
		/// 		<tr>
		/// 			<td><strong>Good</strong></td>
		/// 			<td>4-6</td>
		/// 			<td>Represents a confidence level that is appropriate for most business decisions. At 
		/// this level, positional measurements could be used to make reliable in-route navigation 
		/// suggestions to the user. Satellites involved in a fix are probably distributed evenly 
		/// throughout the sky.</td>
		/// 		</tr>
		/// 		<tr>
		/// 			<td><strong>Moderate</strong></td>
		/// 			<td>7-8</td>
		/// 			<td>Represents a medium confidence level. Positional measurements could be used for 
		/// calculations, but the fix quality could still be improved.  Advise to the user to 
		/// find a more open view of the sky or wait for more satellites to join the fix is 
		/// recommended.</td>
		/// 		</tr>
		/// 		<tr>
		/// 			<td><strong>Fair</strong></td>
		/// 			<td>9-20</td>
		/// 			<td>Represents a low confidence level. Positional measurements can be used for only 
		/// general purposes: displaying the approximate location, approximate distance.</td>
		/// 		</tr>
		/// 		<tr>
		/// 			<td><strong>Poor</strong></td>
		/// 			<td>21-50</td>
		/// 			<td>Represents the minimum confidence level. The satellite fix, if present, is 
		/// unstable. The current position should be reported only to show approximate location.</td>
		/// 		</tr>
		/// 	</table>
		/// 	<para>This property is updated automatically when the <see cref="Value">Value</see>
		/// property changes.</para>
		/// </remarks>
		/// <example>
		///     This example demonstrates how an application could enfore a minimum confidence
		///     level. A variable stores the most recent position which meets the required level,
		///     and the Rating property is used to either discard the new value and keep the
		///     existing one, or to update the variable to the new value. Note that you can use
		///     less-than-or-equal-to operators since enumerative values have integer values as
		///     well. 
		///     <code lang="VB">
		/// Dim WithEvents GPS As New Receiver()
		/// Dim CurrentAcceptablePosition As Position
		/// Dim MinimumRequiredConfidence As DilutionOfPrecisionRating = 
		/// DilutionOfPrecisionRating.Good
		///  
		/// ' Occurs whenever the current position has changed
		/// Private Sub OnPositionChanged(ByVal sender As Object, ByVal e As PositionEventArgs) Handles GPS.PositionChanged
		///  ' Is the current accuracy level good enough
		///  If GPS.MeanDilutionOfPrecision.Rating &lt;= MinimumRequiredConfidence Then
		///      ' Yes. Update the position
		///      CurrentAcceptablePosition = e.Position
		///  Else
		///      ' No.  Do not update the variable, and ignore the new position        
		///  End If
		/// End Sub
		///     </code>
		/// 	<code lang="CS">
		/// Receiver GPS = new Receiver();
		/// Position CurrentAcceptablePosition;
		/// DilutionOfPrecisionRating MinimumRequiredConfidence = DilutionOfPrecisionRating.Good;
		///  
		/// // Occurs whenever the current position has changed
		/// void OnPositionChanged(Object sender, PositionEventArgs e)
		/// {
		///  // Is the current accuracy level good enough?
		///  if (GPS.MeanDilutionOfPrecision.Rating &lt;= MinimumRequiredConfidence)
		///  {
		///      // Yes. Update the position
		///      CurrentAcceptablePosition = e.Position
		///  }
		///  else
		///  {
		///      // No.  Do not update the variable, and ignore the new position        
		///  }
		/// }
		///     </code>
		/// </example>
		/// <seealso cref="DilutionOfPrecisionRating">DilutionOfPrecisionRating Enumeration</seealso>
		/// <seealso cref="Value">Value Property</seealso>
		public DilutionOfPrecisionRating Rating
		{
			get
			{
				double RoundedValue = Math.Floor(pValue);
                // -Infinity --> 0.0
                if (RoundedValue < 0.0)
                {
                    return DilutionOfPrecisionRating.Unknown;
                }
                // 0.1 --> 1.0
                else if (RoundedValue == 1.0 || (RoundedValue == 0.0 && pValue > 0.0))
                {
                    return DilutionOfPrecisionRating.Ideal;
                }
                // 1.1 --> 3.0
                else if (RoundedValue <= 3.0)
                {
                    return DilutionOfPrecisionRating.Excellent;
                }
                else if (RoundedValue <= 6.0)
                {
                    return DilutionOfPrecisionRating.Good;
                }
                else if (RoundedValue <= 8.0)
                {
                    return DilutionOfPrecisionRating.Moderate;
                }
                else if (RoundedValue <= 20.0)
                {
                    return DilutionOfPrecisionRating.Fair;
                }
                else
                {
                    return DilutionOfPrecisionRating.Poor;
                }
			}
		}

//		/// <summary>
//		/// Identifies whether the object is describing mean, horizontal or vertical dilution
//		/// of precision.
//		/// </summary>
//		/// <value>
//		/// A value from the
//		/// <see 
//		/// href="DilutionOfPrecisionType">
//		/// DilutionOfPrecisionType</see> enumeration.
//		/// </value>
//		/// <remarks>
//		/// This property is typically used when handling multiple DOP-related events in one
//		/// method, and some distinction is needed between multiple DOP objects.
//		/// </remarks>
//		[XmlAttribute()]
//		public DilutionOfPrecisionType Type
//		{
//			get
//			{
//				return pType;
//			}
//		}

		/// <summary>Indicates the numeric portion of the measurement</summary>
		/// <value>An <strong>Integer</strong> value.</value>
		/// <remarks>
		/// 	<para>This property is used to determine whether or not a positional measurement
		/// should be ignored due to insufficient accuracy. This property can hold values
		/// between 1.0 and 50.0, where 1.0 means ideal accuracy and 50.0 means measurements
		/// may be off by a factor of fifty.  Typically, DOP values start at fifty when the 
		/// GPS device first begins searching for satellites, then values drop lower once a 
		/// fix has been acquired and more satellites become involved in the fix.</para>
		/// 	<para>When this property changes, the <see cref="ValueChanged">ValueChanged</see>
		/// event is raised.</para>
		/// </remarks>
		/// <example>
		///     This example examines the horizontal dilution of precision to see if positional
		///     measurements are accurate enough for en-route navigation. 
		///     <code lang="VB">
		/// Dim MyReceiver As New Receiver()
		///  
		/// ' Is the horizontal DOP six or less
		/// If MyReceiver.HorizontalDilutionOfPrecision.Value &lt;= 6 Then
		///  ' Yes.  The position is accurate enough for in-car navigation
		/// Else
		///  ' No.  Precision may be too low to make an intelligent navigation recommendation
		/// End If
		///     </code>
		/// 	<code lang="CS">
		/// Receiver MyReceiver = New Receiver();
		///  
		/// // Is the horizontal DOP six or less
		/// if (Receiver.HorizontalDilutionOfPrecision.Value &lt;= 6)
		/// {
		///  // Yes.  The position is accurate enough for in-car navigation
		/// }
		/// else
		/// {
		///  // No.  Precision may be too low to make an intelligent navigation recommendation
		/// }
		///     </code>
		/// </example>
		/// <seealso cref="Rating">Rating Property</seealso>
		/// <seealso cref="DilutionOfPrecision.ValueChanged">ValueChanged Event</seealso>
		public double Value
		{
			get
			{
				return pValue;
			}
		}

		/// <summary>Returns the value of a DOP measurement in a user-friendly format.</summary>
		/// <remarks>
		/// This method is typically used to present the value of a DOP rating to a user
		/// interface.
		/// </remarks>
		/// <returns>A <strong>String</strong> in the following format: "VV (RRRRRR)" where
		/// <strong>V</strong> is the Value property, and <strong>R</strong> is the associated
		/// rating from the <see cref="Rating">Rating</see> property.
		/// </returns>
		/// <example>
		///     This example sets the text of a TextBox to the current mean dilution of precision. 
		///     <code lang="VB">
		/// ' This produces a result such as "1 (Ideal)" or "47 (Poor)", etc.
		/// MyTextBox.Text = Receiver.MeanDilutionOfPrecision.ToString()
		///     </code>
		/// 	<code lang="CS">
		/// // This produces a result such as "1 (Ideal)" or "47 (Poor)", etc.
		/// MyTextBox.Text = Receiver.MeanDilutionOfPrecision.ToString();
		///     </code>
		/// </example>
		public override string ToString()
		{
			return ToString("G", CultureInfo.CurrentCulture);
		}

		public string ToString(string format, CultureInfo culture)
		{
            // Use the default culture if needed
            if (culture == null)
                culture = CultureInfo.CurrentCulture;

			// Use a default format if none was given
			if(format == null || format.Length == 0)
				format = "G";
            
            // If the format is "G", use the default format
            if (String.Compare(format, "G", true, CultureInfo.InvariantCulture) == 0)
				format = "0.0 (R)";

            // Replace "R" with the rating
            format = format.ToUpper(CultureInfo.InvariantCulture).Replace("R", Rating.ToString());
			
            // Now format the value
			return Value.ToString(format, culture);
		}

		#endregion

		#region Operators

		public static DilutionOfPrecision operator +(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.Add(right);
		}

		public static DilutionOfPrecision operator -(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.Subtract(right);
		}

		public static DilutionOfPrecision operator *(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.Multiply(right);
		}

		public static DilutionOfPrecision operator /(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.Divide(right);
		}

		public static bool operator ==(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.Equals(right);
		}

		public static bool operator !=(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return !left.Equals(right);
		}

		public static bool operator >(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.IsGreaterThan(right);
		}

		public static bool operator >=(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.IsGreaterThanOrEqualTo(right);
		}

		public static bool operator <(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.IsLessThan(right);
		}

		public static bool operator <=(DilutionOfPrecision left, DilutionOfPrecision right)
		{
			return left.IsLessThanOrEqualTo(right);
		}

		/// <summary>Returns the current instance increased by one.</summary>
		public DilutionOfPrecision Increment()
		{
			return new DilutionOfPrecision(pValue + 1.0);
		}

		/// <summary>Returns the sum of the current instance with the specified value.</summary>
		public DilutionOfPrecision Add(DilutionOfPrecision value)
		{
			return new DilutionOfPrecision(pValue + value.Value);
		}

		/// <summary>Returns the current instance decreased by one.</summary>
		public DilutionOfPrecision Decrement()
		{
			return new DilutionOfPrecision(pValue - 1.0);
		}

		/// <summary>Returns the current instance decreased by the specified amount.</summary>
		public DilutionOfPrecision Subtract(DilutionOfPrecision value)
		{
			return new DilutionOfPrecision(pValue - value.Value);
		}

		/// <summary>Returns the product of the current instance with the specified value.</summary>
		public DilutionOfPrecision Multiply(DilutionOfPrecision value)
		{
			return new DilutionOfPrecision(pValue * value.Value);
		}

		/// <summary>Returns the current instance divided by the specified value.</summary>
		public DilutionOfPrecision Divide(DilutionOfPrecision value)
		{
			return new DilutionOfPrecision(pValue / value.Value);
		}

		/// <summary>
		/// Indicates whether the current instance is smaller than the specified
		/// value.
		/// </summary>
		public bool IsLessThan(DilutionOfPrecision value)
		{
			return CompareTo(value) < 0;
		}

		/// <summary>
		/// Indicates whether the current instance is smaller than or equivalent to the
		/// specified value.
		/// </summary>
		public bool IsLessThanOrEqualTo(DilutionOfPrecision value)
		{
			return CompareTo(value) < 0 || Equals(value);
		}

		/// <summary>Indicates whether the current instance is larger than the specified value.</summary>
		public bool IsGreaterThan(DilutionOfPrecision value)
		{
			return CompareTo(value) > 0;
		}

		/// <summary>
		/// Indicates whether the current instance is larger than or equivalent to the
		/// specified value.
		/// </summary>
		public bool IsGreaterThanOrEqualTo(DilutionOfPrecision value)
		{
			return CompareTo(value) > 0 || Equals(value);
		}
		#endregion

		public static DilutionOfPrecision Parse(string value)
		{
			return new DilutionOfPrecision(value, CultureInfo.CurrentCulture);
		}

		public static DilutionOfPrecision Parse(string value, CultureInfo culture)
		{
			return new DilutionOfPrecision(value, culture);
		}

		/// <summary>
		/// Compares the current instance to the specified value.
		/// </summary>
		/// <param name="value">A <strong>DilutionOfPrecision</strong> object to compare with.</param>
		/// <returns></returns>
        public int CompareTo(DilutionOfPrecision value)
		{
			return pValue.CompareTo(value.Value);
		}

		#region Conversions

		static public implicit operator DilutionOfPrecision(double value)
		{
			return new DilutionOfPrecision(value);
		}

		static public implicit operator double(DilutionOfPrecision value)
		{
			return value.Value;
		}

		static public implicit operator DilutionOfPrecision(int value)
		{
			return new DilutionOfPrecision(value);
		}

		static public explicit operator DilutionOfPrecision(string value)
		{
			return DilutionOfPrecision.Parse(value, CultureInfo.CurrentCulture);
		}

		static public implicit operator string(DilutionOfPrecision value)
		{
			return value.ToString("g", CultureInfo.CurrentCulture);
		}

		#endregion

        #region IComparable Members

        int IComparable.CompareTo(object obj)
        {
            return CompareTo((DilutionOfPrecision)obj);
        }

#endregion

		#region IFormattable Members

		string System.IFormattable.ToString(string format, IFormatProvider formatProvider)
		{
			return this.ToString(format, (CultureInfo)formatProvider);
		}

		#endregion

        #region IXmlSerializable Members

        System.Xml.Schema.XmlSchema IXmlSerializable.GetSchema()
        {
            return null;
        }

        void IXmlSerializable.ReadXml(XmlReader reader)
        {
            reader.ReadStartElement();
            pValue = double.Parse(reader.ReadString(), CultureInfo.InvariantCulture);
            reader.ReadEndElement();
        }

        void IXmlSerializable.WriteXml(XmlWriter writer)
        {
            writer.WriteStartElement("Value");
            writer.WriteString(pValue.ToString(CultureInfo.InvariantCulture));
            writer.WriteEndElement();
            writer.WriteStartElement("Rating");
            writer.WriteString(Rating.ToString());
            writer.WriteEndElement();
        }

        #endregion


#if !PocketPC || DesignTime
        #region ISerializable Members

        [SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.SerializationFormatter)]        
        void ISerializable.GetObjectData(SerializationInfo info, StreamingContext context)
        {
            info.AddValue("Value", pValue);
        }

        #endregion
#endif

	}

    /// <summary>Represents information about a DOP measurement when an DOP-related event is raised.</summary>
	/// <remarks>This class is typically used for events in the <see cref="DilutionOfPrecision">DilutionOfPrecision</see> class to 
	/// provide notification when hours, minutes, decimal minutes or seconds properties have changed.</remarks>
	/// <example>This example demonstrates how to use this class when raising an event.
	/// <code lang="VB">
	/// ' Declare a new event
	/// Dim MyDilutionOfPrecisionEvent As EventHandler
	/// ' Create a DilutionOfPrecision of 30 horizontal
	/// Dim MyDilutionOfPrecision As New DilutionOfPrecision(DilutionOfPrecisionType.Horizontal, 30.0)
	/// 
	/// Sub Main()
	///   ' Raise our custom event
	///   RaiseEvent MyDilutionOfPrecisionEvent(Me, New DilutionOfPrecisionEventArgs(MyDilutionOfPrecision))
	/// End Sub
	/// </code>
	/// <code lang="C#">
	/// // Declare a new event
	/// EventHandler MyDilutionOfPrecisionEvent;
	/// // Create a DilutionOfPrecision of 30 horizontal
	/// DilutionOfPrecision MyDilutionOfPrecision = new DilutionOfPrecision(DilutionOfPrecisionType.Horizontal, 30.0);
	/// 
	/// void Main()
	/// {
	///   // Raise our custom event
	///   MyDilutionOfPrecisionEvent(this, New DilutionOfPrecisionEventArgs(MyDilutionOfPrecision));
	/// }
	/// </code>
	/// </example>
	/// <seealso cref="DilutionOfPrecision">DilutionOfPrecision Class</seealso>
	/// <seealso cref="EventHandler">EventHandler Delegate</seealso>
	public sealed class DilutionOfPrecisionEventArgs : EventArgs
	{

		private DilutionOfPrecision oDilutionOfPrecision;

		/// <summary>
		/// Creates a new instance with the specified DOP measurement.
		/// </summary>
		/// <param name="DilutionOfPrecision">A DilutionOfPrecision object which is the target of the event.</param>

		public DilutionOfPrecisionEventArgs(DilutionOfPrecision dilutionOfPrecision)
		{
			oDilutionOfPrecision = dilutionOfPrecision;
		}

		/// <summary>
		/// A DilutionOfPrecision object which is the target of the event.
		/// </summary>
		/// <value></value>

		public DilutionOfPrecision DilutionOfPrecision
		{
			get
			{
				return oDilutionOfPrecision;
			}
		}
	}


	/// <summary>
	/// Facilitates events and asynchronous calls which use an
	/// <strong>DilutionOfPrecision</strong> object as a parameter.
	/// </summary>
	/// <example>
	/// This example declares a custom event using this delegate.
	/// <code lang="VB">
	/// Dim MyDilutionOfPRecisionEvent As EventHandler
	/// </code>
	/// <code lang="C#">
	/// EventHandler MyDilutionOfPRecisionEvent;
	/// </code>
	/// </example>
	/// <param name="sender">The object which initiated the call to the delegate.</param>
	/// <param name="e">A <see cref="DilutionOfPrecisionEventArgs">DilutionOfPrecisionEventArgs</see> object containing a <see cref="DilutionOfPrecision">DilutionOfPrecision</see> involved in the event.</param>
	public delegate void DilutionOfPrecisionEventHandler(object sender, DilutionOfPrecisionEventArgs e);
}