using System;
using System.Threading;
using GeoFramework.Gps.Nmea;

namespace GeoFramework.Gps.Emulators
{
	/// <summary>Emulates a DeLorme Earthmate GPS device.</summary>
	/// <remarks>
	/// 	<para>This class gives developers the ability to simulate a DeLorme Earthmate GPS
	///     device. NMEA sentences are generated in the same way that they would with an actual
	///     device. This class can be used as the base stream for an
	///     <strong>NmeaInterpreter</strong> class.</para>
	/// 	<para>GPS device emulators are useful for application testing, as well as for
	///     situations where a real GPS device would not function, such as the interior of a
	///     building or office.</para>
	/// 	<para><em>Earthmate is a registered trademark of DeLorme. Visit DeLorme online
	///     at</em>
	/// 		<a href="http://www.delorme.com"><em>http://www.delorme.com</em></a><em>.</em></para>
	/// </remarks>
	/// <example>
	///     This example creates a new <strong>NmeaInterpreter</strong> using a DeLorme
	///     Earthmate GPS device emulator.
	///     <code lang="VB" title="[New Example]">
	/// ' Create a new DeLorme Earthmate emulator
	/// Dim EarthmateEmulator As New DeLormeEarthmateUsbEmulator()
	/// ' Bind it to an NMEA interpreter
	/// Dim Interpreter As New NmeaInterpreter(EarthmateEmulator)
	/// ' Start processing emulated information
	/// Interpreter.Start()
	/// </code>
	/// 	<code lang="CS" title="[New Example]">
	/// // Create a new DeLorme Earthmate emulator
	/// DeLormeEarthmateUsbEmulator EarthmateEmulator = new DeLormeEarthmateUsbEmulator();
	/// // Bind it to an NMEA interpreter
	/// NmeaInterpreter Interpreter = new NmeaInterpreter(EarthmateEmulator);
	/// // Start processing emulated information
	/// Interpreter.Start();
	///     </code>
	/// </example>
	public sealed class DeLormeEarthmateUsbEmulator : NmeaEmulator
	{
		private DateTime pGpggaLastSent = DateTime.Now;
		private DateTime pGpgsaLastSent = DateTime.Now;
		private DateTime pGpgllLastSent = DateTime.Now;
		private DateTime pGpgsvLastSent = DateTime.Now;
		private DateTime pGprmcLastSent = DateTime.Now;
		private TimeSpan pGpggaInterval = new TimeSpan(0, 0, 1);
        private TimeSpan pGpgsaInterval = new TimeSpan(0, 0, 1);
        private TimeSpan pGpgllInterval = new TimeSpan(0, 0, 1);
        private TimeSpan pGpgsvInterval = new TimeSpan(0, 0, 5);
        private TimeSpan pGprmcInterval = new TimeSpan(0, 0, 1);

		/// <summary>Creates a new instance.</summary>
        public DeLormeEarthmateUsbEmulator() 
			: base("DeLorme Earthmate USB Emulator [NMEA-0183] (http://www.geoframeworks.com)")
		{
		}

        /// <summary>Creates a copy of the current instance.</summary>
        /// <remarks>
        /// This method is typically used by the <strong>VirtualDevice</strong> class to open
        /// a new virtual connection to an Emulator. The properties which have been configured in
        /// the original instance of the emulator are copied over when the new instance is
        /// created.
        /// </remarks>
        public override Emulator Clone()
        {
            DeLormeEarthmateUsbEmulator e = new DeLormeEarthmateUsbEmulator();
            e.AllowRandomRoutes = this.AllowRandomRoutes;
            e.Altitude = this.Altitude;
            e.BaudRate = this.BaudRate;
            e.CurrentPosition = this.CurrentPosition;
            e.DifferentialGps = this.DifferentialGps;
            e.FixMethod = this.FixMethod;
            e.FixMode = this.FixMode;
            e.FixQuality = this.FixQuality;
            e.HorizontalDilutionOfPrecision = this.HorizontalDilutionOfPrecision;
            e.IsFixed = this.IsFixed;
            e.MagneticVariation = this.MagneticVariation;
            e.MeanDilutionOfPrecision = this.MeanDilutionOfPrecision;
            e.Route.AddRange(this.Route);
            e.RouteBoundary = this.RouteBoundary;
            e.Satellites = this.Satellites;
            e.Speed = this.Speed;
#if Framework20 && !PocketPC
            // ThreadPriority is not used on DF2.0 because of stability issues
#else
            e.ThreadPriority = this.ThreadPriority;
#endif
            e.VerticalDilutionOfPrecision = this.VerticalDilutionOfPrecision;
            return e;
        }

		/// <summary>
		/// Generates actual data to send to the client.
		/// </summary>
		/// <remarks>Data is sent according to the behavior of a typical GPS device: $GPGGA,
		/// $GPGSA, $GPRMC, $GPGSV sentences are sent every second, and a $GPGSV sentence 
		/// is sent every five seconds.  
		/// Developers who want to emulate a specific model of GPS device should override this
		/// method and generate the sentences specific to that device.</remarks>
		protected override void OnEmulation()
		{
            base.OnEmulation();

			// $GPGGA
			if(!pGpggaInterval.Equals(TimeSpan.Zero))
			{
				// Has enough time elapsed to send the sentence?
				if(UtcDateTime.Subtract(pGpggaLastSent).TotalMilliseconds > pGpggaInterval.TotalMilliseconds)
				{
					// Yes
                    pGpggaLastSent = UtcDateTime;
                    GpggaSentence GPGGA = new GpggaSentence(UtcDateTime, CurrentPosition, FixQuality, Satellites.FixedSatellites.Count,
						HorizontalDilutionOfPrecision, Altitude, Distance.Empty, DifferentialGps.Age, 
						DifferentialGps.StationId);
					WriteToClient(GPGGA);
				}
			}

			// $GPGSA
			if(!pGpgsaInterval.Equals(TimeSpan.Zero))
			{
				// Has enough time elapsed to send the sentence?
				if(UtcDateTime.Subtract(pGpgsaLastSent).TotalMilliseconds > pGpgsaInterval.TotalMilliseconds)
				{
                    // Yes
                    pGpgsaLastSent = UtcDateTime;
                    GpgsaSentence GPGSA = new GpgsaSentence(FixMode, FixMethod, Satellites.FixedSatellites,
						MeanDilutionOfPrecision, HorizontalDilutionOfPrecision, VerticalDilutionOfPrecision);
					WriteToClient(GPGSA);
					//GPGSA.Dispose();
				}
			}

			// $GPRMC
			if(!pGprmcInterval.Equals(TimeSpan.Zero))
			{
				// Has enough time elapsed to send the sentence?
				if(UtcDateTime.Subtract(pGprmcLastSent).TotalMilliseconds > pGprmcInterval.TotalMilliseconds)
				{
					// Yes
                    pGprmcLastSent = UtcDateTime;
                    GprmcSentence GPRMC = new GprmcSentence(UtcDateTime, IsFixed, CurrentPosition, Speed,
						Bearing, MagneticVariation);
					WriteToClient(GPRMC);		
				}
			}



			// $GPGSV
			if(!pGpgsvInterval.Equals(TimeSpan.Zero))
			{
				// Has enough time elapsed to send the sentence?
				if(UtcDateTime.Subtract(pGpgsvLastSent).TotalMilliseconds > pGpgsvInterval.TotalMilliseconds)
				{
					// Yes
                    pGpgsvLastSent = UtcDateTime;
                    foreach (GpgsvSentence GPGSV in GpgsvSentence.FromSatellites(Satellites))
					{
						WriteToClient(GPGSV);
						//GPGSV.Dispose();
					}
				}
			}

			// $GPGLL
			if(!pGpgllInterval.Equals(TimeSpan.Zero))
			{
				// Has enough time elapsed to send the sentence?
				if(UtcDateTime.Subtract(pGpgllLastSent).TotalMilliseconds > pGpgllInterval.TotalMilliseconds)
				{
					// Yes
                    pGpgllLastSent = UtcDateTime;
                    GpgllSentence GPGLL = new GpgllSentence(CurrentPosition, UtcDateTime, IsFixed);
					WriteToClient(GPGLL);
				}
			}

            // Sleep for a bit
            Thread.Sleep(250);

		}

        public override string ToString()
        {
            return "DeLorme Earthmate USB Emulator (NMEA-0183)";
        }

	}
}
