using System;
using System.Collections;
using System.Drawing;
using System.IO;
using System.Globalization;
using System.Threading;
#if !PocketPC || DesignTime
using System.ComponentModel;
#endif

using GeoFramework;
using GeoFramework.IO.Serial;
using GeoFramework.IO;

namespace GeoFramework.Gps.Emulators
{
	/// <summary>
	/// Represents a base class for building a GPS device emulator.
	/// </summary>
	/// <remarks>
	/// 	<para>This powerful class enables developers to write classes which simulate a real
	///     GPS device. Information such as the current position, speed, bearing and altitude
	///     are all managed automatically. The current position can simulate a traveling along
	///     a course. Programming tasks such as multithreading, thread synchronization and
	///     finalization are all handled automatically. Since this class is derived from the
	///     <strong>Stream</strong> class, any class deriving from this class can be used with
	///     any class deriving from the <strong>Interpreter</strong> class. it can be used as
	///     the data source for any GPS data interpreter.</para>
	/// 	<para>A custom emulator is created by overriding the <strong>OnEmulation</strong>
	///     virtual method and calling the <strong>WriteToClient</strong> method to make
	///     simulated data available to an interpreter.</para>
	/// 	<para>Other more sophisticated base classes are available for designing emulators
	///     which use a specific data protocol. See the <strong>NmeaEmulator</strong>,
	///     <strong>GarminTextEmulator</strong> and <strong>GarminBinaryEmulator</strong>
	///     classes for more information.</para>
	/// </remarks>
	/// <seealso cref="GeoFramework.Gps.Interpreter">Interpreter Class</seealso>
	/// <seealso cref="NmeaEmulator">NmeaEmulator Class</seealso>
	/// <example>
	///     This example creates a custom GPS device emulator which transmits a $GPRMC sentence
	///     every two seconds. Variables like <strong>UtcDateTime</strong>,
	///     <strong>IsFixed</strong>, and others are all properties of the
	///     <strong>Emulator</strong> class which change automatically.
	///     <code lang="VB" title="[New Example]">
	/// Public Class MyEmulator 
	/// Inherits Emulator
	///  
	///     Protected Overrides Sub OnEmulation()
	///         ' Build a $GPRMC sentence using the current simulated location, speed
	///         ' And bearing. 
	///         GprmcSentence Output = New GprmcSentence(UtcDateTime,
	///             IsFixed, CurrentPosition, Speed, Bearing,
	///             Longitude.Empty)
	///  
	///         ' Send the output To any listening clients
	///         WriteToClient(Output)
	///  
	///         ' Sleep For two seconds before sending another sentence
	///         Thread.Sleep(2000)
	///     End Sub        
	/// End Class
	///     </code>
	/// 	<code lang="CS" title="[New Example]">
	/// public class MyEmulator : Emulator
	/// {
	///     protected override void OnEmulation()
	///     {
	///         // Build a $GPRMC sentence using the current simulated location, speed
	///         // and bearing. 
	///         GprmcSentence Output = new GprmcSentence(UtcDateTime,
	///             IsFixed, CurrentPosition, Speed, Bearing,
	///             Longitude.Empty);
	///  
	///         // Send the output to any listening clients
	///         WriteToClient(Output);
	///  
	///         // Sleep for two seconds before sending another sentence
	///         Thread.Sleep(2000);
	///     }
	/// }
	///     </code>
	/// </example>
#if !PocketPC
    [ToolboxBitmap(typeof(Emulator))]
#endif
    public abstract class Emulator : Stream
    {
		private Queue pIncomingStream = new Queue(512, 2.0f);
		//private Object pIncomingLock = new Object();
		private Queue pOutgoingStream = new Queue(512, 2.0f);       
        private int pOutgoingBufferSize = 16384;
		//private Object pOutgoingLock = new Object();
		private bool pIsOpen;
        private bool pIsDisposed;
		private Thread EmulationThread;
		private ManualResetEvent EmulationWaitHandle = new ManualResetEvent(true);
		private string pThreadName;
        private BaudRate pBaudRate = BaudRate.Baud4800;

        private Distance pAltitude = new Distance(400, DistanceUnit.Feet).ToLocalUnitType();
        private Azimuth pBearing = Azimuth.Empty;
        private DateTime pUtcDateTime = DateTime.UtcNow;
        private Position pPosition = GeoFramework.IO.Devices.CurrentPosition;
        private Speed pSpeed = new Speed(20, SpeedUnit.StatuteMilesPerHour).ToLocalUnitType();

        private bool pAllowRandomRoutes = true;
        private int RouteIndex;
        private PositionCollection pRoute = new PositionCollection();
        private GeographicRectangle pRouteBoundary = GeographicRectangle.Maximum;
        
        private static Random Generator = new Random(DateTime.Now.Millisecond);
#if PocketPC && !Framework20
		private bool pIsEmulationThreadAlive;
#endif

#if !PocketPC && Framework20
        private ThreadPriority pThreadPriority = ThreadPriority.Normal;
#else
		private ThreadPriority pThreadPriority = ThreadPriority.BelowNormal;
#endif

#if !PocketPC
        private const int MaximumGracefulShutdownTime = 500;
#else
        private const int MaximumGracefulShutdownTime = 2000;
#endif

		private object ControlSyncRoot = new object();

        /// <summary>
        /// Occurs when the virtual position has reached a waypoint in the virtual route.
        /// </summary>
        public event PositionEventHandler WaypointReached;

		/// <summary>Creates a new instance using the specified thread name.</summary>
		/// <remarks>
		/// 	<para>This constructor is used by classes which inherit from
		///     <strong>Emulator</strong> to specify a name for the emulation thread (i.e. the
		///     Thread.Name property.) In Visual Studio.NET, each thread has it's own name in order
		///     to debugging easier.</para>
		/// 	<para>Typically, the thread name clearly identifies the thread's purpose. It is
		///     also generally a good idea to include a URL of some kind to identify the company
		///     supporting the emulator.</para>
		/// </remarks>
		/// <example>
		///     This example demonstrates the constructor by specifying a thread name "My Custom
		///     Emulator
		///     (<a id="dxCrLf">href="http://www.mycompany.com"&gt;http://www.mycompany.com</a>)".
		///     <code lang="VB" title="[New Example]">
		/// Public Class MyEmulator
		/// Inherits Emulator
		///  
		///     Sub New()
		///         MyBase.New("My Custom Emulator (http://www.mycompany.com)")
		///     End Sub
		/// End Class
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// public class MyEmulator : Emulator
		/// {
		///     MyEmulator()
		///         : base("My Custom Emulator (http://www.mycompany.com)")
		///     {}
		/// }
		///     </code>
		/// </example>
		/// <param name="threadName">
		/// A <strong>String</strong> specifying a name to assign to the emulation
		/// thread.
		/// </param>
        protected Emulator(string threadName)
		{
			pThreadName = threadName;

            // Calculate an imaginary route
            RandomizeRoute();
        }

		protected Emulator() 
            : this("Untitled GPS Device Emulator")
		{
		}

		~Emulator()
		{
			Close();
		}

		#region Overrides

		private void Open()
		{
			// If we're already open, do nothing
			if(pIsOpen) 
				return;

			// Prepare all streams for sending and receiving data
			lock(ControlSyncRoot)
			{
				//pIncomingStream = new MemoryStream();
				//pOutgoingStream = new MemoryStream();

				// Initialize any variables, etc.
				Initialize();

                try
                {
                    // Configure a new thread for interpretation
                    EmulationThread = new Thread(new ThreadStart(EmulationThreadProc));
#if PocketPC || (!PocketPC && !Framework20)
#if !Framework10
                    if (Environment.OSVersion.Platform == PlatformID.WinCE)
#endif
                        EmulationThread.Priority = pThreadPriority;
#endif
#if !PocketPC || Framework20
                    EmulationThread.Name = Name;
                    EmulationThread.IsBackground = true;
#endif
                    // Start the new thread
					pIsOpen = true;
                    EmulationWaitHandle.Reset();
                    EmulationThread.Start();
                    EmulationWaitHandle.WaitOne();

                    // Reset the wait handle for closing later on
                    EmulationWaitHandle.Reset();	
                }
#if !PocketPC
                catch (ThreadAbortException)
                {
                    pIsOpen = false;
                }
#endif
                catch (Exception ex)
                {
                    throw new EmulatorException("The device emulation thread was unable to start.", ex);
                }
			}
		}

		protected virtual void Initialize()
		{
		}

        /// <summary>Returns the name of the GPS data protocol supported by the emulator.</summary>
        /// <value>A <strong>String</strong> indicating the name of the protocol.</value>
        /// <remarks>
        /// This property, when overridden, must return the name of the protocol being
        /// supported. The name of the protocol should match the name of the protocol supported by
        /// an interpreter.
        /// </remarks>
        /// <example>
        ///     This example creates a custom emulator designed to generate NMEA-0183 data.
        ///     <code lang="VB" title="[New Example]">
        /// Imports GeoFramework.Gps.Emulators
        /// Imports GeoFramework.Gps.Nmea
        ///  
        /// Public Class CustomEmulator
        ///     Inherits Emulator
        ///     
        ///     Protected Property Overrides Protocol As String
        ///         Get
        ///             Return NmeaInterpreter.Protocol
        ///         End Get
        ///     End Property
        /// End Class
        ///     </code>
        /// 	<code lang="CS" title="[New Example]">
        /// using GeoFramework.Gps.Emulators
        /// using GeoFramework.Gps.Nmea
        ///  
        /// public class CustomEmulator : Emulator
        /// {    
        ///     protected override string Protocol
        ///     {
        ///         get
        ///         {
        ///             return NmeaInterpreter.Protocol;
        ///         }
        ///     }
        /// }
        ///     </code>
        /// </example>
		public abstract string Protocol { get; }

		public override IAsyncResult BeginRead(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			throw new NotSupportedException("GPS device emulators do not support asynchronous read operations.");
//			if(!pIsOpen)
//				Open();
			
			//return pOutgoingStream.BeginRead(buffer, offset, count, callback, state);
		}

		public override IAsyncResult BeginWrite(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			throw new NotSupportedException("GPS device emulators do not support asynchronous write operations.");
//			if(!pIsOpen)
//				Open();

			//return pIncomingStream.BeginWrite(buffer, offset, count, callback, state);
		}

        private void EmulationThreadProc()
        {
			try
			{
#if PocketPC && !Framework20
                // Flag that we're now simulating
                pIsEmulationThreadAlive = true;
#endif
                
                // Flag that the thread is alive
				EmulationWaitHandle.Set();

				// Remember the last time of the iteration
				//DateTime LastIteration = DateTime.Now;

				// Loop while the stream is open
				while (pIsOpen && !pIsDisposed)
				{
#if PocketPC || (!PocketPC && !Framework20)
                    // Update the thread priority
                    if (EmulationThread.Priority != pThreadPriority)
                        EmulationThread.Priority = pThreadPriority;
#endif

					// Calculate the time sincce the last iteration
					TimeSpan UpdateInterval = DateTime.UtcNow.Subtract(pUtcDateTime);

					// Set the current time
					pUtcDateTime = DateTime.UtcNow;

					// Calculate the new current position by applying speed
					// and bearing to the current position
					pBearing = pPosition.BearingTo(pRoute[RouteIndex]);
					switch (Generator.Next(5))
					{
						case 1:
							pAltitude = pAltitude.Add(new Distance(1, DistanceUnit.Feet)).ToLocalUnitType();
							break;
						case 2:
							pAltitude = pAltitude.Subtract(new Distance(1, DistanceUnit.Feet)).ToLocalUnitType();
							break;
						case 3:
							pSpeed = pSpeed.Add(new Speed(0.5, SpeedUnit.StatuteMilesPerHour)).ToLocalUnitType();
							break;
						case 4:
							pSpeed = pSpeed.Subtract(new Speed(0.5, SpeedUnit.StatuteMilesPerHour)).ToLocalUnitType();
							break;
					}

                    // Create a new Position slightly shifted from the last one
					pPosition = pPosition.TranslateTo(pBearing, Speed.ToDistance(UpdateInterval));

					// Has the next position been reached?
					if (pPosition.Equals(pRoute[RouteIndex], 3))
					{
                        // Notify that a waypoint has been reached
                        OnWaypointReached(new PositionEventArgs(pRoute[RouteIndex]));

						// Move to the next point in the route
						RouteIndex++;
						// Are we at the end of the route?
						if (RouteIndex >= pRoute.Count)
						{
							// Yes.  Start over
							RouteIndex = 0;
						}
					}

					// Request for data to be sent to the client
					try
					{
						OnEmulation();
					}
						//				catch(ObjectDisposedException)
						//				{
						//					// Abort processing
						//					throw;
						//				}

					catch (Exception ex)
					{
						OnEmulationException(new ExceptionEventArgs(ex));
					}
					finally
					{
						// Delay for the update interval second					
#if PocketPC
                        Thread.Sleep(250);
#endif
						Thread.Sleep(0); //1000); //(int)UpdateInterval.TotalMilliseconds);
					}
				}
			}
#if !PocketPC || Framework20
			catch (ThreadAbortException)
			{
				// Just exit!
			}
#endif
			catch
			{
				throw;
			}
			finally
			{
				// Flag that the thread is dead
#if Framework20 && !PocketPC
                if (!EmulationWaitHandle.SafeWaitHandle.IsInvalid)
#else
                if (EmulationWaitHandle.Handle != new IntPtr(-1))
#endif
				{
					try
					{
						EmulationWaitHandle.Set();
					}
					catch (ObjectDisposedException)
					{
						// Do nothing, we're shutting down
					}
				}

#if PocketPC && !Framework20
                // Flag that we're now simulating
                pIsEmulationThreadAlive = false;
#endif

            }
        }

		/// <summary>Occurs when the compass direction associated with 0 has changed.</summary>
		public event ExceptionEventHandler EmulationExceptionOccurred;

		public virtual void OnEmulationException(ExceptionEventArgs e)
		{
			try
			{
				if(EmulationExceptionOccurred != null)
#if PocketPC
				EmulationExceptionOccurred(this, e);
#else
					DynamicEvent.Invoke(EmulationExceptionOccurred, this, e);
#endif
			}
			catch
			{
			}
        }

        /// <summary>Controls the amount of CPU time devoted to the emulation thread.</summary>
		/// <value>
		/// A <strong>ThreadPriority</strong> value. Default is
		/// <strong>AboveNormal</strong>.
		/// </value>
		/// <remarks>
		/// 	<para>This property is provided as a way to control the amount of processor time
		///     devoted to device emulation. This property should be used by advanced developers
		///     only because it can lead to unpredictable results, such as timeouts from
		///     interpreters.</para>
		/// 	<para>Performance tuning studies for this class determined that a priority of
		///     <strong>AboveNormal</strong> should be the minimum.</para>
		/// </remarks>
#if !PocketPC && Framework20
		[Obsolete("Changing thread priorty on Desktop Framework 2.0 can cause instability.  A setting of Normal is recommended for all threads.")]
#endif
		public ThreadPriority ThreadPriority
        {
            get
            {
                return pThreadPriority;
            }
            set
            {
                pThreadPriority = value;
            }
        }

        /// <summary>
        /// Controls the simulated baud rate of data transmitted to and from the
        /// client.
        /// </summary>
        /// <value>
        /// A value from the <strong>BaudRate</strong> enumeration. Default is
        /// <strong>Baud4800</strong>.
        /// </value>
        /// <remarks>
        /// This property controls the precise rate at which characters are sent to the
        /// client. Slower baud rates can simulate older GPS devices, whereas higher baud rates
        /// such as 57600 and higher simulate newer GPS devices, including Bluetooth GPS
        /// devices.
        /// </remarks>
        public virtual BaudRate BaudRate
        {
            get
            {
                return pBaudRate;
            }
            set
            {
                pBaudRate = value;
            }
        }

        public virtual string Name
		{
			get
			{
				return pThreadName;
			}
			set
			{
				pThreadName = value;
#if !PocketPC
				EmulationThread.Name = value;
#endif
			}
		}

		/// <summary>
		/// Processes incoming messages and sends simulated data to the client.
		/// </summary>
		/// <remarks><para>This method is the workhorse of a GPS device emulator.  It is responsible for
		/// generating simulated data and writing it to the outgoing stream for clients to process.
		/// This method is launched on a separate thread automatically, so Thread.Sleep commands are 
		/// allowed.</para>
		/// 	<para>The body of this method should be code which processes any incoming data, sends
		/// emulated data.</para>
		/// </remarks>
		/// <example>
		///     This example demonstrates how the OnEmulation method might look for a basic NMEA
		///     emulator. Notice how the Thread.Sleep statement is called to ensure that the stream
		///     is not quickly flooded. 
		///     <code lang="CS">
		/// protected override void OnEmulation()
		/// {
		///    WriteToClient("$GPRMC,,,,,,,,,,,*04");
		///    Thread.Sleep(1000);
		/// }
		///     </code>
		/// 	<code lang="vb">
		/// Protected Overrides Sub OnEmulation()
		///    WriteToClient("$GPRMC,,,,,,,,,,,*04")
		///    Thread.Sleep(1000)
		/// End Sub
		///     </code>
		/// </example>
		protected abstract void OnEmulation();

		public override bool CanRead
		{
			get
			{
                if (pIsDisposed)
                    return false;
                if (!pIsOpen)
                    Open();

                return pIsOpen;
			}
		}

		public override bool CanWrite
		{
			get
			{
                if (pIsDisposed)
                    return false;
                if (!pIsOpen)
                    Open();

                return pIsOpen;
			}
		}

		public override bool CanSeek
		{
			get
			{
				return false;
			}
		}

#if !PocketPC
        public override System.Runtime.Remoting.ObjRef CreateObjRef(Type requestedType)
        {
            return null;
        }
#endif
#if !Framework20
		protected override WaitHandle CreateWaitHandle()
        {
            return new ManualResetEvent(false);
        }
#endif

		/// <summary>
		/// Stops the generation of new emulated data.
		/// </summary>
		public override void Close()
		{	
			// Are we already closed?
			if (pIsDisposed)
				return;

			pIsDisposed = true;

            // We no longer need to dispose
			GC.SuppressFinalize(this);

			lock(ControlSyncRoot)
			{
				// Flag that we're closed
				pIsOpen = false;

#if !PocketPC || DesignTime || (PocketPC && Framework20)
                    // Abort the thread
                    try
                    {
                        if (EmulationThread != null && !EmulationThread.Join(MaximumGracefulShutdownTime))
                            EmulationThread.Abort();

#if PocketPC && Framework20
                        // On Compact Framework 2.0, extra cleanup is required,
                        // since Thread.Abort does NOT cause finally clauses to execute
                        // Flag that the thread is dead
                        if (EmulationWaitHandle.Handle != new IntPtr(-1) && !pIsDisposed)
                        {
                            try
                            {
                                EmulationWaitHandle.Set();
                            }
                            catch (ObjectDisposedException)
                            {
                                // Do nothing, we're shutting down
                            }
                        }
#endif
                    }
                    catch
                    {
                    }
#else
                // Compact Framework 1.0
                 // Wait up to two seconds for a graceful exit
                for (int i = 0; i < 20; i++)
                {
                    // Was it exited?
                    if (!pIsEmulationThreadAlive)
                        // Yes. 
                        break;

                    // No.  Keep waiting
                    Thread.Sleep(100);
                    Thread.Sleep(0);
                }
#endif

#if Framework20 && !PocketPC
                    if (!EmulationWaitHandle.SafeWaitHandle.IsClosed)
#else
                if (EmulationWaitHandle.Handle != new IntPtr(-1))
#endif
                {
                    try
                    {
                        EmulationWaitHandle.Close();
                    }
                    catch
                    {
                    }
                }
			}
		}

		/// <summary>
		/// Waits for an asynchronous read operation to complete.
		/// </summary>
		/// <param name="asyncResult"></param>
		/// <returns></returns>
		public override int EndRead(IAsyncResult asyncResult)
		{
			throw new NotSupportedException("GPS device emulators do not support asynchronous read operations.");
		}

		/// <summary>
		/// Waits for an asynchronous write operation to complete.
		/// </summary>
		/// <param name="asyncResult"></param>
		public override void EndWrite(IAsyncResult asyncResult)
		{
			throw new NotSupportedException("GPS device emulators do not support asynchronous write operations.");
		}

		/// <summary>
		/// This method has no action.
		/// </summary>
		public override void Flush()
		{
			// Does nothing
		}

		/// <summary>
		/// Returns the number of bytes in the outgoing (emulator-to-client) stream.
		/// </summary>
		public override long Length
		{
			get
			{
                if (pIsDisposed)
                    return 0;
                if (!pIsOpen)
                    Open();

				//lock(pOutgoingStream.SyncRoot)
				{
					return pOutgoingStream.Count;
				}
			}
		}

		/// <summary>
		/// Controls the current position in the outgoing (emulator-to-client) stream.
		/// </summary>
		/// <remarks></remarks>
		public override long Position
		{
			get
			{
                //if (pIsDisposed)
                    return 0;
                //if (!pIsOpen)
                    //Open();

				//lock(pOutgoingLock)
				//{
					//return pOutgoingStream.Position;
				//}
			}
			set
			{
                throw new IOException("Seek operations are not supported by GPS device emulator streams.");
			}
		}

		public override int Read(byte[] buffer, int offset, int count)
		{
            if (pIsDisposed)
                throw new ObjectDisposedException(GetType().ToString(), "Read operations are not allowed because the emulation stream has been closed.");
			if(!pIsOpen)
				Open();

			// Dequeue "count" bytes from the queue
			int ActualRead = 0;
			lock(pOutgoingStream.SyncRoot)
			{
				if(pOutgoingStream.Count == 0)
					return -1;	

				for(int index = 0; index < count; index++)
				{				
					buffer[index + offset] = (byte)pOutgoingStream.Dequeue();
					ActualRead++;
				}
			}
			Thread.Sleep(0);
			return ActualRead;
		}

		public override int ReadByte()
		{
            if (pIsDisposed)
                throw new ObjectDisposedException(GetType().ToString(), "Read operations are not allowed because the emulation stream has been closed.");
            if (!pIsOpen)
				Open();

			// Return a -1 if no data is available
			lock(pOutgoingStream.SyncRoot)
			{
				if(pOutgoingStream.Count == 0)
					return -1;
				else
					return (byte)pOutgoingStream.Dequeue();
			}
			//Thread.Sleep(0);
		}

		public override long Seek(long offset, SeekOrigin origin)
		{
            throw new IOException("Seek operations are not supported by GPS device emulator streams.");
        }

		public override void SetLength(long value)
		{
			throw new EmulatorException("The length of the stream is controlled automatically by the emulator.");
		}

        public override string ToString()
        {
			return ToString("G", CultureInfo.CurrentCulture);
        }

		public virtual string ToString(string format, IFormatProvider formatProvider)
		{
			return "GPS Device Emulator";
		}

		public override void Write(byte[] buffer, int offset, int count)
		{
            if (pIsDisposed)
                throw new IOException("Read operations are not allowed because the emulation stream has been closed.");
            if (!pIsOpen)
                Open();

			// Add the bytes to the incoming stream
			lock(pIncomingStream.SyncRoot)
			{
				for(int index = 0; index < count; index++)
				{
					// Queue the item
					pIncomingStream.Enqueue(buffer[offset + index]);
				}
			}
			Thread.Sleep(0);
        }

        public override void WriteByte(byte value)
        {
            if (pIsDisposed)
                throw new IOException("Read operations are not allowed because the emulation stream has been closed.");
            if (!pIsOpen)
				Open();

			lock(pIncomingStream.SyncRoot)
			{
				// Queue the item
				pIncomingStream.Enqueue(value);
			}
			Thread.Sleep(0);
        }

		#endregion

        #region Protected Members 

        /// <summary>
        /// Represents the stream of data arriving to the emulator from the client.
        /// </summary>
        /// <remarks>The "incoming stream" is the flow of data being sent from the client to the emulator.  The emulator will examine this
        /// stream for commands to process from the client.  Depending on the GPS data protocol in use, this stream may or may not be supported.
        /// For NMEA emulators, no incoming commands are processed.  For SiRF and Garmin Phase Output protocols, incoming commands are processed.</remarks>
        protected Queue IncomingStream
		{
			get
			{
				return pIncomingStream;
			}
		}

        /// <summary>
        /// Represents the stream of data being sent from the emulator to the client.
        /// </summary>
        /// <remarks>The "outgoing stream" is the flow of data being sent from the emulator to the client.  This data is generated internally by the 
        /// emulator to simulate a real GPS device.  Data is added to this buffer via the <strong>WriteToClient</strong> methods, and write operations
        /// will only succeed if the outgoing buffer is not already full.</remarks>
		protected Queue OutgoingStream
		{
			get
			{
				return pOutgoingStream;
			}
		}

		/// <summary>
		/// Sends an entire byte array of simulated data to the client.
		/// </summary>
		/// <param name="buffer">An array of bytes to send.</param>
		protected void WriteToClient(byte[] buffer)
		{
			WriteToClient(buffer, 0, buffer.Length);
		}

		/// <summary>
		/// Sends a portion of a byte array of simulated data to the client.
		/// </summary>
		/// <param name="buffer">An array of bytes to send.</param>
		/// <param name="offset">The index at which sending begins.</param>
		/// <param name="count">the number of bytes to transmit.</param>
		protected void WriteToClient(byte[] buffer, int offset, int count)
		{
			// Write the data to the outgoing stream
			lock(pOutgoingStream.SyncRoot)
			{
                // NEW 11/6/2007: Outgoing buffers are growing so large they crash 
                // the device by eating up memory.  Now, only write the buffer if
                // the buffer will not get overrun.
                if (pOutgoingStream.Count + count >= pOutgoingBufferSize)
                    return;
                
                // Write out the buffer
				for(int index = 0; index < count; index++)
				{
						pOutgoingStream.Enqueue(buffer[index + offset]);
				}
			}
			Thread.Sleep(0);

            // Simulate a serial device by sleeping
            int BaudRateTime = (int)((1000.0 / ((int)pBaudRate / 8.0)) * count) - 100;
            if(BaudRateTime > 0)
	            Thread.Sleep(Math.Min(BaudRateTime, 50));
        }
	
		/// <summary>
		/// Sends a string of simulated data to the client.
		/// </summary>
		/// <param name="buffer">A string containing simulated data to send.</param>
		protected void WriteToClient(string buffer) 
		{
			WriteToClient(System.Text.ASCIIEncoding.ASCII.GetBytes(buffer));
		}

        /// <summary>
        /// Sends a string of simulated data to the client, followed by a carriage return.
        /// </summary>
        /// <param name="buffer">A string containing simulated data to send.</param>
        protected void WriteLineToClient(string buffer)
        {
            WriteToClient(System.Text.ASCIIEncoding.ASCII.GetBytes(buffer + "\r\n"));
        }

        protected void WriteToClient(Packet packet)
        {
            WriteToClient(packet.GetBytes());
        }


        protected int ReadFromClient(byte[] buffer, int offset, int count)
        {
			// Read bytes from the client
			lock(pIncomingStream.SyncRoot)
			{
				for(int index = 0; index < count; index++)
				{
					buffer[index + offset] = (byte)pIncomingStream.Dequeue();
				}
			}
			Thread.Sleep(0);

			// Simulate baud rate by sleeping
			int BaudRateTime = (int)((1000.0 / ((int)pBaudRate / 8.0)) * count);
			if (BaudRateTime > 0)
				Thread.Sleep(BaudRateTime);

			return count;
		}

        #endregion

        /// <summary>
        /// Creates a copy of this emulator.
        /// </summary>
        /// <returns></returns>
		public virtual Emulator Clone()
		{
			// Create a copy of this same type
			return (Emulator)Activator.CreateInstance(this.GetType());
		}

        /// <summary>
        /// Controls the size of the buffer of data being sent to the client.
        /// </summary>
        /// <remarks>In some rare cases, an Emulator class may be generating data, but the data is not being read by the client.
        /// In such cases, memory could be exhausted by the Emulator unless some limit is reached.  This property controls the
        /// maximum size allowed for the outgoing buffer.  If the buffer size is reached, attempts to add data to the buffer
        /// are ignored until the buffer's size decreases.</remarks>
        public int OutgoingBufferSize
        {
            get
            {
                return pOutgoingBufferSize;
            }
            set
            {
                // Make sure the value isn't out of range
                if (value < 0)
#if PocketPC && !Framework20
					throw new ArgumentOutOfRangeException("An emulator's outgoing buffer size must be greater than zero.  A value of 16384 is the default.");
#else
                    throw new ArgumentOutOfRangeException("value", "An emulator's outgoing buffer size must be greater than zero.  A value of 16384 is the default.");
#endif

                // Set the new buffer size
                pOutgoingBufferSize = value;
            }
        }

        /// <summary>
        /// Controls whether the virtual route is random or created manually.
        /// </summary>
        /// <remarks>When set to a value of True, GPS.NET will create a randomized route to
        /// follow by the emulator.  When False, the Route property must be set to a set of
        /// manually-supplied values.  This property is typically set to False to make an emulator 
        /// travel a pre-designed route.</remarks>
        public bool AllowRandomRoutes
        {
            get
            {
                return pAllowRandomRoutes;
            }
            set
            {
                // Has anything actually changed?
                if (pAllowRandomRoutes = value)
                    return;

                // Set the new value
                pAllowRandomRoutes = value;

                // Generate a new route if we need to
                if (pAllowRandomRoutes)
                    RandomizeRoute();
            }
        }

        /// <summary>
        /// Creates a new virtual route to be traveled by the emulator.
        /// </summary>
        public void RandomizeRoute()
        {
            PositionCollection NewRoute = new PositionCollection(20);
            GeoFramework.Position StartPosition = null;
            // Do we have a current position?
            if (CurrentPosition.IsEmpty)
            {
                // No.  Use a random position inside of the boundary
                StartPosition = GeoFramework.Position.Random(pRouteBoundary);
            }
            else
            {
                // Yes.  But is this position outside the boundary?
                if (CurrentPosition.IsWithin(pRouteBoundary))
                    // Yes.  Start it at the current location
                    StartPosition = CurrentPosition;
                else
                    // Yes.  Start in the center of the route boundary
                    StartPosition = pRouteBoundary.Center;
            }
            Position NextPosition = CurrentPosition;
            // Now add 50 positions to the route
            for (int count = 0; count < 50; count++)
            {
                // Shift the position
                for (int tryCount = 0; tryCount < 50; tryCount++)
                {
                    NextPosition = NextPosition.TranslateTo(
                        Azimuth.Random(),
                        new Distance(200, DistanceUnit.Meters));
                    // If the position is not in the route, try again
                    if (pRouteBoundary.IsOverlapping(NextPosition))
                        break;
                }

                // And add it to the route
                NewRoute.Add(NextPosition);
            }
            // Finally, commit the route
			pRoute = NewRoute;

            // If we're using random routes, start the emulator at the first position.
            if (pAllowRandomRoutes)
                pPosition = pRoute[0];
        }

        #region Public Properties

        /// <summary>
        /// Occurs when the virtual GPS position has reached the next position in the virtual route.
        /// </summary>
        /// <param name="e">A <strong>PositionEventArgs</strong> object describing the waypoint which has been reached.</param>
        protected void OnWaypointReached(PositionEventArgs e)
        {
            // Notify of the new waypoint
            if (WaypointReached != null)
#if PocketPC && !Framework20
				WaypointReached(this, e);
#else
                DynamicEvent.Invoke(WaypointReached, this, e);
#endif

        }

        /// <summary>
        /// Indicates whether the emulator is currently processing information.
        /// </summary>
        public bool IsOpen
        {
            get
            {
                return pIsOpen;
            }
        }

        /// <summary>
        /// Indicates whether the emulator has been shut down.
        /// </summary>
		public bool IsDisposed
		{
			get
			{
				return pIsDisposed;
			}
		}

        /// <summary>
        /// Controls the current virtual altitude of the emulator.
        /// </summary>
        public Distance Altitude
        {
            get
            {
                return pAltitude;
            }
            set
            {
				if(pAltitude.Equals(value)) return;
				pAltitude = value;
            }
        }

        /// <summary>
        /// Returns the current virtual direction of travel.
        /// </summary>
        public Azimuth Bearing
        {
            get
            {
                return pBearing;
            }
        }

        /// <summary>
        /// Returns the current virtual date and time of the emulator.
        /// </summary>
        public DateTime UtcDateTime
        {
            get
            {
                return pUtcDateTime;
            }
        }

        /// <summary>
        /// Controls the current virtual location of the emulator.
        /// </summary>
        public Position CurrentPosition
        {
            get
            {
                return pPosition;
            }
            set
            {
                // Has the position actually changed?
				if(pPosition.Equals(value)) 
                    return;

                // Set the new value
				pPosition = value;                
            }
        }

        /// <summary>
        /// Returns the virtual path taken by the emulator.
        /// </summary>
        public PositionCollection Route
        {
            get
            {
                return pRoute;
            }
        }

        /// <summary>
        /// Controls the rectangle inside of which any random routes are generated.
        /// </summary>
        public GeographicRectangle RouteBoundary
        {
            get
            {
                return pRouteBoundary;
            }
            set
            {
                pRouteBoundary = value;
            }
        }

        /// <summary>
        /// Controls the virtual speed that is traveled by the emulator towards the next position in the route.
        /// </summary>
        public Speed Speed
        {
            get
            {
                return pSpeed;
            }
            set
            {
				if(pSpeed.Equals(value)) return;
				pSpeed = value;
            }
        }

        #endregion

	}
}
