using System;


namespace GeoFramework.Gps
{
	/// <summary>
	/// Indicates the likelihood that a GPS satellite fix will be obtained or
	/// sustained.
	/// </summary>
	/// <example>
	///     This example uses the FixLikelihood enumeration to make a judgement call on the
	///     stability of the fix. 
	///     <code lang="VB">
	/// ' Declare a new receiver
	/// Private WithEvents MyReceiver As New Receiver()
	///  
	/// Sub Main()
	///  ' Start listening for messages
	///  MyReceiver.Start
	/// End Sub
	///  
	/// Sub OnFixLikelihoodChanged(ByVal sender As Object, ByVal e As FixLikelihoodEventArgs) Handles MyReceiver.FixLikelihoodChanged
	///  ' Do we have a fix currently
	///  If MyReceiver.IsFixObtained Then
	///      ' Yes.  What's the likelihood that the fix will be sustained
	///      Select Case MyReceiver.FixLikelihood
	///          Case FixLikelihood.Unlikely
	///              Debug.WriteLine("The current fix is about to be lost!")
	///          Case FixLikelihood.Possible
	///              Debug.WriteLine("The current fix is unstable.  Find a more open view of the sky soon.")
	///          Case FixLikelihood.Likely
	///              Debug.WriteLine("The current fix is nearly stable.")
	///          Case FixLikelihood.Certain
	///              Debug.WriteLine("The current fix is stable.")
	///      End Select
	///  Else
	///      ' No. What's the likelihood that a fix will be obtained
	///      Select Case MyReceiver.FixLikelihood
	///          Case FixLikelihood.Unlikely
	///              Debug.WriteLine("A fix is not possible.  Find a more open view of the sky.")
	///          Case FixLikelihood.Possible
	///              Debug.WriteLine("A fix is possible, but satellite signals are still mostly obscured.")
	///          Case FixLikelihood.Likely
	///              Debug.WriteLine("A fix should occur within the next several seconds.")
	///          Case FixLikelihood.Certain
	///              Debug.WriteLine("A fix will occur in a few seconds.")
	///      End Select
	///  End If
	/// End Sub
	///     </code>
	/// 	<code lang="CS">
	/// ' Declare a new receiver
	/// Private Receiver MyReceiver = new Receiver();
	///  
	/// void Main()
	/// {
	///  // Start listening for messages
	///  MyReceiver.Start();
	/// }
	///  
	/// void OnFixLikelihoodChanged(Object sender, FixLikelihoodEventArgs e)
	/// {
	///  // Do we have a fix currently
	///  if (MyReceiver.IsFixObtained)
	///  {
	///      // Yes.  What's the likelihood that the fix will be sustained
	///      switch(MyReceiver.FixLikelihood)
	///      {
	///          case FixLikelihood.Unlikely:
	///              Debug.WriteLine("The current fix is about to be lost!");
	///              break;
	///          case FixLikelihood.Possible:
	///              Debug.WriteLine("The current fix is unstable.  Find a more open view of the sky soon.");
	///              break;
	///          case FixLikelihood.Likely:
	///              Debug.WriteLine("The current fix is nearly stable.");
	///              break;
	///          case FixLikelihood.Certain:
	///              Debug.WriteLine("The current fix is stable.");
	///              break;
	///      }
	///  }
	///  else
	///  {
	///      // No. What's the likelihood that a fix will be obtained
	///      switch(MyReceiver.FixLikelihood)
	///      {
	///          case FixLikelihood.Unlikely:
	///              Debug.WriteLine("A fix is not possible.  Find a more open view of the sky.");
	///              break;
	///          case FixLikelihood.Possible:
	///              Debug.WriteLine("A fix is possible, but satellite signals are still mostly obscured.");
	///              break;
	///          case FixLikelihood.Likely:
	///              Debug.WriteLine("A fix should occur within the next several seconds.");
	///              break;
	///          case FixLikelihood.Certain:
	///              Debug.WriteLine("A fix will occur in a few seconds.");
	///              break;
	///      }
	///  }
	/// }
	///     </code>
	/// </example>
	/// <seealso cref="Receiver.FixLikelihood">FixLikelihood Property (Receiver Class)</seealso>
	public enum FixLikelihood: int
	{
		/// <summary>Indicates that a fix would probably be lost if a fix is acquired.</summary>
		/// <remarks>When this value is returned, nearly all of the available GPS satellite signals are being obscured by buildings, trees, or other solid objects.  The device should be moved into a more open view of the sky.</remarks>
		Unlikely = 40,
		/// <summary>
		/// Indicates that a fix would probably be lost after a short period of time if a fix
		/// is acquired.
		/// </summary>
		/// <remarks>When this value is returned, a few satellite signals are available, but the combined signals are too weak to maintain a fix for long.  The device should be moved into a more open view of the sky.</remarks>
		Possible = 80,
		/// <summary>
		/// Indicates that a fix would probably last for a longer period of time if a fix is
		/// acquired.
		/// </summary>
		/// <remarks>When this value is returned, at least three satellite signals are being received clearly.  If conditions stay the same or improve, a fix is imminent.</remarks>
		Likely = 105,
		/// <summary>
		/// Indicates that a fix is very likely to be sustained given the current satellite
		/// signal strengths.
		/// </summary>
		/// <remarks>When this value is returned, several satellite signals are being received and are strong enough to maintain a fix.</remarks>
		Certain = 135
	}

//	/// <summary>Indicates the name of a company which manufactured the GPS device.</summary>
//	/// <seealso cref="Device.Manufacturer">Manufacturer Property (Device Class)</seealso>
//	/// <seealso cref="Device">Device Class</seealso>
//	public enum Manufacturer: int
//	{
//		/// <summary>The manufacturer is unknown.</summary>
//		Unknown = 0,
//		/// <summary>
//		/// Garmin International, Inc.
//		/// (<a href="http://www.garmin.com">http://www.garmin.com</a>)
//		/// </summary>
//		Garmin = 1,
//		/// <summary>
//		/// Magellan, Inc.
//		/// (<a href="http://www.magellangps.com">http://www.magellangps.com</a>)
//		/// </summary>
//		Magellan = 2,
//		/// <summary>
//		/// DeLorme, Inc.
//		/// (<a href="http://www.delorme.com">http://www.delorme.com</a>)
//		/// </summary>
//		DeLorme = 3,
//		/// <summary>
//		/// Lowrance Electronics, Inc.
//		/// (<a href="http://www.lowrance.com">http://www.lowrance.com</a>)
//		/// </summary>
//		Lowrance = 4,
//		/// <summary>
//		///	 Holux Technology, Inc.
//		///	 (<a href="http://www.holux.com">http://www.holux.com</a>)
//		/// </summary>
//		Holux = 5,
//		/// <summary>
//		/// Leica Geosystems, Inc.
//		/// (<a href="http://www.leica-geosystems.com/corporate/en/lgs_index.htm">http://www.leica-geosystems.com</a>)
//		/// </summary>
//		Leica = 6
//	}



	/// <summary>
	/// Indicates an interpretation of the accuracy of a measurement made by the GPS
	/// device.
	/// </summary>
	/// <remarks>
	/// This enumeration is used by the
	/// <see cref="DilutionOfPrecision.Rating">Rating</see> property of the
	/// <see cref="DilutionOfPrecision">
	/// DilutionOfPrecision</see> class. This interpretation is subject to discussion as to
	/// what precisely constitutes a "good" versus "bad" DOP. Use your own best judgement based
	/// on the needs of your application. Generally speaking, a DOP of six or less is
	/// recommended for en-route navigation, and three or less for highly-precise measurements.
	/// A rating of <strong>Moderate</strong> corresponds with six or better, a rating of
	/// <strong>Excellent</strong> means three or less, and a rating of <strong>Ideal</strong>
	/// means the best value of one.
	/// </remarks>
	public enum DilutionOfPrecisionRating: int
	{
		/// <summary>The rating is unknown or not yet available.</summary>
		Unknown = 0,
		/// <summary>
		/// Represents a value of 1, where the GPS device is making the most accurate
		/// measurements possible.
		/// </summary>
		Ideal = 1,
		/// <summary>
		/// The GPS device is making high-quality measurements, good enough for applications
		/// requiring higher levels of precision. Represents a value of 2 or 3.
		/// </summary>
		Excellent = 3,
		/// <summary>
		/// The GPS device is making measurements accurate enough for en-route navigation.
		/// Represents a value between 4 and 6.
		/// </summary>
		Good = 6,
		/// <summary>
		/// The GPS device is making measurements good enough to indicate the approximate
		/// location, but should be ignored by applications requiring high accuracy. Represents a
		/// value of 7 or 8.
		/// </summary>
		Moderate = 8,
		/// <summary>
		/// The GPS device is making measurements with an accuracy which should only be used
		/// to indicate the approximate location, but is not accurate enough for en-route
		/// navigation. Represents a value between 9 and 20.
		/// </summary>
		Fair = 20,
		/// <summary>
		/// The GPS device is calculating the current location, but the accuracy is extremely
		/// low and may be off by a factor of fifty. Represents a value between 21 and the maximum
		/// possible of 50.
		/// </summary>
		Poor = 50
	}

//	/// <summary>
//	/// Identifies what kind of measurements a dilution of precision value
//	/// describes.
//	/// </summary>
//	/// <remarks>
//	/// This enumeration is used by the
//	/// <see cref="DilutionOfPrecision.Type">Type</see> property of the
//	/// <see cref="DilutionOfPrecisionEventArgs">
//	/// DilutionOfPrecisionEventArgs</see> to allow developers the ability to combine multiple
//	/// DOP events into one method, yet still have a way to tell measurements apart.
//	/// </remarks>
//	public enum DilutionOfPrecisionType: int
//	{
//		/// <summary>
//		/// The type of precision is unknown.
//		/// </summary>
//		Unknown,
//		/// <summary>
//		/// The dilution of precision measurement is an overall measure of accuracy in
//		/// latitude, longitude, or altitude.
//		/// </summary>
//		Mean,
//		/// <summary>
//		/// The dilution of precision measurement is made in regards to latitude and
//		/// longitude.
//		/// </summary>
//		Horizontal,
//		/// <summary>The dilution of precision measurement is made in regards to 
//		/// altitude.</summary>
//		Vertical
//	}
//
//	/// <summary>Indicates the type of error being measured.</summary>
//	/// <remarks>
//	/// This enumeration is used by the
//	/// <see cref="Distance.Type">Type</see> property of the
//	/// <see cref="Distance">Distance</see> class to allow developers the ability to handle multiple position
//	/// error events in one method and be able to tell objects apart.
//	/// </remarks>
//	public enum DistanceType: int
//	{
//		/// <summary>The type of error is unspecified.</summary>
//		Unknown,
//		/// <summary>An overall average of latitude, longitude and altitude measurements.</summary>
//		Mean,
//		/// <summary>A measurement of error as it relates to latitude and longitude.</summary>
//		Horizontal,
//		/// <summary>A measurement of error as it relates to altitude.</summary>
//		Vertical
//	}

	//Private Enum AreaUnit
	//	KilometersSquared
	//	StatuteMilesSquared
	//	MetersSquared
	//	FeetSquared
	//	InchesSquared
	//End Enum

	/// <summary>Indicates a steering direction recommended to correct cross-track 
	/// error.</summary>
	/// <remarks>
	/// This enumeration is most frequently used during flights to maintain the proper
	/// course to a navigation point.
	/// </remarks>
	public enum RecommendedSteeringDirection: int
	{
		/// <summary>
		/// No recommendation is available.
		/// </summary>
		None,
		/// <summary>Steer left to resume course.</summary>
		Left,
		/// <summary>Steer right to resume course.</summary>
		Right
	}

	/// <summary>
	/// Indicates which devices are being used to obtain a fix, other than the GPS device
	/// itself.
	/// </summary>
	/// <remarks>
	/// This enumeration is typically used by the
	/// <see cref="Receiver.FixQuality">FixQuaity</see> property of the
	/// <see cref="Receiver">Receiver</see> class.
	/// </remarks>
	/// <example>
	/// <code lang="VB">
	/// ' Declare a new receiver
	/// Private WithEvents MyReceiver As New Receiver()
	/// 
	/// ' Raised when the fix quality has changed
	/// Private Sub OnFixQualityChanged(ByVal sender As Object, ByVal e As FixQualityEventArgs)
	///	 ' What is the new fix quality
	///	 Select Case e.FixQuality
	///		 Case FixQuality.NoFix
	///			 ' No fix is obtained
	///			 Debug.WriteLine("No fix is currently obtained.")
	///		 Case FixQuality.GpsFix
	///			 ' A fix is present
	///			 Debug.WriteLine("A fix has been obtained using only GPS satellites.")
	///		 Case FixQuality.DifferentialGpsFix
	///			 ' A differential fix is present
	///			 Debug.WriteLine("A fix has been obtained using GPS satellites and correction information from DGPS/WAAS ground stations.")
	///		 Case FixQuality.Estimated
	///			 ' A fix is being estimated (not an actual fix)
	///			 Debug.WriteLine("A fix is currently being estimated. ")
	///	 End Select 
	/// End Sub
	/// </code>
	/// <code lang="C#">
	/// // Declare a new receiver
	/// Receiver MyReceiver = new Receiver();
	/// 
	/// // Raised when the fix quality has changed
	/// void OnFixQualityChanged(Object sender, FixQualityEventArgs e)
	/// {
	///	 // What is the new fix quality
	///	 switch(e.FixQuality)
	///	 {
	///		 case FixQuality.NoFix:
	///			 // No fix is obtained
	///			 Debug.WriteLine("No fix is currently obtained.");
	///		 case FixQuality.GpsFix:
	///			 // A fix is present
	///			 Debug.WriteLine("A fix has been obtained using only GPS satellites.");
	///		 case FixQuality.DifferentialGpsFix:
	///			 // A differential fix is present
	///			 Debug.WriteLine("A fix has been obtained using GPS satellites and correction information from DGPS/WAAS ground stations.");
	///		 case FixQuality.Estimated:
	///			 // A fix is being estimated (not an actual fix)
	///			 Debug.WriteLine("A fix is currently being estimated. ");
	///	 }
	/// }
	/// </code>
	/// </example>
	/// <seealso cref="Receiver.FixQuality">FixQuality Property (Receiver Class)</seealso>
	public enum FixQuality: int
	{
		/// <summary>No fix is currently obtained.</summary>
		NoFix,
		/// <summary>A fix is currently obtained using GPS satellites only.</summary>
		GpsFix,
		/// <summary>A fix is obtained using both GPS satellites and DGPS/WAAS ground 
		/// stations.</summary>
		DifferentialGpsFix,
		/// <summary>Rarely used. The fix is being estimated.</summary>
		Estimated,
		/// <summary>Leica devices only.  A navigation fix is obtained, but it may be out of date.  No real-time position is available.</summary>
		NavigationFixNoRealtimePosition,
		/// <summary>Leica devices only.  A real-time position has been obtained, but there are some unresolved accuracy ambiguities.</summary>
		RealTimePositionAmbiguitiesNotFixed,
		/// <summary>Leica devices only.  A real-time position has been obtained and all accuracy ambiguities have been resolved.</summary>
		RealTimePositionAmbiguitiesFixed
	}

	/// <summary>
	/// Indicates if the GPS device is automatically deciding between a 2-D and a 3-D
	/// fix.
	/// </summary>
	/// <remarks>
	/// This enumeration is used by the
	/// <see cref="Receiver.FixMode">FixMode</see> property of the
	/// <see cref="Receiver">Receiver</see>
	/// class. A vast majority of GPS devices use a setting of Automatic because there is no
	/// complicated math behind figuring out if a 2-D fix or 3-D fix is present. If there are
	/// three satellites involved in a fix, only the latitude and longitude can be calculated,
	/// so the fix is 2-D. If more than three involved, the fix is 3-D.
	/// </remarks>
	/// <seealso cref="Receiver.FixMode">FixMode Property (Receiver Class)</seealso>
	public enum FixMode: int
	{
		/// <summary>Typical value.  The GPS device is automatically deciding between a two- and three-dimensional fix.</summary>
		Automatic,
		/// <summary>Rare value.  The user must specify whether a two- or three-dimensional fix is to be used.</summary>
		Manual
	}

	/// <summary>
	/// Indicates whether a fix is present and if altitude can be calculated along with
	/// latitude and longitude.
	/// </summary>
	/// <remarks>
	/// This enumeration is used by the 
	/// <see cref="Receiver.FixMethod">FixMethod</see> property of the
	/// <see cref="Receiver">Receiver</see>
	/// class to indicate if altitude measurements are possible. Altitude measurements are
	/// possible whenever there are four or more satellites involved in a fix.
	/// </remarks>
	/// <example>
	/// This example demonstrates how to use the FixMethod enumeration to tell if altitude 
	/// measurements can be made.  NOTE: Some devices have built-in altimeters.  These devices
	/// can report altitude even when there is no fix present.  To support such devices, avoid
	/// using this property entirely and use any non-zero altitude measurement.
	/// <code lang="VB">
	/// Private WithEvents MyReceiver As New Receiver()
	/// 
	/// ' Raised whenever the fix method has changed
	/// Private Sub OnFixMethodChanged(ByVal sender As Object, ByVal e As FixMethodEventArgs)
	///	 ' What is the new fix method
	///	 Select Case e.FixMethod
	///		 Case FixMethod.NoFix
	///			 ' We have neither position or altitude
	///			 Debug.WriteLine("Your position is: Not Yet Available")
	///			 Debug.WriteLine("Your altitude is: Not Yet Available")
	///		 Case FixMethod.Fix2D
	///			 ' We have a position but no altitude
	///			 Debug.WriteLine("Your position is: " &amp; MyReceiver.Position.ToString)
	///			 Debug.WriteLine("Your altitude is: Not Yet Available")
	///		 Case FixMethod.Fix3D
	///			 ' We have both position and altitude
	///			 Debug.WriteLine("Your position is: " &amp; MyReceiver.Position.ToString)
	///			 Debug.WriteLine("Your altitude is: " &amp; MyReceiver.Altitude.ToString)
	///	 End Select
	/// End Sub
	/// </code>
	/// <code lang="C#">
	/// // Declare a new receiver
	/// Private WithEvents MyReceiver As New Receiver()
	/// 
	/// // Raised whenever the fix method has changed
	/// void OnFixMethodChanged(Object sender, FixMethodEventArgs e)
	/// {
	///	 // What is the new fix method
	///	 switch(e.FixMethod)
	///	 {
	///		 case FixMethod.NoFix:
	///			 // We have neither position or altitude
	///			 Debug.WriteLine("Your position is: Not Yet Available");
	///			 Debug.WriteLine("Your altitude is: Not Yet Available");
	///		 case FixMethod.Fix2D:
	///			 // We have a position but no altitude
	///			 Debug.WriteLine("Your position is: " &amp; MyReceiver.Position.ToString());
	///			 Debug.WriteLine("Your altitude is: Not Yet Available");
	///		 case FixMethod.Fix3D:
	///			 // We have both position and altitude
	///			 Debug.WriteLine("Your position is: " &amp; MyReceiver.Position.ToString());
	///			 Debug.WriteLine("Your altitude is: " &amp; MyReceiver.Altitude.ToString());
	///	 }
	/// }
	/// </code>
	/// </example>
	/// <seealso cref="Receiver.FixMethod">FixMethod Property (Receiver Class)</seealso>
	public enum FixMethod: int
	{
		/// <summary>The GPS device does not have a fix on the current position.</summary>
		NoFix,
		/// <summary>The GPS device is reporting latitude and longitude.</summary>
		Fix2D,
		/// <summary>The GPS device is reporting latitude, longitude, and altitude.</summary>
		Fix3D
	}

	/// <summary>Indicates an in-English description of the a satellite's radio signal strength</summary>
	/// <remarks>This enumeration is used by the <see cref="SignalToNoiseRatio.Rating">Rating</see> property of the 
	/// <see cref="SignalToNoiseRatio">SignalToNoiseRatio</see> class to give an in-English interpretation 
	/// of satellite radio signal strength.
	/// </remarks>
	public enum SignalToNoiseRatioRating: int
	{
		/// <summary>Represents a value of 0. The radio signal is completely obscured.</summary>
		None,
		/// <summary>Represents a value between 1 and 15.  The radio signal is mostly obscured, such as by a building or tree, but might briefly be part of a fix.</summary>
		Poor,
		/// <summary>Represents a value between 16 and 30.  The radio signal is partially obscured, but could be part of a sustained fix.</summary>
		Moderate,
		/// <summary>Represents a value between 31 and 40.  The radio signal is being received with little interferance and could maintain a reliable fix.</summary>
		Good,
		/// <summary>Represents a value of 41 or above.  The satellite is in direct line of sight from the receiver and can sustain a fix.</summary>
		Excellent
	}

	/// <summary>
	/// Indicates the current state of execution of a processing engine, such as a GPS data interpreter.
	/// </summary>
	[Flags()]
	public enum ExecutionState
	{
		Unknown = 0,
		Stopped = 1,
		Starting = 2,
		Running = 4,
		Pausing = 8,
		Paused = 16,
		Resuming = 32,
		Restarting = 64,
		DeviceSuspending = 128,
		DeviceResuming = 256,
		Stopping = 512,
		Disposing = 1024,
		Disposed = 2048,
		SafeForEvents = Stopped | Running | Paused
	}

}