using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using GeoFramework;
using GeoFramework.IO;
using GeoFramework.IO.Serial;
using GeoFramework.Gps;
using GeoFramework.Gps.Emulators;
using GeoFramework.Gps.Nmea;

namespace GPS.NET_Example
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();

            /* GPS.NET provides the ability to quickly discover GPS devices on the
             * local system.  This feature is known as "Automatic Device Discovery"
             * and greatly simplifies the task of finding a GPS device to communicate
             * with.  Developers can simply call the "Start" method of an interpreter,
             * and GPS.NET will automatically locate the best GPS Device to work with.
             * 
             * During Device Discovery, several events are raised to indicate the
             * progress of detection.  This example hooks into almost all of the events
             * to give users feedback.
             */
            GeoFramework.IO.Devices.DeviceDetectionStarted += new EventHandler(Devices_DeviceDetectionStarted);
            GeoFramework.IO.Devices.DeviceDetected += new GeoFramework.IO.DeviceEventHandler(Devices_DeviceDetected);
            GeoFramework.IO.Devices.DeviceDetectionAttempted += new GeoFramework.IO.DeviceEventHandler(Devices_DeviceDetectionAttempted);
            GeoFramework.IO.Devices.DeviceDetectionCompleted += new EventHandler(Devices_DeviceDetectionCompleted);
            GeoFramework.IO.Devices.DeviceDetectionExceptionOccurred += new GeoFramework.ExceptionEventHandler(Devices_DeviceDetectionExceptionOccurred);
            GeoFramework.IO.Devices.DeviceFailure += new GeoFramework.IO.DeviceEventHandler(Devices_DeviceFailure);
        }

        private void Form1_Closed(object sender, System.EventArgs e)
        {
            /* The Dispose method should also be called for any objects that 
             * implement it.  This prevents memory leaks and can help the application
             * shut down more quickly. */

            Dispose(true);

            /* With everything shut down, the application can now gracefully exit. */

            Application.Exit();
        }

        #region GPS Device Detection Events

        private void SearchButton_Click(object sender, EventArgs e)
        {
            // Start the GPS device detection process.  This operation will complete
            // in the background, on a separate thread.  This method can be called again
            // to attempt to locate new devices.
            GeoFramework.IO.Devices.Detect();
        }

        private void Devices_DeviceDetectionStarted(object sender, EventArgs e)
        {
            StatusBar1.Text = "Detecting GPS devices...";
            SearchButton.Enabled = false;
        }

        private void Devices_DeviceFailure(object sender, GeoFramework.IO.DeviceEventArgs e)
        {
            StatusBar1.Text = "Not Responding: " + e.Device.ToString();
        }

        private void Devices_DeviceDetectionExceptionOccurred(object sender, GeoFramework.ExceptionEventArgs e)
        {
            StatusBar1.Text = "Error: " + e.Exception.ToString();
        }

        private void Devices_DeviceDetectionCompleted(object sender, EventArgs e)
        {
            StatusBar1.Text = "Device detection completed.";
            SearchButton.Enabled = true;
        }

        private void Devices_DeviceDetectionAttempted(object sender, GeoFramework.IO.DeviceEventArgs e)
        {
            StatusBar1.Text = "Detecting " + e.Device.ToString();
        }

        private void Devices_DeviceDetected(object sender, GeoFramework.IO.DeviceEventArgs e)
        {
            StatusBar1.Text = "Found Device: " + e.Device.ToString();

            // Add it to the list box
            ListViewItem DeviceItem = new ListViewItem(e.Device.ToString());
            // Associate the detected device with this list view item, so we can connect 
            // to the device later on
            DeviceItem.Tag = e.Device;
            // Set the image of the new item
            DeviceItem.ImageIndex = 0;
            DevicesListView.Items.Add(DeviceItem);
        }

        #endregion

        #region Connecting and Disconnecting

        private void StartButton_Click(object sender, EventArgs e)
        {
            // What radio button is selected?
            if (this.UseAnyAvailableDeviceRadioButton.Checked)
            {
                // "Use any available device"

                /* To instruct GPS.NET to use any available GPS device,
                 * the interpreter is assigned to a "GpsStream" object. This
                 * object provides data access to the fastest-responding GPS
                 * device on the system.
                 */

                Interpreter.BaseStream = new GpsStream();

            }
            else if (UseDeviceRadioButton.Checked)
            {
                // "Use the following GPS device:"

                /* GPS.NET interpreters can be assigned to any kind of
                 * Stream object, making them very versatile!  FileStream
                 * and NetworkStream objects can be used, as well as GPS.NET's
                 * own SerialStream and BluetoothStream classes.
                 */

                // Find out which item is selected in the list view
                ListViewItem SelectedItem = DevicesListView.Items[DevicesListView.SelectedIndices[0]];

                // The "tag" property holds the device that was detected
                Device SelectedDevice = (Device)SelectedItem.Tag;

                // Open a Stream to the GPS device and bind it to the interpreter
                Interpreter.BaseStream = SelectedDevice.GetHardwareStream();
            }
            else
            {
                // "Use a GPS emulator"

                /* GPS.NET includes a framework for designing your own
                 * GPS device emulator.  GPS.NET also includes some pre-made
                 * emulators which are very useful when building GPS 
                 * applications when no signal is present, such as in an
                 * office building.
                 */

                // Create a new DeLorme Earthmate USB emulator
                Interpreter.BaseStream = new DeLormeEarthmateUsbEmulator();

            }

            try
            {
                // Finally, start the interpreter using the newly-assigned Stream
                Interpreter.Start();

                // Enable the disconnect button
                StopButton.Enabled = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Unable to Connect", MessageBoxButtons.OK, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button1);
            }
        }

        private void UseDeviceRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            // Enable the device list view
            DevicesListView.Enabled = true;
            DevicesListView.Focus();
        }

        private void UseAnyAvailableDeviceRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            // Enable the device list view
            StartButton.Enabled = true;
            DevicesListView.Enabled = false;
            StartButton.Focus();
        }

        private void UseEmulatorRadioButton_CheckedChanged(object sender, EventArgs e)
        {
            // Enable the device list view
            StartButton.Enabled = true;
            DevicesListView.Enabled = false;
            StartButton.Focus();
        }

        private void StopButton_Click(object sender, EventArgs e)
        {
            // Stop any GPS communications
            Interpreter.Stop();

            // And disable the disconnect button
            StopButton.Enabled = false;
            StatusBar1.Text = "Disconnected from GPS device.";
        }

        private void DevicesListView_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Enable the Connect button
            UseDeviceRadioButton.Checked = true;
            StartButton.Enabled = true;
            StartButton.Focus();
        }

        void Interpreter_Disconnected(object sender, EventArgs e)
        {
            StatusBar1.Text = "Disconnected.";
            Cursor.Current = Cursors.Default;
        }

        void Interpreter_Disconnecting(object sender, EventArgs e)
        {
            StatusBar1.Text = "Disconnecting from GPS device...";
            Cursor.Current = Cursors.WaitCursor;
        }

        void Interpreter_Connected(object sender, EventArgs e)
        {
            StatusBar1.Text = "Connected to GPS device!  Reading data...";
            Cursor.Current = Cursors.Default;
        }

        void Interpreter_Connecting(object sender, CancelableEventArgs e)
        {
            Cursor.Current = Cursors.WaitCursor;
            StatusBar1.Text = "Connecting to GPS device...";
        }

        void Interpreter_TimeoutOccurred(object sender, ExceptionEventArgs e)
        {
            StatusBar1.Text = "The device stopped.  Restarting...";
        }

        #endregion

        #region Working with Raw Data

        private void Interpreter_SentenceReceived(object sender, GeoFramework.Gps.Nmea.NmeaSentenceEventArgs e)
        {
            // Keep the list box from getting too big
            if (RawDataListBox.Items.Count > 50)
                RawDataListBox.Items.RemoveAt(0);

            // Add the sentence to the raw data list box
            RawDataListBox.Items.Add(e.Sentence.ToString());
            RawDataListBox.SelectedIndex = RawDataListBox.Items.Count - 1;
        }

        #endregion

        #region Working with the Current Position

        private void Interpreter_PositionChanged(object sender, GeoFramework.PositionEventArgs e)
        {
            // Update the latitude and longitude
            Latitude.Text = e.Position.Latitude.ToString();
            Longitude.Text = e.Position.Longitude.ToString();
            StatusBar1.Text = "Position has changed.";
        }

        #endregion

        #region Working with the Current Altitude

        private void Interpreter_AltitudeChanged(object sender, GeoFramework.DistanceEventArgs e)
        {
            // Update the current altitude
            Altitude.Text = e.Distance.ToLocalUnitType().ToString();
            StatusBar1.Text = "Altitude has changed.";
        }

        #endregion

        #region Working with the Current Speed

        private void Interpreter_SpeedChanged(object sender, GeoFramework.SpeedEventArgs e)
        {
            // Update the current speed
            Speed.Text = e.Speed.ToLocalUnitType().ToString();
            StatusBar1.Text = "Speed has changed.";
        }

        #endregion

        #region Working with Fix Obtained/Lost Events

        private void Interpreter_FixObtained(object sender, EventArgs e)
        {
            // Change the fix status            
            FixStatus.Text = "Fix Obtained!";
            StatusBar1.Text = "A fix has been obtained!";
        }

        private void Interpreter_FixLost(object sender, EventArgs e)
        {
            // Change the fix status
            FixStatus.Text = "Fix Lost";
            StatusBar1.Text = "The current fix was lost.";
        }

        #endregion

        #region Working with the Current Bearing

        private void Interpreter_BearingChanged(object sender, GeoFramework.AzimuthEventArgs e)
        {
            // Output the current bearing as degrees (i.e. 012)
            Bearing.Text = e.Azimuth.ToString();
            // Output the current bearing as a compass direction (i.e. Southwest)
            Direction.Text = e.Azimuth.ToString("c");
            StatusBar1.Text = "Bearing has changed.";
        }

        #endregion

        #region Working with the Current Fix Mode

        private void Interpreter_FixModeChanged(object sender, GeoFramework.Gps.FixModeEventArgs e)
        {
            // Update the current fix mode
            FixMode.Text = e.FixMode.ToString();
            StatusBar1.Text = "Fix mode has changed.";
        }

        #endregion

        #region Working with the Current Fix Quality

        private void Interpreter_FixQualityChanged(object sender, GeoFramework.Gps.FixQualityEventArgs e)
        {
            // Update the current fix quality
            FixQuality.Text = e.FixQuality.ToString();
            StatusBar1.Text = "Fix quality has changed.";
        }

        #endregion

        #region Working with Satellite-Derived Time

        private void Interpreter_UtcDateTimeChanged(object sender, GeoFramework.DateTimeEventArgs e)
        {
            // Update the current satellite-derived time
            SatelliteTime.Text = e.DateTime.ToLocalTime().ToString();
        }

        private void LocalTimeTimer_Tick(object sender, EventArgs e)
        {
            LocalTime.Text = DateTime.Now.ToString();
        }
        #endregion

        #region Enforcing GPS Precision

        private void Interpreter_HorizontalDilutionOfPrecisionChanged(object sender, GeoFramework.Gps.DilutionOfPrecisionEventArgs e)
        {
            // Update the horizontal DOP on the track bar
            CurrentHdopTrackBar.Value = (int)e.DilutionOfPrecision.Value;
            UpdateDopStatus();
            StatusBar1.Text = "Lat/long precision has changed.";
        }

        private void Interpreter_VerticalDilutionOfPrecisionChanged(object sender, GeoFramework.Gps.DilutionOfPrecisionEventArgs e)
        {
            // Update the vertical DOP on the track bar
            CurrentVdopTrackBar.Value = (int)e.DilutionOfPrecision.Value;
            UpdateDopStatus();
            StatusBar1.Text = "Altitude precision has changed.";
        }

        private void UpdateDopStatus()
        {
            // Are readings now precise enough to use?
            if (CurrentHdopTrackBar.Value <= MaximumHdopTrackBar.Value)
                HdopStatusLabel.Text = "Position readings are being accepted.";
            else
                HdopStatusLabel.Text = "Position readings are being ignored.";
            // Are readings now precise enough to use?
            if (CurrentVdopTrackBar.Value <= MaximumVdopTrackBar.Value)
                VdopStatusLabel.Text = "Altitude readings are being accepted.";
            else
                VdopStatusLabel.Text = "Altitude readings are being ignored.";
        }

        private void MaximumHdopTrackBar_ValueChanged(object sender, EventArgs e)
        {
            // Set the new maximum latitude/longitude precision error
            Interpreter.MaximumHorizontalDilutionOfPrecision =
                new DilutionOfPrecision(MaximumHdopTrackBar.Value);
            // And update the UI
            UpdateDopStatus();
        }

        private void MaximumVdopTrackBar_ValueChanged(object sender, EventArgs e)
        {
            // Set the new maximum altitude precision error
            Interpreter.MaximumVerticalDilutionOfPrecision =
                new DilutionOfPrecision(MaximumVdopTrackBar.Value);
            // And update the UI
            UpdateDopStatus();
        }

        #endregion

        #region Handling PocketPC Power Suspend & Resume Events

        void Interpreter_DeviceResumed(object sender, EventArgs e)
        {
            Cursor.Current = Cursors.Default;
            StatusBar1.Text = "Device is no longer suspended!  Restarting...";
        }

        void Interpreter_DeviceSuspended(object sender, EventArgs e)
        {
            StatusBar1.Text = "Device was powered off!";
            Cursor.Current = Cursors.WaitCursor;
        }

        #endregion
     
    }
}