Imports System
Imports System.Drawing
Imports GeoFramework
Imports GeoFramework.Controls

' This small class demonstrates how easy it is to create a double-buffered, multithreaded
' custom control using GPS.NET and the GeoFramework.  See the "Custom Controls" example
' form to see how the clock looks.

Public Class Clock
    Inherits PolarControl

    ' Controls the pen used for small tick marks around the edge
    Private pMinorTickPen As New Pen(Color.Black, 1.0F)
    ' Controls the pen used for larger tick marks around the edge
    Private pMajorTickPen As New Pen(Color.Black, 2.0F)
    ' Controls the brush used for labeling numbers 
    Private pClockLabelBrush As New SolidBrush(Color.Black)
    ' Controls the pen used for the seconds hand
    Private pSecondsPen As New Pen(Color.Red, 1.0F)
    ' Controls the time drawn by the control
    Private pClockTime As DateTime = DateTime.Now

    ' The constructor provides a name for the control.  This name is used as the
    ' name of the thread if multithreading is enabled. 
    Public Sub New()
        MyBase.New("My Custom Clock Control!")

        ' Enable the glass effect
        Effect = PolarControlEffect.Glass
        ' And optimize rendering for the highest quality possible
        OptimizationMode = ControlOptimizationMode.BestQuality
        ' Set the behavior of angles.  "Origin" means that 0 points up,
        ' and "Orientation" states that larger angles go clockwise around the control.
        Origin = Azimuth.North
        Orientation = PolarCoordinateOrientation.Clockwise
    End Sub

    ' This method is called whenever the control must be rendered.  All rendering takes
    ' place off-screen, which prevents flickering.  The "PolarControl" base class automatically
    ' handles tasks such as resizing and smoothing.  All you have to be concerned with is
    ' calculating the polar coordinates to draw.
    Protected Overrides Sub OnPaintOffScreen(ByVal e As CancelablePaintEventArgs)
        ' The GeoFramework comes with a special class named "PolarGraphics," which works
        ' almost identically to the "Graphics" class, except that coordinates are given
        ' in polar form instead of (X, Y).
        ' 
        ' Polar coordinates are given as an angle and a radius.  The angle is between 0
        ' and 360 clockwise from the top of the control.  The radius is zero (0) at the
        ' center of the control, and 100 at the outer edge.  So, a polar coordinate of 90
        ' with a radius of 50 would mark a point straight to the right, half way to the
        ' edge.  Mastering polar coordinates gives you the opportunity to create your
        ' own controls.

        ' Make a PolarGraphics class for easier drawing Imports polar coordinates
        Dim f As PolarGraphics = CreatePolarGraphics(e.Graphics)

        ' Draw sixty small tick marks around the control to show minutes
        Dim value As Double
        For value = 0 To 59
            ' There are 120 tick marks in 360
            Dim TickAngle As New Angle(value * (360 / 60))

            ' Get the coordinate of the line's start and end
            Dim StartPoint As New PolarCoordinate(96, TickAngle, Azimuth.North, PolarCoordinateOrientation.Clockwise)
            Dim EndPoint As New PolarCoordinate(100, TickAngle, Azimuth.North, PolarCoordinateOrientation.Clockwise)

            ' And draw a line
            f.DrawLine(pMinorTickPen, StartPoint, EndPoint)
        Next value

        ' Draw twelve tick marks, one for each hour
        For value = 1 To 12
            ' Convert the value to an angle
            Dim TickAngle As New Angle(value * (360 / 12))

            ' Get the coordinate of the line's start
            Dim StartPoint As New PolarCoordinate(93, TickAngle, Azimuth.North, PolarCoordinateOrientation.Clockwise)
            Dim EndPoint As New PolarCoordinate(100, tickangle, Azimuth.North, PolarCoordinateOrientation.Clockwise)

            ' And draw the tick mark
            f.DrawLine(pMajorTickPen, startpoint, endpoint)

            ' Label the clock position around the circle
            Dim s As String = Convert.ToString(value)
            f.DrawCenteredString(s, Me.Font, pClockLabelBrush, _
                New PolarCoordinate(85, tickAngle, Azimuth.North, PolarCoordinateOrientation.Clockwise))
        Next value

        ' In order to make the control more appealing, we'll make the hours, minutes, and
        ' seconds hand as accurate as possible.  For example, if the time is 5:30, the hours
        ' hand should be halfway between 5 and 6, or 5.5.  To get these values, we'll
        ' calculate "fractional" hours, minutes and seconds Imports the TimeSpan structure.  
        ' 
        ' when you look at the control, you'll see that the minutes hand moves slowly but
        ' smoothly as time progresses.  With smoothing enabled, you almost can't tell it's 
        ' actually moving.

        ' Calculate the time elapsed since midnight
        Dim Midnight As New DateTime(pClockTime.Year, pClockTime.Month, pClockTime.Day)
        Dim TimeSinceMidnight As TimeSpan = pClockTime.Subtract(Midnight)

        ' There are twelve hours in 360 of a circle
        Dim HourAngle As New Angle(TimeSinceMidnight.TotalHours * (360 / 12))

        ' Draw the hour "needle"
		f.DrawLine(pMajorTickPen, PolarCoordinate.Center, _
			new PolarCoordinate(30, HourAngle, Azimuth.North, PolarCoordinateOrientation.Clockwise))

        ' There are sixty minutes in 360 of a circle
        Dim MinuteAngle As New Angle(TimeSinceMidnight.TotalMinutes * (360 / 60))

        ' Draw the Minute "needle"
        f.DrawLine(pMajorTickPen, PolarCoordinate.Center, _
            New PolarCoordinate(80, MinuteAngle, Azimuth.North, PolarCoordinateOrientation.Clockwise))

        ' There are sixty seconds in 360 of a circle
        Dim SecondAngle As New Angle(TimeSinceMidnight.TotalSeconds * (360 / 60))

        ' Draw the Seconds "needle"
        f.DrawLine(pSecondsPen, PolarCoordinate.Center, _
            New PolarCoordinate(90, SecondAngle, Azimuth.North, PolarCoordinateOrientation.Clockwise))

    End Sub

    Protected Overloads Overrides Sub Dispose(ByVal disposing As Boolean)
        ' Free unmanaged GDI resources (very important!)
        pMinorTickPen.Dispose()
        pMajorTickPen.Dispose()
        pClockLabelBrush.Dispose()
        pSecondsPen.Dispose()
        ' And continue disposing
        MyBase.Dispose(disposing)
    End Sub

    ' Returns/set the time drawn by the control
    Public Property ClockTime() As DateTime
        Get
            Return pClockTime
        End Get
        Set(ByVal Value As DateTime)
            pClockTime = Value

            ' This method makes the control redraw.  Notice that we are not using
            ' the Invalidate() or Refresh() methods, because rendering takes place
            ' off-screen to eliminate flickering.  Use the InvokePaintOffScreen()
            ' method to redraw the control.

            InvokePaintOffScreen()
        End Set
    End Property
End Class
