using System;
using System.IO;
using System.Threading;
using System.Net.Sockets;
using GeoFramework.IO;
using System.Security.Permissions;

namespace GeoFramework.IO.Bluetooth
{
	/// <summary>
	/// Summary description for BluetoothStream.
	/// </summary>
    public sealed class BluetoothStream : Stream
	{
		private NetworkStream pBaseStream;
		private BluetoothAddress pAddress;
		private Guid pServiceGuid;
		private BluetoothSocket pBluetoothSocket = new BluetoothSocket();
		internal static object BluetoothSyncRoot = new object();
        private object SyncLock = new object(); 
        private bool pIsClosed = true;

		~BluetoothStream()
		{
			Close();
		}

		public BluetoothStream(BluetoothAddress address, Guid serviceGuid)
			: this(address, serviceGuid, FileAccess.ReadWrite)
		{
		}

		public BluetoothStream(BluetoothAddress address, Guid serviceGuid, FileAccess access)
		{
			pAddress = address;
			pServiceGuid = serviceGuid;

			// Allow only one connection to this address at a time
			try
			{
				lock(BluetoothSyncRoot)
				{
					pBluetoothSocket.Connect(new BluetoothEndPoint(pAddress, pServiceGuid));
				}
				// Return a stream which owns the socket
                pIsClosed = false;
                pBaseStream = new NetworkStream(pBluetoothSocket, access, true);
			}
			catch (FileNotFoundException ex)
			{
				// Record the failure
				throw new IOException("A Bluetooth stream cannot be opened at the specified address.  The device is not responding.", ex);
			}
			catch(SocketException ex)
			{
				//				try
				//				{
				//					lock(BluetoothSyncRoot)
				//					{
				//						pBluetoothSocket.Connect(new BluetoothEndPoint(pAddress, pServiceGuid));						
				//					}
				//					// Return a stream which owns the socket
				//					pBaseStream = new NetworkStream(pBluetoothSocket, access, true);		
				//				}
				//				catch (FileNotFoundException ex2)
				//				{
				//					throw new IOException("A Bluetooth stream cannot be opened at the specified address.  The device is not responding.", ex2);
				//				}
				//				catch(SocketException ex2)
				//				{
//				pIsResponding = false;
				throw new IOException("A Bluetooth stream cannot be opened at the specified address.  The device may be turned off or in use by another Smartphone.", ex);
				//				}	
			}	
		}

		#region Overrides

		/// <summary>Returns the underlying stream for a GPS device.</summary>
		public Stream BaseStream
		{
			get
			{
				return pBaseStream;
			}
		}

		public BluetoothStream Clone()
		{
			return new BluetoothStream(pAddress, pServiceGuid);
		}


		/// <summary>
		/// Returns the name of the stream along with the name of the underlying
		/// stream.
		/// </summary>
		public override string ToString()
		{
			return "Bluetooth Stream on " + pBluetoothSocket.ToString();
		}


		/// <summary>Begins an asynchronous read on the underlying stream.</summary>
		public override IAsyncResult BeginRead(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			return pBaseStream.BeginRead (buffer, offset, count, callback, state);
		}

		/// <summary>Begins an asynchronous write on the underlying stream.</summary>
		public override IAsyncResult BeginWrite(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			return pBaseStream.BeginWrite (buffer, offset, count, callback, state);
		}

		/// <summary>Indicates if read operations are allowed.</summary>
		public override bool CanRead
		{
			get
			{
				lock(SyncLock)
				{
					return pBaseStream.CanRead;
				}
			}
		}

		/// <summary>Indicates if seek operations are allowed.</summary>
		public override bool CanSeek
		{
			get
			{
				return false;
			}
		}

		/// <summary>Indicates if write operations are allowed.</summary>
		public override bool CanWrite
		{
			get
			{
				lock(SyncLock)
				{
					return pBaseStream.CanWrite;
				}
			}
		}

		/// <summary>Closes this stream and the underlying stream.</summary>
		public override void Close()
		{
			if(pIsClosed)
				return;

			pIsClosed = true;
			
			GC.SuppressFinalize(this);

			////lock(SyncLock)
			{
				// Close and terminate the stream
				try
				{
					lock(BluetoothSyncRoot)
					{
						pBluetoothSocket.Shutdown(SocketShutdown.Both);
						pBaseStream.Close();
					}

					// Tell the device associated with this stream to disconnect soon					
				}
				catch
				{
				}
				finally
				{
//					pIsOpen = false;
					pIsClosed = true;
				}
			}
		}

		public BluetoothAddress Address
		{
			get
			{
				return pAddress;
			}
		}

		public Guid ServiceGuid
		{
			get
			{
				return pServiceGuid;
			}
		}

		public bool IsClosed
		{
			get
			{
				return pIsClosed;
			}
		}


#if !Framework20
		/// <summary>Obsolete. Returns a wait handle used for synchronizing asynchronous operations.</summary>
		protected override WaitHandle CreateWaitHandle()
		{
			return new ManualResetEvent(false);
		}
#endif

		/// <summary>Completes an asynchronous read operation.</summary>
		public override int EndRead(IAsyncResult asyncResult)
		{
			return pBaseStream.EndRead (asyncResult);
		}

		/// <summary>Completes an asynchronous write operation.</summary>
		public override void EndWrite(IAsyncResult asyncResult)
		{
			pBaseStream.EndWrite (asyncResult);
		}

		/// <summary>Empties any pending data sending or receiving buffers.</summary>
		public override void Flush()
		{
			lock(SyncLock)
			{
//				if(!pIsOpen)
//					Open();

				pBaseStream.Flush();
			}
		}

		/// <summary>Returns the number of available bytes in the underlying stream.</summary>
		public override long Length
		{
			get
			{
				lock(SyncLock)
				{
//					if(!pIsOpen)
//						Open();
                    return pBaseStream.DataAvailable ? 1 : 0;
					//return pBaseStream.Length;
				}
			}
		}

		/// <summary>Returns the underlying stream's current position.</summary>
		public override long Position
		{
			get
			{
				// Always return zero because seek operations are not supported
				return 0;
			}
			set
			{
				lock(SyncLock)
				{
//					if(!pIsOpen)
//						Open();

					pBaseStream.Position = value;
				}
			}
		}

		/// <summary>Reads a number of bytes from the underlying stream.</summary>
		public override int Read(byte[] buffer, int offset, int count)
		{
			lock(SyncLock)
			{
//				if(!pIsOpen)
//					Open();

				return pBaseStream.Read(buffer, offset, count);
			}
		}

		/// <summary>Reads a single byte from the underlying stream.</summary>
		public override int ReadByte()
		{
			lock(SyncLock)
			{
				return pBaseStream.ReadByte();
			}
		}

		/// <summary>Changes the position of the underlying stream.</summary>
		public override long Seek(long offset, SeekOrigin origin)
		{
			lock(SyncLock)
			{
//				if(!pIsOpen)
//					Open();

				return pBaseStream.Seek(offset, origin);
			}
		}

		/// <summary>Changes the length of the underlying stream.</summary>
		public override void SetLength(long value)
		{
			lock(SyncLock)
			{
//				if(!pIsOpen)
//					Open();

				pBaseStream.SetLength(value);
			}
		}

		/// <summary>Writes a number of bytes to the underlying stream.</summary>
		public override void Write(byte[] buffer, int offset, int count)
		{
			lock(SyncLock)
			{
//				if(!pIsOpen)
//					Open();

				pBaseStream.Write(buffer, offset, count);
			}
		}

		/// <summary>Writes a single byte to the underlying stream.</summary>
		public override void WriteByte(byte value)
		{
			lock(SyncLock)
			{
//				if(!pIsOpen)
//					Open();

				pBaseStream.WriteByte (value);
			}
		}

		#endregion

	}
}
