using System;
using System.IO;
using System.Threading;
using GeoFramework;
using GeoFramework.Gps;
#if !PocketPC || DesignTime
using System.ComponentModel;
#endif

namespace GeoFramework.IO
{
    /// <summary>Represents a base class for building GPS device objects.</summary>
    /// <remarks>
    /// 	<para>This class is a base class for designing specific GPS devices. Devices
    ///     provide rather limited functionality and are designed primarily to organize
    ///     information and configuration settings. Devices are also able to carry out GPS data
    ///     protocol detection requests.</para>
    /// 	<para>Typically, this class will not be used for custom device objects since major
    ///     ones such as <strong>SerialDevice</strong> and <strong>BluetoothDevice</strong>
    ///     have already been created. Creating a new class is beneficial if it encapsulates a
    ///     unique form of device I/O, such as USB, networks, or other technology.</para>
    /// </remarks>
    public abstract class Device :
#if !PocketPC && Framework20
        System.Runtime.ConstrainedExecution.CriticalFinalizerObject ,
#endif
        IFormattable, IDisposable
    {
        private bool pIsPreviouslyDetected;
        private bool pIsLastUsedGpsDevice;
        private string pProtocol;
        private bool pIsGpsDevice;
#if PocketPC && !Framework20 && !Smartphone
        private bool pIsDetectionThreadAlive;
#endif
        private bool pIsDetectionInProgress;
        private ManualResetEvent DetectionWaitHandle = new ManualResetEvent(false);
        private Thread DetectionThread;
        private Type pInterpreterType;
        private Type pParsingStreamType;

#if !PocketPC
        private const int MaximumGracefulShutdownTime = 500;
#else
        private const int MaximumGracefulShutdownTime = 2000;
#endif

        /// <summary>
        /// Returns the open stream used to identify the device as a GPS device.
        /// </summary>
        /// <remarks>To speed up the process of connecting to a GPS device, GPS.NET will re-use an
        /// open data stream whenever possible.  When Automatic GPS Device Detection locates a device,
        /// it sets this property to the open stream used to identify the device.  Then, requests for
        /// GPS data will use this stream.  This property is used by the GetHardwareStream method.  If
        /// an immediate stream is available, GetHardwareStream returns it, otherwise a new connection
        /// is made.</remarks>
        private Stream _ImmediateNeedStream;
        private bool pIsImmediateNeedStreamAvailable;

#if PocketPC
        public void SetImmediateNeedStream(Stream input)
        {
            _ImmediateNeedStream = input;
        }
#endif
       
        /// <summary>
        /// Returns a stream from a recently-detected GPS device.
        /// </summary>
        /// <remarks>When Automatic Device Detection takes place, GPS.NET will check to see if
        /// a call to the "Start" method of an Interpreter object is currently waiting for data.
        /// If so, the stream of a detected GPS device is left open so that it can be immediately used to 
        /// process data.  If data is not needed immediately, the streams of detected GPS devices
        /// are closed.</remarks>
        public Stream ImmediateNeedStream
        {
            get
            {
                return _ImmediateNeedStream;
            }
        }

        /// <summary>
        /// Controls whether or not an open stream is available for immediate use.
        /// </summary>
        public bool IsImmediateNeedStreamAvailable
        {
            get
            {
                return pIsImmediateNeedStreamAvailable;
            }
            set
            {
                pIsImmediateNeedStreamAvailable = value;
            }
        }

        /// <summary>Creates a new instance.</summary>
        protected Device()
        {
        }

        ~Device()
        {
            Dispose();
        }

        public void Dispose()
        {
            // No need to finalize anymore
            GC.SuppressFinalize(this);

            try
            {
                if (ImmediateNeedStream != null)
                    ImmediateNeedStream.Close();
            }
            catch
            {
            }

#if Framework20 && !PocketPC
            if (!DetectionWaitHandle.SafeWaitHandle.IsClosed)
#else
                if (DetectionWaitHandle.Handle != new IntPtr(-1))
#endif
            {
                try
                {
                    DetectionWaitHandle.Close();
                }
                catch
                {
                }
            }
        }

        void IDisposable.Dispose()
        {
            Dispose();
        }

        /// <summary>Indicates the GPS data protocol last detected on the device.</summary>
        /// <value>
        /// If the device has been recognized as a GPS device, a <strong>String</strong>
        /// indicating the GPS data protocol detected on the device, otherwise null.
        /// </value>
        /// <remarks>
        /// This property will return a null value until the device has been successfully
        /// recognized by GPS interpreter. Once recognized, this property is set to the
        /// <strong>Protocol</strong> property of the <strong>Interpreter</strong> object which
        /// recognized the device. For example, if this device was recognized by an
        /// <strong>NmeaInterpreter</strong> object, this property will return the
        /// <strong>NmeaInterpreter</strong>'s own <strong>Protocol</strong> property value:
        /// "NMEA-0183."
        /// </remarks>
        public virtual string Protocol
        {
            get
            {
                return pProtocol;
            }
        }

        internal void SetProtocol(string protocol)
        {
            pProtocol = protocol;
        }

        internal void SetIsGpsDevice(bool isGpsDevice)
        {
            pIsGpsDevice = isGpsDevice;
        }

        /// <summary>
        /// Indicates if the device has previously been recognized as a GPS device.
        /// </summary>
        public bool IsPreviouslyDetected
        {
            get
            {
                return pIsPreviouslyDetected;
            }
        }

        internal void SetIsPreviouslyDetected(bool isPreviouslyDetected)
        {
            pIsPreviouslyDetected = isPreviouslyDetected;
        }

        /// <summary>
        /// Indicates if the device was the most recent device used for GPS communications.
        /// </summary>
        public bool IsLastUsedGpsDevice
        {
            get
            {
                return pIsLastUsedGpsDevice;
            }
        }

        internal void SetIsLastUsedGpsDevice(bool isLastUsedGpsDevice)
        {
            pIsLastUsedGpsDevice = isLastUsedGpsDevice;
        }

        /// <summary>
        /// Returns a new <strong>Stream</strong> object for communicating with the
        /// device.
        /// </summary>
        /// <returns>A <strong>Stream</strong> object able to communicate with the device.</returns>
        /// <remarks>
        /// This method, when overridden, returns an object descended from
        /// <strong>Stream</strong> which can communicate with the device. For example, the
        /// <strong>SerialDevice</strong> returns a <strong>SerialStream</strong> object when this
        /// method is called.
        /// </remarks>
        /// <example>
        /// 	<code lang="VB" title="[New Example]">
        /// Public Function GetSerialGpsStream() As Stream
        /// Dim Device As SerialDevice
        ///     ' Examine the collection of serial devices on the current computer
        ///     For Each Device In GeoFramework.IO.Devices.SerialDevices
        ///         ' Is the device a GPS device?
        ///         If Device.IsGpsDevice Then
        ///             ' Yes.  Get the stream for this GPS device 
        ///             Return Device.GetHardwareStream()
        ///         End If
        ///     Next Device
        /// End Function
        ///     </code>
        /// 	<code lang="CS" title="[New Example]">
        /// public Stream GetSerialGpsStream()
        /// {
        ///     foreach(SerialDevice device in GeoFramework.IO.Devices.SerialDevices)
        ///     {
        ///         // Is the device a GPS device?
        ///         if(Device.IsGpsDevice)
        ///             // Yes.  Get the stream for this GPS device 
        ///             return Device.GetHardwareStream();
        ///     }
        /// }
        ///     </code>
        /// </example>
        public abstract Stream GetHardwareStream();

        /// <summary>
        /// Returns a <strong>Stream</strong> capable of interpreting data from the
        /// device.
        /// </summary>
        /// <returns>
        /// If data from the device has been successfully identified, a
        /// <strong>Stream</strong> object capable of interpreting raw data from the device,
        /// otherwise null.
        /// </returns>
        /// <remarks>
        /// 	<para>When automatic GPS device detection occurs, each GPS data interpreter is
        ///     tested against data from the device. If an interpreter recognizes the data, it
        ///     becomes associated with the device and this method can return a valid parsing
        ///     stream.</para>
        /// 	<para>As an example: let's say that a GPS device is connected to COM3. Automatic
        ///     device detection will test data from COM3 with an <strong>NmeaInterpreter</strong>
        ///     object. If the data is recognized, the interpreter's base
        ///     <strong>NmeaStream</strong> is recorded as the parsing stream, and any call to
        ///     <strong>GetParsingStream</strong> will return an <strong>NmeaStream</strong>.
        ///     Likewise, any call to <strong>GetInterpreter</strong> would return an
        ///     <strong>NmeaInterpreter</strong>.</para>
        /// </remarks>
        /// <example>
        ///     This example scans through the list of available serial devices on the local
        ///     machine, then examines each device to see if it is a GPS device. If one is found,
        ///     the device's associated parsing stream is created. 
        ///     <code lang="VB" title="[New Example]">
        /// Public Function GetSerialParsingStream() As Stream
        /// Dim Device As SerialDevice
        ///     ' Examine the collection of serial devices on the current computer
        ///     For Each Device In GeoFramework.IO.Devices.SerialDevices
        ///         ' Is the device a GPS device?
        ///         If Device.IsGpsDevice Then
        ///             ' Yes.  Get a new parsing stream
        ///             Return Device.GetParsingStream()
        ///         End If
        ///     Next Device
        /// End Function
        ///     </code>
        /// 	<code lang="CS" title="[New Example]">
        /// public Stream GetSerialParsingStream()
        /// {
        ///     foreach(SerialDevice device in GeoFramework.IO.Devices.SerialDevices)
        ///     {
        ///         // Is the device a GPS device?
        ///         if(Device.IsGpsDevice)
        ///             // Yes.  Get a new interpreter
        ///             return Device.GetParsingStream();
        ///     }
        /// }
        ///     </code>
        /// </example>
        public Stream GetParsingStream()
        {
#if PocketPC
			return (Stream)Activator.CreateInstance(pParsingStreamType);
#else
            return (Stream)Activator.CreateInstance(pParsingStreamType, new object[] { GetHardwareStream() });
#endif
        }

        /// <remarks>
        /// 	<para>When automatic GPS device detection occurs, each GPS data interpreter is
        ///     tested against data from the device. If an interpreter recognizes the data, it
        ///     becomes associated with the device and this method can return a valid parsing
        ///     stream.</para>
        /// 	<para>As an example: let's say that a GPS device is connected to COM3. Automatic
        ///     device detection will test data from COM3 with an <strong>NmeaInterpreter</strong>
        ///     object. If the data is recognized, the interpreter's base
        ///     <strong>NmeaStream</strong> is recorded as the parsing stream, and any call to
        ///     <strong>GetInterpreter</strong> will return an <strong>NmeaInterpreter</strong>.
        ///     Likewise, any call to <strong>GetParsingStream</strong> would return a new
        ///     <strong>NmeaStream</strong> object.</para>
        /// </remarks>
        /// <returns>
        /// An <strong>Interpreter</strong> object able to process data from the
        /// device.
        /// </returns>
        /// <example>
        ///     This example scans through the list of available serial devices on the local
        ///     machine, then examines each device to see if it is a GPS device. If one is found,
        ///     the device's associated interpreter is created.
        ///     <code lang="VB" title="[New Example]">
        /// Public Function GetSerialInterpreter() As Interpreter
        /// Dim Device As SerialDevice
        ///     ' Examine the collection of serial devices on the current computer
        ///     For Each Device In GeoFramework.IO.Devices.SerialDevices
        ///         ' Is the device a GPS device?
        ///         If Device.IsGpsDevice Then
        ///             ' Yes.  Get a new interpreter
        ///             Return Device.GetInterpreter()
        ///         End If
        ///     Next Device
        /// End Function
        ///     </code>
        /// 	<code lang="CS" title="[New Example]">
        /// public Interpreter GetSerialInterpreter()
        /// {
        ///     foreach(SerialDevice device in GeoFramework.IO.Devices.SerialDevices)
        ///     {
        ///         // Is the device a GPS device?
        ///         if(Device.IsGpsDevice)
        ///             // Yes.  Get a new interpreter
        ///             return Device.GetInterpreter();
        ///     }
        /// }
        ///     </code>
        /// </example>
        public Interpreter GetInterpreter()
        {
#if PocketPC
			Interpreter Result = (Interpreter)Activator.CreateInstance(pInterpreterType);
			Result.BaseStream = GetHardwareStream();
			return Result;
#else
            return (Interpreter)Activator.CreateInstance(pInterpreterType, new object[] { GetHardwareStream() });
#endif
        }

        //		public virtual string Text
        //		{
        //			get
        //			{
        //				return "Generic Device";
        //			}
        //		}

        /// <summary>Attempts to recognize data arriving from the device as GPS data.</summary>
        /// <remarks>
        /// 	<para>This powerful method begins the process of identifying the device as a GPS
        ///     device using any available GPS data interpreter. Device detection takes place on a
        ///     background. If an interpreter successfully recognizes the device, the
        ///     <strong>IsGpsDevice</strong> property becomes <strong>True</strong>, the
        ///     <strong>Protocol</strong> is set to the recognized protocol name, and calls to
        ///     <strong>GetHardwareStream</strong>, <strong>GetParsingStream</strong> or
        ///     <strong>GetInterpreter</strong> will be successful.</para>
        /// 	<para>This method is called automatically if device detection is started via the
        ///     <strong>Detect</strong> method of the <strong>Devices</strong> class. If detection
        ///     is already in progress, this method has no effect. If this method is called, and it
        ///     has already been recognized as a GPS device, detection is restarted.</para>
        /// </remarks>
        /// <example>
        ///     This example will start automatic protocol detection on the serial device on COM3.
        ///     <code lang="VB" title="[New Example]">
        /// ' Start protocol detection on COM3
        /// SerialDevice.Com3.DetectProtocol()
        /// ' Wait for it to complete
        /// SerialDevice.Com3.WaitForProtocolDetection()
        /// ' Is it a GPS device?
        /// If SerialDevice.Com3.IsGpsDevice Then
        ///     Debug.WriteLine("Found a GPS device on COM3")
        /// Else
        ///     Console.WriteLine("This is not a GPS device.")
        /// End If    
        /// </code>
        /// 	<code lang="CS" title="[New Example]">
        /// // Start protocol detection on COM3
        /// SerialDevice.Com3.DetectProtocol();
        /// // Wait for it to complete
        /// SerialDevice.Com3.WaitForProtocolDetection();
        /// // Is it a GPS device?
        /// if(SerialDevice.Com3.IsGpsDevice)
        ///     Console.WriteLine("Found a GPS device on COM3");
        /// else
        ///     Console.WriteLine("This is not a GPS device.");
        ///     </code>
        /// </example>
        public void DetectProtocol()
        {
            try
            {
                // Is detection already in progress?
                if (pIsDetectionInProgress || pIsGpsDevice)
                    return;                

                // Signal that detection is in progress
                pIsDetectionInProgress = true;

                // Reset the wait state
                DetectionWaitHandle.Reset();

                // The priority of this thread depends on the platform
                switch (Platform.HostPlatformID)
                {
                    case HostPlatformID.Desktop:
                        // Detect the protocol on a separate thread
                        DetectionThread = new Thread(new ThreadStart(DetectThreadProc));

                        // No priority boost is given. ALL threads run at Normal!
                        break;

                    case HostPlatformID.Smartphone:
                        // Detect the protocol on a separate thread
                        DetectionThread = new Thread(new ThreadStart(DetectThreadProc));

                        // No priority boost is given. ALL threads run at Normal!
                        break;

                        //// Detect the protocol on the SAME thread to reduce overhead.
                        //// Detection gets hung up on CF1 when too many threads run.
                        //DetectThreadProc();
                        //return;

                    case HostPlatformID.WindowsCE:
                    case HostPlatformID.PocketPC:

                        // Detect the protocol on a separate thread
                        DetectionThread = new Thread(new ThreadStart(DetectThreadProc));
#if Framework20
                        // A boost is given if the device has been detected before
                        if (IsPreviouslyDetected)
                            DetectionThread.Priority = ThreadPriority.AboveNormal;
                        else
                            DetectionThread.Priority = ThreadPriority.BelowNormal;
#else
                        // A boost is given if the device has been detected before
                        if (IsPreviouslyDetected)
                            DetectionThread.Priority = ThreadPriority.Highest;
                        else
                            DetectionThread.Priority = ThreadPriority.AboveNormal;
#endif
                        break;
                }

#if !PocketPC || Framework20
                DetectionThread.Name = "GPS.NET GPS Protocol Detector on " + ToString() + " (http://www.geoframeworks.com)";
                DetectionThread.IsBackground = true;
#endif

                // Start the thread
                DetectionThread.Start();

                // Let it start up  
                DetectionWaitHandle.WaitOne();
                DetectionWaitHandle.Reset();
            }
            catch
            {
                // signal that detection is in progress
                pIsDetectionInProgress = false;

                throw;
            }
        }

        private void DetectThreadProc()
        {
#if PocketPC && !Framework20 && !Smartphone
            pIsDetectionThreadAlive = true;
#endif

            try
            {
                try
                {
                    // Signal that we're alive
                    DetectionWaitHandle.Set();
                }
                catch (ObjectDisposedException)
                {
                    // We're already shutting down!  Abort.
                    pIsDetectionInProgress = false;
                    return;
                }

                // Now loop through each interpreter and see if it recognizes us

                foreach (Type type in Devices.Interpreters)
                {
                    Interpreter interpreter = null;
                    try
                    {
                        // Create a new GPS data interpreter
                        interpreter = (Interpreter)Activator.CreateInstance(type);

                        // Have we cancelled?
                        if (!pIsDetectionInProgress)
                            // Yes.  Exit immediately
                            return;

                        // Signal that detection has started
                        Devices.OnDeviceDetectionAttempted(this);

                        // Perform detection on the device
                        DeviceRecognitionResult Result = interpreter.IsDeviceRecognizable(this);

                        // Was the device identified as a GPS device?
                        if (Result.IsGpsDevice)
                        {
                            // Yes!  GPS device discovered!

                            // Alter the thread priority depending on the platform
                            switch (Platform.HostPlatformID)
                            {                                                                     
                                case HostPlatformID.PocketPC:
                                case HostPlatformID.WindowsCE:
                                    // Max out the priority!
                                    Thread.CurrentThread.Priority = ThreadPriority.Highest;
                                    break;
                            }

                            // Yes!  Flag this as a GPS device
                            pIsGpsDevice = true;
                            pProtocol = interpreter.Protocol;
                            // Is a device needed immediately?
                            if (Devices.IsDeviceNeededImmediately)
                            {
                                _ImmediateNeedStream = Result.DeviceStream;
                                //XX GC.ReRegisterForFinalize(this);
                                IsImmediateNeedStreamAvailable = true;
                                // Suppress the interpreter's finalizer
                                interpreter.DisposeWithoutClosingStream();
                            }
                            else
                            {
                                // Close the stream
                                Result.DeviceStream.Close();

                                // There's no stream available
                                _ImmediateNeedStream = null;
                                IsImmediateNeedStreamAvailable = false;

                                // Dispose of the interpreter
                                interpreter.Dispose();
                            }

                            // Signal which interpreter made the discovery
                            OnProtocolDetected(interpreter);

                            // And register it for automatic device detection
                            Devices.RegisterGpsDevice(this);
                            // No need to continue detection, so exit
                            return;
                        }
                        else
                        {
                            // Nope, this is not a device.  The thread is now useless.

                            // Demote the thread priority depending on the platform
                            switch (Platform.HostPlatformID)
                            {
                                case HostPlatformID.PocketPC:
                                case HostPlatformID.WindowsCE:
                                    // This thread is no longer important at all :( wah
                                    Thread.CurrentThread.Priority = ThreadPriority.Lowest;
                                    break;
                            }
                        }
                    }
#if !PocketPC || Framework20
                    catch (ThreadAbortException)
                    {
                        // Just abort!
                        break;
                    }
#endif
                    catch 
                    {
                        throw;
                    }        
                }
            }
#if !PocketPC || Framework20
            catch (ThreadAbortException)
            {
                // Just abort!
            }
#endif
            catch (Exception ex)
            {
                // Channel the exception through an event to prevent an unhandled
                // exception.
                Devices.OnDeviceDetectionExceptionOccurred(this, ex);
            }
            finally
            {
                // Signal that we're shutting down now
                pIsDetectionInProgress = false;

                // Try to set the detection handle
#if Framework20 && !PocketPC
                if (!DetectionWaitHandle.SafeWaitHandle.IsInvalid)
#else
                if (DetectionWaitHandle.Handle != new IntPtr(-1))
#endif
                {
                    try
                    {
                        DetectionWaitHandle.Set();
                    }
                    catch (ObjectDisposedException)
                    {
                        // Do nothing
                    }
                }

#if PocketPC && !Framework20 && !Smartphone
                // On Compact Framework 1.0, we need special signals that indicate
                // if the thread is guaranteed to exit.  On CF2 and DF, we can simply
                // abort any non-responding thread.
                pIsDetectionThreadAlive = false;
#endif
            }
        }

        /// <summary>Indicates if protocol detection is currently in progress.</summary>
        public bool IsDetectionInProgress
        {
            get
            {
                return pIsDetectionInProgress;
            }
        }

        /// <summary>Indicates if the protocol has been successfully identified.</summary>
        public bool IsGpsDevice
        {
            get
            {
                return pIsGpsDevice;
            }
        }

        /// <summary>Cancels any current attempt to recognize the protocol.</summary>
        /// <remarks>
        /// 	<para>This method is used to stop any protocol detection attempt in progress. A
        ///     short delay may occur while the detection thread is stopped. If protocol detection
        ///     is not in progress, this method has no effect.</para>
        /// 	<para>This method is called automatically if the <strong>CancelDetection</strong>
        ///     method of the <strong>Devices</strong> class is called.</para>
        /// </remarks>
        /// <example>
        ///     This example will start detection on all serial devices, then immediately cancel
        ///     detection for each device. 
        ///     <code lang="VB" title="[New Example]">
        /// Imports GeoFramework.IO
        ///  
        /// Public Sub DetectAllSerialDevices()
        /// Dim Device As SerialDevice
        ///     ' Start detection for each device
        ///     For Each Device In Devices.SerialDevices
        ///         Device.DetectProtocol()
        ///     Next Device
        ///     ' Now cancel detection for each device
        ///     For Each Device In Devices.SerialDevices
        ///         Device.CancelProtocolDetection()
        ///     Next Device
        /// End Sub
        ///     </code>
        /// 	<code lang="CS" title="[New Example]">
        /// using GeoFramework.IO;
        ///  
        /// public void DetectAllSerialDevices()
        /// {
        ///     // Start detection for each device
        ///     foreach(SerialDevice Device in Devices.SerialDevices)
        ///         Device.DetectProtocol();
        ///     // Now cancel detection for each device
        ///     foreach(SerialDevice Device in Devices.SerialDevices)
        ///         Device.CancelProtocolDetection();
        /// }
        ///     </code>
        /// </example>
        public void CancelProtocolDetection()
        {
            // Are we not detecting?  If so, there's nothing to do
            if (!pIsDetectionInProgress)
                return;

            // Cancel the thread in progress
            pIsDetectionInProgress = false;

#if !PocketPC || DesignTime || (PocketPC && Framework20)
            // Abort the thread
            try
            {
#if !Smartphone
                if (DetectionThread != null && !DetectionThread.Join(MaximumGracefulShutdownTime))
                    DetectionThread.Abort();
#endif

                // On CF2, extra cleanup is required because Thread.Abort will not
                // cause "finally" sections to execute!
#if PocketPC && Framework20
                // And shut down
                pIsDetectionInProgress = false;

                if (DetectionWaitHandle.Handle != new IntPtr(-1))
                {
                    try
                    {
                        DetectionWaitHandle.Set();
                    }
                    catch (ObjectDisposedException)
                    {
                        // Do nothing
                    }
                }
#endif
            }
            catch
            {
            }
#endif
        }


        /// <summary>Waits for all attempts to recognize the protocol to complete.</summary>
        /// <remarks>
        /// This method will cause the current thread to pause until protocol detection has
        /// completed for the device, or until a timeout has been reached. If detection is not
        /// currently in progress, this method has no effect. The length of timeout time is
        /// determined by the <strong>DetectionTimeout</strong> property of the
        /// <strong>Devices</strong> class.
        /// </remarks>
        /// <example>
        ///     This example starts device detection for all serial devices then waits for
        ///     detection to complete for each device. 
        ///     <code lang="VB" title="[New Example]">
        /// Imports GeoFramework.IO
        ///  
        /// Public Sub DetectAllSerialDevices()
        /// Dim Device As SerialDevice
        ///     ' Start detection for each device
        ///     For Each Device In Devices.SerialDevices
        ///         Device.DetectProtocol()
        ///     Next Device
        ///     ' Now wait for detection to complete
        ///     For Each Device In Devices.SerialDevices
        ///         Device.WaitForProtocolDetection()
        ///     Next Device
        /// End Sub
        ///     </code>
        /// 	<code lang="CS" title="[New Example]">
        /// using GeoFramework.IO;
        ///  
        /// public void DetectAllSerialDevices()
        /// {
        ///     // Start detection for each device
        ///     foreach(SerialDevice Device in Devices.SerialDevices)
        ///         Device.DetectProtocol();
        ///     // Now cancel detection for each device
        ///     foreach(SerialDevice Device in Devices.SerialDevices)
        ///         Device.WaitForProtocolDetection();
        /// }
        ///     </code>
        /// </example>
        public void WaitForProtocolDetection()
        {
            while (pIsDetectionInProgress && !pIsGpsDevice)
            {
                // Wait 100ms
                Thread.Sleep(100);

                // Wait for any additional heavy-duty task
                Thread.Sleep(0);
            }
        }

        //		/// <summary>Waits for for the specified time to recognize the GPS protocol.</summary>
        //		/// <example>
        //		///     This example starts device detection for all serial devices then waits up to ten
        //		///     seconds for detection to complete for each device. 
        //		///     <code lang="VB" title="[New Example]">
        //		/// Imports GeoFramework.IO
        //		///  
        //		/// Public Sub DetectAllSerialDevices()
        //		/// Dim Device As SerialDevice
        //		///     ' Start detection for each device
        //		///     For Each Device In Devices.SerialDevices
        //		///         Device.DetectProtocol()
        //		///     Next Device
        //		///     ' Now wait for detection to complete
        //		///     For Each Device In Devices.SerialDevices
        //		///         Device.WaitForProtocolDetection(TimeSpan.FromSeconds(10))
        //		///     Next Device
        //		/// End Sub
        //		///     </code>
        //		/// 	<code lang="CS" title="[New Example]">
        //		/// using GeoFramework.IO;
        //		///  
        //		/// public void DetectAllSerialDevices()
        //		/// {
        //		///     // Start detection for each device
        //		///     foreach(SerialDevice Device in Devices.SerialDevices)
        //		///         Device.DetectProtocol();
        //		///     // Now cancel detection for each device
        //		///     foreach(SerialDevice Device in Devices.SerialDevices)
        //		///         Device.WaitForProtocolDetection(TimeSpan.FromSeconds(10));
        //		/// }
        //		///     </code>
        //		/// </example>
        //		/// <remarks>
        //		/// 	<para>This method will cause the current thread to pause until protocol detection
        //		///     has completed for the device, or until a timeout has been reached. If detection is
        //		///     not currently in progress, this method has no effect. Typically, at least one to
        //		///     two seconds are required for the protocol to be detected. However, a significantly
        //		///     longer period of time (such as twenty seconds) may be required when working with
        //		///     several different GPS data protocols.</para>
        //		/// 	<para>If a GPS device should be recognized but is not being recognized, try
        //		///     increasing the <strong>Timeout</strong> parameter.</para>
        //		/// </remarks>
        //		/// <param name="timeout">
        //		/// A <strong>TimeSpan</strong> indicating the maximum amount of time to wait for
        //		/// detection to complete.
        //		/// </param>
        //		public void WaitForProtocolDetection(TimeSpan timeout)
        //		{
        //#if PocketPC
        //            if (!pIsDetectionInProgress)
        //                return;
        //			// Wait for it to finish
        //			DetectionWaitHandle.WaitOne();
        //#else
        //			// Are we not detecting?  If so, there's nothing to do
        //			if(!pIsDetectionInProgress && (DetectionThread == null || !DetectionThread.IsAlive)) 
        //				return;
        //			// Wait for it to finish
        //			DetectionWaitHandle.WaitOne(timeout, false);
        //#endif
        //		}

        /// <summary>Outputs a name identifying the device.</summary>
        /// <returns>A <strong>String</strong> indicating the name of the device.</returns>
        /// <remarks>
        /// 	<para>This property can return a wide variety of values depending on whether or not
        ///     the device has been identified, and what GPS data protocol the device is using.
        ///     Some protocols, such as the Garmin binary protocol, are able to identify a device
        ///     very specifically (e.g. "Garmin eTrex Vista"), whereas other data protocols, such
        ///     as NMEA-0183 cannot identify the device.</para>
        /// 	<para>Generally speaking, if the specific device name is unknown, but its protocol
        ///     is known, a generic string such as "Generic NMEA Device on COM5 at 4800" is
        ///     returned.</para>
        /// </remarks>
        public override string ToString()
        {
            return "Generic Device";
        }

        /// <summary>Occurs when the protocol has been successfully detected.</summary>
        /// <remarks>
        /// This virtual method is called when GPS device detection successfully identifies
        /// the device. In most cases, the interpreter is used to make calls to
        /// GetParsingStream.
        /// </remarks>
        /// <param name="interpreter">
        /// The <strong>Interpreter</strong> object which successfully recognized the
        /// device.
        /// </param>
        protected virtual void OnProtocolDetected(Interpreter interpreter)
        {
            pIsGpsDevice = true;
            pInterpreterType = interpreter.GetType();
            pParsingStreamType = interpreter.ParsingType;
        }

        /// <summary>Returns the type associated with the parsing stream for this device.</summary>
        /// <value>
        /// A <strong>Type</strong> object indicating the parsing stream for this
        /// device.
        /// </value>
        /// <remarks>
        /// This property is used by the <strong>GetParsingStream</strong> method to create a
        /// new instance of the parsing stream responsible for data arriving from the
        /// device.
        /// </remarks>
        protected Type ParsingStreamType
        {
            get
            {
                return pParsingStreamType;
            }
        }

        /// <remarks>
        /// This property is used by the <strong>GetInterpreter</strong> method to create a
        /// new instance of the parsing stream responsible for data arriving from the
        /// device.
        /// </remarks>
        /// <value>A <strong>Type</strong> object indicating the interpreter for this device.</value>
        /// <summary>Returns the type associated with the interpreter for this device.</summary>
        protected Type InterpreterType
        {
            get
            {
                return pInterpreterType;
            }
        }


        #region IFormattable Members

        public virtual string ToString(string format, IFormatProvider formatProvider)
        {
            return ToString();
        }

        string IFormattable.ToString(string format, IFormatProvider formatProvider)
        {
            return this.ToString(format, formatProvider);
        }

        #endregion

    }

    public sealed class DeviceEventArgs : EventArgs
    {
        private Device pDevice;

        public new static readonly DeviceEventArgs Empty = new DeviceEventArgs(null);

        public DeviceEventArgs(Device device)
        {
            pDevice = device;
        }

        public Device Device
        {
            get
            {
                return pDevice;
            }
        }
    }
}
