using System;
using System.IO;
using System.Net.Sockets;
using System.Threading;
using GeoFramework.IO.Serial;
#if !PocketPC || DesignTime
using System.ComponentModel;
#endif

namespace GeoFramework.IO
{
	/// <summary>Represents a stream of GPS data for a specific data protocol.</summary>
	/// <remarks>
	/// 	<para>This class uses automatic GPS device detection to locate a
	///     <strong>Stream</strong> object associated with an active GPS device. This class
	///     frees developers from having to manually determine the correct serial port or
	///     Bluetooth" settings to connect to a GPS device.</para>
	/// 	<para>This class is typically used by the <strong>Interpreter</strong> class to
	///     automatically locate GPS data if no data stream is specified. If no GPS device is
	///     found, automatic device detection is performed.</para>
	/// </remarks>
	/// <example>
	///     This example demonstrates how to use the <strong>GpsStream</strong> class to
	///     automatically locate a GPS device. NOTE: If you create an Interpreter with no
	///     stream, it will automatically create its own <strong>GpsStream</strong> object. 
	///     <code lang="VB" title="[New Example]">
	/// Imports GeoFramework.IO
	/// Imports GeoFramework.Gps.Nmea
	///  
	/// ' Create a new interpreter using the first
	/// ' available GPS device
	/// Dim MyInterpreter As New NmeaInterpreter(New GpsStream())
	///  
	/// ' This declaration will also use the first
	/// ' available GPS device.
	/// Dim MyInterpreter2 As New NmeaInterpreter()
	///     </code>
	/// 	<code lang="CS" title="[New Example]">
	/// using GeoFramework.IO;
	/// using GeoFramework.Gps.Nmea;
	///  
	/// // Create a new interpreter using the first
	/// // available GPS device
	/// NmeaInterpreter MyInterpreter = new NmeaInterpreter(new GpsStream());
	///  
	/// // This declaration will also use the first
	/// // available GPS device.
	/// NmeaInterpreter MyInterpreter2 = new NmeaInterpreter();
	///     </code>
	/// </example>
#if !PocketPC
	[TypeConverter(typeof(ExpandableObjectConverter))]
#endif
	public sealed class GpsStream : Stream
	{
		private Device pCurrentDevice;
        private Stream pRawStream;
        private Stream pBaseStream;
        private string pProtocol;
		//private object SyncLock = new object();
		private bool pIsClosed;

		/// <summary>Creates a new instance configured to detect NMEA-0183 GPS devices.</summary>
		public GpsStream() 
			: this(Gps.Nmea.NmeaInterpreter.ProtocolName)
		{
		}

		/// <summary>Creates a new instance using the specified GPS data protocol.</summary>
		/// <remarks>
		/// 	<para>This constructor is used in special situations where a specific kind of GPS
		///     data is needed. For example, to automatically detect a GPS device using the Garmin
		///     binary protocol, this constructor would be used with "Garmin binary" as the
		///     parameter.</para>
		/// 	<para>To determine the correct value for the Protocol parameter.</para>
		/// </remarks>
		/// <param name="protocol">A <strong>String</strong> identifying the desired GPS data protocol.</param>
		public GpsStream(string protocol)
		{
			pProtocol = protocol;
		}


		/// <summary>
		/// Locates the first available GPS device matching the requested GPS data
		/// protocol.
		/// </summary>
		/// <remarks>
		/// 	<para>This method examines the <strong>GpsDevices</strong> collection on the
		///     <strong>Devices</strong> class in an attempt to locate a device whose protocol
		///     matched the current requested protocol. If a device is found, it's hardware stream
		///     is used as this classes base stream, and all calls to <strong>Read</strong>,
		///     <strong>ReadByte</strong>, <strong>Write</strong>, <strong>WriteByte</strong> and
		///     other stream properties are redirected to the base stream.</para>
		/// 	<para>This method is called automatically if no base stream has yet been
		///     located.</para>
		/// </remarks>
		public void Refresh()
		{
			lock(this)
			{
				if(pBaseStream != null)
					return;
				// If the stream is closed, force them to create a new one
				if(pIsClosed)
					throw new ObjectDisposedException("The automatic GPS stream has been closed.  Create a new instance of a GpsStream to resume the flow of GPS data.");                

				// Get the first responding GPS stream matching this protocol
                pRawStream = Devices.GetStreamByProtocol(Protocol);

                // Was anything found?
                if (pRawStream == null)
				{
                    switch (Platform.HostPlatformID)
                    {
                        case HostPlatformID.Smartphone:
                            throw new TimeoutException("No GPS devices could be detected on the system.  Please ensure that Bluetooth devices are paired with your Smartphone before attempting to connect.  Pairing can be performed from your phone's Bluetooth menu.");
                        case HostPlatformID.WindowsCE:
                        case HostPlatformID.PocketPC:
                            throw new TimeoutException("No GPS devices could be detected on the system.  Please verify that devices are powered on and configured to use the NMEA-0183, Garmin text, or Garmin binary protocols.  Some USB devices require USB-to-serial drivers before they will work with GPS.NET.");
                        case HostPlatformID.Desktop:
                            throw new TimeoutException("No GPS devices could be detected on the system.  Please verify that devices are powered on and configured to use the NMEA-0183, Garmin text, or Garmin binary protocols.  Some USB devices require USB-to-serial drivers before they will work with GPS.NET.");
                    }
				}

                // Start a buffered, shared read stream
                pBaseStream = new SharedReadStream(pRawStream);

                /* If we get here, we found a GPS device */
				pCurrentDevice = Devices.GetDeviceByProtocol(Protocol);
			}
		}

		#region Overrides

		/// <summary>Returns the underlying stream for a GPS device.</summary>
		public Stream BaseStream
		{
			get
			{
				return pBaseStream;
			}
		}

		/// <summary>Returns the current device responsible for the underlying stream.</summary>
		public Device CurrentDevice
		{
			get
			{
				return pCurrentDevice;
			}
		}

		/// <summary>
		/// Returns the name of the stream along with the name of the underlying
		/// stream.
		/// </summary>
		public override string ToString()
		{
            if (BaseStream == null)
                return "Automatically-Discovered GPS Device (None found yet)";
            else
                return BaseStream.ToString() + " (Automatically-Discovered " + Protocol + " GPS Device)";
		}


		/// <summary>Begins an asynchronous read on the underlying stream.</summary>
		public override IAsyncResult BeginRead(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			if(pBaseStream == null)
				Refresh();
			return pBaseStream.BeginRead (buffer, offset, count, callback, state);
		}

		/// <summary>Begins an asynchronous write on the underlying stream.</summary>
		public override IAsyncResult BeginWrite(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			if(pBaseStream == null)
				Refresh();
			return pBaseStream.BeginWrite (buffer, offset, count, callback, state);
		}

		/// <summary>Indicates if read operations are allowed.</summary>
		public override bool CanRead
		{
			get
			{
					if (pBaseStream == null)
						Refresh();
				//lock(SyncLock)
				{
					return pBaseStream.CanRead;
				}
			}
		}

		/// <summary>Indicates if seek operations are allowed.</summary>
		public override bool CanSeek
		{
			get
			{
					if (pBaseStream == null)
						Refresh();
				//lock(SyncLock)
				{
					return pBaseStream.CanSeek;
				}
			}
		}

		/// <summary>Indicates if write operations are allowed.</summary>
		public override bool CanWrite
		{
			get
			{
					if(pBaseStream == null)
						Refresh();
				//lock(SyncLock)
				{
					return pBaseStream.CanWrite;
				}
			}
		}

		///// <summary>
		///// Compares the underlying stream to the specified object.
		///// </summary>
		///// <param name="obj"></param>
		///// <returns></returns>
		//public override bool Equals(object obj)
		//{
		//    if(pBaseStream == null && obj == null)
		//        return true;
		//    if(pBaseStream == null)
		//    if(obj is SerialStream)
		//        return p
		//}

		/// <summary>Closes this stream and the underlying stream.</summary>
		public override void Close()
		{
			if(pIsClosed)
				return;

			////lock(SyncLock)
			{
				// Close and terminate the stream
				try
				{
					if (pBaseStream != null)
						pBaseStream.Close();
				}
				catch
				{
				}
				finally
				{
					pIsClosed = true;
				}
			}
		}

		public GpsStream Clone()
		{
			return new GpsStream(Protocol);
		}

		public bool IsClosed
		{
			get
			{
				return pIsClosed;
			}
		}

#if !Framework20
		/// <summary>Returns a wait handle used for synchronizing asynchronous operations.</summary>
		protected override WaitHandle CreateWaitHandle()
		{
			return new ManualResetEvent(false);
		}
#endif

		/// <summary>Completes an asynchronous read operation.</summary>
		public override int EndRead(IAsyncResult asyncResult)
		{
			if(pBaseStream == null)
				Refresh();
			return pBaseStream.EndRead (asyncResult);
		}

		/// <summary>Completes an asynchronous write operation.</summary>
		public override void EndWrite(IAsyncResult asyncResult)
		{
			if(pBaseStream == null)
				Refresh();
			pBaseStream.EndWrite (asyncResult);
		}

		/// <summary>Empties any pending data sending or receiving buffers.</summary>
		public override void Flush()
		{
				if(pBaseStream == null)
					Refresh();
			//lock(SyncLock)
			{
				pBaseStream.Flush();
			}
		}

		/// <summary>Returns the number of available bytes in the underlying stream.</summary>
		public override long Length
		{
			get
			{
					if(pBaseStream == null)
						Refresh();
				//lock(SyncLock)
				{
					if(pRawStream is NetworkStream)
						return ((NetworkStream)pRawStream).DataAvailable ? 1 : 0;
					return pBaseStream.Length;
				}
			}
		}

		/// <summary>Returns the name of the GPS data protocol to use during binding.</summary>
		public string Protocol
		{
			get
			{
				return pProtocol;
			}
		}

		/// <summary>Returns the underlying stream's current position.</summary>
		public override long Position
		{
			get
			{
					if(pBaseStream == null)
						Refresh();
				//lock(SyncLock)
				{
					return pBaseStream.Position;
				}
			}
			set
			{
					if(pBaseStream == null)
						Refresh();
				//lock(SyncLock)
				{
					pBaseStream.Position = value;
				}
			}
		}

		/// <summary>Reads a number of bytes from the underlying stream.</summary>
		public override int Read(byte[] buffer, int offset, int count)
		{			
			if(pBaseStream == null)
				Refresh();
			//lock(SyncLock)
			{
					return pBaseStream.Read(buffer, offset, count);
			}
		}

		/// <summary>Reads a single byte from the underlying stream.</summary>
		public override int ReadByte()
		{
				if(pBaseStream == null)
					Refresh();
				try
				{
					//lock(SyncLock)
					{
						return pBaseStream.ReadByte();
					}
				}
				catch(ObjectDisposedException)
				{
					return -1;
				}
			
		}

		/// <summary>Changes the position of the underlying stream.</summary>
		public override long Seek(long offset, SeekOrigin origin)
		{
				if(pBaseStream == null)
					Refresh();
			//lock(SyncLock)
			{
				return pBaseStream.Seek(offset, origin);
			}
		}

		/// <summary>Changes the length of the underlying stream.</summary>
		public override void SetLength(long value)
		{
				if(pBaseStream == null)
					Refresh();
			//lock(SyncLock)
			{
				pBaseStream.SetLength(value);
			}
		}

		/// <summary>Writes a number of bytes to the underlying stream.</summary>
		public override void Write(byte[] buffer, int offset, int count)
		{
				if(pBaseStream == null)
					Refresh();
			//lock(SyncLock)
			{
				pBaseStream.Write(buffer, offset, count);
			}
		}

		/// <summary>Writes a single byte to the underlying stream.</summary>
		public override void WriteByte(byte value)
		{
			//lock(SyncLock)
			{
				if(pBaseStream == null)
					Refresh();
				pBaseStream.WriteByte (value);
			}
		}

		#endregion
		
	}
}
