using System;

namespace GeoFramework.IO.Serial
{
	/// <summary>
	/// Indicates the technique used to control the transmission of serial port data.
	/// </summary>
	public enum FlowControl 
	{
		/// <summary>
		/// No flow control is used.
		/// </summary>
		None, 
		/// <summary>
		/// Flow control is controlled via "XOn" and "XOff" characters.
		/// </summary>
		XOnXOff, 
		/// <summary>
		/// Flow control is managed by the serial port hardware.
		/// </summary>
		Hardware
	}

	/// <summary>
	/// Indicates the speed of serial port communications.
	/// </summary>
	/// <remarks>
	/// <para>This enumeration is used by the <see cref="BaudRate">BaudRate</see> property of the <see cref="Device">Device</see>
	/// and <see cref="Receiver">Receiver</see> classes.  A value of <strong>Baud4800</strong> is used by default for the Receiver class, and is the most
	/// commonly-used baud rate.  Some newer devices will transmit NMEA data at much higher baud rates, but
	/// 4800 baud handles a majority of devices on the market.  A value of <strong>Baud9600</strong> is used for
	/// Garmin devices set to use the Garmin binary protocol.</para>
	/// </remarks>
	/// <seealso cref="Device.BaudRate">BaudRate Property (Device Class)</seealso>
	/// <seealso cref="Receiver.BaudRate">BaudRate Property (Receiver Class)</seealso>
	public enum BaudRate: int
	{
		/// <summary>The baud rate is not yet known.</summary>
		Unknown = 0,
		/// <summary>4,800 bits per second.  Recommended value for devices transmitting GPS data using NMEA-0183.</summary>
		Baud4800 = 4800,
		/// <summary>9,600 bits per second.  Recommended value for devices transmitting GPS data using the Garmin binary protocol.</summary>
		Baud9600 = 9600,
		/// <summary>14,400 bits per second.</summary>
		Baud14400 = 14400,
		/// <summary>19,200 bits per second.</summary>
		Baud19200 = 19200,
		/// <summary>28,800 bits per second.</summary>
		Baud28800 = 28800,
		/// <summary>38,400 bits per second.</summary>
		Baud38400 = 38400,
		/// <summary>56,700 bits per second.</summary>
		Baud57600 = 57600,
		/// <summary>115,200 bits per second.</summary>
		Baud115200 = 115200
	}

    /// <summary>
    /// Describes the number of bits of data sent by a serial device after each stop bit.
    /// </summary>
    /// <remarks>This enumeration is used by the <see cref="Receiver.DataBits">DataBits</see> 
    /// property of the <see cref="Receiver">Receiver</see> class to control how many bits are 
    /// sent between each stop bit.  Typically, GPS devices use <strong>Eight</strong>, which 
    /// is the default of the Receiver class.</remarks>
    /// <seealso cref="Receiver.DataBits">DataBits Property (Receiver Class)</seealso>
    /// <seealso cref="Device.DataBits">DataBits Property (Device Class)</seealso>
    public enum DataBits : int
    {
        /// <summary>Four bytes of data are sent after each stop bit.</summary>
        Four = 4,
        /// <summary>Five bytes of data are sent after each stop bit.</summary>
        Five = 5,
        /// <summary>Six bytes of data are sent after each stop bit.</summary>
        Six = 6,
        /// <summary>Seven bytes of data are sent after each stop bit.</summary>
        Seven = 7,
        /// <summary>Eight bytes of data are sent after each stop bit.  The default value used by the Receiver and a majority of GPS devices.</summary>
        Eight = 8
    }

	/// <summary>
	/// Describes the method of simple error-checking used during serial communications.
	/// </summary>
	/// <remarks>This enumeration is used by the <see cref="Receiver.Parity">Parity</see> property 
	/// of the <see cref="Receiver">Receiver</see> class. A parity bit affords a small (yet 
	/// crude by today's standards) amount of error checking, to help detect data corruption that might occur during 
	/// transmission. You can choose either even parity, odd parity, mark parity, space parity or 
	/// none at all.   Typically, GPS devices use a value of <strong>None</strong>, which is the 
	/// default used by the Receiver class.</remarks>
	/// <seealso cref="Receiver.Parity">Parity Property (Receiver Class)</seealso>
	/// <seealso cref="Device.Parity">Parity Property (Device Class)</seealso>
	public enum Parity
	{
		/// <summary>No parity is used.  The default value used by the Receiver and a majority of GPS devices.</summary>
		None,
		/// <summary>Returns 1 if the counted number of on bits is odd.</summary>
		Odd,
		/// <summary>Returns 1 if the counted number of on bits is even.</summary>
		Even,
		/// <summary>Sends a constant "mark" signal.  Rarely used.</summary>
		Mark,
		/// <summary>Sends a constant "space" signal.  Rarely used.</summary>
		Space
	}

    /// <summary>
	/// Describes the length of time that a stop bit is sent, in bit periods.
	/// </summary>
	/// <remarks>This enumeration is used by the <see cref="Receiver.StopBits">StopBits</see> 
	/// property of the <see cref="Receiver">Receiver</see> class to indicate how many stop 
	/// bits are being sent between blocks of data to and from a serial device.  Typically, most 
	/// GPS devices use a value of <strong>One</strong>, which is the default of the 
	/// <see cref="Receiver">Receiver</see> class.</remarks>
	/// <seealso cref="Receiver.StopBits">StopBits Property (Receiver Class)</seealso>
	/// <seealso cref="Device.StopBits">StopBits Property (Device Class)</seealso>
	public enum StopBits: int
	{
		/// <summary>The stop bit is sent for one bit period.</summary>
		One = 1,
		/// <summary>The stop bit is sent for two bit periods.</summary>
        Two = 2,
        /// <summary>The stop bit is sent for one-and-a-half bit periods.</summary>
        OnePointFive = 3
    }

    /// <summary>An enumeration used to specify which serial port a GPS device is connected to.</summary>
	/// <remarks>A valid COM port must be specified in order to begin working with GPS devices.  
	/// This enumeration of typically used by the
	/// <see cref="Receiver.Port">CommunicationPort</see> property of the
	/// <see cref="Receiver">Receiver</see> and
	/// <see cref="Device">Device</see> classes.
	/// </remarks>
	/// <seealso cref="Receiver.Port">CommunicationPort Property (Receiver Class)</seealso>
	/// <seealso cref="Device.Port">CommunicationPort Property (Device Class)</seealso>
	public enum CommunicationPort: int
	{
		/// <summary>The serial port of the serial device is not yet known.</summary>
		Unknown = -1,
		/// <summary>The serial device is connected to COM0.</summary>
		Com0 = 0,
		/// <summary>The serial device is connected to COM1.</summary>
		Com1 = 1,
		/// <summary>The serial device is connected to COM2.</summary>
		Com2 = 2,
		/// <summary>The serial device is connected to COM3.</summary>
		Com3 = 3,
		/// <summary>The serial device is connected to COM4.</summary>
		Com4 = 4,
		/// <summary>The serial device is connected to COM5.</summary>
		Com5 = 5,
		/// <summary>The serial device is connected to COM6.</summary>
		Com6 = 6,
		/// <summary>
		/// The serial device is connected to COM7. A common default for PocketPC devices, such
		/// as Bluetooth, along with COM8.
		/// </summary>
		Com7 = 7,
		/// <summary>
		/// The serial device is connected to COM8. A common default for PocketPC devices, such
		/// as Bluetooth, along with COM7.
		/// </summary>
		Com8 = 8,
		/// <summary>The serial device is connected to COM9.</summary>
		Com9 = 9,
		/// <summary>The serial device is connected to COM10.</summary>
		Com10 = 10,
		/// <summary>The serial device is connected to COM11.</summary>
		Com11 = 11,
		/// <summary>The serial device is connected to COM12.</summary>
		Com12 = 12,
		/// <summary>The serial device is connected to COM13.</summary>
		Com13 = 13,
		/// <summary>The serial device is connected to COM14.</summary>
		Com14 = 14,
		/// <summary>The serial device is connected to COM15.</summary>
		Com15 = 15,
		/// <summary>The serial device is connected to COM16.</summary>
		Com16 = 16,
		/// <summary>The serial device is connected to COM17.</summary>
		Com17 = 17,
		/// <summary>The serial device is connected to COM18.</summary>
		Com18 = 18,
		/// <summary>The serial device is connected to COM19.</summary>
		Com19 = 19,
		/// <summary>The serial device is connected to COM20.</summary>
		Com20 = 20,
		/// <summary>The serial device is connected to COM21.</summary>
		Com21 = 21,
		/// <summary>The serial device is connected to COM22.</summary>
		Com22 = 22,
		/// <summary>The serial device is connected to COM23.</summary>
		Com23 = 23,
		/// <summary>The serial device is connected to COM24.</summary>
		Com24 = 24,
		/// <summary>The serial device is connected to COM25.</summary>
		Com25 = 25,
		/// <summary>The serial device is connected to COM26.</summary>
		Com26 = 26,
		/// <summary>The serial device is connected to COM27.</summary>
		Com27 = 27,
		/// <summary>The serial device is connected to COM28.</summary>
		Com28 = 28,
		/// <summary>The serial device is connected to COM29.</summary>
		Com29 = 29,
		/// <summary>The serial device is connected to COM30.</summary>
		Com30 = 30
	}

	#region Internal Enumerations

	/// <summary>
	/// Communication escapes
	/// </summary>
	internal enum CommunicationEscapeCode : uint
	{
		/// <summary>
		/// Causes transmission to act as if an XOFF character has been received.
		/// </summary>
		SETXOFF		= 1,
		/// <summary>
		/// Causes transmission to act as if an XON character has been received.
		/// </summary>
		SETXON		= 2,
		/// <summary>
		/// Sends the RTS (Request To Send) signal.
		/// </summary>
		SETRTS		= 3,
		/// <summary>
		/// Clears the RTS (Request To Send) signal
		/// </summary>
		CLRRTS		= 4,
		/// <summary>
		/// Sends the DTR (Data Terminal Ready) signal.
		/// </summary>
		SETDTR		= 5,
		/// <summary>
		/// Clears the DTR (Data Terminal Ready) signal.
		/// </summary>
		CLRDTR		= 6,
		/// <summary>
		/// Suspends character transmission and places the transmission line in a break state until the ClearCommBreak function is called (or EscapeCommFunction is called with the CLRBREAK extended function code). The SETBREAK extended function code is identical to the SetCommBreak function. This extended function does not flush data that has not been transmitted.
		/// </summary>
		SETBREAK	= 8,
		/// <summary>
		/// Restores character transmission and places the transmission line in a nonbreak state. The CLRBREAK extended function code is identical to the ClearCommBreak function
		/// </summary>
		CLRBREAK	= 9,
		///Set the port to IR mode.
		SETIR		= 10,
		/// <summary>
		/// Set the port to non-IR mode.
		/// </summary>
		CLRIR		= 11
	}

	internal enum PurgeBuffers
	{
		RXAbort = 0x2,
		RXClear = 0x8,
		TXAbort = 0x1,
		TXClear = 0x4
	}

	[Flags()]
	internal enum DCBFlags : int
	{
		// Specifies if binary mode is enabled. The Win32 API does not support 
		// nonbinary mode transfers, so this member must be TRUE. Using FALSE will 
		// not work. 
		Binary = 1, // YES
		// Specifies if parity checking is enabled. If this member is TRUE, 
		// parity checking is performed and errors are reported. 
		Parity = 2, // NO
		// Specifies if the CTS (clear-to-send) signal is monitored for output 
		// flow control. If this member is TRUE and CTS is turned off, output 
		// is suspended until CTS is sent again. 
		OutXCTSFlow = 4, // NO
		// Specifies if the DSR (data-set-ready) signal is monitored for output 
		// flow control. If this member is TRUE and DSR is turned off, output is 
		// suspended until DSR is sent again. 
		OutXDSRFlow = 8, // NO
		// Specifies the DTR (data-terminal-ready) flow control. This member can 
		// be one of the following values:
		// DTR_CONTROL_DISABLE - Disables the DTR line when the device is opened and leaves it disabled.
		// DTR_CONTROL_ENABLE - Enables the DTR line when the device is opened and leaves it on. 
		// DTR_CONTROL_HANDSHAKE - Enables DTR HandshakeMode. If HandshakeMode is enabled, it is an error for the application to adjust the line by using the EscapeCommFunction function.
		DTRControl1 = 16, // YES
		DTRControl2 = 32, // NO
		// Specifies if the communications driver is sensitive to the state of 
		// the DSR signal. If this member is TRUE, the driver ignores any bytes 
		// received, unless the DSR modem input line is high. 
		DSRSensitivity = 64, // NO
		// Specifies if transmission stops when the input buffer is full and the 
		// driver has transmitted the XoffChar character. If this member is TRUE, 
		// transmission continues after the input buffer has come within XoffLim 
		// bytes of being full and the driver has transmitted the XoffChar character 
		// to stop receiving bytes. If this member is FALSE, transmission does not 
		// continue until the input buffer is within XonLim bytes of being empty and 
		// the driver has transmitted the XonChar character to resume reception. 
		TXContinueOnXOff = 128, // YES
		// Specifies if XON/XOFF flow control is used during transmission. If this 
		// member is TRUE, transmission stops when the XoffChar character is received 
		// and starts again when the XonChar character is received. 
		OutX = 256, // NO
		// Specifies if XON/XOFF flow control is used during reception. If this member 
		// is TRUE, the XoffChar character is sent when the input buffer comes within 
		// XoffLim bytes of being full, and the XonChar character is sent when the 
		// input buffer comes within XonLim bytes of being empty. 
		InX = 512, // NO
		// Specifies if bytes received with parity errors are replaced with the 
		// character specified by the ErrorChar member. If this member is TRUE and 
		// the ParityCheck member is TRUE, replacement occurs. 
		ErrorChar = 1024, // NO
		// Specifies if null bytes are discarded. If this member is TRUE, null bytes 
		// are discarded when received. 
		Null = 2048, // NO
		// Specifies the RTS (request-to-send) flow control. If this value is zero, 
		// the default is RTS_CONTROL_HANDSHAKE. This member can be one of the 
		// following values:
		// RTS_CONTROL_DISABLE - Disables the RTS line when the device is opened and leaves it disabled.
		// RTS_CONTROL_ENABLE - Enables the RTS line when the device is opened and leaves it on.
		// RTS_CONTROL_HANDSHAKE - Enables RTS HandshakeMode. The driver raises the RTS line when the type-ahead (input) buffer is less than one-half full and lowers the RTS line when the buffer is more than three-quarters full. If HandshakeMode is enabled, it is an error for the application to adjust the line by using the EscapeCommFunction function.
		// RTS_CONTROL_TOGGLE - Specifies that the RTS line will be high if bytes are available for transmission. After all buffered bytes have been sent, the RTS line will be low.
		RTSControl1 = 4096, // YES
		RTSControl2 = 8192, // NO
		// Specifies if read and write operations are terminated if an error occurs. 
		// If this member is TRUE, the driver terminates all read and write operations 
		// with an error status if an error occurs. The driver will not accept any 
		// further communications operations until the application has acknowledged 
		// the error by calling the ClearCommError function. 
		AbortOnError = 16384 // YES
		// Reserved; do not use
		//Dummy2 = 17 bits
	}

	[Flags()]
	internal enum COMSTATFlags : int
	{
		// If this member is TRUE, transmission is waiting for the CTS (clear-to-send) 
		// signal to be sent. 
		CTSHold = 1,
		// If this member is TRUE, transmission is waiting for the DSR (data-set-ready)
		// signal to be sent. 
		DSRHold = 2,
		// If this member is TRUE, transmission is waiting for the RLSD (receive-line-
		// signal-detect) signal to be sent. 
		RLSDHold = 4,
		// If this member is TRUE, transmission is waiting because the XOFF character
		// was received. 
		XOffHold = 8,
		// If this member is TRUE, transmission is waiting because the XOFF character
		// was transmitted. (Transmission halts when the XOFF character is transmitted
		// to a system that takes the next character as XON, regardless of the actual
		// character.) 
		XOffSent = 16,
		// If this member is TRUE, the end-of-file (EOF) character has been received. 
		EOF = 32,
		// If this member is TRUE, there is a character queued for transmission that
		// has come to the communications device by way of the TransmitCommChar function.
		// The communications device transmits such a character ahead of other
		// characters in the device's output buffer. 
		Txim = 64
		// Reserved; do not use. 
		//Reserved = 25 bits
	}

	[Flags]
	internal enum SerialMode : uint
	{
		GenericRead = 0x80000000,
		GenericWrite = 0x40000000,
		GenericExecute = 0x20000000,
		GenericAll = 0x10000000,
	}

	[Flags]
	internal enum SerialShareMode : uint
	{
		None = 0x00000000,
		Read = 0x00000001,
		Write = 0x00000002,
		Delete = 0x00000004,
	}

	internal enum SerialCreationDisposition : uint
	{
		New = 1,
		CreateAlways = 2,
		OpenExisting = 3,
		OpenAlways = 4,
		TruncateExisting = 5,
	}

	[Flags]
	internal enum SerialFileAttributes : uint
	{
		Readonly = 0x00000001,
		Hidden = 0x00000002,
		System = 0x00000004,
		Directory = 0x00000010,
		Archive = 0x00000020,
		Device = 0x00000040,
		Normal = 0x00000080,
		Temporary = 0x00000100,
		SparseFile = 0x00000200,
		ReparsePoint = 0x00000400,
		Compressed = 0x00000800,
		Offline= 0x00001000,
		NotContentIndexed = 0x00002000,
		Encrypted = 0x00004000,
		Write_Through = 0x80000000,
		Overlapped = 0x40000000,
		NoBuffering = 0x20000000,
		RandomAccess = 0x10000000,
		SequentialScan = 0x08000000,
		DeleteOnClose = 0x04000000,
		BackupSemantics = 0x02000000,
		PosixSemantics = 0x01000000,
		OpenReparsePoint = 0x00200000,
		OpenNoRecall = 0x00100000,
		FirstPipeInstance = 0x00080000
	}

	[Flags()]
	internal enum PurgeFlags : int
	{
		CancelReceiving = 0x2,
		PurgeReceiving = 0x8,
		CancelSending = 0x1,
		PurgeSending = 0x4
	}

	#endregion

}
