using System;
using System.Security.Permissions;
using System.IO;
using System.Globalization;
using System.Text;
using System.Threading;
using GeoFramework;
using GeoFramework.Gps;
#if !PocketPC || Framework20
using Microsoft.Win32;
#endif

namespace GeoFramework.IO.Serial
{
	/// <summary>
	/// Represents a device which communicates with applications via a serial
	/// port.
	/// </summary>
	/// <remarks>
	/// 	<para>This class serves as a gateway to serial port device information and
	///     communications. This class is used to provide information regarding installed
	///     serial devices to developers, and can also create <strong>SerialStream</strong>
	///     objects for valid serial devices via the <strong>GetHardwareStream</strong>
	///     method.</para>
	/// 	<para>Instances of this object are partially thread-safe. The <strong>Port</strong>
	///     property is read-only, but other properties are mutable.</para>
	/// </remarks>
	/// <example>
	///     This example enumerates through each valid serial device on the local machine. 
	///     <code lang="VB" title="[New Example]">
	/// Imports GeoFramework.IO
	///  
	/// Sub Main()
	///     Dim FoundDevice As SerialDevice
	///     For Each FoundDevice In Devices.SerialDevices
	///         ' Write the device info out to the Output window
	///         Debug.WriteLine(FoundDevice.ToString())
	///     Next FoundDevice
	/// End Sub
	///     </code>
	/// 	<code lang="CS" title="[New Example]">
	/// using GeoFramework.IO;
	///  
	/// void Main()
	/// {
	///     foreach(SerialDevice FoundDevice in Devices.SerialDevices)
	///     {
	///         // Write the device info out to the Output window
	///         Console.WriteLine(FoundDevice.ToString());
	///     }
	/// }
	///     </code>
	/// </example>
#if !PocketPC
    [RegistryPermission(SecurityAction.Assert, Read = @"HKEY_LOCAL_MACHINE\Software\Widcomm\Connections")]
    [RegistryPermission(SecurityAction.Assert, Read = @"HKEY_CURRENT_USER\Software\Widcomm\Connections")]
#endif
	public sealed class SerialDevice : Device
	{
		private string pDeviceName;
		private string pPortName;
        private CommunicationPort pPort;
        private BaudRate pBaudRate;
        private DataBits pDataBits;
        private StopBits pStopBits;
        private Parity pParity;
        private FlowControl pFlowControl;
        private bool pIsBluetoothDevice;

		/// <summary>Represents a serial device connected to COM1.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM1.</value>
        public static readonly SerialDevice Com1 = new SerialDevice(CommunicationPort.Com1);
		/// <summary>Represents a serial device connected to COM2.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM2.</value>
		public static readonly SerialDevice Com2 = new SerialDevice(CommunicationPort.Com2);
		/// <summary>Represents a serial device connected to COM3.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM3.</value>
		public static readonly SerialDevice Com3 = new SerialDevice(CommunicationPort.Com3);
		/// <summary>Represents a serial device connected to COM4.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM4.</value>
		public static readonly SerialDevice Com4 = new SerialDevice(CommunicationPort.Com4);
		/// <summary>Represents a serial device connected to COM5.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM5.</value>
		public static readonly SerialDevice Com5 = new SerialDevice(CommunicationPort.Com5);
		/// <summary>Represents a serial device connected to COM6.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM6.</value>
		public static readonly SerialDevice Com6 = new SerialDevice(CommunicationPort.Com6);
		/// <summary>Represents a serial device connected to COM7.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM7.</value>
		public static readonly SerialDevice Com7 = new SerialDevice(CommunicationPort.Com7);
		/// <summary>Represents a serial device connected to COM8.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM8.</value>
		public static readonly SerialDevice Com8 = new SerialDevice(CommunicationPort.Com8);
		/// <summary>Represents a serial device connected to COM9.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM9.</value>
		public static readonly SerialDevice Com9 = new SerialDevice(CommunicationPort.Com9);
		/// <summary>Represents a serial device connected to COM10.</summary>
		/// <value>A <strong>SerialDevice</strong> object associated with COM10.</value>
		public static readonly SerialDevice Com10 = new SerialDevice(CommunicationPort.Com10);

		/// <summary>
		/// Creates a new instance using the specified COM: port.
		/// </summary>
		/// <param name="port">A <strong>CommunicationPort</strong> value indicating which port to associate with this device.</param>
		public SerialDevice(CommunicationPort port) 
			: this(port, BaudRate.Baud57600, DataBits.Eight, Parity.None, StopBits.One, FlowControl.None)
		{
		}

		public SerialDevice(CommunicationPort port, BaudRate baudRate) 
			: this(port, baudRate, DataBits.Eight, Parity.None, StopBits.One, FlowControl.None)
		{
		}

		public SerialDevice(CommunicationPort port, BaudRate baudRate,
			DataBits dataBits, Parity parity, StopBits stopBits, FlowControl flowControl)
            : this(port, baudRate, dataBits, parity, stopBits, flowControl, null)
		{}

        public SerialDevice(CommunicationPort port, BaudRate baudRate,
                            DataBits dataBits, Parity parity, StopBits stopBits, FlowControl flowControl, string deviceName)   
        {
            pPort = port;

            if (port != CommunicationPort.Unknown)
            {
                int PortNumber = Convert.ToInt32(port);
                if (PortNumber >= 0 && PortNumber < 10)
                {
                    pPortName = "COM" + PortNumber.ToString() + ":";
                }
                else
                {
#if PocketPC
					if(Environment.OSVersion.Platform == PlatformID.WinCE)
					{
						pPortName = @"COM" + PortNumber.ToString() + ":";
					}
					else
					{
						pPortName = @"\\.\COM" + PortNumber.ToString();
					}
#else
                    pPortName = @"\\.\COM" + PortNumber.ToString();
#endif
                }
            }

            pBaudRate = baudRate;
            pDataBits = dataBits;
            pParity = parity;
            pStopBits = stopBits;
            pFlowControl = flowControl;

            SetDeviceName();
        }

		public SerialDevice(string portName)
			:this(portName, BaudRate.Baud57600, DataBits.Eight, Parity.None, StopBits.One, FlowControl.None)
		{}

		public SerialDevice(string portName, BaudRate baudRate)
			:this(portName, baudRate, DataBits.Eight, Parity.None, StopBits.One, FlowControl.None)
		{}

		/// <summary>
		/// Creates a new instance using a special named port (i.e. "GPD1:") for the port instead of "COM:".
		/// </summary>
		/// <param name="portName">A <strong>String</strong> indicating the name of the port.</param>
		/// <param name="baudRate"></param>
		/// <param name="dataBits"></param>
		/// <param name="parity"></param>
		/// <param name="stopBits"></param>
		/// <param name="flowControl"></param>
        public SerialDevice(string portName, BaudRate baudRate,
            DataBits dataBits, Parity parity, StopBits stopBits, FlowControl flowControl)
            : this(portName, baudRate, dataBits, parity, stopBits, flowControl, null)
        {}

        public SerialDevice(string portName, BaudRate baudRate,
			DataBits dataBits, Parity parity, StopBits stopBits, FlowControl flowControl, string deviceName)
        {
			// Store the speciaiized port name
			pPortName = portName;
            pPort = CommunicationPort.Unknown;

			// Try to interpret the port
            int ComStart = pPortName.IndexOf("COM");
            if (ComStart != -1)
            {
                int ComEnd = pPortName.IndexOf(":", ComStart);
                if (ComEnd != -1)
                {
                    pPortName = pPortName.Substring(ComStart, ComEnd - ComStart).Trim();
                    string InterpretedPort = pPortName.Replace("COM", "").Replace(":", "");
                    try
                    {
                        int PortNumber = Convert.ToInt32(InterpretedPort);
#if PocketPC && !Framework20
				switch(PortNumber)
				{
					case 0:
						pPort = CommunicationPort.Com0;
						break;
					case 1:
						pPort = CommunicationPort.Com1;
						break;
					case 2:
						pPort = CommunicationPort.Com2;
						break;
					case 3:
						pPort = CommunicationPort.Com3;
						break;
					case 4:
						pPort = CommunicationPort.Com4;
						break;
					case 5:
						pPort = CommunicationPort.Com5;
						break;
					case 6:
						pPort = CommunicationPort.Com6;
						break;
					case 7:
						pPort = CommunicationPort.Com7;
						break;
					case 8:
						pPort = CommunicationPort.Com8;
						break;
					case 9:
						pPort = CommunicationPort.Com9;
						break;
					case 10:
						pPort = CommunicationPort.Com10;
						break;
					case 11:
						pPort = CommunicationPort.Com11;
						break;
					case 12:
						pPort = CommunicationPort.Com12;
						break;
					case 13:
						pPort = CommunicationPort.Com13;
						break;
					case 14:
						pPort = CommunicationPort.Com14;
						break;
					case 15:
						pPort = CommunicationPort.Com15;
						break;
					case 16:
						pPort = CommunicationPort.Com16;
						break;
					case 17:
						pPort = CommunicationPort.Com17;
						break;
					case 18:
						pPort = CommunicationPort.Com18;
						break;
					case 19:
						pPort = CommunicationPort.Com19;
						break;
					case 20:
						pPort = CommunicationPort.Com20;
						break;
					case 21:
						pPort = CommunicationPort.Com21;
						break;
					case 22:
						pPort = CommunicationPort.Com22;
						break;
					case 23:
						pPort = CommunicationPort.Com23;
						break;
					case 24:
						pPort = CommunicationPort.Com24;
						break;
					case 25:
						pPort = CommunicationPort.Com25;
						break;
					case 26:
						pPort = CommunicationPort.Com26;
						break;
					case 27:
						pPort = CommunicationPort.Com27;
						break;
					case 28:
						pPort = CommunicationPort.Com28;
						break;
					case 29:
						pPort = CommunicationPort.Com29;
						break;
					case 30:
						pPort = CommunicationPort.Com30;
						break;
				}
#else
                        pPort = (CommunicationPort)Enum.Parse(typeof(CommunicationPort), PortNumber.ToString(), true); // TODO
#endif

                        // If the port is greater than 10, change the port name
                        if (PortNumber < 10)
                        {
                            pPortName = @"COM" + PortNumber.ToString() + ":";
                        }
                        else
                        {
                            pPortName = @"\\.\COM" + PortNumber.ToString();
                        }

                    }
                    catch
                    {
                    }
                }
            }

			pBaudRate = baudRate;
			pDataBits = dataBits;
			pParity = parity;
			pStopBits = stopBits;
			pFlowControl = flowControl;

            SetDeviceName();
		}

        internal void SetDeviceName(string name)
        {
            pDeviceName = name;
        }

        internal void SetDeviceName()
        {
            #region Bluetooth Name Determination (Pass 1)

            if (pDeviceName == null)
            {
                RegistryKey AutoConnect = null;
                try
                {
                    AutoConnect = Registry.LocalMachine.OpenSubKey(@"Software\Widcomm\BTConfig\AutoConnect\" + Convert.ToInt32(pPort).ToString(CultureInfo.InvariantCulture).PadLeft(4, '0'), false);

                    if (AutoConnect != null)
                    {
                        // Read the "BDName" value for the device name
                        object FriendlyObject = AutoConnect.GetValue("BDName");
                        byte[] FriendlyBytes = FriendlyObject as byte[];
                        string FriendlyString = FriendlyObject as string;
                        if (FriendlyString != null)
                        {
                            // The device name is a string
                            pDeviceName = FriendlyString;
                            pIsBluetoothDevice = true;
                        }
                        else if (FriendlyBytes != null)
                        {
                            // The device name may in fact be a byte array. Convert it
                            // while stripping off the null at the ned
                            pDeviceName = System.Text.ASCIIEncoding.ASCII.GetString(FriendlyBytes, 0, FriendlyBytes.Length - 1);
                            pIsBluetoothDevice = true;
                        }
                    }
                }
                catch
                {
                    // Ignore problems!
                }
                finally
                {
                    if (AutoConnect != null)
                        AutoConnect.Close();
                }
            }

            #endregion

            #region Bluetooth Name Determination (Pass 2)

            if (pDeviceName == null)
            {
                // Look for WidComm Bluetooth devices
                RegistryKey Widcomm = null;
                try
                {
#if !Framework10
                    if (Environment.OSVersion.Platform == PlatformID.WinCE)
                    {
                        Widcomm = Registry.LocalMachine.OpenSubKey(@"Software\Widcomm\Connections", false);
                    }
                    else
                    {
#endif
                        Widcomm = Registry.CurrentUser.OpenSubKey(@"Software\Widcomm\Connections", false);
#if !Framework10
                    }
#endif

                    if (Widcomm != null)
                    {
                        // Enumerate all WIDCOMM devices
                        string[] SubKeys = Widcomm.GetSubKeyNames();
                        foreach (string subKey in SubKeys)
                        {
                            // Open the sub-key  
                            RegistryKey WidcommConnectionKey = Widcomm.OpenSubKey(subKey);
                            try
                            {
                                // Skip if this key can't be opened
                                if (WidcommConnectionKey == null)
                                    continue;

                                // Look for "ComPortNumber"
                                string ServiceGuid = WidcommConnectionKey.GetValue("GUID") as string;
                                int DevicePortNumber = Convert.ToInt32(WidcommConnectionKey.GetValue("ComPortNumber"));
                                object FriendlyNameValue = WidcommConnectionKey.GetValue("BDName");
                                string FriendlyName = null;

                                if (FriendlyNameValue != null)
                                {
#if !Framework10
                                    if (Environment.OSVersion.Platform == PlatformID.WinCE)
                                    {
                                        FriendlyName = FriendlyNameValue as string;
                                    }
                                    else
                                    {
#endif
                                        byte[] FriendlyNameBytes = (byte[])FriendlyNameValue;
                                        FriendlyName = System.Text.ASCIIEncoding.ASCII.GetString(FriendlyNameBytes, 0, FriendlyNameBytes.Length).Replace("\0", "");
#if !Framework10
                                    }
#endif
                                }

                                // If the service GUID is right, use it
                                if (ServiceGuid == "{00001101-0000-1000-8000-00805F9B34FB}")
                                {
                                    // Set the port
#if PocketPC && !Framework20
									CommunicationPort DeviceComPort = CommunicationPort.Com0;
									switch(DevicePortNumber)
									{
										case 0:
											DeviceComPort = CommunicationPort.Com0;
											break;
										case 1:
											DeviceComPort = CommunicationPort.Com1;
											break;
										case 2:
											DeviceComPort = CommunicationPort.Com2;
											break;
										case 3:
											DeviceComPort = CommunicationPort.Com3;
											break;
										case 4:
											DeviceComPort = CommunicationPort.Com4;
											break;
										case 5:
											DeviceComPort = CommunicationPort.Com5;
											break;
										case 6:
											DeviceComPort = CommunicationPort.Com6;
											break;
										case 7:
											DeviceComPort = CommunicationPort.Com7;
											break;
										case 8:
											DeviceComPort = CommunicationPort.Com8;
											break;
										case 9:
											DeviceComPort = CommunicationPort.Com9;
											break;
										case 10:
											DeviceComPort = CommunicationPort.Com10;
											break;
										case 11:
											DeviceComPort = CommunicationPort.Com11;
											break;
										case 12:
											DeviceComPort = CommunicationPort.Com12;
											break;
										case 13:
											DeviceComPort = CommunicationPort.Com13;
											break;
										case 14:
											DeviceComPort = CommunicationPort.Com14;
											break;
										case 15:
											DeviceComPort = CommunicationPort.Com15;
											break;
										case 16:
											DeviceComPort = CommunicationPort.Com16;
											break;
										case 17:
											DeviceComPort = CommunicationPort.Com17;
											break;
										case 18:
											DeviceComPort = CommunicationPort.Com18;
											break;
										case 19:
											DeviceComPort = CommunicationPort.Com19;
											break;
										case 20:
											DeviceComPort = CommunicationPort.Com20;
											break;
										case 21:
											DeviceComPort = CommunicationPort.Com21;
											break;
										case 22:
											DeviceComPort = CommunicationPort.Com22;
											break;
										case 23:
											DeviceComPort = CommunicationPort.Com23;
											break;
										case 24:
											DeviceComPort = CommunicationPort.Com24;
											break;
										case 25:
											DeviceComPort = CommunicationPort.Com25;
											break;
										case 26:
											DeviceComPort = CommunicationPort.Com26;
											break;
										case 27:
											DeviceComPort = CommunicationPort.Com27;
											break;
										case 28:
											DeviceComPort = CommunicationPort.Com28;
											break;
										case 29:
											DeviceComPort = CommunicationPort.Com29;
											break;
										case 30:
											DeviceComPort = CommunicationPort.Com30;
											break;										
									}
#else
                                CommunicationPort DeviceComPort = (CommunicationPort)Enum.Parse(typeof(CommunicationPort), DevicePortNumber.ToString(), false);
#endif

                                    // Was it found?
                                    if (pPort == DeviceComPort)
                                    {
                                        // Set the friendly name of the device
                                        pDeviceName = FriendlyName;

                                        // And flag it as a Bluetooth serial port
                                        pIsBluetoothDevice = true;
                                    }
                                }
                            }
                            catch
                            {
                            }
                            finally
                            {
                                if (WidcommConnectionKey != null)
                                    WidcommConnectionKey.Close();
                            }
                        }
                    }
                }
                catch
                {
                }
                finally
                {
                    if (Widcomm != null)
                        Widcomm.Close();
                }
            }
            #endregion

            #region Bluetooth Name Determination (Pass 3)


            // Get the device name (?)
            if (pDeviceName == null)
            {
                RegistryKey IncomingPort = null;
                try
                {
                    IncomingPort = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Bluetooth\Serial\Ports", false);
                    if (IncomingPort != null)
                    {
                        // Enumerate sub-keys
                        foreach (string subKey in IncomingPort.GetSubKeyNames())
                        {
                            // If it's "Incoming Port", skip it
                            if (subKey == "Incoming")
                                continue;

                            // Otherwise, this is a device!  
                            RegistryKey DeviceKey = null;
                            try
                            {
                                DeviceKey = IncomingPort.OpenSubKey(subKey, false);
                                if (DeviceKey != null)
                                {
                                    // Get the port name
                                    string BluetoothPortName = DeviceKey.GetValue("Port") as string;
                                    if (BluetoothPortName == null)
                                        return;

                                    // Convert the port name to an integer
                                    int PortNumber = Convert.ToInt32(BluetoothPortName.Substring(3));

                                    // Do we have a match?
                                    if ((int)pPort == PortNumber)
                                    {
                                        // Set as a bluetooth device
                                        pIsBluetoothDevice = true;

                                        // Yes.  Look up the name!
                                        RegistryKey FriendlyNameKey = null;
                                        try
                                        {
                                            FriendlyNameKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Bluetooth\Device\" + subKey.Substring(4), false);
                                            if (FriendlyNameKey != null)
                                            {
                                                string FriendlyName = FriendlyNameKey.GetValue("name") as string;
                                                if (FriendlyName != null)
                                                    pDeviceName = FriendlyName;
                                            }
                                        }
                                        catch
                                        {
                                        }
                                        finally
                                        {
                                            if (FriendlyNameKey != null)
                                                FriendlyNameKey.Close();
                                        }
                                    }
                                }
                            }
                            catch
                            {
                            }
                            finally
                            {
                                if (DeviceKey != null)
                                    DeviceKey.Close();
                            }
                        }
                    }
                }
                catch
                {
                }
                finally
                {
                    if (IncomingPort != null)
                        IncomingPort.Close();
                }
            }
            #endregion

            #region Bluetooth Name Determination (Pass 4)

#if PocketPC
            if (pDeviceName == null)
            {
                RegistryKey AutoConnect = null;
                try
                {
                    AutoConnect = Registry.CurrentUser.OpenSubKey(@"Software\Widcomm\BTConfig\AutoConnect\" + Convert.ToInt32(pPort).ToString(CultureInfo.InvariantCulture).PadLeft(4, '0'), false);
                

                    if (AutoConnect != null)
                    {
                        // Read the "BDName" value for the device name
                        string FriendlyName = AutoConnect.GetValue("BDName") as string;
                        if (FriendlyName != null)
                        {
                            pDeviceName = FriendlyName;
                            pIsBluetoothDevice = true;
                        }
                    }
                }
                catch
                {
                    // Ignore problems!
                }
                finally
                {
                    if (AutoConnect != null)
                        AutoConnect.Close();
                }
            }

#endif
            #endregion
        }


		/// <summary>
		/// Returns whether the device is actually a Bluetooth device.
		/// </summary>
		/// <remarks>
		/// In order to maintain backwards compatibility with legacy applications, several
		/// Bluetooth GPS devices will make a virtual serial port available.  In some cases,
		/// connections to Bluetooth devices may not be preferable, such as when the Bluetooth
		/// service is entirely turned off.  This property returns <strong>True</strong> if
		/// system settings indicate that this serial device is a virtual device.  When this
		/// property is <strong>True</strong>, and the <strong>AllowSerialConnections</strong>
		/// property of the <strong>Devices</strong> class is also <strong>True</strong> connections 
		/// to the device are avoided during device detection.
		/// </remarks>
		public bool IsBluetoothDevice
		{
			get
			{
                return pIsBluetoothDevice;
			}
		}

		#region Protocol Detection

        /// <summary>
        /// Returns the friendly name for the device.
        /// </summary>
        /// <remarks>During Automatic GPS Device detection, GPS.NET will make attempts to determine
        /// the best possible label for a device.  In some cases, such as Bluetooth devices exposed
        /// via a virtual serial port, the exact name of the device can be found.  If no information
        /// is found, a generic name such as "Serial Device on COM5:" or "NMEA-0183 GPS Device on COM4: at 38400 baud"
        /// is used.  This property is typically used when presenting GPS device information to the end user
        /// and produces the same output as the ToString method.</remarks>
        public string DeviceName
        {
            get
            {
                if (IsBluetoothDevice)
                {
                    if(pDeviceName == null)
                        return "Bluetooth Device on " + pPortName;
                    else
                        return pDeviceName + " on " + pPortName;
                }
                else
                {
                    if (pDeviceName == null)
                    {
                        if (Protocol == null)
                        {
                            return "Serial Device on " + pPortName + " at " + (int)pBaudRate + " baud";
                        }
                        else
                        {
                            return Protocol + " GPS Device on " + pPortName + " at " + (int)pBaudRate + " baud";
                        }
                    }
                    else
                    {
                        return pDeviceName + " on " + pPortName + " at " + (int)pBaudRate + " baud";
                    }
                }
            }
        }

		/// <summary>Outputs the name of the device.</summary>
		public override string ToString()
		{
            return DeviceName;  	
		}

		/// <summary>
		/// Returns a new <strong>SerialStream</strong> for communicating with the
		/// device.
		/// </summary>
		public override Stream GetHardwareStream()
		{
			if(IsImmediateNeedStreamAvailable)
			{
				IsImmediateNeedStreamAvailable = false;
				return ImmediateNeedStream;
			}
			else
			{
				return new SerialStream(pPortName, pBaudRate, pDataBits, pParity, pStopBits, pFlowControl);
			}
		}

		protected override void OnProtocolDetected(Interpreter interpreter)
		{
            if (pDeviceName == null)
                pDeviceName = interpreter.DeviceName;
		}

		/// <summary>Returns the communications port of the device.</summary>
		/// <value>A <strong>CommunicationPort</strong> value.</value>
		/// <remarks>
		/// This property is read only. All calls to <strong>GetHardwareStream</strong> will
		/// use this port as a parameter.
		/// </remarks>
        public CommunicationPort Port
        {
            get
            {
                return pPort;
            }
        }

		/// <summary>
		/// Returns the name of the port used to open connections.
		/// </summary>
		/// <remarks>This property behaves very similar to the Port property,
		/// returning the communication port, except as a string instead of a CommunicationPort
		/// enumeration.  In some situations, a special port name such as "GPD1:" or any other value
		/// than "COM:" must be used.  For this reason, the PortName property should be used if
		/// the current port must be displayed in a user interface.</remarks>
		public string PortName
		{
			get
			{
				return pPortName;
			}
		}

		/// <summary>Controls the current communication speed of the port.</summary>
		/// <remarks>
		/// All calls to <strong>GetHardwareStream</strong> will use this value as a
		/// parameter.
		/// </remarks>
		/// <value>A <strong>BaudRate</strong> value. Default is <strong>Baud4800</strong>.</value>
        public BaudRate BaudRate
        {
            get
            {
                return pBaudRate;
            }
			set
			{
				if(pBaudRate.Equals(value)) 
                    return;
				pBaudRate = value;
			}        
		}

		/// <summary>Controls the error correction technique for the port.</summary>
		/// <value>A <strong>Parity</strong> value. Default is <strong>None</strong>.</value>
		/// <remarks>
		/// All calls to <strong>GetHardwareStream</strong> will use this value as a
		/// parameter.
		/// </remarks>
        public Parity Parity
        {
            get
            {
                return pParity;
            }
			set
			{
				if(pParity == value) return;
				pParity = value;
			}           
		}

		/// <summary>Controls the number of data bits sent per stop bit.</summary>
		/// <remarks>
		/// All calls to <strong>GetHardwareStream</strong> will use this value as a
		/// parameter.
		/// </remarks>
		/// <value>A <strong>DataBits</strong> value. Default is <strong>Eight</strong>.</value>
        public DataBits DataBits
        {
            get
            {
                return pDataBits;
            }
			set
			{
				if(pDataBits == value) return;
				pDataBits = value;
			}           
		}

		/// <summary>Controls the way read and write operations are managed.</summary>
		/// <remarks>
		/// All calls to <strong>GetHardwareStream</strong> will use this value as a
		/// parameter.
		/// </remarks>
		/// <value>A <strong>FlowControl</strong> value. Default is <strong>None</strong>.</value>
        public FlowControl FlowControl
        {
            get
            {
                return pFlowControl;
            }
			set
			{
				if(pFlowControl == value) return;
				pFlowControl = value;
			}           
		}

		/// <summary>Returns the amount of delay before sending more data bits.</summary>
		/// <remarks>
		/// All calls to <strong>GetHardwareStream</strong> will use this value as a
		/// parameter.
		/// </remarks>
		/// <value>A <strong>StopBits</strong> value. Default is <strong>One</strong>.</value>
        public StopBits StopBits
        {
            get
            {
                return pStopBits;
            }
			set
			{
				if(pStopBits == value) return;
				pStopBits = value;
			}           
		}

		#endregion
	}
}
