using System.Text;
using System.Collections;
using System.Threading;
using System.Globalization;
using System;
using System.IO;
using System.Diagnostics; 
using System.Runtime.InteropServices;
using System.Security;
using GeoFramework;
using System.ComponentModel;

#if !PocketPC || Framework20
using Microsoft.Win32;
#endif
#if !PocketPC
using System.Security.Permissions;
#endif
#if !PocketPC && Framework20
using Microsoft.Win32.SafeHandles;
#endif

namespace GeoFramework.IO.Serial
{
	/// <summary>
	/// Provides communication services for a serial port.
	/// </summary>
	/// <remarks>
	/// 	<para>This class provides communication services for any serial device. This class
	///     supports all versions of the .NET framework, including Visual Studio.NET 2005 and
	///     the special unsupported "Deploy to My Computer" emulator. All data is provided in
	///     the form of a <strong>Stream</strong>, allowing this class to be used as the data
	///     provider for GPS data interpreters such as <strong>NmeaInterpreter</strong>,
	///     <strong>GarminTextInterpreter</strong> and
	///     <strong>GarminBinaryInterpreter</strong>.</para>
	/// 	<para>This class is typically used when correct port settings are known ahead of
	///     time. For situations where the correct port is not known, use the
	///     <strong>GpsStream</strong> class instead of this class. <strong>GpsStream</strong>
	///     will scan the local computer for GPS devices and bind itself to the first
	///     responding GPS device, whether it's a serial or Bluetooth" device.</para>
	/// 	<para>Like other GeoFrameworks classes, this class uses a "Just In Time" design,
	///     where the port is not actually opened until an attempt is made to read from or
	///     write to the port using the <strong>Read</strong>, <strong>ReadByte</strong>,
	///     <strong>Write</strong> or <strong>WriteByte</strong> methods.</para>
	/// 	<para>Instances of this class are guaranteed to be thread-safe because all
	///     interaction with the serial port is serialized using thread synchronization, with
	///     the exception of asynchronous methods such as <strong>BeginRead</strong> and
	///     <strong>BeginWrite</strong>.</para>
	/// </remarks>
	/// <example>
	///     This example creates a new <strong>SerialStream</strong> on COM4 at 4800 baud, and
	///     assigns it to an <strong>NmeaInterpreter</strong> . 
	///     <code lang="VB" title="[New Example]">
	/// ' Create a serial connection on COM4 at 4800 baud
	/// Dim MyStream As New SerialStream(CommunicationsPort.Com4, BaudRate.Baud4800)
	/// ' Bind the port to a GPS interpreter
	/// Dim MyInterpreter As New NmeaInterpreter(MyStream)
	/// ' And start processing data
	/// MyInterpreter.Start()
	///     </code>
	/// 	<code lang="CS" title="[New Example]">
	/// // Create a serial connection on COM4 at 4800 baud
	/// SerialStream MyStream = new SerialStream(CommunicationsPort.Com4, BaudRate.Baud4800);
	/// // Bind the port to a GPS interpreter
	/// NmeaInterpreter MyInterpreter = new NmeaInterpreter(MyStream);
	/// // And start processing data
	/// MyInterpreter.Start();
	///     </code>
	/// </example>
#if !PocketPC
	[TypeConverter(typeof(ExpandableObjectConverter))]
    [SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode=true)]
    [RegistryPermission(SecurityAction.Assert, Read = @"HKEY_LOCAL_MACHINE\Software\Widcomm\BtConfig\AutoConnect")]
    [RegistryPermission(SecurityAction.Assert, Create = @"HKEY_LOCAL_MACHINE\Software\Widcomm\BtConfig\AutoConnect")]
    [RegistryPermission(SecurityAction.Assert, Write = @"HKEY_LOCAL_MACHINE\Software\Widcomm\BtConfig\AutoConnect")]
#endif
	public sealed class SerialStream : Stream
	{
        #region Is Widcomm Bluetooth supported?

        private static bool pIsWidcommBluetoothSupported = false;

        static SerialStream()
        {
            RegistryKey BluetoothKey = null;
            try
            {
                // Try to open the key.  Is there anything?
                BluetoothKey = Registry.LocalMachine.OpenSubKey(@"Software\Widcomm", false);
                pIsWidcommBluetoothSupported = (BluetoothKey != null);
            }
            catch
            {
            }
            finally
            {
                if (BluetoothKey != null)
                    BluetoothKey.Close();
            }
        }


        #endregion

        #region Private Variables
        private SafeFileHandle pFileHandle; // This value holds the CreateFile handle.
		private COMSTAT pCommunicationStatus;
		private DCB pDCB = new DCB();
		private string pPortName;
		private CommunicationPort pPort;
		private BaudRate pBaudRate;
		private Parity pParity;
		private DataBits pDataBits;
		private StopBits pStopBits;
		private FlowControl pFlowControl;
        private uint pLength = 128;

		private bool pCanRead;
		private bool pCanWrite;
		private bool pIsStreamClosed;
        private TimeSpan pReadTimeout = DefaultReadTimeout;
        private TimeSpan pWriteTimeout = DefaultWriteTimeout;
        private TimeSpan pCommandTimeout = DefaultCommandTimeout;


		// Set Timeouts
		private SerialTimeouts ctm = new SerialTimeouts();
        private string pDeviceName;
        private bool pIsBluetoothDevice;

        // Commonly re-used variables
        uint CharactersRead;
        uint CharactersWritten;
        byte[] SingleByteArray = new byte[1];
		uint ErrorCode;
#if !PocketPC
        NativeOverlapped pOverlapped = new NativeOverlapped();
#else
        IntPtr pOverlapped = new IntPtr(0);
#endif


		#endregion

        #region Static Members

        internal static Hashtable SerialPortSyncLocks = new Hashtable();
        public static TimeSpan DefaultReadTimeout = TimeSpan.FromSeconds(5);
        public static TimeSpan DefaultCommandTimeout = TimeSpan.FromSeconds(10);
        public static TimeSpan DefaultWriteTimeout = TimeSpan.FromSeconds(5);
        /// <summary>
        /// Controls the number of bytes that we assume are ready to be read.
        /// </summary>
        /// <value>An <strong>Integer</strong>.  Defaults to <strong>128</strong>.</value>
        /// <remarks>This property is only in use when GPS.NET has been instructed to assume data is always waiting to be read
        /// from the GPS device.  Some manufacturers erroneously always return zero even if data is available.  A higher value
        /// can improve performance by reducing the number of reads made from a device.  A value which is too high, however, can
        /// introduce latency issues as the buffer becomes filled with old data.  A value equivalent to the length of one or two 
        /// NMEA sentences (approximately 128 bytes) is ideal.</remarks>
        /// <seealso cref="Devices.IsSerialDeviceQueriedBeforeReading"/>
        /// <seealso cref="Length"/>
        public static int DefaultReadBufferSize = 128;
        //		public static readonly SerialStream Com1 = new SerialStream(CommunicationPort.Com1);
        //		public static readonly SerialStream Com2 = new SerialStream(CommunicationPort.Com2);
        //		public static readonly SerialStream Com3 = new SerialStream(CommunicationPort.Com3);
        //		public static readonly SerialStream Com4 = new SerialStream(CommunicationPort.Com4);
        //		public static readonly SerialStream Com5 = new SerialStream(CommunicationPort.Com5);
        //        public static readonly SerialStream Com6 = new SerialStream(CommunicationPort.Com6);
        //        public static readonly SerialStream Com7 = new SerialStream(CommunicationPort.Com7);
        //        public static readonly SerialStream Com8 = new SerialStream(CommunicationPort.Com8);
        //        public static readonly SerialStream Com9 = new SerialStream(CommunicationPort.Com9);
        //        public static readonly SerialStream Com10 = new SerialStream(CommunicationPort.Com10);

        #endregion

		#region  Constructor & Destructor

		/// <summary>Creates a new instance on the specified port.</summary>
		/// <remarks>
		/// This constructor will create a new instance and use default settings of 4800
		/// baud, no parity, eight data bits, one stop bit, and no flow control. This constructor
		/// is typically used when communicating with an NMEA-compliant GPS device.
		/// </remarks>
		/// <param name="port">A <strong>CommunicationPort</strong> value specifying which port to use.</param>
		public SerialStream(CommunicationPort port) 
			: this(port, BaudRate.Baud4800, DataBits.Eight, Parity.None, StopBits.One, FlowControl.None)
		{}

		/// <summary>Creates a new instance on the specified port at the specified baud rate.</summary>
		/// <remarks>
		/// This constructor will create a new instance and set the baud rate. Other settings
		/// are set to the defaults: no parity, eight data bits, one stop bit, and no flow control.
		/// This constructor is typically used when communicating with a GPS device set to a
		/// certain speed, such as a Garimn device, which typically transmits data at 9600
		/// baud.
		/// </remarks>
		/// <example>
		///     This example creates a new <strong>SerialStream</strong> on COM1 at 9600 baud. 
		///     <code lang="VB" title="[New Example]">
		/// ' Create a stream on COM1 at 9600 baud
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1, BaudRate.Baud9600)
		/// ' Bind the stream to a Garmin binary interpreter
		/// Dim Interpreter As New GarminBinaryInterpreter(MyStream)
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create a stream on COM1 at 9600 baud
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1, BaudRate.Baud9600);
		/// // Bind the stream to a Garmin binary interpreter
		/// GarminBinaryInterpreter Interpreter = new GarminBinaryInterpreter(MyStream);
		///     </code>
		/// </example>
		public SerialStream(CommunicationPort port, BaudRate baudRate) 
			: this(port, baudRate, DataBits.Eight, Parity.None, StopBits.One, FlowControl.None)
		{}

		/// <summary>Creates a new instance on the specified port at the specified settings.</summary>
		/// <remarks>
		/// This constructor is typically used for rare cases where data bit, parity, stop
		/// bit, or flow control settings other than the defaults are required. This situation
		/// usually occurs when communicating with older devices. At GeoFrameworks, we haven't yet
		/// encountered a GPS device which required settings other than the defaults.
		/// </remarks>
		/// <example>
		/// 	<code lang="VB" title="[New Example]">
		/// ' Create a stream on COM1 using highly customized settings
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1, _
		///     BaudRate.Baud9600, DataBits.Seven, Parity.Mark, _
		///     StopBits.Two, FlowControl.Hardware)
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create a stream on COM1 using highly customized settings
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1, 
		///     BaudRate.Baud9600, DataBits.Seven, Parity.Mark, StopBits.Two, 
		///     FlowControl.Hardware)
		///     </code>
		/// </example>
		public SerialStream(CommunicationPort port, BaudRate baudRate, DataBits dataBits, Parity parity, StopBits stopBits, FlowControl flowControl)
		{
            // Preserve serial port settings
			pPort = port;

            if (port != CommunicationPort.Unknown)
            {
                int PortNumber = Convert.ToInt32(port);
                if (PortNumber >= 0 && PortNumber < 10)
                {
                    pPortName = "COM" + PortNumber.ToString() + ":";
                }
                else
                {
#if PocketPC
					if(Environment.OSVersion.Platform == PlatformID.WinCE)
					{
						pPortName = @"COM" + PortNumber.ToString() + ":";
					}
					else
					{
						pPortName = @"\\.\COM" + PortNumber.ToString();
					}
#else
                    pPortName = @"\\.\COM" + PortNumber.ToString();
#endif
				}
            }

			pBaudRate = baudRate;
			pParity = parity;
			pDataBits = dataBits;
			pStopBits = stopBits;
			pFlowControl = flowControl;

            Initialize();

#if !PocketPC
			// We don't need to finalize until we actually
			// open a connection. On the PocketPC, messing with
			// the GC a lot causes fatal exceptions, which is
			// why this is only done on the desktop.
			GC.SuppressFinalize(this);
#endif
        }

		public SerialStream(string portName)
			:this(portName, BaudRate.Baud4800, DataBits.Eight, Parity.None, StopBits.One, FlowControl.None)
		{}

		public SerialStream(string portName, BaudRate baudRate)
			:this(portName, baudRate, DataBits.Eight, Parity.None, StopBits.One, FlowControl.None)
		{}


		public SerialStream(string portName, BaudRate baudRate, DataBits dataBits, Parity parity, StopBits stopBits, FlowControl flowControl)
		{
            // If the port name is missing, complain
            if (portName == null || portName.Length == 0)
                throw new ArgumentException("A port name must be specified in order to open a serial connection.  Port names are typically three letters followed by a number, such as \"COM7:\".");

            // Store the speciaiized port name
            pPortName = portName.ToUpper(CultureInfo.InvariantCulture);
            pPort = CommunicationPort.Unknown;

            // Try to interpret the port
            int ComStart = pPortName.IndexOf("COM");
            if (ComStart != -1)
            {
                int ComEnd = pPortName.IndexOf(":", ComStart);
                if (ComEnd != -1)
                {
                    pPortName = pPortName.Substring(ComStart, ComEnd - ComStart).Trim();
                    string InterpretedPort = pPortName.Replace("COM", "").Replace(":", "");
                    try
                    {
                        int PortNumber = Convert.ToInt32(InterpretedPort);
#if PocketPC && !Framework20
				switch(PortNumber)
				{
					case 0:
						pPort = CommunicationPort.Com0;
						break;
					case 1:
						pPort = CommunicationPort.Com1;
						break;
					case 2:
						pPort = CommunicationPort.Com2;
						break;
					case 3:
						pPort = CommunicationPort.Com3;
						break;
					case 4:
						pPort = CommunicationPort.Com4;
						break;
					case 5:
						pPort = CommunicationPort.Com5;
						break;
					case 6:
						pPort = CommunicationPort.Com6;
						break;
					case 7:
						pPort = CommunicationPort.Com7;
						break;
					case 8:
						pPort = CommunicationPort.Com8;
						break;
					case 9:
						pPort = CommunicationPort.Com9;
						break;
					case 10:
						pPort = CommunicationPort.Com10;
						break;
					case 11:
						pPort = CommunicationPort.Com11;
						break;
					case 12:
						pPort = CommunicationPort.Com12;
						break;
					case 13:
						pPort = CommunicationPort.Com13;
						break;
					case 14:
						pPort = CommunicationPort.Com14;
						break;
					case 15:
						pPort = CommunicationPort.Com15;
						break;
					case 16:
						pPort = CommunicationPort.Com16;
						break;
					case 17:
						pPort = CommunicationPort.Com17;
						break;
					case 18:
						pPort = CommunicationPort.Com18;
						break;
					case 19:
						pPort = CommunicationPort.Com19;
						break;
					case 20:
						pPort = CommunicationPort.Com20;
						break;
					case 21:
						pPort = CommunicationPort.Com21;
						break;
					case 22:
						pPort = CommunicationPort.Com22;
						break;
					case 23:
						pPort = CommunicationPort.Com23;
						break;
					case 24:
						pPort = CommunicationPort.Com24;
						break;
					case 25:
						pPort = CommunicationPort.Com25;
						break;
					case 26:
						pPort = CommunicationPort.Com26;
						break;
					case 27:
						pPort = CommunicationPort.Com27;
						break;
					case 28:
						pPort = CommunicationPort.Com28;
						break;
					case 29:
						pPort = CommunicationPort.Com29;
						break;
					case 30:
						pPort = CommunicationPort.Com30;
						break;
				}
#else
                        pPort = (CommunicationPort)Enum.Parse(typeof(CommunicationPort), PortNumber.ToString(), false); 
#endif

                        // If the port is greater than 10, change the port name
                        if (PortNumber < 10)
                        {
                            pPortName = @"COM" + PortNumber.ToString() + ":";
                        }
                        else
                        {
#if PocketPC
							if(Environment.OSVersion.Platform == PlatformID.WinCE)
							{
								pPortName = @"COM" + PortNumber.ToString() + ":";
							}
							else
							{
								pPortName = @"\\.\COM" + PortNumber.ToString();
							}
#else
							pPortName = @"\\.\COM" + PortNumber.ToString();
#endif
                        }

                    }
                    catch
                    {
                    }
                }
            }


			pBaudRate = baudRate;
			pParity = parity;
			pDataBits = dataBits;
			pStopBits = stopBits;
			pFlowControl = flowControl;

            Initialize();

#if !PocketPC
            // We don't need to finalize until we actually
            // open a connection. On the PocketPC, messing with
            // the GC a lot causes fatal exceptions, which is
            // why this is only done on the desktop.
            GC.SuppressFinalize(this);
#endif
		}
       
		~SerialStream()
		{
			Close();
        }

        internal void SetDeviceName()
        {
            #region Bluetooth Name Determination (Pass 1)

            if (pDeviceName == null)
            {
                RegistryKey AutoConnect = null;
                try
                {
                    AutoConnect = Registry.LocalMachine.OpenSubKey(@"Software\Widcomm\BTConfig\AutoConnect\" + Convert.ToInt32(pPort).ToString(CultureInfo.InvariantCulture).PadLeft(4, '0'), false);

                    if (AutoConnect != null)
                    {
                        // Read the "BDName" value for the device name
                        object FriendlyObject = AutoConnect.GetValue("BDName");
                        byte[] FriendlyBytes = FriendlyObject as byte[];
                        string FriendlyString = FriendlyObject as string;
                        if (FriendlyString != null)
                        {
                            // The device name is a string
                            pDeviceName = FriendlyString;
                            pIsBluetoothDevice = true;
                        }
                        else if (FriendlyBytes != null)
                        {
                            // The device name may in fact be a byte array. Convert it
                            // while stripping off the null at the ned
                            pDeviceName = System.Text.ASCIIEncoding.ASCII.GetString(FriendlyBytes, 0, FriendlyBytes.Length - 1);
                            pIsBluetoothDevice = true;
                        }
                    }
                }
                catch
                {
                    // Ignore problems!
                }
                finally
                {
                    if (AutoConnect != null)
                        AutoConnect.Close();
                }
            }

            #endregion

            #region Bluetooth Name Determination (Pass 2)

            if (pDeviceName == null)
            {
                // Look for WidComm Bluetooth devices
                RegistryKey Widcomm = null;
                try
                {
#if !Framework10
                    if (Environment.OSVersion.Platform == PlatformID.WinCE)
                    {
                        Widcomm = Registry.LocalMachine.OpenSubKey(@"Software\Widcomm\Connections", false);
                    }
                    else
                    {
#endif
                        Widcomm = Registry.CurrentUser.OpenSubKey(@"Software\Widcomm\Connections", false);
#if !Framework10
                    }
#endif

                    if (Widcomm != null)
                    {
                        // Enumerate all WIDCOMM devices
                        string[] SubKeys = Widcomm.GetSubKeyNames();
                        foreach (string subKey in SubKeys)
                        {
                            // Open the sub-key  
                            RegistryKey WidcommConnectionKey = Widcomm.OpenSubKey(subKey);
                            try
                            {
                                // Skip if this key can't be opened
                                if (WidcommConnectionKey == null)
                                    continue;

                                // Look for "ComPortNumber"
                                string ServiceGuid = WidcommConnectionKey.GetValue("GUID") as string;
                                int DevicePortNumber = Convert.ToInt32(WidcommConnectionKey.GetValue("ComPortNumber"));
                                object FriendlyNameValue = WidcommConnectionKey.GetValue("BDName");
                                string FriendlyName = null;

                                if (FriendlyNameValue != null)
                                {
#if !Framework10
                                    if (Environment.OSVersion.Platform == PlatformID.WinCE)
                                    {
                                        FriendlyName = FriendlyNameValue as string;
                                    }
                                    else
                                    {
#endif
                                        byte[] FriendlyNameBytes = (byte[])FriendlyNameValue;
                                        FriendlyName = System.Text.ASCIIEncoding.ASCII.GetString(FriendlyNameBytes, 0, FriendlyNameBytes.Length).Replace("\0", "");
#if !Framework10
                                    }
#endif
                                }

                                // If the service GUID is right, use it
                                if (ServiceGuid == "{00001101-0000-1000-8000-00805F9B34FB}")
                                {
                                    // Set the port
#if PocketPC && !Framework20
									CommunicationPort DeviceComPort = CommunicationPort.Com0;
									switch(DevicePortNumber)
									{
										case 0:
											DeviceComPort = CommunicationPort.Com0;
											break;
										case 1:
											DeviceComPort = CommunicationPort.Com1;
											break;
										case 2:
											DeviceComPort = CommunicationPort.Com2;
											break;
										case 3:
											DeviceComPort = CommunicationPort.Com3;
											break;
										case 4:
											DeviceComPort = CommunicationPort.Com4;
											break;
										case 5:
											DeviceComPort = CommunicationPort.Com5;
											break;
										case 6:
											DeviceComPort = CommunicationPort.Com6;
											break;
										case 7:
											DeviceComPort = CommunicationPort.Com7;
											break;
										case 8:
											DeviceComPort = CommunicationPort.Com8;
											break;
										case 9:
											DeviceComPort = CommunicationPort.Com9;
											break;
										case 10:
											DeviceComPort = CommunicationPort.Com10;
											break;
										case 11:
											DeviceComPort = CommunicationPort.Com11;
											break;
										case 12:
											DeviceComPort = CommunicationPort.Com12;
											break;
										case 13:
											DeviceComPort = CommunicationPort.Com13;
											break;
										case 14:
											DeviceComPort = CommunicationPort.Com14;
											break;
										case 15:
											DeviceComPort = CommunicationPort.Com15;
											break;
										case 16:
											DeviceComPort = CommunicationPort.Com16;
											break;
										case 17:
											DeviceComPort = CommunicationPort.Com17;
											break;
										case 18:
											DeviceComPort = CommunicationPort.Com18;
											break;
										case 19:
											DeviceComPort = CommunicationPort.Com19;
											break;
										case 20:
											DeviceComPort = CommunicationPort.Com20;
											break;
										case 21:
											DeviceComPort = CommunicationPort.Com21;
											break;
										case 22:
											DeviceComPort = CommunicationPort.Com22;
											break;
										case 23:
											DeviceComPort = CommunicationPort.Com23;
											break;
										case 24:
											DeviceComPort = CommunicationPort.Com24;
											break;
										case 25:
											DeviceComPort = CommunicationPort.Com25;
											break;
										case 26:
											DeviceComPort = CommunicationPort.Com26;
											break;
										case 27:
											DeviceComPort = CommunicationPort.Com27;
											break;
										case 28:
											DeviceComPort = CommunicationPort.Com28;
											break;
										case 29:
											DeviceComPort = CommunicationPort.Com29;
											break;
										case 30:
											DeviceComPort = CommunicationPort.Com30;
											break;										
									}
#else
                                    CommunicationPort DeviceComPort = (CommunicationPort)Enum.Parse(typeof(CommunicationPort), DevicePortNumber.ToString(), false);
#endif

                                    // Was it found?
                                    if (pPort == DeviceComPort)
                                    {
                                        // Set the friendly name of the device
                                        pDeviceName = FriendlyName;

                                        // And flag it as a Bluetooth serial port
                                        pIsBluetoothDevice = true;
                                    }
                                }
                            }
                            catch
                            {
                            }
                            finally
                            {
                                if (WidcommConnectionKey != null)
                                    WidcommConnectionKey.Close();
                            }
                        }
                    }
                }
                catch
                {
                }
                finally
                {
                    if (Widcomm != null)
                        Widcomm.Close();
                }
            }
            #endregion

            #region Bluetooth Name Determination (Pass 3)


            // Get the device name (?)
            if (pDeviceName == null)
            {
                RegistryKey IncomingPort = null;
                try
                {
                    IncomingPort = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Bluetooth\Serial\Ports", false);
                    if (IncomingPort != null)
                    {
                        // Enumerate sub-keys
                        foreach (string subKey in IncomingPort.GetSubKeyNames())
                        {
                            // If it's "Incoming Port", skip it
                            if (subKey == "Incoming")
                                continue;

                            // Otherwise, this is a device!  
                            RegistryKey DeviceKey = null;
                            try
                            {
                                DeviceKey = IncomingPort.OpenSubKey(subKey, false);
                                if (DeviceKey != null)
                                {
                                    // Get the port name
                                    string BluetoothPortName = DeviceKey.GetValue("Port") as string;
                                    if (BluetoothPortName == null)
                                        return;

                                    // Convert the port name to an integer
                                    int PortNumber = Convert.ToInt32(BluetoothPortName.Substring(3));

                                    // Do we have a match?
                                    if ((int)pPort == PortNumber)
                                    {
                                        // Set as a bluetooth device
                                        pIsBluetoothDevice = true;

                                        // Yes.  Look up the name!
                                        RegistryKey FriendlyNameKey = null;
                                        try
                                        {
                                            FriendlyNameKey = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\Microsoft\Bluetooth\Device\" + subKey.Substring(4), false);
                                            if (FriendlyNameKey != null)
                                            {
                                                string FriendlyName = FriendlyNameKey.GetValue("name") as string;
                                                if (FriendlyName != null)
                                                    pDeviceName = FriendlyName;
                                            }
                                        }
                                        catch
                                        {
                                        }
                                        finally
                                        {
                                            if (FriendlyNameKey != null)
                                                FriendlyNameKey.Close();
                                        }
                                    }
                                }
                            }
                            catch
                            {
                            }
                            finally
                            {
                                if (DeviceKey != null)
                                    DeviceKey.Close();
                            }
                        }
                    }
                }
                catch
                {
                }
                finally
                {
                    if (IncomingPort != null)
                        IncomingPort.Close();
                }
            }
            #endregion

            #region Bluetooth Name Determination (Pass 4)

#if PocketPC
            if (pDeviceName == null)
            {
                RegistryKey AutoConnect = null;
                try
                {
                    AutoConnect = Registry.CurrentUser.OpenSubKey(@"Software\Widcomm\BTConfig\AutoConnect\" + Convert.ToInt32(pPort).ToString(CultureInfo.InvariantCulture).PadLeft(4, '0'), false);
                

                    if (AutoConnect != null)
                    {
                        // Read the "BDName" value for the device name
                        string FriendlyName = AutoConnect.GetValue("BDName") as string;
                        if (FriendlyName != null)
                        {
                            pDeviceName = FriendlyName;
                            pIsBluetoothDevice = true;
                        }
                    }
                }
                catch
                {
                    // Ignore problems!
                }
                finally
                {
                    if (AutoConnect != null)
                        AutoConnect.Close();
                }
            }

#endif
            #endregion
        }

        private void Initialize()
        {           

            #region DCB Configuration

            /* The DCB controls the settings for the serial port.  Configure it
			 * now with the desired settings.
			 */

            pDCB.XonLimit = 2048;
            pDCB.XoffLimit = 512;
            pDCB.XonChar = 0x11; // Ctrl-Q
            pDCB.XoffChar = 0x13; // Ctrl-S
            pDCB.DCBlength = 28;
            pDCB.Binary = pDCB.DtrControl = pDCB.TXContinueOnXoff = 1;

            // Set the DCB baud rate
            pDCB.BaudRate = (uint)(int)pBaudRate;

            // Set the number of data bits
            pDCB.ByteSize = (byte)(int)pDataBits;

            // Set the correct stop bit
            switch (pStopBits)
            {
                case GeoFramework.IO.Serial.StopBits.One:
                    pDCB.StopBits = 0;
                    break;
                case GeoFramework.IO.Serial.StopBits.OnePointFive:
                    pDCB.StopBits = 1;
                    break;
                case GeoFramework.IO.Serial.StopBits.Two:
                    pDCB.StopBits = 2;
                    break;
            }


            // Set the correct parity
            switch (pParity)
            {
                case GeoFramework.IO.Serial.Parity.Even:
                    pDCB.Parity = 2;
                    break;
                case GeoFramework.IO.Serial.Parity.Mark:
                    pDCB.Parity = 3;
                    break;
                case GeoFramework.IO.Serial.Parity.None:
                    pDCB.Parity = 0;
                    break;
                case GeoFramework.IO.Serial.Parity.Odd:
                    pDCB.Parity = 1;
                    break;
                case GeoFramework.IO.Serial.Parity.Space:
                    pDCB.Parity = 4;
                    break;
            }

            // Update fParity as well
            pDCB.fParity = (pParity > 0) ? 1U : 0U;

            // Also configure first-time DCB information
            // Flow control
            pDCB.OutxCtsFlow = pDCB.AbortOnError = (pFlowControl == FlowControl.Hardware) ? 1U : 0U;
            pDCB.OutX = pDCB.InX = (pFlowControl == FlowControl.XOnXOff) ? 1U : 0U;
            pDCB.RtsControl = (pFlowControl == FlowControl.Hardware) ? 2U : 1U;

            #endregion

            #region Timeout information

            // Changes setup on the fly
            ctm.ReadIntervalTimeout = 0xFFFF; //(int)pReadTimeout.TotalMilliseconds;
            ctm.ReadTotalTimeoutMultiplier = 50;
            ctm.ReadTotalTimeoutConstant = (int)pReadTimeout.TotalMilliseconds;
            ctm.WriteTotalTimeoutMultiplier = 50;
            ctm.WriteTotalTimeoutConstant = (int)pWriteTimeout.TotalMilliseconds;

            #endregion

            // Get the name of the device
            SetDeviceName();
        }

#if Framework20
        public override bool CanTimeout
        {
            get
            {
                return true;
            }
        }
#endif

		#endregion



		#region Public Members

		/// <summary>Returns the communications port used by the stream.</summary>
		/// <value>A <strong>CommunicationPort</strong> value.</value>
		/// <remarks>
		/// This property is read only. This property is set during construction of the
		/// stream.
		/// </remarks>
		public CommunicationPort Port
		{
			get
			{
				return pPort;
			}
		}

		/// <summary>
		/// Returns the name of the port used to open connections.
		/// </summary>
		/// <remarks>This property behaves very similar to the Port property,
		/// returning the communication port, except as a string instead of a CommunicationPort
		/// enumeration.  In some situations, a special port name such as "GPD1:" or any other value
		/// than "COM:" must be used.  For this reason, the PortName property should be used if
		/// the current port must be displayed in a user interface.</remarks>
		public string PortName
		{
			get
			{
				return pPortName;
			}
		}

		/// <summary>Returns the communication speed for the serial device.</summary>
		/// <value>A <strong>BaudRate</strong> value.</value>
		/// <remarks>
		/// This property is read only. This property is set during construction of the
		/// stream.
		/// </remarks>
		public BaudRate BaudRate
		{
			get
			{
				return pBaudRate;
			}
			set
			{
				if(pBaudRate == value) 
                    return;
				pBaudRate = value;

                // Set the DCB baud rate
                pDCB.BaudRate = (uint)(int)pBaudRate;

				ConfigureDCB();
			}
		}

		/// <remarks>
		/// This property is read only. This property is set during construction of the
		/// stream.
		/// </remarks>
		public Parity Parity
		{
			get
			{
				return pParity;
			}
			set
			{
				if(pParity == value) 
                    return;
				pParity = value;

                // Set the correct parity
                switch (pParity)
                {
                    case GeoFramework.IO.Serial.Parity.Even:
                        pDCB.Parity = 2;
                        break;
                    case GeoFramework.IO.Serial.Parity.Mark:
                        pDCB.Parity = 3;
                        break;
                    case GeoFramework.IO.Serial.Parity.None:
                        pDCB.Parity = 0;
                        break;
                    case GeoFramework.IO.Serial.Parity.Odd:
                        pDCB.Parity = 1;
                        break;
                    case GeoFramework.IO.Serial.Parity.Space:
                        pDCB.Parity = 4;
                        break;
                }

                // Update fParity as well
                pDCB.fParity = (pParity > 0) ? 1U : 0U;

				ConfigureDCB();
			}
		}

		/// <remarks>
		/// This property is read only. This property is set during construction of the
		/// stream.
		/// </remarks>
		public DataBits DataBits
		{
			get
			{
				return pDataBits;
			}
			set
			{
				if(pDataBits == value) 
                    return;
				pDataBits = value;

                // Set the number of data bits
                pDCB.ByteSize = (byte)(int)pDataBits;

				ConfigureDCB();
            }
		}

		/// <remarks>
		/// This property is read only. This property is set during construction of the
		/// stream.
		/// </remarks>
		public FlowControl FlowControl
		{
			get
			{
				return pFlowControl;
			}
			set
			{
				if(pFlowControl == value) 
                    return;
				pFlowControl = value;

                // Update the DCB with flow control information
                pDCB.OutxCtsFlow = pDCB.AbortOnError = (pFlowControl == FlowControl.Hardware) ? 1U : 0U;
                pDCB.OutX = pDCB.InX = (pFlowControl == FlowControl.XOnXOff) ? 1U : 0U;
                pDCB.RtsControl = (pFlowControl == FlowControl.Hardware) ? 2U : 1U;

				ConfigureDCB();
            }
		}

		/// <remarks>
		/// This property is read only. This property is set during construction of the
		/// stream.
		/// </remarks>
		public StopBits StopBits
		{
			get
			{
				return pStopBits;
			}
			set
			{
				if(pStopBits == value) 
                    return;
				pStopBits = value;

                // Set the correct stop bit
                switch (pStopBits)
                {
                    case GeoFramework.IO.Serial.StopBits.One:
                        pDCB.StopBits = 0;
                        break;
                    case GeoFramework.IO.Serial.StopBits.OnePointFive:
                        pDCB.StopBits = 1;
                        break;
                    case GeoFramework.IO.Serial.StopBits.Two:
                        pDCB.StopBits = 2;
                        break;
                }

				ConfigureDCB();
			}
		}

		/// <summary>Returns the name of the port and its current baud rate.</summary>
		/// <returns>
		/// A <strong>String</strong> in the form "<em>[Port]</em> at [<em>baud]</em>
		/// baud."
		/// </returns>
		/// <remarks>This method is used to return the friendly name of the port.</remarks>
		/// <example>
		///     This example opens a new <strong>SerialStream</strong> on COM1, then outputs its
		///     friendly name. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// Debug.WriteLine("This stream is " &amp; MyStream.ToString())
		/// ' Output: This stream is COM1 at 4800 baud
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// Console.WriteLine("This stream is " + MyStream.ToString());
		/// // Output: This stream is COM1 at 4800 baud
		///     </code>
		/// </example>
		public override string ToString()
		{
            if(pDeviceName == null)
                return pPortName + " at " + (int)pBaudRate + " baud";
            else
			    return pDeviceName + " at " + (int)pBaudRate + " baud";
		}

		/// <summary>Indicates if read operations are allowed for the serial device.</summary>
		/// <value>
		/// A <strong>Boolean</strong>, <strong>True</strong> if read operations are
		/// allowed.
		/// </value>
		/// <remarks>
		/// This property will become <strong>True</strong> when a connection attempt to a
		/// device is successful, then become <strong>False</strong> when the port is
		/// closed.
		/// </remarks>
		public override bool CanRead
		{
			get
			{
				if (!pIsStreamClosed && !IsOpen)
					return TryOpen(false);

				return pCanRead;
			}
		}

		/// <value>
		/// A <strong>Boolean</strong>, <strong>True</strong> if write operations are
		/// allowed.
		/// </value>
		/// <summary>Indicates if write operations are allowed for the serial device.</summary>
		public override bool CanWrite
		{
			get
			{
				if (!pIsStreamClosed && !IsOpen)
					return TryOpen(false);

				return pCanWrite;
			}
		}

		/// <remarks>
		/// This property is always <strong>False</strong> because seek operations are not
		/// supported by this class. Read operations from a serial device occur in a forward-only
		/// manner.
		/// </remarks>
		/// <summary>Indicates if the current stream position can be changed.</summary>
		/// <value>A <strong>Boolean</strong>, always <strong>False</strong>.</value>
		public override bool CanSeek
		{
			get
			{
				return false;
			}
		}

		/// <summary>Reads all available characters from the stream.</summary>
		/// <returns>An array of bytes containing all available data.</returns>
		/// <remarks>
		/// This method is typically used to add available data to an existing array for the
		/// purposes of buffering, where it is later parsed.
		/// </remarks>
		/// <example>
		///     This example opens a <strong>SerialStream</strong> on COM1 and reads any available
		///     data. 
		///     <code lang="VB" title="[New Example]">
		/// ' Create a new serial stream on COM1 at 4800 baud
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Read all available bytes into a buffer
		/// Dim Buffer As Byte() = MyStream.ReadToEnd()
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create a new serial stream on COM1 at 4800 baud
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Read all available bytes into a buffer
		/// byte[] Buffer = MyStream.ReadToEnd();
		///     </code>
		/// </example>
		public byte[] ReadToEnd()
		{
			byte[] Result = new byte[(int)Length];
			Read(Result, 0, Result.Length);
			return Result;
		}

		/// <summary>Reads the specified number of bytes into an array.</summary>
		/// <returns>An array of bytes read from the serial device.</returns>
		/// <remarks>
		/// This method is used to read data from the serial port. Typically, the
		/// <strong>Length</strong> property is examined to determine how many bytes to read before
		/// this method is called.
		/// </remarks>
		/// <example>
		///     This example will open a <strong>SerialStream</strong> on COM1, examine the number
		///     of bytes available, then read them from the device. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Is any data available on the device?
		/// Dim AvailableByteCount As Long = MyStream.Length
		/// If AvailableByteCount &gt; 0 Then
		///     ' Yes. Read any available bytes into an array
		///     Dim Buffer As New Byte(AvailableByteCount)
		///     MyStream.Read(Buffer, 0, Buffer.Length)
		/// End If
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Is any data available on the device?
		/// long AvailableByteCount = MyStream.Length;
		/// if(AvailableByteCount &gt; 0)
		/// {
		///     // Yes. Read any available bytes into an array
		///     byte[] Buffer = new Byte[AvailableByteCount];
		///     MyStream.Read(Buffer, 0, Buffer.Length);
		/// }
		///     </code>
		/// </example>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public override int Read(byte[] buffer, int offset, int count)
		{
			// Is there an open port?
            if (!IsOpen)
            {
                if (pIsStreamClosed)
                    throw new ObjectDisposedException(GetType().ToString(), "A read operation cannot be performed because the serial stream is closed.");
                if (!TryOpen(true))
                    return -1;
            }

			// Get a buffer for reading
            byte[] ReadBuffer = new byte[count];

            // Read it in
#if !PocketPC
            bool Success = NativeMethods.ReadFile(pFileHandle, ReadBuffer, (uint)count, out CharactersRead, ref pOverlapped);
#else
            bool Success = NativeMethods.ReadFile(pFileHandle, ReadBuffer, (uint)count, out CharactersRead, pOverlapped);
#endif

            // And return the number of bytes read
			if(Success && CharactersRead > 0)
            {
                // Copy the buffer to the parameter
                Array.Copy(ReadBuffer, 0, buffer, offset, count);

                return Convert.ToInt32(CharactersRead);
            }
            else
            {
				return -1;
            }
		}

		/// <summary>Reads a single byte from the serial device.</summary>
		/// <returns>
		/// A <strong>Byte</strong> read from the serial port if one is available, otherwise
		/// -1 is returned.
		/// </returns>
		/// <remarks>
		/// This method is typically the most common method used to read data from a serial
		/// device. If no data is available to read, this method returns negative one (-1).
		/// </remarks>
		/// <example>
		///     This example reads any available data on COM1 into an array one byte at a time. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Keep reading bytes until -1 is returned
		/// Dim ReceivedByte As Byte = MyStream.ReadByte()
		/// Do While ReceivedByte &lt;&gt; -1
		///     ' Write the byte to the console
		///     Debug.Write(ReceivedByte.ToString())
		///     ' And read another byte
		///     ReceivedByte = MyStream.ReadByte()
		/// Loop
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Keep reading bytes until -1 is returned
		/// byte ReceivedByte = MyStream.ReadByte();
		/// while(ReceivedByte != -1)
		/// {
		///     // Write the byte to the console
		///     Console.Write(ReceivedByte.ToString());
		///     // And read another byte
		///     ReceivedByte = MyStream.ReadByte();
		/// }
		///     </code>
		/// </example>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public override int ReadByte()
		{
			// Is there an open port?
            if (!pCanRead)
            {
                if (pIsStreamClosed)
                    throw new ObjectDisposedException("SerialStream", "A read operation cannot be performed because the serial stream is closed.");
                if (!TryOpen(true))
                    return -1;
            }

#if !PocketPC
			NativeMethods.ReadFile(pFileHandle, SingleByteArray, 1, out CharactersRead, ref pOverlapped);
#else
			NativeMethods.ReadFile(pFileHandle, SingleByteArray, 1, out CharactersRead, pOverlapped);
#endif

			if(CharactersRead > 0)
                return SingleByteArray[0];
            else
				return -1;
		}

		/// <summary>
		/// Reads a single line of data from the serial device.
		/// </summary>
		/// <returns>A <strong>String</strong> containing one line of data.</returns>
		/// <remarks>
		/// This method will read data from the device, byte by byte, until an end-of-line
		/// character is recieved. If over 16,384 bytes of data are read from the device without
		/// receiving an end-of-line character, an exception is thrown. An exception is also thrown
		/// if an end-of-line character is not received in the amount of time specified in
		/// <strong>ReadTimeout</strong>.
		/// </remarks>
		/// <example>
		///     This example uses the <strong>ReadLine</strong> method to read ten lines of data
		///     from COM1. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Write ten lines of data to the console
		/// Dim Count As Integer
		/// For Count = 1 To 10
		///     Console.WriteLine(MyStream.ReadLine())
		/// Next Count
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Write ten lines of data to the console
		/// for(int Count = 1; Count &lt;= 10; Count++)
		///     Console.WriteLine(MyStream.ReadLine());
		///     </code>
		/// </example>
		public string ReadLine()
		{
			return ReadLine(true);
		}

		/// <summary>
		/// Reads a single line of data from the serial device.
		/// </summary>
		/// <returns>A <strong>String</strong> containing one line of data.</returns>
		/// <remarks>
		/// This method will read data from the device, byte by byte, until an end-of-line
		/// character is recieved. If over 16,384 bytes of data are read from the device without
		/// receiving an end-of-line character, an exception is thrown. An exception is also thrown
		/// if an end-of-line character is not received in the amount of time specified in
		/// <strong>ReadTimeout</strong>.
		/// </remarks>
		/// <example>
		///     This example uses the <strong>ReadLine</strong> method to read ten lines of data
		///     from COM1. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Write ten lines of data to the console
		/// Dim Count As Integer
		/// For Count = 1 To 10
		///     Console.WriteLine(MyStream.ReadLine())
		/// Next Count
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Write ten lines of data to the console
		/// for(int Count = 1; Count &lt;= 10; Count++)
		///     Console.WriteLine(MyStream.ReadLine());
		///     </code>
		/// </example>
		public string ReadLine(bool allowExceptions)
		{
			try
			{
				if (pIsStreamClosed)
					throw new ObjectDisposedException(GetType().ToString(), "A wait for data operation cannot begin because the stream has been closed.");
				// We have data!  Now read it in.
				StringBuilder Result = new StringBuilder((int)pLength);
				int BadByteCount = 0;
				DateTime TimeoutDateTime = DateTime.MinValue;
				// And loop while we're able to read
				while(CanRead)
				{
					// Read in a byte
					int Data = ReadByte();

					// Is the data valid?
					if(Data > 0)
					{
						// Yes. Add it to the list
						Result.Append((char)Data);
						// If it's a line feed, we're done
						if(Data == 10)
							return Result.ToString();
						// If the length is OK, keep going
						if(Result.Length <= pLength)
							continue;
						// The buffer is full of bad data
						if(allowExceptions)
							throw new FormatException("The incoming buffer overflowed while attempting to read a line of data.  The baud rate may be incorrect.  Try changing the baud rate or increase the buffer size using the SetLength() property.");
						else
							return Result.ToString();
					}

					/* If we get here, we encountered either a null (0)
					 * or -1, which means no data was available.  Increase
					 * the "bad byte" count.
					 */
					BadByteCount++;

					// If the count is over the buffer, flag failure
					if (BadByteCount > 512)
					{
						if (allowExceptions)
							throw new TimeoutException("The buffer has become full trying to read a single line of data.  The baud rate may be incorrect.  Try changing the baud rate, or use the SetLength method to increase the size of the buffer.");
						else
							return null;
					}

					// If we have no timeout set, set it now
					if(TimeoutDateTime.Equals(DateTime.MinValue))
						TimeoutDateTime = DateTime.Now.Add(pReadTimeout);

					// If the read timeout has passed, throw a timeout
					if(DateTime.Now > TimeoutDateTime)
					{
						if(allowExceptions)
							// No. flag an exception
							throw new TimeoutException("The serial device did not respond in a timely manner with any data.");
						else
							return null;
					}

					// If we get here, we have a bad byte, but not too
					// many to get in a fuss over.  Sleep for a moment
					Thread.Sleep(100);
					Thread.Sleep(0);
				}
				return null;

//				if (pIsStreamClosed)
//					throw new ObjectDisposedException(GetType().ToString(), "Data cannot be read from the device because the stream has been closed.");
//				// Specify a read timeout
//				DateTime TimeoutTime = DateTime.Now.Add(pReadTimeout);
//				// Read until a CrLf is encountered
//				StringBuilder Result = new StringBuilder();
//				int Character;
//				while(TimeoutTime > DateTime.Now && CanRead)
//				{
//					// Wait for data to arrive
//					if (!WaitForData(pReadTimeout))
//						break;
//
//					// Reset the timeout
//					TimeoutTime = DateTime.Now.Add(pReadTimeout);
//
//					// Read information in until we have a complete string
//					Character = ReadByte();
//					
//					// If it's not -1, it's real data
//					while(Character != -1)
//					{
//						char char1 = (char)Character;
//						// Append the character to the sentence
//						Result.Append(char1);
//						// If there's a carriage return, we have a complete sentence!
//						if(char1 == '\n') //Result.IndexOf("\r\n") != -1)
//						{
//							return Result.Replace("\r\n", "").ToString();
//						}
//						// If the buffer is huge, the data may not be recognizable
//						if(Result.Length > 16384)
//							throw new InvalidOperationException("A large amount of data was read from the device but no end-of-line marker was found.  the port may not be set to the correct baud rate.");	
//						// Read in another byte
//						Character = ReadByte();
//						// And sleep for a tiny bit if there's no data
//						if(Character == -1)
//							Thread.Sleep(0); // (50);
//					}
//					// If we get here, nothing is responding.  Sleep a bit
//					// to give other threads a chance to work.
//					//Thread.Sleep(10);
//				}
//				// If we've timed out, throw an exception
//				if(TimeoutTime < DateTime.Now && CanRead)
//				{
//					// And flag an exception
//					throw new TimeoutException("The GPS device did not respond in a timely manner with any data.");
//				}
//				else
//					return Result.ToString();
			}
#if !PocketPC
			catch(ThreadAbortException)
			{
				// Just exit
				return "";
			}
#endif
			catch
			{
				throw;
			}
			//			finally
			//			{
			//				// And let other threads work
			//				//Thread.Sleep(10);
			//			}
		}

		/// <summary>Blocks execution until data has been received from the device.</summary>
		/// <returns>
		/// A <strong>Boolean</strong>, <strong>True</strong> if data is available from the
		/// device, otherwise <strong>False</strong>.
		/// </returns>
		/// <remarks>
		/// This method will wait for the amount of time specified in
		/// <strong>ReadTimeout</strong> for data to become available. If data is detected, a value
		/// of <strong>True</strong> is returned. If no data is received, or the operation is
		/// aborted due to the port being closed or a thread aborting, a value of
		/// <strong>False</strong> is returned.
		/// </remarks>
		/// <example>
		///     This example opens a connection to COM1, waits for data to become available, then
		///     reads a line of data and writes it to the Output window. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// If MyStream.WaitforData() Then
		///     Debug.WriteLine(MyStream.ReadLine())
		/// Else
		///     Debug.WriteLine("No data is available.")
		/// End If
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// if(MyStream.WaitforData())
		///     Debug.WriteLine(MyStream.ReadLine());
		/// else
		///     Debug.WriteLine("No data is available.");
		///     </code>
		/// </example>
		public bool WaitForData()
		{
			return WaitForData(pReadTimeout);
		}

		/// <remarks>
		/// This method will wait for the amount of time specified in the
		/// <strong>Timeout</strong> parameter for data to become available. If data is detected, a
		/// value of <strong>True</strong> is returned. If no data is received, or the operation is
		/// aborted due to the port being closed or a thread aborting, a value of
		/// <strong>False</strong> is returned.
		/// </remarks>
		/// <summary>
		/// Blocks execution for the specified time until data has been received from the
		/// device.
		/// </summary>
		/// <returns>
		/// A <strong>Boolean</strong>, <strong>True</strong> if data is available from the
		/// device, otherwise <strong>False</strong>.
		/// </returns>
		/// <example>
		///     This example opens a connection to COM1, waits a maximum of three seconds for data
		///     to become available, then reads a line of data and writes it to the Output window. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// If MyStream.WaitforData(TimeSpan.FromSeconds(3)) Then
		///     Debug.WriteLine(MyStream.ReadLine())
		/// Else
		///     Debug.WriteLine("No data is available.")
		/// End If
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// if(MyStream.WaitforData(TimeSpan.FromSeconds(3)))
		///     Debug.WriteLine(MyStream.ReadLine());
		/// else
		///     Debug.WriteLine("No data is available.");
		///     </code>
		/// </example>
		public bool WaitForData(TimeSpan timeout)
		{
			// Open the port if necessary
			if(!IsOpen)
				if (!TryOpen(true))
					return false;
			// Calculate the time to stop waiting
			DateTime AbortTime = DateTime.Now.Add(timeout);
			// And wait for either the timeout, data to arrive, or aborted processing
			while(CanRead && Length == 0 && DateTime.Now <= AbortTime)
			{
				Thread.Sleep(0); //(50);
			}
			// Return whether there is data waiting
			return Length != 0;
		}

#if !PocketPC

		private delegate int ReadDelegate(byte[] buffer, int offset, int count);
		private delegate void WriteDelegate(byte[] buffer, int offset, int count);

		private ReadDelegate pReadDelegate;
		private WriteDelegate pWriteDelegate;

		/// <example>
		/// 	<code lang="CS" title="[New Example]">
		/// 	</code>
		/// 	<code lang="CS" title="[New Example]">
		/// 	</code>
		/// </example>
		/// <summary>Begins an asynchronous read operation.</summary>
		/// <returns>
		/// An <strong>IAsyncResult</strong> object used to monitor the asynchronous read for
		/// completion.
		/// </returns>
		/// <remarks>
		/// This method will call the <strong>Read</strong> method using an asynchronous
		/// delegate. This method is generally not recommended because of issues that can occur if
		/// multiple calls to <strong>BeginRead</strong> are made before the prevoius operation
		/// completes. Consider using the <strong>Read</strong> method on a separate thread
		/// instead.
		/// </remarks>
		public override IAsyncResult BeginRead(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			// Invoke the read operation
			pReadDelegate = new ReadDelegate(Read);
			return pReadDelegate.BeginInvoke(buffer, offset, count, callback, state);
		}

		/// <summary>Completes an asynchronous read operation.</summary>
		/// <returns>
		/// An <strong>Integer</strong> indicating the number of bytes read, or -1 if no data
		/// was available.
		/// </returns>
		/// <remarks>
		/// This method is generally not recommended because of issues that can occur if
		/// multiple calls to <strong>BeginRead</strong> are made before the prevoius operation
		/// completes. Consider using the <strong>Read</strong> method on a separate thread
		/// instead.
		/// </remarks>
		/// <param name="asyncResult">
		/// An <strong>IAsyncResult</strong> object returned during a call to
		/// <strong>BeginRead</strong>.
		/// </param>
		public override int EndRead(IAsyncResult asyncResult)
		{
			while(!asyncResult.IsCompleted)
			{
				Thread.Sleep(50);
			}
			return pReadDelegate.EndInvoke(asyncResult);
		}

		/// <remarks>
		/// This method will call the <strong>Write</strong> method using an asynchronous
		/// delegate. This method is generally not recommended because of issues that can occur if
		/// multiple calls to <strong>BeginWrite</strong> are made before the prevoius operation
		/// completes. Consider using the <strong>Write</strong> method on a separate thread
		/// instead.
		/// </remarks>
		/// <summary>Begins an asynchronous write operation.</summary>
		/// <returns>
		/// An <strong>IAsyncResult</strong> object used to monitor the state of the write
		/// operation.
		/// </returns>
		/// <param name="buffer">An array of bytes to transmit to the serial device.</param>
		/// <param name="offset">
		/// An <strong>Integer</strong> indicating the index of the first byte in the array
		/// to transmit.
		/// </param>
		/// <param name="count">An <strong>Integer</strong> indicating the number of bytes to transmit.</param>
		/// <param name="callback">A <strong>Delegate</strong> to execute when the write operation completes.</param>
		/// <param name="state">An <strong>Object</strong> indicating miscellaneous state for the write.</param>
		public override IAsyncResult BeginWrite(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			// Invoke the Write operation
			pWriteDelegate = new WriteDelegate(Write);
			return pWriteDelegate.BeginInvoke(buffer, offset, count, callback, state);
		}

		/// <remarks>
		/// This method will call the <strong>Write</strong> method using an asynchronous
		/// delegate. This method is generally not recommended because of issues that can occur if
		/// multiple calls to <strong>BeginWrite</strong> are made before the prevoius operation
		/// completes. Consider using the <strong>Write</strong> method on a separate thread
		/// instead.
		/// </remarks>
		/// <returns>
		/// An <strong>IAsyncResult</strong> object used to monitor the state of the write
		/// operation.
		/// </returns>
		/// <summary>Writes an entire array of bytes to the device.</summary>
		/// <param name="buffer">An array of bytes to transmit to the serial device.</param>
		/// <param name="callback">A <strong>Delegate</strong> to execute when the write operation completes.</param>
		/// <param name="state">An <strong>Object</strong> indicating miscellaneous state for the write.</param>
		public IAsyncResult BeginWrite(byte[] buffer, AsyncCallback callback, object state)
		{
			return BeginWrite(buffer, 0, buffer.Length, callback, state);
		}

		/// <summary>Completes an asynchronous write operation.</summary>
		/// <returns>
		/// An <strong>Integer</strong> indicating the number of bytes sent to the
		/// device.
		/// </returns>
		/// <remarks>
		/// This method is generally not recommended because of issues that can occur if
		/// multiple calls to <strong>BeginWrite</strong> are made before the prevoius operation
		/// completes. Consider using the <strong>Write</strong> method on a separate thread
		/// instead.
		/// </remarks>
		/// <param name="asyncResult">
		/// An <strong>IAsyncResult</strong> object returned during a call to
		/// <strong>BeginWrite</strong>.
		/// </param>
		public override void EndWrite(IAsyncResult asyncResult)
		{
			pWriteDelegate.EndInvoke(asyncResult);
		}

        /// <summary>Controls the amount of time allowed to connect or disconnect from a serial port.
        public TimeSpan CommandTimeout
        {
            get
            {
                return pCommandTimeout;
            }
            set
            {
                pCommandTimeout = value;
            }
        }
#endif

		/// <summary>Controls the amount of time allowed for a read operation to complete.</summary>
		/// <value>A <strong>TimeSpan</strong> object. Default is three seconds.</value>
		/// <remarks>
		/// This property controls the maximum amount of time allowed to either receive data
		/// or to wait for data to arrive. Changing this property changes the serial port settings
		/// for the device. This property is typically left at its default since delays
		/// <em>during</em> a data transmission are rare.
		/// </remarks>
		/// <example>
		///     This example instructs the serial device to allow up to five seconds for data to
		///     arrive. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// MyStream.ReadTimeout = TimeSpan.FromSeconds(5)
		/// If MyStream.WaitForData() Then
		///     Debug.Write(MyStream.ReadToEnd().ToString())
		/// Else
		///     Debug.Write("No data is available within the specified timeout.")    
		/// End If
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// MyStream.ReadTimeout = TimeSpan.FromSeconds(5);
		/// if(MyStream.WaitForData())
		///     Debug.Write(MyStream.ReadToEnd().ToString());
		/// else
		///     Debug.Write("No data is available within the specified timeout.");
		///     </code>
		/// </example>
		public 
#if Framework20
			new 
#endif
			TimeSpan ReadTimeout
		{
			get
			{
				return pReadTimeout;
			}
			set
			{
				pReadTimeout = value;
				SetTimeout();
			}
		}

		/// <remarks>
		/// This property controls the maximum amount of time allowed to wait for outgoing
		/// characters to be sent once sending begins. Changing this property changes the serial
		/// port settings for the device. This property is typically left at its default since
		/// delays <em>during</em> a data transmission are rare.
		/// </remarks>
		/// <summary>Controls the amount of time allowed for a send operation to complete.</summary>
		/// <value>A <strong>TimeSpan</strong> object. Default is three seconds.</value>
		public 
#if Framework20
			new 
#endif
			TimeSpan WriteTimeout
		{
			get
			{
				return pWriteTimeout;
			}
			set
			{
				pWriteTimeout = value;
				SetTimeout();
			}
		}

#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        private void SetTimeout()
		{
			// Changes setup on the fly
			ctm.ReadTotalTimeoutConstant = (int)pReadTimeout.TotalMilliseconds;
			ctm.WriteTotalTimeoutConstant = (int)pWriteTimeout.TotalMilliseconds;

			NativeMethods.SetCommTimeouts(pFileHandle, ref ctm);
		}

		/// <summary>Indicates if a connection to the device is currently open.</summary>
		/// <value>
		/// A <strong>Boolean</strong>, <strong>True</strong> if an active connection to the
		/// device exists.
		/// </value>
		/// <remarks>
		/// This property becomes <strong>True</strong> when a connection attempt is made,
		/// then returns to <strong>False</strong> after the connection is closed.
		/// </remarks>
		/// <example>
		///     This example opens a connection to COM1, closes it, then confirms it is closed by
		///     examining the <strong>IsOpen</strong> property. 
		///     <code lang="VB" title="[New Example]">
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1);
		/// ' Attempt to open the port
		/// If MyStream.TryOpen(False) Then
		///     ' The port was opened.  Now close it
		///     MyStream.Close()
		///     ' Confirm that it was closed
		///     If MyStream.IsOpen Then
		///         ' This will never happen.
		///         Debug.WriteLine("The port is still open even after closing!")
		///     Else
		///         Debug.WriteLine("The port was closed properly.")
		///     End If
		/// End If
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Attempt to open the port
		/// if(MyStream.TryOpen(false))
		/// {
		///     // The port was opened.  Now close it
		///     MyStream.Close();
		///     // Confirm that it was closed
		///     if(MyStream.IsOpen)
		///         // This will never happen.
		///         Console.WriteLine("The port is still open even after closing!");
		///     else
		///         Console.WriteLine("The port was closed properly.");
		/// }
		///     </code>
		/// </example>
		public bool IsOpen
		{
			get
			{
				return pFileHandle != null && !pFileHandle.IsInvalid && !pFileHandle.IsClosed;
			}
		}

		/// <summary>Returns the number of characters waiting to be read from the device.</summary>
		/// <value>
		/// A <strong>Long</strong> indicating the number of bytes available, otherwise
		/// zero.
		/// </value>
		/// <remarks>
		/// This property is used during read operations to see if any data is available.
		/// This property will return zero if the port is not open.
		/// </remarks>
		/// <example>
		///     This example uses the <strong>Length</strong> property to examine COM1 to see if
		///     more than fifty bytes are available to read, and, if so, reads them from the
		///     device.
		///     <code title="[New Example]">
		/// 	</code>
		/// 	<code title="[New Example]">
		/// 	</code>
		/// </example>
        public override long Length
		{
			get
            {
                // 2.3.20:  Some devices such as the HTC TYTN II and AT&T Tilt have flaws causing ClearCommError to report 0 bytes available.
                //          So, we now must check to see if we should even call the API at all.
                if (!Devices.IsSerialDeviceQueriedBeforeReading)
                    return DefaultReadBufferSize;

				// If we're not open, open
                if (!IsOpen)
                {
                    // If the stream is closed,
                    if (pIsStreamClosed)
                        return 0;
                    if (!TryOpen(true))
                        return 0;
                }

				// Get the number of bytes waiting to be transmitted				
				NativeMethods.ClearCommError(pFileHandle, out ErrorCode, out pCommunicationStatus);

#if PocketPC && !Smartphone
				/* If we're using the HW6515, ClearCommError always returns 0
				 * even is data is actually available.  Return 1 to allow read
				 * operations to continue.
				 */
                if(pCommunicationStatus.InQue == 0
                    && GeoFramework.IO.HewlettPackard.Ipaq.IsSupported)
				{
                    return DefaultReadBufferSize;
				}
#endif
                // Get the number of bytes in the input buffer
                return pCommunicationStatus.InQue;
            }
		}

		/// <summary>Indicates the current position in the stream.</summary>
		/// <value>A <strong>Long</strong> value, always zero.</value>
		/// <remarks>
		/// This property always returns zero because seek operations are not allowed. Data
		/// is read from the serial device in a forward-only manner.
		/// </remarks>
		public override long Position
		{
			get
			{
				return 0;
			}
			set
			{
				//				// Is it?
				//				if(pPosition == value) return;
				//				// Is it out of bounds?
				//				if(value < 0 || value > Length - 1)
				//				{
				//					throw new IOException("Cannot seek past the end or before the start of a serial port stream.");
				//				}
				//				// Where is the position relative to the current?
				//				System.Diagnostics.Debugger.Break();
				throw new IOException("Seek operations are not currently supported on serial port streams.");
			}
		}

		/// <summary>Not supported. Moves the current pointer to the specified location.</summary>
		/// <returns>
		/// A <strong>Long</strong> indicating the current position of the stream after the
		/// seek operation.
		/// </returns>
		/// <remarks>
		/// 	<font color="red">Seek operations are not currently supported by this class
		/// because data is discarded as soon as it is read from the device. As an alternative,
		/// consider using a <strong>BufferedStream</strong> wrapper or
		/// <strong>MemoryStream</strong> to store data read from the serial port.</font>
		/// </remarks>
		/// <param name="offset">A <strong>Long</strong> indicating the relative number of bytes to move.</param>
		/// <param name="origin">
		/// A <strong>SeekOrigin</strong> indicating the starting position of the
		/// seek.
		/// </param>
		public override long Seek(long offset, SeekOrigin origin)
		{
			//			// Change the position
			//			switch(origin)
			//			{
			//				case SeekOrigin.Begin:
			//				{
			//					if(offset > Length)
			//						throw new IOException("The seek would go past the end of the stream.");
			//					else
			//					{
			//						Position = offset;
			//					}
			//					return pPosition;
			//				}
			//				case SeekOrigin.Current:
			//				{
			//					if(pPosition + offset > Length)
			//						throw new IOException("The seek would go past the end of the stream.");
			//					else
			//					{
			//						Position = pPosition + offset;
			//					}
			//					return pPosition;
			//				}
			//				case SeekOrigin.End:
			//				{
			//					if(pPosition + Length > Length)
			//						throw new IOException("The seek would go past the end of the stream.");
			//					else
			//					{
			//						Position = Length + offset;
			//					}
			//					return pPosition;
			//				}
			//				default:
			//					throw new InvalidOperationException("The specified SeekOrigin \"" + origin.ToString() + "\" is not supported by serial port streams.");
			//			}
			throw new IOException("Seek operations are not currently supported on serial port streams.");
		}

		/// <remarks>
		/// 	<para>This property is used to control the size of the serial port buffer. The
		///     buffer size controls the maximum number of bytes that will be queued for receiving
		///     by the operating system. Using a larger buffer size is recommended if the device
		///     will be accessed in a read-only fashion, such as a GPS device. This class assumes
		///     that most communications will be with a GPS device, and therefore uses a high
		///     buffer size as the default.</para>
		/// 	<para>A value which is a power of two is recommended (i.e. 1024, 2048, 4096, 16384,
		///     etc.).</para>
		/// </remarks>
		/// <summary>Controls the maximum size of incoming and outgoing buffers.</summary>
		/// <returns>
		/// A <strong>Long</strong> indicating the current size of read and write
		/// buffers.
		/// </returns>
		/// <remarks>
		/// This property is used to control the size of the serial port buffer. The buffer
		/// size controls the maximum number of bytes that will be queued for receiving by the
		/// operating system. Using a larger buffer size is recommended if the device will be
		/// accessed in a read-only fashion, such as a GPS device.
		/// </remarks>
		/// <example>
		///     This example changes the default buffer size to 1024 bytes on COM1. 
		///     <code lang="VB" title="[New Example]">
		/// ' Open a connection to COM1
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// MyStream.SetLength(1024)
		/// ' Proceed to read and write data...
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Open a connection to COM1
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// MyStream.SetLength(1024);
		/// // Proceed to read and write data...
		///     </code>
		/// </example>
		/// <param name="value">
		/// A <strong>Long</strong> specifying the new length of the stream. Default is
		/// <strong>32,768</strong>.
		/// </param>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public override void SetLength(long value)
		{
			// If the stream is closed,
			if (pIsStreamClosed)
				return;

            // Remember the buffer length
            pLength = (uint)value;

			// Set up the input/output buffer sizes
            NativeMethods.SetupComm(pFileHandle, pLength, pLength);
		}

		/// <summary>Sends the specified data to the serial device.</summary>
		/// <returns>
		/// An <strong>Integer</strong> indicating the number of bytes successfully sent to
		/// the device.
		/// </returns>
		/// <remarks>
		/// This method sends an arbitrary number of bytes to the serial device. If the
		/// operation completes successfully, the number of bytes sent is returned.
		/// </remarks>
		/// <example>
		/// 	<code lang="VB" title="[New Example]">
		/// Imports System.Text.ASCIIEncoding
		///  
		/// Public Sub Main()
		///     ' Open a connection to COM1
		///     Dim MyStream As New SerialStream(CommunicationPort.Com1)
		///     ' Get a byte array from a string
		///     Dim BytesToSend As Byte() = ASCII.GetBytes("Hello, world!")
		///     ' Transmit the array to the device
		///     MyStream.Write(BytesToSend, 0, BytesToSend.Length)
		/// End Sub
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// using System.Text.ASCIIEncoding;
		///  
		/// void Main()
		/// {
		///     // Open a connection to COM1
		///     SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		///     // Get a byte array from a string
		///     byte[] BytesToSend = ASCII.GetBytes("Hello, world!");
		///     // Transmit the array to the device
		///     MyStream.Write(BytesToSend, 0, BytesToSend.Length);
		/// }
		///     </code>
		/// </example>
		public override void Write(byte[] buffer, int offset, int count)
		{
			if(count == buffer.Length)
				Write(buffer);
			else
			{
				// Create an array to send
				byte[] TrimmedBuffer = new byte[count];
				Array.Copy(buffer, offset, TrimmedBuffer, 0, count);
				Write(TrimmedBuffer);
			}
		}

		/// <summary>Writes a single byte of data to the serial device.</summary>
		/// <remarks>This method writes a single byte of data to the device.</remarks>
		/// <example>
		///     This example uses the <strong>WriteByte</strong> method to write a few bytes to
		///     COM1. 
		///     <code lang="VB" title="[New Example]">
		/// ' Open a connection to COM1
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Write a carriage return to the device byte-by-byte
		/// MyStream.WriteByte(13)
		/// MyStream.WriteByte(10)
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Open a connection to COM1
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Write a carriage return to the device byte-by-byte
		/// MyStream.WriteByte(13);
		/// MyStream.WriteByte(10);
		///     </code>
		/// </example>
		public override void WriteByte(byte value)
		{
			if (pIsStreamClosed)
				throw new ObjectDisposedException(GetType().ToString(), "A read operation cannot be performed because the serial stream is closed.");
            
			byte[] buffer = new byte[1];
			buffer[0] = value;
			Write(buffer, 0, 1);
		}

		/// <summary>Writes a single line of data to the serial device.</summary>
		/// <remarks>
		/// This method is used to send ASCII data to a serial device. The string is
		/// converted into a byte array, then the <strong>Write</strong> method is called to send
		/// the entire array.
		/// </remarks>
		/// <example>
		///     This example sends "Hello, World" to the serial device on COM1. 
		///     <code lang="VB" title="[New Example]">
		/// ' Open a connection to COM1
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Send "Hello, World"
		/// MyStream.WriteLine("Hello, world")
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Open a connection to COM1
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Send "Hello, World"
		/// MyStream.WriteLine("Hello, world");
		///     </code>
		/// </example>
		public void WriteLine(string data)
		{
			Write(System.Text.ASCIIEncoding.ASCII.GetBytes(data));
		}

		/// <remarks>
		/// This method sends an entire byte array to the serial device. If the operation
		/// completes successfully, the number of bytes sent is returned.
		/// </remarks>
		/// <returns>
		/// An <strong>Integer</strong> indicating the number of bytes successfully sent to
		/// the device.
		/// </returns>
		/// <example>
		///     This example opens a connection to COM1 and transmits an array of bytes. 
		///     <code lang="VB" title="[New Example]">
		/// Imports System.Text.ASCIIEncoding
		///  
		/// Public Sub Main()
		///     ' Open a connection to COM1
		///     Dim MyStream As New SerialStream(CommunicationPort.Com1)
		///     ' Get a byte array from a string
		///     Dim BytesToSend As Byte() = ASCII.GetBytes("Hello, world!")
		///     ' Transmit the entire array to the device
		///     MyStream.Write(BytesToSend)
		/// End Sub
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// using System.Text.ASCIIEncoding;
		///  
		/// void Main()
		/// {
		///     // Open a connection to COM1
		///     SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		///     // Get a byte array from a string
		///     byte[] BytesToSend = ASCII.GetBytes("Hello, world!");
		///     // Transmit the entire array to the device
		///     MyStream.Write(BytesToSend);
		/// }
		///     </code>
        /// </example>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public void Write(byte[] buffer)
		{
			// Is there an open port?
            if (!IsOpen)
            {
                if (pIsStreamClosed)
                    throw new ObjectDisposedException(GetType().ToString(), "A read operation cannot be performed because the serial stream is closed.");
                if (!TryOpen(true))
                    return;
            }
			
			if (!NativeMethods.WriteFile(pFileHandle, buffer, (uint)buffer.Length, out CharactersWritten, pOverlapped))
			{
				throw new System.ComponentModel.Win32Exception(Marshal.GetLastWin32Error());
			}

            //if(iBytesWritten > 0 && IsNoDataReceived)
            //{
            //    IsNoDataReceived = false;
            //    OnConnected();
            //}

            //// Yes.  Flag the last received date
            ////LastReceivedTime = DateTime.Now;

            //OnDataSent(buffer);
		}
		#endregion

        #region  Private Members

#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        private void ConfigureDCB()
		{
            // Set the baud rate, parity, data/stop bits
            if (!IsOpen)
                return;

            // Configure the DCB
			NativeMethods.SetCommState(pFileHandle, ref pDCB);

            // And purge any existing data
            Purge();
		}

		/// <summary>
		/// Not supported. Empties any local buffers of data waiting to be sent or
		/// received.
		/// </summary>
		/// <remarks>
		/// This method is not supported because no buffering features exist for the
		/// <strong>SerialStream</strong>. Data is transmitted immediately. To use buffering
		/// features, consider wrapping a <strong>SerialStream</strong> with a
		/// <strong>BufferedStream</strong> object.
		/// </remarks>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public override void Flush() 
		{
            // Is there an open port?
            if (!IsOpen)
                return;

			NativeMethods.FlushFileBuffers(pFileHandle);

			// Also, purge the read/write buffers
			Purge();
		}

		/// <summary>Discards any available data waiting to be read from the device.</summary>
		/// <remarks>
		/// This method is typically called when a connection is first opened to the device
		/// to remove any unwanted characters. This can be a good practice if, for example, a port
		/// was set to the wrong baud rate and several garbage characters exist in the buffer. If
		/// no data exists in the read buffer, this method has no effect.
		/// </remarks>
		/// <example>
		///     This example discards all data waiting to be read from the device. 
		///     <code lang="VB" title="[New Example]">
		/// ' Create a new stream on COM1
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Get rid of any possible garbage characters
		/// MyStream.PurgeRead()
		/// ' Read a line of fresh data from the device
		/// Debug.WriteLine(MyStream.ReadLine())
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create a new stream on COM1
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Get rid of any possible garbage characters
		/// MyStream.PurgeRead();
		/// // Read a line of fresh data from the device
		/// Console.WriteLine(MyStream.ReadLine());
		///     </code>
		/// </example>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public bool PurgeRead() 
		{
            // Is there an open port?
            if (!IsOpen)
                return false;

			return NativeMethods.PurgeComm(pFileHandle, (uint)PurgeFlags.PurgeReceiving);
		}

		/// <remarks>
		/// This method is typically called when a transmission of data sent to the device is
		/// to be aborted. If no data exists in the outgoing buffer, this method has no effect.
		/// This method would have to be called almost instantly after a write operation to clear
		/// the outgoing buffer because data is transmitted instantly to the device.
		/// </remarks>
		/// <summary>Discards any bytes waiting to be sent to the serial device.</summary>
		/// <example>
		/// 	<code lang="VB" title="[New Example]">
		/// ' Create a new stream on COM1
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Clear the outgoing buffer
		/// MyStream.PurgeWrite()
		/// ' Write a line of fresh data to the device
		/// MyStream.WriteLine("Here's some fresh data")
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create a new stream on COM1
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Clear the outgoing buffer
		/// MyStream.PurgeWrite();
		/// // Write a line of fresh data to the device
		/// MyStream.WriteLine("Here's some fresh data");
		///     </code>
		/// </example>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public bool PurgeWrite() 
		{
            // Is there an open port?
            if (!IsOpen)
                return false;

			return NativeMethods.PurgeComm(pFileHandle, (uint)PurgeFlags.PurgeSending);
		}

		/// <summary>Discards any data waiting to be read from or written to the device.</summary>
		/// <returns>
		/// A <strong>Boolean</strong>, <strong>True</strong> if all buffers were
		/// successfully cleared.
		/// </returns>
		/// <remarks>
		/// This method has the same effect of calling <strong>PurgeRead</strong> and
		/// <strong>PurgeWrite</strong> methods individually. All data waiting to be read or
		/// written is discarded. This method is typically called when a connection is first opened
		/// to the device to remove any unwanted characters. This can be a good practice if, for
		/// example, a port was set to the wrong baud rate and several garbage characters exist in
		/// the buffer. If no data exists in either buffer, this method has no effect.
		/// </remarks>
		/// <example>
		///     This example discards all data in both incoming and outgoing buffers on COM1,
		///     writes a string to the device, then reads a line of data and displays it in the
		///     Output window. 
		///     <code lang="VB" title="[New Example]">
		/// ' Create a new stream on COM1
		/// Dim MyStream As New SerialStream(CommunicationPort.Com1)
		/// ' Clear both incoming and outgoing buffers
		/// MyStream.Purge()
		/// ' Write a line of fresh data to the device
		/// MyStream.WriteLine("Here's some fresh data")
		/// ' Read a line of fresh data from the device
		/// Debug.WriteLine(MyStream.ReadLine())
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create a new stream on COM1
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com1);
		/// // Clear both incoming and outgoing buffers
		/// MyStream.Purge();
		/// // Write a line of fresh data to the device
		/// MyStream.WriteLine("Here's some fresh data");
		/// // Read a line of fresh data from the device
		/// Console.WriteLine(MyStream.ReadLine());
		///     </code>
        /// </example>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public bool Purge() 
		{
			// Is there an open port?
			if (!IsOpen)
				return false;

			return NativeMethods.PurgeComm(pFileHandle, (uint)(PurgeFlags.PurgeReceiving | PurgeFlags.PurgeSending));
		}

#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public bool CancelRead() 
		{
			// Is there an open port?
			if (!IsOpen)
				return false;

			return NativeMethods.PurgeComm(pFileHandle, (uint)PurgeFlags.CancelReceiving);
		}

#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public bool CancelWrite() 
		{
			// Is there an open port?
			if (!IsOpen)
				return false;

			return NativeMethods.PurgeComm(pFileHandle, (uint)PurgeFlags.CancelSending);
		}

#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public bool Cancel() 
		{
			// Is there an open port?
			if (!IsOpen)
				return false;

			return NativeMethods.PurgeComm(pFileHandle, (uint)(PurgeFlags.CancelSending | PurgeFlags.CancelReceiving));
		}


		/// <summary>Controls whether the Bluetooth browser is displayed during connection.</summary>
		/// <value>
		/// A <strong>Boolean</strong>, <strong>True</strong> if the Bluetooth browser is
		/// enabled.
		/// </value>
		/// <remarks>
		/// 	<para>Used only rarely, this property controls whether or not the Bluetooth"
		///     browser appears. The Bluetooth browser gives users the opportunity to select a
		///     Bluetooth device to associate withe the port. In a vast majority of cases, only one
		///     device will appear in the list, making the browser pointless to display. This
		///     property is only supported on devices using the Widcomm Bluetooth stack, such as
		///     Hewlett-Packard iPAQ devices. Setting this property causes the registry to be
		///     modified to enabled or suppress the Bluetooth browser.</para>
		/// 	<para>This property is automatically set to <strong>False</strong> during a
		///     connection attempt on mobile devices to temporarily disable the browser. After the
		///     connection attempt (or upon failure), the original value is restored.</para>
		/// 	<para>The Bluetooth browser will only appear if a connection is made to a
		///     Bluetooth device using the device's virtual serial port. For devices which do not
		///     use a Widcomm stack, such as the Dell Axim, this property has no affect.</para>
		/// 	<para><em>iPAQ is a registered trademark of Hewlett-Packard Corporation
		///     (<a href="http://www.hp.com">www.hp.com</a>). Axim is a registered trademark of
		///     Dell Corporation (<a href="http://www.dell.com">www.dell.com</a>).</em></para>
		/// </remarks>
		/// <example>
		///     This example turns the Bluetooth browser on, then attempts a connection. 
		///     <code lang="VB" title="[New Example]">
		/// ' Create a new stream on COM5
		/// Dim MyStream As New SerialStream(CommunicationPort.Com5)
		/// ' Enable the Bluetooth browser
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// 	</code>
		/// </example>
		public bool IsBluetoothBrowserEnabled
		{
			get
			{
                // If Widcomm isn't supported, skip
                if (!pIsWidcommBluetoothSupported)
                    return false;

                // Get the value from the registry
                RegistryKey BluetoothKey = null;
                try
                {
                    // Is there a Widcomm key?
                    BluetoothKey = Registry.LocalMachine.OpenSubKey(@"Software\Widcomm\BtConfig\AutoConnect\" + ((int)pPort).ToString().PadLeft(4, '0'), true);
                    if (BluetoothKey == null)
                    {
                        // No.  Default to True
                        return true;
                    }
                    else
                    {
                        // Return the existing BtBrowser value
                        object Result = BluetoothKey.GetValue("BtBrowserEnabled");
                        if (Result != null)
                            return (int)Result == 1;
                        else
                            return true;
                    }
                }
                catch
                {
                    return false;
                }
                finally
                {
                    // Clean up the registry key
                    if (BluetoothKey != null)
                        BluetoothKey.Close();
                }
			}
			set
			{
                // If Widcomm isn't supported, skip this
                if (!pIsWidcommBluetoothSupported)
                    return;

                RegistryKey BluetoothKey = null;
                try
                {
                    // Try to open the key.  Is there anything?
                    BluetoothKey = Registry.LocalMachine.OpenSubKey(@"Software\Widcomm\BtConfig\AutoConnect", false);
                    if(BluetoothKey == null)
                    {
                        // Ignore completely!
                        return;
                    }
                    BluetoothKey.Close();
                    BluetoothKey = Registry.LocalMachine.OpenSubKey(@"Software\Widcomm\BtConfig\AutoConnect\" + ((int)pPort).ToString().PadLeft(4, '0'), true);
                    if (BluetoothKey == null)
                        // No.  Create a new key
                        BluetoothKey = Registry.LocalMachine.CreateSubKey(@"Software\Widcomm\BtConfig\AutoConnect\" + ((int)pPort).ToString().PadLeft(4, '0'));
                    // Set the value
                    BluetoothKey.SetValue("BtBrowserEnabled", (value ? 1 : 0));
                }
                catch
                {
                }
                finally
                {
                    if (BluetoothKey != null)
                        BluetoothKey.Close();
                }
			}
		}

		/// <summary>Attempts to open a connection to the device.</summary>
		/// <returns>
		/// A <strong>Boolean</strong>, <strong>True</strong> if the connection attempt was
		/// successful.
		/// </returns>
		/// <remarks>
		/// 	<para>This method provides a way to try a connection attempt when there is some
		///     doubt as to whether the connection would succeed. This method is called
		///     automatically if any calls are made to <strong>Length</strong>,
		///     <strong>Read</strong>, <strong>ReadByte</strong>, <strong>ReadLine</strong>,
		///     <strong>Write</strong>, <strong>WriteByte</strong>, <strong>WriteLine</strong>, or
		///     any other method that requires an open connection.</para>
		/// 	<para>If the communications port is already open, this method has no effect.</para>
		/// 	<para>NOTE: On most Windows systems, a connection to a communications port can be
		///     successful even if no device is actually connected to the port. For this reason, a
		///     successful connection should be indicated by a successful read or write attempt to
		///     the device.</para>
		/// </remarks>
		/// <example>
		///     This example tests to see if a connection succeeds on COM5. 
		///     <code lang="VB" title="[New Example]">
		/// ' Create a new stream on COM5
		/// Dim MyStream As New SerialStream(CommunicationPort.Com5)
		/// ' See if a connection attempt succeeds
		/// If MyStream.TryOpen(false) Then
		///     Debug.WriteLine("The port was successfuly opened!")
		/// Else
		///     Debug.WriteLine("The port could not be opened.")
		/// End If
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create a new stream on COM5
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com5);
		/// // See if a connection attempt succeeds
		/// if(MyStream.TryOpen(false))
		///     Console.WriteLine("The port was successfuly opened!");
		/// else
		///     Console.WriteLine("The port could not be opened.");
		///     </code>
		/// </example>
		/// <exception cref="IOException" caption="IOException">The serial port could not be opened.</exception>
		/// <exception cref="Win32Exception" caption="Win32Exception">Access is denied.  The port is already in use.</exception>
        /// <exception cref="MissingMethodException" caption="MissingMethodException">(Rare)  The specified port is invalid for serial communications or unsupported by the computer.</exception>
#if !PocketPC
		[SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode = true)]
#endif
        public bool TryOpen(bool allowExceptions)
		{
            int Win32ErrorCode; 
            bool IsLocked = false;
            // Initializes Com Port
            try
            {
                // Has a SyncRoot been created to sync access to ports?
                object SyncLock = SerialPortSyncLocks[PortName];
                if (SyncLock == null)
                {
                    // No. Create a new one now
                    lock (SerialPortSyncLocks.SyncRoot)
                    {
                        // The synclock only needs to be a basic Object
                        SyncLock = new object();
                        SerialPortSyncLocks.Add(PortName, SyncLock);
                    }
                }

#if !PocketPC
                IsLocked = Monitor.TryEnter(SyncLock, CommandTimeout);
#elif PocketPC && Framework20
                IsLocked = Monitor.TryEnter(SyncLock);
#elif PocketPC && !Framework20
                Monitor.Enter(SyncLock);
                IsLocked = true;
#endif

                if (!IsLocked)
                    throw new InvalidOperationException("A connection attempt cannot be made on port \"" + PortName + "\" because another thread is attempting to connect or disconnect.  Increase the ReadTimeout property to allow more time for connection and disconnection operations.");

                // Are we already open?
                if (IsOpen)
                    return true;

                // Signal that no operations are allowed
                pCanRead = false;
                pCanWrite = false;

#if !Framework10
                // Create a COM Port stream handle 
                if (Environment.OSVersion.Platform == PlatformID.WinCE)
                {
                    // Remember the bluetooth browser settings
                    bool WasBluetoothBrowserEnabled = IsBluetoothBrowserEnabled;

                    // Disable the Bluetooth browser
                    IsBluetoothBrowserEnabled = false;

                    //// Signal that connection is imminent
                    //OnConnecting();



                    // Lock the instance and connect
                    pFileHandle = NativeMethods.CreateFile(pPortName, SerialMode.GenericRead | SerialMode.GenericWrite, SerialShareMode.Read | SerialShareMode.Write, IntPtr.Zero, SerialCreationDisposition.OpenExisting, SerialFileAttributes.Normal, IntPtr.Zero);
                    Win32ErrorCode = Marshal.GetLastWin32Error();

                    // Restore the original bluetooth browser setting
                    IsBluetoothBrowserEnabled = WasBluetoothBrowserEnabled;
                }
                else
                {
#endif
                    //OnConnecting();

#if DEBUG
                    //Console.WriteLine("Opening port " + pPortName);
#endif

                    pFileHandle = NativeMethods.CreateFile(pPortName, SerialMode.GenericRead | SerialMode.GenericWrite, SerialShareMode.Read | SerialShareMode.Write, IntPtr.Zero, SerialCreationDisposition.OpenExisting, SerialFileAttributes.Normal, IntPtr.Zero);
                    Win32ErrorCode = Marshal.GetLastWin32Error();
#if !Framework10
                }
#endif

                // Did we just at the same time close the stream?
                if (pIsStreamClosed)
                {
#if DEBUG
                    //Console.WriteLine("Stream closed! " + pPortName);
#endif

                    // Yes.  Close this connection and indicate failure since
                    // the stream is closed down.
                    return false;
                }

                // Was the port opened?
                if (!pFileHandle.IsInvalid)
                {
#if DEBUG
                    //Console.WriteLine("Opened " + pPortName);

                    //if (pPort == CommunicationPort.Com6)
                    //    System.Diagnostics.Debugger.Break();
#endif

#if !PocketPC
                    // We will need to finalize/dispose now that the port is open
                    GC.ReRegisterForFinalize(this);
#endif
                    // Clear the BREAK state to resume communications
                    NativeMethods.EscapeCommFunction(pFileHandle, CommunicationEscapeCode.CLRBREAK);

                    // Signal that we're ready to send data (?)
                    NativeMethods.EscapeCommFunction(pFileHandle, CommunicationEscapeCode.SETDTR);

                    // Signal that we're ready to receive data (?)
                    NativeMethods.EscapeCommFunction(pFileHandle, CommunicationEscapeCode.SETRTS);

                    // Finally, clear any errors and check the state
                    NativeMethods.ClearCommError(pFileHandle, out ErrorCode, out pCommunicationStatus);

                    // Configure the DCB
                    NativeMethods.SetCommState(pFileHandle, ref pDCB);

                    // Set up the input/output buffer sizes
                    NativeMethods.SetupComm(pFileHandle, pLength, pLength);

                    // Set character timeouts
                    NativeMethods.SetCommTimeouts(pFileHandle, ref ctm);

                    // Finally, clear any errors and check the state
                    NativeMethods.ClearCommError(pFileHandle, out ErrorCode, out pCommunicationStatus);

                    // Yes.  Notify of success
                    pCanRead = true;
                    pCanWrite = true;
                    return true;
                }
                else
                {

#if DEBUG
                    //Win32Exception t = new Win32Exception(Win32ErrorCode);

                    //Console.WriteLine("Port opening problem! {0}", t.Message);
#endif
                    // The stream cannot be opened.
                    pIsStreamClosed = true;

                    // Raise Initialization problems
                    if (allowExceptions)
                        throw new Win32Exception(Win32ErrorCode);
                    else
                        return false;
                }
            }
            catch (IOException ex)
            {
                if (allowExceptions)
                    throw ex;
                else
                    return false;
            }
            catch (System.ComponentModel.Win32Exception ex)
            {
                if (allowExceptions)
                    throw new IOException("The serial port could not be opened.", ex);
                else
                    return false;
            }
            catch (MissingMethodException ex)
            {
                if (allowExceptions)
                    throw new IOException("The serial port could not be opened.", ex);
                else
                    return false;
            }
#if !PocketPC || DesignTime || (PocketPC && Framework20)
            catch (ThreadAbortException)
            {
                // Just exit!
                return false;
            }
#endif
            catch (Exception ex)
            {
                if (allowExceptions)
                    throw new IOException(ex.Message, ex);
                else
                    return false;
            }
            finally
            {
                if (IsLocked)
                    Monitor.Exit(SerialPortSyncLocks[PortName]);
            }
		}

		/// <summary>Attempts to open the communications port.</summary>
		/// <remarks>
		/// 	<para>This method attempts to open the communications port. If the attempt fails,
		///     an exception is thrown. This method is called automatically if any calls are made
		///     to <strong>Length</strong>, <strong>Read</strong>, <strong>ReadByte</strong>,
		///     <strong>ReadLine</strong>, <strong>Write</strong>, <strong>WriteByte</strong>,
		///     <strong>WriteLine</strong>, or any other method that requires an open
		///     connection.</para>
		/// 	<para>If the communications port is already open, this method has no effect.</para>
		/// 	<para>NOTE: On most Windows systems, a connection to a communications port can be
		///     successful even if no device is actually connected to the port. For this reason, a
		///     successful connection should be indicated by a successful read or write attempt to
		///     the device.</para>
		/// </remarks>
		/// <example>
		///     This example tries to open COM5 and outputs any exception information if the
		///     attempt fails. 
		///     <code lang="CS" title="[New Example]">
		/// ' Create a new stream on COM5
		/// Dim MyStream As New SerialStream(CommunicationPort.Com5)
		/// ' See if a connection attempt succeeds
		/// Try
		///     MyStream.Open()
		///     Debug.WriteLine("The port was successfuly opened!")
		/// Catch ex As IOException
		///     Debug.WriteLine("The port could not be opened.")
		/// Catch ex As Win32Exception
		///     Debug.WriteLine("The port is already in use.")
		/// Catch ex As NotSupportedException
		///     Debug.WriteLine("The port does not support serial communications.")
		/// End Try
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create a new stream on COM5
		/// SerialStream MyStream = new SerialStream(CommunicationPort.Com5);
		/// // See if a connection attempt succeeds
		/// try
		/// {
		///     MyStream.Open();
		///     Console.WriteLine("The port was successfuly opened!");
		/// }
		/// catch(IOException ex)
		/// {
		///     Debug.WriteLine("The port could not be opened.");
		/// }
		/// catch(Win32Exception ex)
		/// {
		///     Debug.WriteLine("The port is already in use.");
		/// }
		/// catch(NotSupportedException ex)
		/// {
		///     Debug.WriteLine("The port does not support serial communications.");
		/// }
		///     </code>
		/// </example>
		/// <exception cref="IOException" caption="IOException">The serial port could not be opened.</exception>
		/// <exception cref="Win32Exception" caption="Win32Exception">Access is denied.  The port is already in use.</exception>
        /// <exception cref="MissingMethodException" caption="MissingMethodException">(Rare)  The specified port is invalid for serial communications or unsupported by the computer.</exception>
        public void Open()
		{
			try
			{
				if (pIsStreamClosed)
					throw new ObjectDisposedException(GetType().ToString(), "The serial stream has been closed and cannot be reopened.  Reopen the port by creating a new or cloned SerialStream object.");

				TryOpen(true);
			}
			catch
			{
				throw;
			}
		}

#if !Framework20
		/// <summary>
		/// Obsolete. Creates a wait handle used for synchronizing asynchronous
		/// operations.
		/// </summary>
		/// <returns>A <strong>ManualResetEvent</strong> object.</returns>
		/// <remarks>
		/// This property has been made obsolete as of Visual Studio.NET 2005. It is not
		/// supported for this class.
		/// </remarks>
		protected override WaitHandle CreateWaitHandle()
		{
			return new ManualResetEvent(false);
		}
#endif

		/// <summary>Closes an open connection to the serial device.</summary>
		/// <remarks>
		/// 	<para>If a connection is open, this method will close the connection and release
		///     the unmanaged file handle used during communications. After this method is called,
		///     the serial port becomes available to other applications.</para>
		/// 	<para>If the communications port is already closed, this method has no
		///     effect.</para>
		/// </remarks>
		/// <example>
		///     This example attempts to open a connection to COM1, then closes it if the attempt
		///     is successful.
		///     <code title="[New Example]">
		/// 	</code>
		/// 	<code title="[New Example]">
		/// 	</code>
		/// </example>
#if !PocketPC && Framework20
        [SuppressUnmanagedCodeSecurity]
#endif
        public override void Close()
        {
            // Is the stream already closed?
            if (pIsStreamClosed)
                return;

            object SyncLock = null;
            bool IsLocked = false;

            try
            {
                if (pPortName == null)
                    IsLocked = false;
                else
                {
                    SyncLock = SerialPortSyncLocks[pPortName];
                    if (SyncLock == null)
                        IsLocked = false;
                    else
                    {
#if !PocketPC   

                        IsLocked = Monitor.TryEnter(SyncLock, CommandTimeout);                
#elif PocketPC && Framework20
                        IsLocked = Monitor.TryEnter(SerialPortSyncLocks[PortName]);
#elif PocketPC && !Framework20
                        Monitor.Enter(SerialPortSyncLocks[PortName]);
                        IsLocked = true;
#endif
                    }
                }

                // Signal that read/write operations are no longer allowed
                pCanRead = false;
                pCanWrite = false;

                // Get rid of the file handle
                if (pFileHandle != null && !pFileHandle.IsInvalid && !pFileHandle.IsClosed)
                {

#if Framework20 && !PocketPC
                    //// Remove incoming and outgoing data
                    //Purge();

                    // Cancel all I/O
                    NativeMethods.CancelIo(pFileHandle);

                    //// Clear any error condition
                    //NativeMethods.ClearCommError(pFileHandle, out ErrorCode, out pCommunicationStatus);
#endif

                    // Close the handle
                    pFileHandle.Close();

                    // signal that we're closed
                    pIsStreamClosed = true;

                    // No longer need to finalize
                    GC.SuppressFinalize(this);

                    // If this is a Bluetooth device, force a pause for the stack to catch up
                    if (pIsBluetoothDevice)
                    {
                        Thread.Sleep(1000);
                    }

                }

            }
            catch
            {

                throw;
            }
            finally
            {
                if (IsLocked)
                    Monitor.Exit(SyncLock);

                // signal that we're closed
                pIsStreamClosed = true;

                // Signal that read/write operations are no longer allowed
                pCanRead = false;
                pCanWrite = false;
            }
        }

		/// <summary>
		/// Creates a new SerialStream using the same configuration settings as the current instance.
		/// </summary>
		/// <returns>A <strong>SerialStream</strong> object.</returns>
		/// <remarks>This method is typically used when the current instance has been closed and
		/// a new, ready-to-use Stream is required.  This method is used by the GPS.NET interpreter
		/// to restart itself if a serial GPS device stops responding.</remarks>
		public SerialStream Clone()
		{
			return new SerialStream(Port, BaudRate, DataBits, Parity, StopBits, FlowControl);
		}

		#endregion

		#region Public Members
		//        public BaudRate CurrentBaudRate
		//		{
		//			get
		//			{
		//				return pCurrentBaudRate;
		//			}
		//		}
		//
		//		public CommunicationPort CurrentCommunicationPort
		//		{
		//			get
		//			{
		//				return pCurrentCommunicationPort;
		//			}
		//		}

		/// <summary>Returns the handle to the serial device.</summary>
		/// <value>A <strong>SafeFileHandle</strong> object.</value>
		/// <remarks>
		/// Communications with a serial device are done through unmanaged platform invoke
		/// API's. In order to make this class as safe as possible, all unmanaged handles are
		/// managed very strictly to ensure that they are properly deallocated. This property is
		/// provided for advanced developers which need to perform custom platform invocation not
		/// directly supported by this class.
		/// </remarks>
		public SafeFileHandle Handle
		{
			get
			{
				return pFileHandle;
			}
		}

		#endregion

		#region  Serial Port Structures 

		private const Int32 PURGE_TXCLEAR = 0X4;
		private const uint GENERIC_READ = 0x80000000;
		private const uint GENERIC_WRITE = 0x40000000;
		private const Int32 OPEN_EXISTING = 3;
		private const Int32 FILE_ATTRIBUTE_NORMAL = 0X80;
		private const UInt32 FILE_FLAG_OVERLAPPED = 0x40000000;

		#endregion

	}
}

