using System;
#if PocketPC 
using System.IO;
using System.Globalization;
using System.Text;
#if !PocketPC || Framework20
using Microsoft.Win32;
#endif

using GeoFramework;
using GeoFramework.IO.Serial;
using GeoFramework.Gps;
using GeoFramework.Gps.Nmea;

namespace GeoFramework.IO
{
    /// <summary>
    /// Represents the GPS Intermediate Driver on Windows Mobile 5.0 hand-held devices.
    /// </summary>
    /// <remarks>Beginning with the Windows Mobile 5.0 mobile operating system, Microsoft has
    /// provided the "GPS Intermediate Driver," which provides shared-appication access to GPS
    /// devices, along with a crude API for accessing data.  GPS.NET is designed to detect when
    /// shared-application mode is enabled by use of this class.  This class can also be used to
    /// configure the intermediate driver from managed code.</remarks>
    public sealed class SharedGpsDevice : Device 
    {
        private static bool pIsSupported;
        private string pDeviceName;
        private SerialDevice pSourceDevice;
        //private bool _IsQualCommDriver;
        private const string IntermediateDriverKey = @"System\CurrentControlSet\GPS Intermediate Driver";
        private const string ControlPanelKey = IntermediateDriverKey + @"\Drivers\Control Panel Configured Device";
        private const string QualCommKey = IntermediateDriverKey + @"\Drivers\GpsOneDevice";

        /// <summary>
        /// Returns the GPS Intermediate Driver device is one is present on the system.
        /// </summary>
        /// <returns></returns>
        internal SharedGpsDevice()
        {
            Refresh();
        }

        static SharedGpsDevice()
        {
            if (Platform.HostPlatformID != HostPlatformID.PocketPC)
                return;

            // Test to see if we can configure the driver
            RegistryKey SourceKey = null;
            try
            {
                /* Try to create the key which stores GID settings.  This will test to make
                 * sure we have permissions enough to actually use the GID at all!  If any
                 * part of this test fails, we can't use the GID.
                 */

                /* A Windows Mobile 5.0 device WILL have a "GPS Intermediate Driver" registry 
                 * key, even after a hard reset.  This is the best indicator that the GID is supported
                 * at all.
                 */

                // Try to create the key as well as settings
                SourceKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey, false);
                if (SourceKey != null)
                {
                    // Close the key
                    SourceKey.Close();

                    // Re-open it for writing
                    SourceKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey, true);

                    // Set the interface type to "COMM"                    
                    SourceKey.SetValue("GPS.NET", "Testing write registry permissions.  It's safe to delete this value.", RegistryValueKind.String);

                    // Delete the valuye
                    SourceKey.DeleteValue("GPS.NET");

                    // If that works, we have permissions enough to proceed
                    pIsSupported = true;
                }
                else
                {
                    // Not supported!
                    pIsSupported = false;
                    return;
                }

            }
            catch (UnauthorizedAccessException)
            {
                pIsSupported = false;
                return;
            }
            catch (FileNotFoundException)
            {
                pIsSupported = false;
                return;
            }
            finally
            {
                if (SourceKey != null)
                    SourceKey.Close();
            }

            // Determine whether this feature is supported
            RegistryKey GPSDriverKey = null;
            try
            {
                GPSDriverKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey, false);
                pIsSupported = (GPSDriverKey != null);
            }
            catch
            {
                pIsSupported = false;
            }
            finally
            {
                if (GPSDriverKey != null)
                    GPSDriverKey.Close();
            }
        }

        /// <summary>
        /// Returns whether the host device is using a proprietary driver for GPS communications.
        /// </summary>
        /// <remarks>During the Summer of 2007, reports were being received about mobile devices which broke the
        /// standard GPS.NET expects for GPS Intermediate Driver (GPSID) information.  Devices such as the HTC TTYN II and HTC Kaiser
        /// ignored the standard and implemented their own customized driver.  To accommodate these devices, GPS.NET now examines
        /// the GPSID to see what driver it uses.  If any other than the default driver is in use, this property will return
        /// <strong>True</strong>.  When a non-standard driver is in play, GPS.NET will change its behavior to try and use the
        /// custom driver instead of (ahem) following GPSID standards.</remarks>
        public bool IsProprietaryDriver
        {
            get
            {
                // Check to see which DLL is being used.  This value will be "GPSID.DLL" on regular devices.  The HTC TYTN II and HTC Kaiser
                // have values such as "GPSID_QCT.DLL".  We'll return True if the dll is named "GPSID.DLL"... that's about the only way to get
                // this detected reliably right now (11/4/07)
                RegistryKey SettingsKey = null;
                try
                {
                    SettingsKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey + @"\ActiveDevice", false);
                    if (SettingsKey != null)
                    {
                        // Get the name of the DLL used by the driver
                        string AssemblyName = Convert.ToString(SettingsKey.GetValue("DLL")).ToUpper(CultureInfo.InvariantCulture);

                        // Return True if the assembly name is "GPSID.DLL"
                        return AssemblyName != "GPSID.DLL";
                    }
                    else
                    {
                        // No driver name exists at all?  WTF.  This shouldn't happen.  Return true
                        return true;
                    }
                }
                catch
                {
                    // Any problems here suggest a proprietary configuration
                    return true;
                }
                finally
                {
                    if (SettingsKey != null)
                        SettingsKey.Close();
                }
            }
        }

        /// <summary>
        /// Controls whether the "GPS" Control Panel is enabled in the device's Settings.
        /// </summary>
        /// <remarks>On come hand-held devices, the "GPS" Control Panel is hidden by default
        /// in the Settings.  Setting this property to True will cause the Control Panel icon
        /// (a blue globe with four compass points) to become visible.  It is considered good practice
        /// to show this icon whenever possible and to avoid hiding the icon since it prevents the
        /// end-user from making changes or manual repairs to GPS Intermediate Driver settings.</remarks>
        public bool IsControlPanelVisible
        {
            get
            {
                // Enable the control panel icon
                RegistryKey SettingsKey = null;
                try
                {
                    SettingsKey = Registry.LocalMachine.OpenSubKey(@"ControlPanel\GPS Settings");
                    if (SettingsKey != null)
                        return Convert.ToBoolean(SettingsKey.GetValue("Hide", 0));
                    else
                        return true;
                }
                catch
                {
                    return true;
                }
                finally
                {
                    if (SettingsKey != null)
                        SettingsKey.Close();
                }
            }
            set
            {
                // Enable the control panel icon
                RegistryKey SettingsKey = null;
                try
                {
                    SettingsKey = Registry.LocalMachine.OpenSubKey(@"ControlPanel\GPS Settings");
                    if (SettingsKey != null)
                        SettingsKey.SetValue("Hide", 0);
                }
                catch
                {
                }
                finally
                {
                    if (SettingsKey != null)
                        SettingsKey.Close();
                }
            }
        }

        /// <summary>
        /// Applies changes to settings for the GPS Intermediate Driver.
        /// </summary>
        /// <remarks>Whenever configuration changes are made, or GID settings are manually changed
        /// in the registry, the GID driver must be refreshed.  This method should be called any
        /// time properties are changed to ensure that the driver uses the new settings.
        /// </remarks>
        public void Refresh()
        {
            // If we're not on a mobile device, skip
            if (!pIsSupported || Environment.OSVersion.Platform != PlatformID.WinCE)
                return;

            // If we're not using a proprietary driver, kick the GPSID, telling it to refresh itself using the latest
            // registry settings.
            if (!IsProprietaryDriver)
            {
                // Tell the device to reconfigure from the registry
                SafeFileHandle x = null;
                try
                {
                    x = Serial.NativeMethods.CreateFile(
                            "GPD0:",
                        // Use default settings for opening a serial connection
                            SerialMode.GenericRead,
                            SerialShareMode.Read | SerialShareMode.Write,
                            IntPtr.Zero,
                            SerialCreationDisposition.OpenExisting,
                            SerialFileAttributes.Normal,
                            IntPtr.Zero);
                    if (!x.IsInvalid)
                    {
                        int Result = NativeMethods.DeviceIoControl(x, NativeMethods.IOCTL_SERVICE_REFRESH);
                    }
                }
                catch
                {
                }
                finally
                {
                    if (x != null)
                        x.Close();
                }
            }
            else
            { 
                // Devices with proprietary drivers are (as of this writing) also subject to manufacturing flaws which cause GPS.NET
                // to timeout.  As a result, we will tell GPS.NET to bypass checking serial ports for available bytes when a proprietary
                // driver is present.
                Devices.IsSerialDeviceQueriedBeforeReading = false;

                /* Let's see if this is one of those INCREDIBLY AWESOME HTC TyTn devices who decided to take a shit all over
                 * well-established industry standards in favor of their own shitty QualComm card which produces malformed sentences
                 * and only works at 4800 baud.  As in THE FUCKING 1990'S.
                 */

                RegistryKey HTCkey = null;
                try
                {
                    HTCkey = Registry.LocalMachine.OpenSubKey(QualCommKey, false);

                    // Does this key exist?
                    if (HTCkey != null)
                    {
                        // Yes!  Joy!
                        //_IsQualCommDriver = true;

                        // The source device will ALWAYS be COM4: at 4800 baud.
                        pSourceDevice = new SerialDevice(CommunicationPort.Com4, BaudRate.Baud4800);

                        // Set it to the NMEA protocol
                        pSourceDevice.SetProtocol(NmeaInterpreter.ProtocolName);

                        // Get the name of the device
                        string friendlyName = HTCkey.GetValue("FriendlyName") as string;
                        if(friendlyName != null)
                            pSourceDevice.SetDeviceName(friendlyName);

                        // Finally, increase the NmeaStream default timeout
                        NmeaStream.DefaultReadTimeout = TimeSpan.FromSeconds(15);

                        // All done
                        return;
                    }
                }
                catch
                {
                    pSourceDevice = null;
                }
                finally
                {
                    if (HTCkey != null)
                    {
                        HTCkey.Close();
                    }
                }
            }
         
            #region Source GPS Device

            // Read settings
            RegistryKey GPSDriverKey = null;
            try
            {
                // Open the control panel key for reading
                GPSDriverKey = Registry.LocalMachine.OpenSubKey(ControlPanelKey, false);

                // If the control panel key is not present, there's nothing to read
                if (GPSDriverKey == null)
                    return;

                // Set the port and baud rate, etc.
                string PortNameValue = Convert.ToString(GPSDriverKey.GetValue("CommPort"));                
                int BaudRateValue = Convert.ToInt32(GPSDriverKey.GetValue("Baud"));

                // Oops!  John Spaith accidentally forgot to include 38,400 as the baud rate.  Easy to fix.
                if (BaudRateValue == 33600)
                    BaudRateValue = 38400;

                // If the port is empty, return null
                if (PortNameValue.Length > 0)
                {
                    pSourceDevice = new SerialDevice(PortNameValue, (BaudRate)BaudRateValue);

                    // Set it to the NMEA protocol
                    pSourceDevice.SetProtocol(NmeaInterpreter.ProtocolName);
                }
                else
                {
                    pSourceDevice = null;
                }
            }
            catch
            {
                pSourceDevice = null;
            }
            finally
            {
                if (GPSDriverKey != null)
                {
                    GPSDriverKey.Close();
                }
            }

            #endregion

        }

        /// <summary>Outputs the friendly name for the device.</summary>
        public string DeviceName
        {
            get
            {
                if (pDeviceName == null)
                {
                    if (pSourceDevice == null)
                    {
                        return "Shared GPS Device (Not yet configured)";
                    }
                    else
                    {
                        return "Shared GPS Device (" + pSourceDevice.ToString() + ")";
                    }
                }
                else
                {
                    return "Shared GPS Device (" + pDeviceName + " on " + pSourceDevice.ToString() + ")";
                }
            }
        }

        public override string ToString()
        {
            return DeviceName;
        }

        /// <summary>
        /// Returns a new <strong>SerialStream</strong> for communicating with the
        /// device.
        /// </summary>
        public override Stream GetHardwareStream()
        {
            if (IsImmediateNeedStreamAvailable)
            {
                IsImmediateNeedStreamAvailable = false;
                return ImmediateNeedStream;
            }
            else
            {
                // Is this some strange proprietary driver?  If so, they may be more strict than the standard allows.
                // Stupid fucking morons always trying to reinvent the wheel, and it's ALWAYS a square.
                if (IsProprietaryDriver)
                {
                    // Yeah, it's proprietary.  Use the baud rate of the source device
                    return new SerialStream(SharedPortName, pSourceDevice.BaudRate);
                }
                else
                {
                    // It's safe to open a connection to the GPSID at the highest baud rate available
                    return new SerialStream(SharedPortName, BaudRate.Baud115200);
                }
            }
        }

        public override string Protocol
        {
            get
            {
                return pSourceDevice.Protocol;
            }
        }

        protected override void OnProtocolDetected(Interpreter interpreter)
        {
            pDeviceName = interpreter.DeviceName;
        }       

        /// <summary>
        /// Controls whether the GPS Intermediate Driver is managing GPS devices.
        /// </summary>
        /// <remarks>The GPS Intermediate Driver can automatically share a GPS device
        /// with multiple applications by multiplexing the raw data from a GPS Device.  This
        /// property controls whether the GPS device is shared.  Setting this property to False
        /// will disable the driver.  When disabled, applications cannot share a GPS device, and connections
        /// are exclusive.  This property should be enabled as a best practice since it allows
        /// applications to share GPS devices.</remarks>
        public bool IsEnabled
        {
            get
            {
                // Load information from the registry
                RegistryKey GPSDriverKey = null;
                try
                {
                    GPSDriverKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey, false);

                    // If no key was found, there's no intermediate driver.  Return false
                    if (GPSDriverKey == null)
                        return false;

                    // Found!  Is the multiplexer active?
                    object IsEnabledValue = GPSDriverKey.GetValue("IsEnabled");
                    /* If no value was found, then the multiplexer is **ON**
                     * by default.
                     */
                    if (IsEnabledValue == null)
                    {
                        // Set the value to true! The GPSID is ON by default.
                        return true;
                    }
                    else
                    {
                        // Read the enabled/diabled value
                        return Convert.ToBoolean(IsEnabledValue);
                    }
                }
                catch
                {
                    return false;
                }
                finally
                {
                    if (GPSDriverKey != null)
                        GPSDriverKey.Close();
                }
            }
            set
            {
                // If it's not supported, don't make the key
                if (!pIsSupported)
                    throw new InvalidOperationException("The GPS Intermediate Driver is currently not supported by this device."); 

                RegistryKey GPSDriverKey = null;
                try
                {
                    GPSDriverKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey, true);
                    if (GPSDriverKey == null)
                        GPSDriverKey = Registry.LocalMachine.CreateSubKey(IntermediateDriverKey);
                    {
                        // Found!  Is the multiplexer active?
#if Framework20
                        if (value)
                            GPSDriverKey.SetValue("IsEnabled", 1, RegistryValueKind.DWord);
                        else
                            GPSDriverKey.SetValue("IsEnabled", 0, RegistryValueKind.DWord);
#else
						if (value)
							GPSDriverKey.SetValue("IsEnabled", 1);
						else
							GPSDriverKey.SetValue("IsEnabled", 0);
#endif
                    }
                }
                catch
                {
                }
                finally
                {
                    if (GPSDriverKey != null)
                        GPSDriverKey.Close();
                }
            }
        }

        /// <summary>
        /// Returns whether the device supports the Microsoft GPS Intermediate Driver.
        /// </summary>
        /// <remarks>The GPS Intermediate Driver is only supported on hand-held devices
        /// running the Windows Mobile 5.0 (WM5) operating system.  Some hand-helds also exist which run
        /// WM5 do not support the Intermediate Driver as well.  This property checks the local device
        /// to see if the driver is indeed supported.</remarks>
        public static bool IsSupported
        {
            get
            {
                return pIsSupported;
            }
        }

        /// <summary>
        /// Disables the GPS Intermediate Driver.
        /// </summary>
        public void Enable()
        {
            IsEnabled = true;
        }

        /// <summary>
        /// Activates the GPS Intermediate Driver.
        /// </summary>
        public void Disable()
        {
            IsEnabled = false;
        }

        /// <summary>
        /// Controls the actual GPS device responsible for providing raw data to the GPS Intermediate Driver.
        /// </summary>
        /// <remarks>This property is used to control which serial device will be shared.  Setting this
        /// property will cause the GID to be reconfigured in the registry.  The <strong>Refresh</strong>
        /// method should be called after setting this property to ensure that new settings are utilized by the driver.</remarks>
        public SerialDevice SourceDevice
        {
            get
            {
                return pSourceDevice;
            }
            set
            {
                // If it's not supported, don't make the key
                if (!pIsSupported)
                    throw new InvalidOperationException("The GPS Intermediate Driver is currently not supported by this device."); 

                // Set the current device
                pSourceDevice = value;

                // Write settings
                RegistryKey SourceKey = null;
                try
                {
                    // IMPORTANT!  On some devices such as the HTC Kaiser and HTC TYTN II, a proprietary driver is used.  In this case, it
                    // is CRITICAL to avoid writing any "Control Panel Configured Device" key!  This will cause a hang and force end users to
                    // perform a HARD RESET in order to undo these changes.  As a result, this key is only written if a non-proprietary driver
                    // is in use.

                    if (!IsProprietaryDriver)
                    {
                        // Config the driver to use a GPS Device
                        SourceKey = Registry.LocalMachine.OpenSubKey(ControlPanelKey, true);
                        if (SourceKey == null)
                            SourceKey = Registry.LocalMachine.CreateSubKey(ControlPanelKey);

                        // Set the port and baud rate, etc.
#if Framework20
                        SourceKey.SetValue("CommPort", value.PortName, RegistryValueKind.String);
#else
					    SourceKey.SetValue("CommPort", value.PortName);
#endif

#if Framework20
                        SourceKey.SetValue("Baud", (int)value.BaudRate, RegistryValueKind.DWord);
                        SourceKey.SetValue("InterfaceType", "COMM", RegistryValueKind.String);
#else
					SourceKey.SetValue("Baud", (int)value.BaudRate);
					SourceKey.SetValue("InterfaceType", "COMM");
#endif
                        
                        SourceKey.Close();

                        // Set the driver to use the control panel
                        SourceKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey + @"\Drivers", true);
                        if (SourceKey != null)
                        {
#if Framework20
                            SourceKey.SetValue("CurrentDriver", "Control Panel Configured Device", RegistryValueKind.String);
#else
						SourceKey.SetValue("CurrentDriver", "Control Panel Configured Device");
#endif
                        }

                        SourceKey.Close();
                    }
                }
                catch (FileNotFoundException)
                {
                    // Errors while writing to the key :P  disable support
                    pIsSupported = false;
                }
                catch (UnauthorizedAccessException)
                {
                    // Errors while writing to the key :P  disable support
                    pIsSupported = false;
                }
                catch
                {
                    throw;
                }
                finally
                {
                    if (SourceKey != null)
                        SourceKey.Close();
                }
            }
        }

        /// <summary>
        /// Controls the name of the shared port which should be used by applications.
        /// </summary>
        public string SharedPortName
        {
            get
            {
                RegistryKey GPSDriverKey = null;
                try
                {
                    GPSDriverKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey + @"\Multiplexer", false);

                    // If no key was found, the driver is not present.  
                    if (GPSDriverKey != null)
                    {
                        // Found!  Get the multiplexer port name
                        object Result = GPSDriverKey.GetValue("DriverInterface");
                        if (Result == null)
                            return null;
                        else
                            return Convert.ToString(Result);
                    }
                    else
                    {
                        return null;
                    }

                }
                catch
                {
                    throw;
                }
                finally
                {
                    if (GPSDriverKey != null)
                        GPSDriverKey.Close();
                }
            }
            set
            {
                // If it's not supported, don't make the key
                if (!pIsSupported)
                    throw new InvalidOperationException("The GPS Intermediate Driver is currently not supported by this device.");

                // If there's a proprietary driver, skip changing any of this information
                if (IsProprietaryDriver)
                    return;

                // Force the value to upper-case
                value = value.ToUpper(CultureInfo.InvariantCulture).Trim();

                RegistryKey GPSDriverKey = null;
                try
                {
                    // Open a handle to the GPSID Multiplexer key
                    GPSDriverKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey + @"\Multiplexer", true);

                    // If it doesn't exist, create it now
                    if (GPSDriverKey == null)
                        GPSDriverKey = Registry.LocalMachine.CreateSubKey(IntermediateDriverKey + @"\Multiplexer");

                    // Set the shared port name
                    GPSDriverKey.SetValue("DriverInterface", value);
                    GPSDriverKey.Close();

                    // Extract the prefix and index
                    string Prefix = value.Substring(0, 3);
                    string Index = value.Replace(Prefix, "").Replace(":", "").Trim();

                    // Config the GPSID to use this port name and index (COM4: is also known as "Prefix 'COM', Index '4'")
                    GPSDriverKey = Registry.LocalMachine.OpenSubKey(IntermediateDriverKey + @"\Multiplexer\ActiveDevice", true);
                    if (GPSDriverKey == null)
                        GPSDriverKey = Registry.LocalMachine.CreateSubKey(IntermediateDriverKey + @"\Multiplexer\ActiveDevice");
                    GPSDriverKey.SetValue("Index", Convert.ToInt32(Index));
                    GPSDriverKey.SetValue("Prefix", Prefix);
                }
                catch
                {
                    throw;
                }
                finally
                {
                    if (GPSDriverKey != null)
                        GPSDriverKey.Close();
                }
            }
        }
    }
}

#endif