using System;
using System.Text;
using GeoFramework;
using GeoFramework.Gps;

namespace GeoFramework.Gps.Nmea
{
	/// <summary>
	/// Represents a $GPGGQ sentence containing time, position, and fix quality
	/// information.
	/// </summary>
	public sealed class GpggqSentence : NmeaSentence
    {
        #region Private variables

        private DateTime pUtcDateTime;
		private Position pPosition;
		private FixQuality pFixQuality;
		private int pFixedSatelliteCount;
		private DilutionOfPrecision pMeanDilutionOfPrecision;
		private Distance pAltitude;

        #endregion

        #region Static members

        /// <summary>
        /// Controls the number of digits of precision written in latitude and longitude measurements.
        /// </summary>
        /// <remarks>
        /// This property allows sentences to include greater precision when they are written out as strings.
        /// By default, latitude and longitude values are written to only four digits of precision, yet as
        /// many as nine digits are required to preserve sub-millimeter accuracy.  Changing this value from
        /// its default of four to a number as high as 12 will maximize precision.  This property defaults to
        /// four in order to comply with the NMEA standard.
        /// </remarks>
        public static int MaximumPrecisionDigits = 4;

        #endregion

        #region Constructors

        public GpggqSentence(string sentence) 
            : base(sentence)
		{
			// Is the checksum valid?
			if (!IsValid && !IsPartialSentenceParsingAllowed)
			{
				pUtcDateTime = DateTime.MinValue;
				pPosition = Position.Empty;
				pFixQuality = FixQuality.NoFix;
				pMeanDilutionOfPrecision = DilutionOfPrecision.Maximum;
				pAltitude = Distance.Empty;
				// No.  Do no parsing
				return;
			}

			//			$--GGQ Header including talker ID
			//hhmmss.ss UTC time of position
			//mmddyy UTC date
			//			llll.ll(Latitude)
			//			a(Hemisphere, North Or South)
			//			yyyyy.yy(Longitude)
			//			a(East Or West)
			//x GPS quality indicator
			//0 = Fix not available or invalid
			//1 = No real-time position, navigation fix
			//2 = Real-time position, ambiguities not fixed
			//3 = Real-time position, ambiguities fixed
			//xx Number of satellites in use, 00 to 12.
			//x.x Coordinate quality
			//F-15 GPS1200 NMEA Message Formats
			//Example $GPGGQ,123208.00,093003,4724.5400344,N,00937.0472306,E,1,06,14.195,1174.674,M
			//*0
			//x.x Altitude of position marker above/below mean sea level in metres. If no orthometric
			//height is available the local ellipsoidal height will be exported. If the
			//local ellipsoidal height is not available either, the WGS 1984 ellipsoidal
			//height will be exported.
			//M Units of altitude as fixed text M

			// Set the time
			try
			{
				// Parse the current UTC date/time
				if(WordCount > 2 && Words[1].Length != 0 && Words[2].Length != 0)
				{
                    int UTCMonths = int.Parse(Words[2].Substring(0, 2), NmeaCultureInfo);
                    int UTCDays = int.Parse(Words[2].Substring(2, 2), NmeaCultureInfo);
                    int UTCYears = int.Parse(Words[2].Substring(4, 2), NmeaCultureInfo) + 2000;
                    int UTCHours = int.Parse(Words[1].Substring(0, 2), NmeaCultureInfo);
					int UTCMinutes = int.Parse(Words[1].Substring(2, 2), NmeaCultureInfo);

                    // Parse seconds and milliseconds as a float
                    float UTCSecondsRaw = float.Parse(Words[1].Substring(4), NmeaCultureInfo);

                    // Extract the integer portion of the seconds value
#if Framework20 && !PocketPC
                    int UTCSeconds = Convert.ToInt32(Math.Truncate(UTCSecondsRaw)) % 60;
#else
                    int UTCSeconds = Truncate(UTCSecondsRaw) % 60;
#endif
                    // Parse the milliseconds portion of the value
                    int UTCMilliseconds = Convert.ToInt32((UTCSecondsRaw - UTCSeconds) * 1000.0);

                    // Build a date and time, including milliseconds
#if Framework20
                    pUtcDateTime = new System.DateTime(UTCYears, UTCMonths, UTCDays, UTCHours, UTCMinutes, UTCSeconds, UTCMilliseconds, DateTimeKind.Utc);
#else
                    pUtcDateTime = new System.DateTime(UTCYears, UTCMonths, UTCDays, UTCHours, UTCMinutes, UTCSeconds, UTCMilliseconds);
#endif
				}
				else
				{
					pUtcDateTime = DateTime.MinValue;
				}

				// Update the GPS position
                if (WordCount > 6 && Words[3].Length != 0 && Words[4].Length != 0 && Words[5].Length != 0 && Words[6].Length != 0)
                {
                    pPosition = new Position(
                        new Latitude(int.Parse(Words[3].Substring(0, 2), NmeaCultureInfo),
                            double.Parse(Words[3].Substring(2), NmeaCultureInfo),
                            Words[4] == "N" ? LatitudeHemisphere.North : LatitudeHemisphere.South)
                        ,
                        new Longitude(int.Parse(Words[5].Substring(0, 3), NmeaCultureInfo),
                            double.Parse(Words[5].Substring(3), NmeaCultureInfo),
                            Words[6] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West)
                        );

                    //pPosition = new Position(
                    //    Latitude.Parse(Words[3].Substring(0, 2) + "" + Words[3].Substring(2) + "'" + Words[4], NmeaCultureInfo),
                    //    Longitude.Parse(Words[5].Substring(0, 3) + "" + Words[5].Substring(3) + "'" + Words[6], NmeaCultureInfo));
                }
                else
                {
                    pPosition = GeoFramework.Position.Empty;
                }

				// Parse fix quality
				if(WordCount > 7 && Words[7].Length != 0)
					{
						switch (int.Parse(Words[7], NmeaCultureInfo))
						{
							case 0:
								pFixQuality = FixQuality.NoFix;
								break;
							case 1:
								pFixQuality = FixQuality.NavigationFixNoRealtimePosition;
								break;
							case 2:
								pFixQuality = FixQuality.RealTimePositionAmbiguitiesNotFixed;
								break;
							case 3:
								pFixQuality = FixQuality.RealTimePositionAmbiguitiesFixed;
								break;
						}
					}
				else
				{
					pFixQuality = FixQuality.NoFix;
				}

				// Parse fixed satellite count
				if(WordCount > 8 && Words[8].Length != 0)
						pFixedSatelliteCount = int.Parse(Words[8], NmeaCultureInfo);
				else
					pFixedSatelliteCount = 0;
	
				// Parse mean Dilution of Precision
				if(WordCount > 9 && Words[9].Length != 0)
						pMeanDilutionOfPrecision = new DilutionOfPrecision(double.Parse(Words[9], NmeaCultureInfo));
				else
					pMeanDilutionOfPrecision = DilutionOfPrecision.Maximum;
	
				// Update the Altitude
				if(WordCount > 10 && Words[10].Length != 0)
						pAltitude = new Distance(double.Parse(Words[10], NmeaCultureInfo), DistanceUnit.Meters).ToLocalUnitType();
				else
					pAltitude = Distance.Empty;
			}
			catch (Exception ex)
			{
				throw new NmeaSentenceException(this, "A $GPGGQ sentence contained unrecognizable data.", ex);
			}
		}

		public GpggqSentence(byte[] bytes) 
            : this(ASCIIEncoding.ASCII.GetString(bytes, 0, bytes.Length))
		{}

		public GpggqSentence(DateTime utcDateTime, Position position, FixQuality fixQuality, int fixedSatelliteCount,
			DilutionOfPrecision meanDilutionOfPrecision, Distance altitude)
		{
			pUtcDateTime = utcDateTime;
			pPosition = RoundToNmea(position);
			pFixQuality = fixQuality;
			pFixedSatelliteCount = fixedSatelliteCount;
			pMeanDilutionOfPrecision = meanDilutionOfPrecision;
			pAltitude = altitude;
			// Build a sentence from the values
			// Output the information as a GPGGQ sentence (!!)
			StringBuilder Sentence = new StringBuilder();
			Sentence.Append("$GPGGQ,");
			// Convert UTC data time to a string
			Sentence.Append(UtcDateTime.Hour.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Minute.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Second.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(".");
			Sentence.Append(UtcDateTime.Millisecond.ToString(NmeaCultureInfo).PadLeft(3, '0'));
			Sentence.Append(",");
			// Append latitude
            Sentence.Append(Position.Latitude.ToString("hhmm." + "".PadLeft(MaximumPrecisionDigits, 'm'), NmeaCultureInfo).Replace("-", ""));
            Sentence.Append(",");
			Sentence.Append(Position.Latitude.ToString("i"));
			Sentence.Append(",");
			// Append longitude
            Sentence.Append(Position.Longitude.ToString("hhhmm." + "".PadLeft(MaximumPrecisionDigits, 'm'), NmeaCultureInfo).Replace("-", ""));
            Sentence.Append(",");
			Sentence.Append(Position.Longitude.ToString("i"));
			Sentence.Append(",");
			// Append fix quality
			switch(pFixQuality)
			{
				case FixQuality.NoFix:
					Sentence.Append("0");
					break;
				case FixQuality.NavigationFixNoRealtimePosition:
					Sentence.Append("1");
					break;
				case FixQuality.RealTimePositionAmbiguitiesNotFixed:
					Sentence.Append("2");
					break;
				case FixQuality.RealTimePositionAmbiguitiesFixed:
					Sentence.Append("3");
					break;
				default:
					Sentence.Append("0");
					break;
			}
			Sentence.Append(",");
			// Append fixed satellite count
			Sentence.Append(FixedSatelliteCount.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(",");
			// Append MDOP
			Sentence.Append(MeanDilutionOfPrecision.Value.ToString("#.#", NmeaCultureInfo));
			Sentence.Append(",");
			// Append altitude
			Sentence.Append(Altitude.ToMeters().ToString("v.v", NmeaCultureInfo));
			Sentence.Append(",M");
			// Set the sentence
			SetSentence(Sentence.ToString());
			SetChecksum();
        }

        #endregion

        #region Public Members

        /// <summary>Returns the current satellite-derived time.</summary>
		public DateTime UtcDateTime
		{
			get
			{
				return pUtcDateTime;
			}
		}

		/// <summary>Returns the current location on Earth.</summary>
		public Position Position
		{
			get
			{
				return pPosition;
			}
		}

		/// <summary>Returns the quality level of the current fix.</summary>
		public FixQuality FixQuality
		{
			get
			{
				return pFixQuality;
			}
		}

		/// <summary>Returns the number of satellites involved in the current fix.</summary>
		public int FixedSatelliteCount
		{
			get
			{
				return pFixedSatelliteCount;
			}
		}

		/// <summary>
		/// Indicates whether a satellite fix is currently in progress.
		/// </summary>
		public bool IsFixObtained
		{
			get
			{
				// Update the fix flag
				return 
					pFixQuality == FixQuality.DifferentialGpsFix 
					| pFixQuality == FixQuality.GpsFix 
					| pFixQuality == FixQuality.RealTimePositionAmbiguitiesFixed 
					| pFixQuality == FixQuality.RealTimePositionAmbiguitiesNotFixed;
			}
		}

		/// <summary>
		/// Returns the amount of precision as it relates to latitude, longitude and
		/// altitude.
		/// </summary>
		public DilutionOfPrecision MeanDilutionOfPrecision
		{
			get
			{
				return pMeanDilutionOfPrecision;
			}
		}

		/// <summary>Returns the current altitude above sea level.</summary>
		public Distance Altitude
		{
			get
			{
				return pAltitude;
			}
		}

		/// <summary>
		/// Indicates that this sentence is supported by GPS.NET.
		/// </summary>
		public override bool IsSupported
		{
			get
			{
				return true;
			}
		}

        #endregion
    }
}
