
using System;
using System.Text;
using GeoFramework;

namespace GeoFramework.Gps.Nmea
{
	/// <summary>Represents a $GPGLL sentence containing latitude and longitude.</summary>
	public sealed class GpgllSentence : NmeaSentence
    {
        #region Private variables

        private DateTime pUtcDateTime;
		private Position pPosition;
		private bool pIsFixObtained;

        #endregion

        #region Static members

        /// <summary>
        /// Controls the number of digits of precision written in latitude and longitude measurements.
        /// </summary>
        /// <remarks>
        /// This property allows sentences to include greater precision when they are written out as strings.
        /// By default, latitude and longitude values are written to only four digits of precision, yet as
        /// many as nine digits are required to preserve sub-millimeter accuracy.  Changing this value from
        /// its default of four to a number as high as 12 will maximize precision.  This property defaults to
        /// four in order to comply with the NMEA standard.
        /// </remarks>
        public static int MaximumPrecisionDigits = 4;

        /// <summary>Represents a $GPGLL sentence with no values.</summary>
        public static readonly GpgllSentence Empty = new GpgllSentence(Position.Empty, DateTime.UtcNow, false);

        /// <summary>Returns a new instance by interpreting the specified string.</summary>
        public static GpgllSentence Parse(string sentence)
        {
            return new GpgllSentence(sentence);
        }

        #endregion

		#region Constructors

		public GpgllSentence(string sentence) 
            : base(sentence)
		{
			// Is the checksum valid?
			if(!IsValid && !IsPartialSentenceParsingAllowed)
			{
				pUtcDateTime = DateTime.MinValue;
				pPosition = Position.Empty;
				// No.  Do no parsing
				return;
			}

			//Where:
			//	 4916.46,N	Latitude 49 deg. 16.45 min. North
			//	 12311.12,W   Longitude 123 deg. 11.12 min. West
			//	 225444	   Fix taken at 22:54:44 UTC
			//	 A			Fix valid or V (void)
			//	 *23		  checksum data

			try
			{
				// Update the GPS position
                if (WordCount > 4 && Words[1].Length != 0 && Words[2].Length != 0 && Words[3].Length != 0 && Words[4].Length != 0)
                {
#if PocketPC && !Framework20
                    int a = int.Parse(Words[1].Substring(0, 2), NmeaCultureInfo);
                    double b = double.Parse(Words[1].Substring(2), NmeaCultureInfo);
                    LatitudeHemisphere c = (Words[2] == "N" ? LatitudeHemisphere.North : LatitudeHemisphere.South);
                    Latitude d = new Latitude(int.Parse(Words[1].Substring(0, 2), NmeaCultureInfo),
                        double.Parse(Words[1].Substring(2), NmeaCultureInfo),
                        Words[2] == "N" ? LatitudeHemisphere.North : LatitudeHemisphere.South);
							

                    int e = int.Parse(Words[3].Substring(0, 3), NmeaCultureInfo);
                    double f = double.Parse(Words[3].Substring(3), NmeaCultureInfo);
                    LongitudeHemisphere g = (Words[4] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West);
                    Longitude h = new Longitude(int.Parse(Words[3].Substring(0, 3), NmeaCultureInfo),
                        double.Parse(Words[3].Substring(3), NmeaCultureInfo),
                        Words[4] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West);

                    pPosition = new Position(d, h);
#else
                    pPosition = new Position(
                        new Latitude(int.Parse(Words[1].Substring(0, 2), NmeaCultureInfo),
                        double.Parse(Words[1].Substring(2), NmeaCultureInfo),
                        Words[2] == "N" ? LatitudeHemisphere.North : LatitudeHemisphere.South)
                        ,
                        new Longitude(int.Parse(Words[3].Substring(0, 3), NmeaCultureInfo),
                        double.Parse(Words[3].Substring(3), NmeaCultureInfo),
                        Words[4] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West)
                        );
#endif
                    //pPosition = new Position(
                    //    Latitude.Parse(Words[1].Substring(0, 2) + "" + Words[1].Substring(2) + "'" + Words[2], NmeaCultureInfo),
                    //    Longitude.Parse(Words[3].Substring(0, 3) + "" + Words[3].Substring(3) + "'" + Words[4], NmeaCultureInfo));

                }
                else
                {
                    pPosition = GeoFramework.Position.Empty;
                }

				// Update the fix time
				if(WordCount > 5 && Words[5].Length != 0)
				{
					int UTCHours = Convert.ToInt32(Words[5].Substring(0, 2), NmeaCultureInfo);
					int UTCMinutes = Convert.ToInt32(Words[5].Substring(2, 2), NmeaCultureInfo);

                    // Parse seconds and milliseconds as a float
                    float UTCSecondsRaw = float.Parse(Words[5].Substring(4), NmeaCultureInfo);

                    // Extract the integer portion of the seconds value
#if Framework20 && !PocketPC
                    int UTCSeconds = Convert.ToInt32(Math.Truncate(UTCSecondsRaw)) % 60;
#else
                    int UTCSeconds = Truncate(UTCSecondsRaw) % 60;
#endif
                    // Parse the milliseconds portion of the value
                    int UTCMilliseconds = Convert.ToInt32((UTCSecondsRaw - UTCSeconds) * 1000.0);

					// Get the current satellite-derived date
					DateTime pToday = GeoFramework.IO.Devices.CurrentUtcDateTime;

					// Build a new date/time from the values
#if Framework20
                    pUtcDateTime = new System.DateTime(pToday.Year, pToday.Month, pToday.Day, UTCHours, UTCMinutes, UTCSeconds, UTCMilliseconds, DateTimeKind.Utc);
#else
					pUtcDateTime = new System.DateTime(pToday.Year, pToday.Month, pToday.Day, UTCHours, UTCMinutes, UTCSeconds, UTCMilliseconds);
#endif
				}
				else
				{
					pUtcDateTime = DateTime.MinValue;
				}
				// Is a fix currently obtained?
				if(WordCount > 6 && Words[6].Length != 0)
				{
					pIsFixObtained = Words[6] == "A" ? true : false;
				}
				else
				{
					pIsFixObtained = false;
				}
			}
			catch(Exception ex)
			{
				throw new NmeaSentenceException(this, "A $GPGLL sentence contained unrecognizable data.", ex);
			}
		}

		public GpgllSentence(byte[] bytes) 
            : this(ASCIIEncoding.ASCII.GetString(bytes, 0, bytes.Length))
		{}	

		public GpgllSentence(Position position, DateTime utcDateTime, bool isFixObtained)
		{
			pUtcDateTime = utcDateTime;
			pPosition = RoundToNmea(position);
			pIsFixObtained = isFixObtained;
			// Build the sentence from the position information
			StringBuilder Sentence = new StringBuilder();
			Sentence.Append("$GPGLL,");
			// Append latitude
            if (Position != null)
            {
                Sentence.Append(Position.Latitude.ToString("hhmm." + "".PadLeft(MaximumPrecisionDigits, 'm'), NmeaCultureInfo).Replace("-", ""));
            }
			Sentence.Append(",");
			if(Position != null)
				Sentence.Append(Position.Latitude.ToString("i"));
			Sentence.Append(",");
			// Append longitude
            if (Position != null)
            {
                Sentence.Append(Position.Longitude.ToString("hhhmm." + "".PadLeft(MaximumPrecisionDigits, 'm'), NmeaCultureInfo).Replace("-", ""));
            }
			Sentence.Append(",");
			if(Position != null)
				Sentence.Append(Position.Longitude.ToString("i"));
			Sentence.Append(",");
			// Convert UTC data time to a string
			Sentence.Append(UtcDateTime.Hour.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Minute.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Second.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(".");
			Sentence.Append(UtcDateTime.Millisecond.ToString(NmeaCultureInfo).PadLeft(3, '0'));
			Sentence.Append(",");
			// Write fix information
			if(IsFixObtained)
				Sentence.Append("A");
			else
				Sentence.Append("V");
			// Set the sentence
			SetSentence(Sentence.ToString());
			SetChecksum();
		}

		#endregion

		#region Public Members

		/// <summary>Returns whether the GPS device has a fix on the current location.</summary>
		public bool IsFixObtained
		{
			get
			{
				return pIsFixObtained;
			}
		}

		/// <summary>Returns the current satellite-derived time.</summary>
		public DateTime UtcDateTime
		{
			get
			{
				return pUtcDateTime;
			}
		}

		/// <summary>Returns the current location on Earth's surface.</summary>
		public Position Position
		{
			get
			{
				return pPosition;
			}
		}

		
		/// <summary>
		/// Indicates that this sentence is supported by GPS.NET.
		/// </summary>
		public override bool IsSupported
		{
			get
			{
				return true;
			}
		}

		#endregion
	}
}
