
using System;
using System.Text;
using GeoFramework;

namespace GeoFramework.Gps.Nmea
{
	/// <summary>
	/// Leica devices only. Represents a $GPLLK sentence describing satellite-derived
	/// time, location, and fix quality information.
	/// </summary>
	public sealed class GpllkSentence : NmeaSentence
	{
		#region Private Variables

		private DateTime pUtcDateTime;
		private Position pPosition;
		private FixQuality pFixQuality;
		private int pFixedSatelliteCount;
		private DilutionOfPrecision pMeanDilutionOfPrecision;
		private Distance pAltitude;

		#endregion

		#region Constructors

		public GpllkSentence(string sentence) : base(sentence)
		{
			// Is the checksum valid?
			if (!IsValid && !IsPartialSentenceParsingAllowed)
			{
				pUtcDateTime = DateTime.MinValue;
				pPosition = Position.Empty;
				pFixQuality = FixQuality.NoFix;
				pMeanDilutionOfPrecision = DilutionOfPrecision.Maximum;
				pAltitude = Distance.Empty;
				// No.  Do no parsing
				return;
			}

			//$--LLK,hhmmss.ss,mmddyy,eeeeee.eee,M,nnnnnn.nnn,M,x,xx,x.x,x.x,M*hh<CR><LF>
			//hhmmss.ss UTC time of position
			//mmddyy UTC date
			//eeeeee.eee Grid Easting in metres
			//M Units of grid Easting as fixed text M
			//nnnnnn.nnn Grid Northing in metres
			//M Units of grid Northing as fixed text M
			//x GPS quality
			//0 = Fix not available or invalid
			//1 = No real-time position, navigation fix
			//2 = Real-time position, ambiguities not fixed
			//3 = Real-time position, ambiguities fixed
			//xx Number of satellites used in computation
			//			x.x(GDOP)
			//x.x Altitude of position marker above/below mean sea level in metres. If no orthometric
			//height is available the local ellipsoidal height will be exported.
			//M Units of altitude as fixed text M
			// Set the manufacturer
			//Receiver.Device.SetManufacturer(Manufacturer.Leica);
			// Set the time
			if(WordCount > 2)
			{
                if (Words[1].Length != 0 & Words[2].Length != 0)
                {
                    int UTCMonths = int.Parse(Words[2].Substring(0, 2), NmeaCultureInfo);
                    int UTCDays = int.Parse(Words[2].Substring(2, 2), NmeaCultureInfo);
                    int UTCYears = int.Parse(Words[2].Substring(4, 2), NmeaCultureInfo) + 2000;
                    int UTCHours = int.Parse(Words[1].Substring(0, 2), NmeaCultureInfo);
                    int UTCMinutes = int.Parse(Words[1].Substring(2, 2), NmeaCultureInfo);

                    // Parse seconds and milliseconds as a float
                    float UTCSecondsRaw = float.Parse(Words[1].Substring(4), NmeaCultureInfo);

                    // Extract the integer portion of the seconds value
#if Framework20 && !PocketPC
                    int UTCSeconds = Convert.ToInt32(Math.Truncate(UTCSecondsRaw)) % 60;
#else
                    int UTCSeconds = Truncate(UTCSecondsRaw) % 60;
#endif
                    // Parse the milliseconds portion of the value
                    int UTCMilliseconds = Convert.ToInt32((UTCSecondsRaw - UTCSeconds) * 1000.0);

                    // Build a date/time from the values, including milliseconds
#if Framework20
                    pUtcDateTime = new System.DateTime(UTCYears, UTCMonths, UTCDays, UTCHours, UTCMinutes, UTCSeconds, UTCMilliseconds, DateTimeKind.Utc);
#else
                    pUtcDateTime = new System.DateTime(UTCYears, UTCMonths, UTCDays, UTCHours, UTCMinutes, UTCSeconds, UTCMilliseconds);
#endif
                }
			}

			if(WordCount > 6)
			{
				if (Words[3].Length != 0 & Words[4].Length != 0 & Words[5].Length != 0 & Words[6].Length != 0)
				{
					double NewEasting = double.Parse(Words[3], NmeaCultureInfo);
					double NewNorthing = double.Parse(Words[5], NmeaCultureInfo);
					pPosition = new Position(ZoneLetter.Unknown, 0, new Distance(NewEasting, DistanceUnit.Meters), new Distance(NewNorthing, DistanceUnit.Meters));
				}
			}

			if(WordCount > 7)
			{
				if (Words[7].Length != 0)
				{
						switch (int.Parse(Words[7], NmeaCultureInfo))
						{
							case 0:
								pFixQuality = FixQuality.NoFix;
								break;
							case 1:
								pFixQuality = FixQuality.NavigationFixNoRealtimePosition;
								break;
							case 2:
								pFixQuality = FixQuality.RealTimePositionAmbiguitiesNotFixed;
								break;
							case 3:
								pFixQuality = FixQuality.RealTimePositionAmbiguitiesFixed;
								break;
						}
				}
			}

			// Update the satellite count
			if(WordCount > 8)
				if(Words[8].Length != 0)
					pFixedSatelliteCount = int.Parse(Words[8], NmeaCultureInfo);

			// Update mean DOP
			if(WordCount > 9)
				if(Words[9].Length != 0)
					pMeanDilutionOfPrecision = new DilutionOfPrecision(double.Parse(Words[9], NmeaCultureInfo));
	
			// Update the Altitude
			if(WordCount > 10)
				if (Words[10].Length != 0)
					pAltitude = new Distance(double.Parse(Words[10], NmeaCultureInfo), DistanceUnit.Meters).ToLocalUnitType();
		}

		//[CLSCompliant(false)]
		public GpllkSentence(DateTime utcDateTime, Position position, FixQuality fixQuality,
			int fixedSatelliteCount, DilutionOfPrecision meanDilutionOfPrecision,
			Distance altitude)
		{
			// Set variables
			pUtcDateTime = utcDateTime;
			pPosition = position;
			pFixQuality = fixQuality;
			pFixedSatelliteCount = fixedSatelliteCount;
			pMeanDilutionOfPrecision = meanDilutionOfPrecision;
			pAltitude = altitude;
			// Build the sentence from the position information
			// Output the information as a GPLLK sentence (!!)
			StringBuilder Sentence = new StringBuilder();
			Sentence.Append("$GPLLK,");
			// Convert UTC time to a string
			Sentence.Append(UtcDateTime.Hour.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Minute.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Second.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(".");
			Sentence.Append(UtcDateTime.Millisecond.ToString(NmeaCultureInfo).PadLeft(3, '0'));
			Sentence.Append(",");
			// Convert UTC time to a string
			Sentence.Append(UtcDateTime.Month.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Day.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Year.ToString(NmeaCultureInfo).Substring(2, 2).PadLeft(2, '0'));
            Sentence.Append(",");
			// Append easting
			Sentence.Append(Position.ToString("eeeeee.eee"));
			Sentence.Append(",M,");
			Sentence.Append(Position.ToString("nnnnnn.nnn"));
			Sentence.Append(",M,");
			// Append fix quality
			switch(FixQuality)
			{
				case FixQuality.NoFix:
					Sentence.Append("0");
					break;
				case FixQuality.NavigationFixNoRealtimePosition:
					Sentence.Append("1");
					break;
				case FixQuality.RealTimePositionAmbiguitiesNotFixed:
					Sentence.Append("2");
					break;
				case FixQuality.RealTimePositionAmbiguitiesFixed:
					Sentence.Append("3");
					break;
				default:
					Sentence.Append("0");
					break;
			}
			Sentence.Append(",");
			// Append fixed satellite count
			Sentence.Append(FixedSatelliteCount.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(",");
			// Append PDOP
			Sentence.Append(MeanDilutionOfPrecision.Value.ToString("#.#", NmeaCultureInfo));
			Sentence.Append(",");
			// Append altitude
			Sentence.Append(Altitude.ToMeters().ToString("v.v", NmeaCultureInfo));
			Sentence.Append(",M");
			// Set the sentence
			SetSentence(Sentence.ToString());
			SetChecksum();
		}

		#endregion

		#region Public Members

		/// <summary>Returns the current satellite-derived time.</summary>
		public DateTime UtcDateTime
		{
			get
			{
				return pUtcDateTime;
			}
		}

		/// <summary>Returns the current location as a UTM coordinate.</summary>
		public Position Position
		{
			get
			{
				return pPosition;
			}
		}

		/// <summary>Returns the quality level of the current fix.</summary>
		public FixQuality FixQuality
		{
			get
			{
				return pFixQuality;
			}
		}

		/// <summary>Returns the number of satellites involved in the current fix.</summary>
		public int FixedSatelliteCount
		{
			get
			{
				return pFixedSatelliteCount;
			}
		}

		/// <summary>
		/// Indicates whether a satellite fix is currently in progress.
		/// </summary>
		public bool IsFixObtained
		{
			get
			{
				// Update the fix flag
				return 
					pFixQuality == FixQuality.DifferentialGpsFix 
					| pFixQuality == FixQuality.GpsFix 
					| pFixQuality == FixQuality.RealTimePositionAmbiguitiesFixed 
					| pFixQuality == FixQuality.RealTimePositionAmbiguitiesNotFixed;
			}
		}

		/// <summary>
		/// Returns the current level of precision as it relates to UTM easting, northing,
		/// and altitude.
		/// </summary>
		public DilutionOfPrecision MeanDilutionOfPrecision
		{
			get
			{
				return pMeanDilutionOfPrecision;
			}
		}

		/// <summary>Returns the current distance above sea level.</summary>
		public Distance Altitude
		{
			get
			{
				return pAltitude;
			}
		}

		/// <summary>
		/// Indicates that this sentence is supported by GPS.NET.
		/// </summary>
		public override bool IsSupported
		{
			get
			{
				return true;
			}
		}

		#endregion
	}
}
