
using System;
using System.Text;
using GeoFramework;

namespace GeoFramework.Gps.Nmea
{
	/// <summary>
	/// Represents a $GPRMC sentence describing common position and precision
	/// information.
	/// </summary>
	/// <remarks>This sentence, known as the "Recommended Minimum" sentence, contains nearly
	/// every important piece of GPS information, with the expection of precision.  </remarks>
	public sealed class GprmcSentence : NmeaSentence
	{
		private DateTime pUtcDateTime;
		private bool pIsFixObtained;
		private Position pPosition;
		private Speed pSpeed;
		private Azimuth pBearing;
		private Longitude pMagneticVariation;

        /// <summary>
        /// Controls the number of digits of precision written in latitude and longitude measurements.
        /// </summary>
        /// <remarks>
        /// This property allows sentences to include greater precision when they are written out as strings.
        /// By default, latitude and longitude values are written to only four digits of precision, yet as
        /// many as nine digits are required to preserve sub-millimeter accuracy.  Changing this value from
        /// its default of four to a number as high as 12 will maximize precision.  This property defaults to
        /// four in order to comply with the NMEA standard.
        /// </remarks>
        public static int MaximumPrecisionDigits = 4;

        #region Constructors

        public GprmcSentence(string sentence) 
            : base(sentence)
		{
			// Is the checksum valid?
			if (!IsValid && !IsPartialSentenceParsingAllowed)
			{
				pUtcDateTime = DateTime.MinValue;
				pPosition = Position.Empty;
				pSpeed = Speed.Empty;
				pBearing = Azimuth.Empty;
				pMagneticVariation = Longitude.Empty;
				// No.  Do no parsing
				return;
			}


			try
			{

				// Set the time
				if(WordCount > 1 && Words[1].Length != 0)
				{
					int UTCHours = int.Parse(Words[1].Substring(0, 2), NmeaCultureInfo);
					int UTCMinutes = int.Parse(Words[1].Substring(2, 2), NmeaCultureInfo);


                    // Parse seconds and milliseconds as a float
                    float UTCSecondsRaw = float.Parse(Words[1].Substring(4), NmeaCultureInfo);

                    // Extract the integer portion of the seconds value
#if Framework20 && !PocketPC
                    int UTCSeconds = Convert.ToInt32(Math.Truncate(UTCSecondsRaw)) % 60;
#else
                    int UTCSeconds = Truncate(UTCSecondsRaw) % 60;
#endif
                    // Parse the milliseconds portion of the value
                    int UTCMilliseconds = Convert.ToInt32((UTCSecondsRaw - UTCSeconds) * 1000.0);
                 
                    // Extract the UTC date
                    int UTCDays = 0;
                    int UTCMonths = 0;
                    int UTCYears = 0;
                    if (Words[9].Length > 0)
                    {
                        UTCDays = int.Parse(Words[9].Substring(0, 2), NmeaCultureInfo);
                        UTCMonths = int.Parse(Words[9].Substring(2, 2), NmeaCultureInfo);
                        UTCYears = int.Parse(Words[9].Substring(4, 2), NmeaCultureInfo) + 2000;
                    }
                    else
                    {
                        // Use the local system clock's date :(
                        DateTime UtcNow = DateTime.UtcNow;
                        UTCDays = UtcNow.Day;
                        UTCMonths = UtcNow.Month;
                        UTCYears = UtcNow.Year;
                    }

                    // Build a date/time from the values, including milliseconds
#if Framework20
                    pUtcDateTime = new System.DateTime(UTCYears, UTCMonths, UTCDays, UTCHours, UTCMinutes, UTCSeconds, UTCMilliseconds, DateTimeKind.Utc);
#else
                    pUtcDateTime = new System.DateTime(UTCYears, UTCMonths, UTCDays, UTCHours, UTCMinutes, UTCSeconds, UTCMilliseconds);
#endif
				}
				else
					pUtcDateTime = DateTime.MinValue;
				
				// Set the GPS Validity status
				if(WordCount > 2 && Words[2].Length != 0)
					pIsFixObtained = Words[2] == "A" ? true : false;
				
				// Update the GPS position
				if(WordCount > 6 && Words[6].Length != 0)
				{
#if PocketPC && !Framework20
                    int a = int.Parse(Words[3].Substring(0, 2), NmeaCultureInfo);
                    double b = double.Parse(Words[3].Substring(2), NmeaCultureInfo);
                    LatitudeHemisphere c = (Words[4] == "N" ? LatitudeHemisphere.North : LatitudeHemisphere.South);
                    Latitude d = new Latitude(int.Parse(Words[3].Substring(0, 2), NmeaCultureInfo),
                        double.Parse(Words[3].Substring(2), NmeaCultureInfo),
                        Words[4] == "N" ? LatitudeHemisphere.North : LatitudeHemisphere.South);
						
                    int e = int.Parse(Words[5].Substring(0, 3), NmeaCultureInfo);
                    string jkl = Words[5].Substring(3);
                    double f = double.Parse(Words[5].Substring(3), NmeaCultureInfo);
                    LongitudeHemisphere g = (Words[6] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West);
                    Longitude h = new Longitude(int.Parse(Words[5].Substring(0, 3), NmeaCultureInfo),
                        double.Parse(Words[5].Substring(3), NmeaCultureInfo),
                        Words[6] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West);
						
                    pPosition = new Position(d, h);
#else
                    pPosition = new Position(
                        new Latitude(int.Parse(Words[3].Substring(0, 2), NmeaCultureInfo),
                        double.Parse(Words[3].Substring(2), NmeaCultureInfo),
                        Words[4] == "N" ? LatitudeHemisphere.North : LatitudeHemisphere.South)
                        ,
                        new Longitude(int.Parse(Words[5].Substring(0, 3), NmeaCultureInfo),
                        double.Parse(Words[5].Substring(3), NmeaCultureInfo),
                        Words[6] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West)
                        );
#endif
					//pPosition = new Position(
					//    Latitude.Parse(Words[3].Substring(0, 2) + "" + Words[3].Substring(2) + "'" + Words[4], NmeaCultureInfo),
					//    Longitude.Parse(Words[5].Substring(0, 3) + "" + Words[5].Substring(3) + "'" + Words[6], NmeaCultureInfo));
				}
				else
					pPosition = Position.Empty;							
				
				// Update the speed over ground
				if(WordCount > 7 && Words[7].Length != 0)
					// Parse the speed measurement 
					pSpeed = new Speed(double.Parse(Words[7], NmeaCultureInfo), SpeedUnit.Knots).ToLocalUnitType();
				else
					pSpeed = Speed.Empty;
				
				// Update the direction of travel
				if(WordCount > 8 && Words[8].Length != 0)
					pBearing = Azimuth.Parse(Words[8], NmeaCultureInfo);
				else
					pBearing = Azimuth.Empty;
				
				// Update the magnetic variation
				if(WordCount > 11 && Words[10].Length != 0 && Words[11].Length != 0)
				{
					pMagneticVariation = new Longitude(
						double.Parse(Words[10], NmeaCultureInfo),
						Words[11] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West);
				}
				else
				{
					pMagneticVariation = Longitude.Empty;
				}
			}
			catch(Exception ex)
			{
				throw new NmeaSentenceException(this, "A $GPRMC sentence contained unrecognizable data.", ex);
			}
		}

		public GprmcSentence(byte[] bytes) 
            : this(ASCIIEncoding.ASCII.GetString(bytes, 0, bytes.Length))
		{
		}

		//[CLSCompliant(false)]
		public GprmcSentence(DateTime utcDateTime, bool isFixObtained, Position position,
			Speed speed, Azimuth bearing, Longitude magneticVariation)
		{
			pUtcDateTime = utcDateTime;
			pIsFixObtained = isFixObtained;
			pPosition = RoundToNmea(position);
			pSpeed = speed;
			pBearing = bearing;
			pMagneticVariation = magneticVariation;
			// Build the sentence from the position information
			// Output the information as a GPGGA sentence (!!)
			StringBuilder Sentence = new StringBuilder();
			Sentence.Append("$GPRMC,");
			// Convert UTC data time to a string
			Sentence.Append(UtcDateTime.Hour.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Minute.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Second.ToString(NmeaCultureInfo).PadLeft(2, '0'));
            Sentence.Append(".");
            Sentence.Append(UtcDateTime.Millisecond.ToString(NmeaCultureInfo).PadLeft(3, '0'));
			Sentence.Append(",");
			// Write fix information
			if(pIsFixObtained)
				Sentence.Append("A");
			else
				Sentence.Append("V");
			Sentence.Append(",");
			// Append latitude
            Sentence.Append(Position.Latitude.ToString("hhmm." + "".PadLeft(MaximumPrecisionDigits, 'm'), NmeaCultureInfo).Replace("-", ""));
            Sentence.Append(",");
			Sentence.Append(Position.Latitude.ToString("i"));
			Sentence.Append(",");
			// Append longitude
            Sentence.Append(Position.Longitude.ToString("hhhmm." + "".PadLeft(MaximumPrecisionDigits, 'm'), NmeaCultureInfo).Replace("-", ""));
            Sentence.Append(",");
			Sentence.Append(Position.Longitude.ToString("i"));
			Sentence.Append(",");
			// Append speed
			Sentence.Append(Speed.ToKnots().ToString("v.v", NmeaCultureInfo));
			Sentence.Append(",");
			// Append bearing
			Sentence.Append(Bearing.ToString("d.d", NmeaCultureInfo));
			Sentence.Append(",");
			// Append the date
			Sentence.Append(UtcDateTime.Day.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Month.ToString(NmeaCultureInfo).PadLeft(2, '0'));
			Sentence.Append(UtcDateTime.Year.ToString(NmeaCultureInfo).PadLeft(4, '0').Substring(2, 2));
			Sentence.Append(",");
			// Append magnetic variation
			Sentence.Append(MagneticVariation.ToString("d.d", NmeaCultureInfo));
			// Set the sentence
			SetSentence(Sentence.ToString());
			SetChecksum();
        }

        #endregion

        #region Static Members

        /// <summary>Returns a $GPRMC sentence containing empty values.</summary>
		public static readonly GprmcSentence Empty = new GprmcSentence(DateTime.MinValue,
			false, Position.Empty, Speed.Empty, Azimuth.Empty, Longitude.Empty);

		/// <summary>Returns a new instance by interpreting the specified string.</summary>
		public static GprmcSentence Parse(string sentence)
		{
			return new GprmcSentence(sentence);
		}

		#endregion

		#region Public Members

		/// <summary>Returns the current satellite-derived time.</summary>
		public DateTime UtcDateTime
		{
			get
			{
				return pUtcDateTime;
			}
		}

		/// <summary>Returns whether the GPS device can calculate the current location.</summary>
		public bool IsFixObtained
		{
			get
			{
				return pIsFixObtained;
			}
		}

		/// <summary>Returns the current location on Earth's surface.</summary>
		public Position Position
		{
			get
			{
				return pPosition;
			}
		}

		/// <summary>Returns the current rate of travel.</summary>
		public Speed Speed
		{
			get
			{
				return pSpeed;
			}
		}

		/// <summary>Returns the current direction of travel.</summary>
		public Azimuth Bearing
		{
			get
			{
				return pBearing;
			}
		}

		/// <summary>Returns the current difference between magnetic north and true north.</summary>
		public Longitude MagneticVariation
		{
			get
			{
				return pMagneticVariation;
			}
		}

		/// <summary>
		/// Indicates that this sentence is supported by GPS.NET.
		/// </summary>
		public override bool IsSupported
		{
			get
			{
				return true;
			}
		}

		#endregion
	}
}
