
using System;
using System.Text;
using GeoFramework;
using GeoFramework.Gps;

namespace GeoFramework.Gps.Nmea
{
	/// <summary>
	/// Garmin devices only. Represents an $HCHDG sentence containing the current
	/// magnetic heading, deviation and variation.
	/// </summary>
    public class HchdgSentence : NmeaSentence
	{
		#region Private Variables

		private Azimuth pMagneticBearing = Azimuth.Empty;
		private Longitude pMagneticDeviation = Longitude.Empty;
		private Longitude pMagneticVariation = Longitude.Empty;

		#endregion

		#region Constructors

		public HchdgSentence(string sentence) : base(sentence)
		{
			// Is the checksum valid?
            if (!IsValid && !IsPartialSentenceParsingAllowed)
				// No.  Do no parsing
				return;

			// Garmin-only sentence:
			//This sentence is used on Garmin eTrex summit, Vista and GPS76S receivers to output the value of the internal flux-gate compass. Only the magnetic heading and magnetic variation is shown in the message. 

			//  $HCHDG,101.1,,,7.1,W*3C
			//where:
			//	 HCHDG	Magnetic heading, deviation, variation
			//	 101.1	heading
			//	,,	   deviation (no data)
			//	 7.1,W	variation

			try
			{
				if(WordCount > 1)
					if(Words[1].Length != 0)
						pMagneticBearing = Azimuth.Parse(Words[1], NmeaCultureInfo);

				if(WordCount > 3)
                    if (Words[2].Length != 0 && Words[3].Length != 0)
                    {
                        pMagneticVariation = new Longitude(
                            double.Parse(Words[2], NmeaCultureInfo),
                            Words[3] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West);

                        //pMagneticDeviation = Longitude.Parse(Words[2] + "" + Words[3], NmeaCultureInfo);
                    }

				if(WordCount > 5)
				{
					// Get the magnetic variation
					if (Words[4].Length != 0 & Words[5].Length != 0)
					{
						// Get the degree hemisphere
                        pMagneticVariation = new Longitude(
                            double.Parse(Words[4], NmeaCultureInfo),
                            Words[5] == "E" ? LongitudeHemisphere.East : LongitudeHemisphere.West);

                        //switch (Words[5].ToUpper(NmeaCultureInfo))
                        //{
                        //    case "E":
                        //        pMagneticVariation = new Longitude(double.Parse(Words[4], NmeaCultureInfo));
                        //        break;
                        //    case "W":
                        //        pMagneticVariation = new Longitude(-double.Parse(Words[4], NmeaCultureInfo));
                        //        break;
                        //}
					}

				}
			}
			catch(Exception ex)
			{
				throw new NmeaSentenceException(this, "A $GPGGA sentence contained unrecognizable data.", ex);
			}

		}

		public HchdgSentence(byte[] bytes) 
            : this(ASCIIEncoding.ASCII.GetString(bytes, 0, bytes.Length))
		{}

		//[CLSCompliant(false)]
		public HchdgSentence(Azimuth magneticBearing, Longitude magneticDeviation,
			Longitude magneticVariation)
		{
			pMagneticBearing = magneticBearing;
			pMagneticDeviation = magneticDeviation;
			pMagneticVariation = magneticVariation;
			// Build a sentence
			StringBuilder Sentence = new StringBuilder();
			Sentence.Append("$HCHDG,");
			// Append mangetic bearing
			Sentence.Append(MagneticBearing.ToString("d.d", NmeaCultureInfo));
			Sentence.Append(",");
			// Append deviation
			Sentence.Append(MagneticDeviation.ToString("d.d", NmeaCultureInfo));
			Sentence.Append(",");
			// Append variation
			Sentence.Append(MagneticVariation.ToString("d.d", NmeaCultureInfo));
			Sentence.Append(",");
			Sentence.Append(MagneticVariation.ToString("i", NmeaCultureInfo));
			// Append the checksum 
			SetSentence(Sentence.ToString());
			SetChecksum();
		}

		#endregion

		#region Public Members

		/// <summary>Returns the current direction of travel in relation to Magnetic North.</summary>
		public Azimuth MagneticBearing
		{
			get
			{
				return pMagneticBearing;
			}
		}

		public Longitude MagneticDeviation
		{
			get
			{
				return pMagneticDeviation;
			}
		}

		/// <summary>Returns the difference between Magnetic North and True North.</summary>
		public Longitude MagneticVariation
		{
			get
			{
				return pMagneticVariation;
			}
		}

		/// <summary>
		/// Indicates that this sentence is supported by GPS.NET.
		/// </summary>
		public override bool IsSupported
		{
			get
			{
				return true;
			}
		}

		#endregion

		#region Static Members

		/// <summary>Represents an $HCHDG sentence with empty values.</summary>
		public static readonly HchdgSentence Empty = new HchdgSentence(Azimuth.Empty, Longitude.Empty,
			Longitude.Empty);

		/// <summary>Returns a new instance by interpreting the specified string.</summary>
		public static HchdgSentence Parse(string sentence)
		{
			return new HchdgSentence(sentence);
		}

		#endregion
	}
}
