using System;
using System.Security.Permissions;
using System.Collections;
using System.IO;
using System.Threading;
using System.Windows.Forms;
using System.ComponentModel;
using GeoFramework;
using GeoFramework.Gps;
using GeoFramework.IO.Serial;
using GeoFramework.IO;

namespace GeoFramework.Gps.Nmea
{
	/// <summary>Represents a multithreaded, event-driven NMEA-0183 data interpreter.</summary>
	/// <remarks>
	/// 	<para>This class is probably the most frequently-used class in the GPS.NET
	///     namespace. This class interprets sentences of NMEA-0183 data and converts the data
	///     into an object-oriented form.</para>
	/// 	<para>This class uses "layered stream" technology to process data. The
	///     <strong>BaseStream</strong> property is set to any <strong>Stream</strong> object,
	///     such as a <strong>SerialStream</strong>, <strong>BluetoothStream</strong>, or
	///     <strong>FileStream</strong>. If no stream is specified, a
	///     <strong>GpsStream</strong> object is used to automatically locate an NMEA-compliant
	///     GPS device.</para>
	/// 	<para>The interpreter will begin processing information and raising events after
	///     the <strong>Start</strong> method is called. Processing can be temporarily
	///     suspended using the <strong>Pause</strong> method. When data is no longer needed,
	///     the <strong>Stop</strong> method is called. As data is received, such as the
	///     current location, bearing and speed, properties in this class are updated and
	///     events are raised to notify of any changes.</para>
	/// </remarks>
	/// <example>
	///     The first example creates a new NMEA interpreter using any available source of NMEA
	///     data. The second example associates an interpreter with a
	///     <strong>FileStream</strong> in order to process a log file. 
	///     <code lang="VB" title="[New Example]" description="This example creates an NMEA interpreter using automatic GPS device detection.">
	/// ' Create an interpreter without specifying a base stream
	/// Dim Interpreter As New NmeaInterpreter()
	/// ' Start the interpreter, performing automatic device
	/// ' detection if necessary.
	/// Interpreter.Start()
	///     </code>
	/// 	<code lang="CS" title="[New Example]" description="This example creates an NMEA interpreter using automatic GPS device detection.">
	/// // Create an interpreter without specifying a base stream
	/// NmeaInterpreter Interpreter = new NmeaInterpreter();
	/// // Start the interpreter, performing automatic device
	/// // detection if necessary.
	/// Interpreter.Start();
	///     </code>
	/// 	<code lang="VB" title="[New Example]" description="This example creates an NMEA interpreter to process an entire log file.">
	/// ' Create an interpreter without specifying a base stream
	/// Dim LogFileStream As New FileStream("C:\MyLogFile.txt")
	/// Dim Interpreter As New NmeaInterpreter(LogFileStream)
	/// ' Start the interpreter, which processes all data in the
	/// ' file.
	/// Interpreter.Start()
	///     </code>
	/// 	<code lang="CS" title="[New Example]" description="This example creates an NMEA interpreter to process an entire log file.">
	/// // Create an interpreter without specifying a base stream
	/// FileStream LogFileStream = new FileStream(@"C:\MyLogFile.txt");
	/// NmeaInterpreter Interpreter = new NmeaInterpreter(LogFileStream);
	/// // Start the interpreter, which processes all data in the
	/// // file.
	/// Interpreter.Start();
	///     </code>
	/// </example>
	public sealed class NmeaInterpreter : Interpreter
	{
		#region Private Variables

		private NmeaStream pNmeaStream;
		private bool pIsFixed;
		private Distance pAltitude;
		private Distance pAltitudeAboveEllipsoid;
		private Azimuth pBearing;
		private DifferentialGps pDifferentialGps;
		private FixMethod pFixMethod;
		private FixMode pFixMode;
		private FixQuality pFixQuality;
        private FixLikelihood pFixLikelihood;
		private Distance pGeoidalSeparation;
		private DilutionOfPrecision pHorizontalDilutionOfPrecision;
		private Distance pHorizontalPositionError;
		private Longitude pMagneticVariation;
		private DilutionOfPrecision pMeanDilutionOfPrecision;
		private Distance pMeanPositionError;
		private DateTime pUtcDateTime;
		private Position pPosition;
		private SatelliteCollection pSatellites;
		private Speed pSpeed;
		private DilutionOfPrecision pVerticalDilutionOfPrecision;
		private Distance pVerticalPositionError;

		// Precision filters
		private DilutionOfPrecision pMaximumMeanDilutionOfPrecision = DilutionOfPrecision.Maximum;
		private DilutionOfPrecision pMaximumHorizontalDilutionOfPrecision = DilutionOfPrecision.Maximum;
		private DilutionOfPrecision pMaximumVerticalDilutionOfPrecision = DilutionOfPrecision.Maximum;
        private bool pIsMeanDilutionOfPrecisionSupported;
        private bool pIsHorizontalDilutionOfPrecisionSupported;
        private bool pIsVerticalDilutionOfPrecisionSupported;
        private bool pIsFixRequired;

        // Vertical speed and horizontal acceleration
        private DateTime pPreviousUtcDateTime;
        private Distance pPreviousAltitude;
        private Speed pAcceleration;
        private Speed pPreviousSpeed;
        private Speed pVerticalSpeed;

        // Position averaging 
        private PositionCollection pPositionAveragingItems;

		#endregion

		#region Static Members

        /// <summary>
        /// Returns the name of the protocol handled by this interpreter.
        /// </summary>
        public static readonly string ProtocolName = "NMEA-0183";

#if EMULATORS
		/// <summary>
		/// Creates a new NMEA interpreter using simulated NMEA-0183 data.
		/// </summary>
		/// <returns>An <strong>NmeaInterpreter</strong> configured to use a Emulator.</returns>
		/// <remarks>This method is useful for situations where no GPS device is present, or
		/// accuracy conditions are too poor to use an actual GPS device.  An interpreter is
		/// created and the Emulator is configured to behave like a typical NMEA-compliant GPS 
		/// device.</remarks>
		public static NmeaInterpreter FromEmulator()
		{
			return new NmeaInterpreter(new GeoFramework.Gps.Emulators.DeLormeEarthmateUsbEmulator());
		}
#endif

//		/// <summary>
//		/// Creates a new NMEA interpreter using the first available GPS device.
//		/// </summary>
//		/// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//		/// device.  The COM port is configured for 4800 baud.</remarks>
//		public static readonly NmeaInterpreter Automatic = new NmeaInterpreter();
//		/// <summary>
//		/// Creates a new NMEA interpreter using COM1.
//		/// </summary>
//		/// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//		/// device.  The COM port is configured for 4800 baud.</remarks>
//		public static readonly NmeaInterpreter Com1 = new NmeaInterpreter(new SerialPortStream(CommunicationPort.Com1, BaudRate.Baud4800));
//		/// <summary>
//		/// Creates a new NMEA interpreter using COM2.
//		/// </summary>
//		/// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//		/// device.  The COM port is configured for 4800 baud.</remarks>
//		public static readonly NmeaInterpreter Com2 = new NmeaInterpreter(new SerialPortStream(CommunicationPort.Com2, BaudRate.Baud4800));
//		/// <summary>
//		/// Creates a new NMEA interpreter using COM3.
//		/// </summary>
//		/// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//		/// device.  The COM port is configured for 4800 baud.</remarks>
//		public static readonly NmeaInterpreter Com3 = new NmeaInterpreter(new SerialPortStream(CommunicationPort.Com3, BaudRate.Baud4800));
//		/// <summary>
//		/// Creates a new NMEA interpreter using COM4.
//		/// </summary>
//		/// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//		/// device.  The COM port is configured for 4800 baud.</remarks>
//		public static readonly NmeaInterpreter Com4 = new NmeaInterpreter(new SerialPortStream(CommunicationPort.Com4, BaudRate.Baud4800));
//		/// <summary>
//		/// Creates a new NMEA interpreter using COM5.
//		/// </summary>
//		/// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//		/// device.  The COM port is configured for 4800 baud.</remarks>
//		public static readonly NmeaInterpreter Com5 = new NmeaInterpreter(new SerialPortStream(CommunicationPort.Com5, BaudRate.Baud4800));
//        /// <summary>
//        /// Creates a new NMEA interpreter using COM6.
//        /// </summary>
//        /// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//        /// device.  The COM port is configured for 4800 baud.</remarks>
//        public static readonly NmeaInterpreter Com6 = new NmeaInterpreter(new SerialPortStream(CommunicationPort.Com6, BaudRate.Baud4800));
//        /// <summary>
//        /// Creates a new NMEA interpreter using COM7.
//        /// </summary>
//        /// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//        /// device.  The COM port is configured for 4800 baud.</remarks>
//        public static readonly NmeaInterpreter Com7 = new NmeaInterpreter(new SerialPortStream(CommunicationPort.Com7, BaudRate.Baud4800));
//        /// <summary>
//        /// Creates a new NMEA interpreter using COM8.
//        /// </summary>
//        /// <remarks>This member is useful for quickly stating an interpreter using a serial GPS
//        /// device.  The COM port is configured for 4800 baud.</remarks>
//        public static readonly NmeaInterpreter Com8 = new NmeaInterpreter(new SerialPortStream(CommunicationPort.Com8, BaudRate.Baud4800));

		#endregion

		#region Constructors

		/// <summary>Creates a new instance.</summary>
		/// <remarks>
		/// 	<para>This constructor creates a new instance using no specific base stream. If no
		///     base stream is specified when the <strong>Start</strong> method is called, a
		///     <strong>GpsStream</strong> object is used to automatically discover an
		///     NMEA-compliant GPS device.</para>
		/// 	<para>This constructor is recommended for all applications because it allows the
		///     application to discover GPS devices on any computer.</para>
		/// </remarks>
		/// <example>
		///     This example creates an <strong>NmeaInterpreter</strong> set to automatically use
		///     the first available NMEA-compliant GPS device. 
		///     <code lang="VB" title="[New Example]">
		/// ' Create an interpreter without specifying a base stream
		/// Dim Interpreter As New NmeaInterpreter()
		/// ' Start the interpreter, performing automatic device
		/// ' detection if necessary.
		/// Interpreter.Start()
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create an interpreter without specifying a base stream
		/// NmeaInterpreter Interpreter = new NmeaInterpreter();
		/// // Start the interpreter, performing automatic device
		/// // detection if necessary.
		/// Interpreter.Start();
		///     </code>
		/// </example>
		public NmeaInterpreter() 
			: base("GPS.NET Interpreter [NMEA-0183] (http://www.geoframeworks.com)")
		{
		}

		/// <summary>Creates a new instance using the specified stream for data.</summary>
		/// <returns>
		/// A <strong>Stream</strong> object identifying the data source of the
		/// interpreter.
		/// </returns>
		/// <remarks>
		/// This constructor should only be used when a specific data source is needed, such
		/// as a log file. To use any available GPS device, regardless of it's specific type,
		/// specify no stream or use the constructor with no parameters.
		/// </remarks>
		/// <example>
		///     This example create an <strong>NmeaInterpreter</strong> using a log file as its
		///     data source. The entire log file is then processed as rapidly as possible. 
		///     <code lang="VB" title="[New Example]">
		/// ' Create an interpreter without specifying a base stream
		/// Dim LogFileStream As New FileStream("C:\MyLogFile.txt")
		/// Dim Interpreter As New NmeaInterpreter(LogFileStream)
		/// ' Start the interpreter, which processes all data in the
		/// ' file.
		/// Interpreter.Start()
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// // Create an interpreter without specifying a base stream
		/// FileStream LogFileStream = new FileStream(@"C:\MyLogFile.txt");
		/// NmeaInterpreter Interpreter = new NmeaInterpreter(LogFileStream);
		/// // Start the interpreter, which processes all data in the
		/// // file.
		/// Interpreter.Start();
		///     </code>
		/// </example>
        public NmeaInterpreter(Stream input) 
			: base("GPS.NET Interpreter [NMEA-0183] (http://www.geoframeworks.com)",
			       input)
		{
		}

		#endregion

		#region Events

        /// <summary>Occurs when the distance above sea level has changed.</summary>
#if DesignTime || !PocketPC
		[Description("Occurs when the distance above sea level has changed.")]
#endif
        public event DistanceEventHandler AltitudeChanged;

        /// <summary>Occurs when a reading of the current distance above sea level is received from the underlying stream.</summary>
#if DesignTime || !PocketPC
		[Description("Occurs when a reading of the current distance above sea level is received from the underlying stream.")]
#endif
        public event DistanceEventHandler AltitudeReceived;
        /// <summary>Occurs when the distance above the ellipsoid has changed.</summary>
		public event DistanceEventHandler AltitudeAboveEllipsoidChanged;
        /// <summary>Occurs when a reading of the current distance above sea level is received from the underlying stream.</summary>
        public event DistanceEventHandler AltitudeAboveEllipsoidReceived;
        /// <summary>Occurs when the current direction of travel has changed.</summary>
		public event AzimuthEventHandler BearingChanged;
        /// <summary>Occurs when a reading of the current direction of travel is received from the underlying stream.</summary>
        public event AzimuthEventHandler BearingReceived;
        /// <summary>Occurs when the current satellite fix has been lost.</summary>
		public event EventHandler FixLost;
		/// <summary>Occurs when a satellite fix becomes active.</summary>
		public event EventHandler FixObtained;
		/// <summary>Occurs when the current fix method has changed.</summary>
		public event FixMethodEventHandler FixMethodChanged;
		/// <summary>Occurs when the current fix mode has changed.</summary>
		public event FixModeEventHandler FixModeChanged;
		/// <summary>Occurs when the current fix quality has changed.</summary>
		public event FixQualityEventHandler FixQualityChanged;
        /// <summary>Occurs when the current fix stability has changed.</summary>
        public event FixLikelihoodEventHandler FixLikelihoodChanged;
        /// <summary>Occurs when the distance between the Earth's surface and the ellipsoid's surface has changed.</summary>
		public event DistanceEventHandler GeoidalSeparationChanged;
		/// <summary>
		/// Occurs when the current level of precision as it relates to latitude and
		/// longitude has changed.
		/// </summary>
		public event DilutionOfPrecisionEventHandler HorizontalDilutionOfPrecisionChanged;
		/// <summary>Occurs when the current amount of positional error has changed.</summary>
		public event DistanceEventHandler HorizontalPositionErrorChanged;
		/// <summary>Occurs when the magnetic variation for the current location becomes known.</summary>
		public event LongitudeEventHandler MagneticVariationAvailable;
		/// <summary>
		/// Occurs when the current level of precision as it relates to latitude and
		/// longitude has changed.
		/// </summary>
		public event DilutionOfPrecisionEventHandler MeanDilutionOfPrecisionChanged;
		/// <summary>Occurs when the overall amount of positional error has changed.</summary>
		public event DistanceEventHandler MeanPositionErrorChanged;
		/// <summary>Occurs when the current location has changed.</summary>
		public event PositionEventHandler PositionChanged;
		/// <summary>Occurs when a positional reading is received from the underlying stream.</summary>
		public event PositionEventHandler PositionReceived;
		/// <summary>Occurs when a satellite's signal can be used in the current fix.</summary>
		public event SatelliteEventHandler SatelliteFixObtained;
		/// <summary>Occurs when a satellite signal becomes too weak to use in the current fix.</summary>
		public event SatelliteEventHandler SatelliteFixLost;
		/// <summary>Occurs when a GPS satellite's location has changed.</summary>
		public event SatelliteEventHandler SatellitePositionChanged;
		/// <summary>Occurs when a GPS satellite signal becomes more or less obscured.</summary>
		public event SatelliteEventHandler SatelliteSignalStrengthChanged;		
        /// <summary>Occurs when an NMEA sentence has been received for interpretation.</summary>		
        public event NmeaSentenceEventHandler SentenceReceived;
		/// <summary>Occurs when an NMEA sentence is successfully interpreted.</summary>
		public event NmeaSentenceEventHandler SentenceInterpreted;
		/// <summary>Occurs when an NMEA sentence has been sent to the recording stream.</summary>
		public event NmeaSentenceEventHandler SentenceRecorded;
		/// <summary>Occurs when an invalid NMEA sentence is detected.</summary>
		public event NmeaSentenceExceptionEventHandler SentenceExceptionOccurred;
		/// <summary>Occurs when the current rate of travel has changed.</summary>
		public event SpeedEventHandler SpeedChanged;
        /// <summary>Occurs when a speed reading is received from the underlying stream.</summary>
        public event SpeedEventHandler SpeedReceived;
        /// <summary>Occurs when the current satellite-derived time has changed.</summary>
		public event DateTimeEventHandler UtcDateTimeChanged;
		/// <summary>
		/// Occurs when the current level of precision as it relates to altitude has
		/// changed.
		/// </summary>
		public event DilutionOfPrecisionEventHandler VerticalDilutionOfPrecisionChanged;
		/// <summary>Occurs when the current amount of positional accuracy has changed.</summary>
		public event DistanceEventHandler VerticalPositionErrorChanged;
		/// <summary>
		/// Occurs when the maximum allowed precision for latitude, longitude and altitude
		/// has changed.
		/// </summary>
		public event DilutionOfPrecisionEventHandler MaximumMeanDilutionOfPrecisionChanged;
		/// <summary>
		/// Occurs when the maximum allowed precision for latitude and longitude has
		/// changed.
		/// </summary>
		public event DilutionOfPrecisionEventHandler MaximumHorizontalDilutionOfPrecisionChanged;
		/// <summary>Occurs when the maximum allowed precision for altitude has changed.</summary>
		public event DilutionOfPrecisionEventHandler MaximumVerticalDilutionOfPrecisionChanged;
		#endregion

		protected override void Dispose(bool disposing)
        {
            // No need to finalize anymore
            GC.SuppressFinalize(this);

            // Forcefully unsubscribe all event handlers so that we can finalize
			AltitudeChanged = null;
			AltitudeAboveEllipsoidChanged = null;
			BearingChanged = null;
			FixLost = null;
			FixObtained = null;
			FixMethodChanged = null;
			FixModeChanged = null;
			FixQualityChanged = null;
            FixLikelihoodChanged = null;
			GeoidalSeparationChanged = null;
			HorizontalDilutionOfPrecisionChanged = null;
			HorizontalPositionErrorChanged = null;
			MagneticVariationAvailable = null;
			MeanDilutionOfPrecisionChanged = null;
			MeanPositionErrorChanged = null;
			PositionChanged = null;
			PositionReceived = null;
			SatelliteFixObtained = null;
			SatelliteFixLost = null;
			SatellitePositionChanged = null;
			SatelliteSignalStrengthChanged = null;
			SentenceReceived = null;
			SentenceInterpreted = null;
			SentenceRecorded = null;
			SentenceExceptionOccurred = null;
			SpeedChanged = null;
			UtcDateTimeChanged = null;
			VerticalDilutionOfPrecisionChanged = null;
			VerticalPositionErrorChanged = null;
			MaximumMeanDilutionOfPrecisionChanged = null;
			MaximumHorizontalDilutionOfPrecisionChanged = null;
			MaximumVerticalDilutionOfPrecisionChanged = null;

            // The NMEA base stream is NOT disposed of since it may still be needed!
            // pBaseStream.Close();  pBaseStream.Dispose();  

			base.Dispose (disposing);
		}


		#region Public Members

		/// <summary>Returns the current direction of travel.</summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns the current direction of travel.")]
#endif
		public Azimuth Bearing
		{
            get
			{
				return pBearing;
			}
		}

		/// <summary>Occurs when the data source for the interpreter has changed.</summary>
		/// <returns>
		/// A <strong>Stream</strong> indicating the new data source for the
		/// interpreter.
		/// </returns>
		/// <remarks>
		/// This virtual method is called when the <strong>BaseStream</strong> property has
		/// changed. This method gives custom interpreters an opportunity to make adjustments. For
		/// this interpreter, a new <strong>NmeaStream</strong> object is created using the
		/// BaseStream as its own data source.
		/// </remarks>
		/// <param name="baseStream">The new value of the interpreter's <strong>BaseStream</strong> property.</param>
		protected override void OnBaseStreamChanged(Stream baseStream)
		{
			if(pNmeaStream != null)
				pNmeaStream.Close();

			if(baseStream == null)
				pNmeaStream = null;
			else
				pNmeaStream = new NmeaStream(baseStream);

			// Update the thread name
            if(baseStream == null)
                SetThreadName("GPS.NET NMEA Interpreter [No data source] (http://www.geoframeworks.com)");
            else
                SetThreadName("GPS.NET NMEA Interpreter [" + baseStream.ToString() + "] (http://www.geoframeworks.com)");

		}

		/// <summary>
		/// Returns the Type reponsible for processing raw NMEA-0183 data.
		/// </summary>
		public override Type ParsingType
		{
			get
			{
				return typeof(NmeaStream);
			}
		}

		/// <summary>Returns the difference between True North and Magnetic North.</summary>
		/// <remarks>True North represents the physical location of the North Pole: 90N, 0E.  
		/// However, a magnetic compass may actually point towards slightly away from True North
		/// because Earth's magnetic field is not perfectly aligned to the poles.  As a result, 
		/// what a magnetic compass shows to be North may actually differ from what a GPS device
		/// would show to be North.  Magnetic variation measures the number of degrees between True North 
		/// and Magnetic North so that longitudes can be converted to either form.</remarks>
#if DesignTime || !PocketPC
		[Category("Precision")]
		[Description("Returns the difference between True North and Magnetic North.")]
#endif
		public Longitude MagneticVariation
		{
			get
			{
				return pMagneticVariation;
			}
		}

        /// <summary>
        /// Controls whether or not position or altitude measurements are ignored based on
        /// the current level of precision.
        /// </summary>
#if !PocketPC
		[Category("Precision")]
		[DefaultValue(typeof(DilutionOfPrecision), "50.0")]
		[Description("Controls the maximum amount of overall precision error as it relates to altitude, latitude or longitude.")]
#endif
#if Framework20 && !PocketPC
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
#endif
        [Obsolete("Mean DOP is a measurement which is too ambiguous to enforce for a professional GPS application.  Replace this line with code which enforces horizontal DOP, vertical DOP, or both, instead.")]
		public DilutionOfPrecision MaximumMeanDilutionOfPrecision
		{
			get
			{
				return pMaximumMeanDilutionOfPrecision;
			}
			set
			{
				if(pMaximumMeanDilutionOfPrecision.Equals(value)) return;
				pMaximumMeanDilutionOfPrecision = value;
				OnMaximumMeanDilutionOfPrecisionChanged(pMaximumMeanDilutionOfPrecision);
			}
		}

        /// <summary>
        /// Controls whether or not position measurements are ignored based on the current
        /// level of precision.
        /// </summary>
#if !PocketPC
		[Category("Precision")]
		[DefaultValue(typeof(DilutionOfPrecision), "50.0")]
		[Description("Controls the maximum amount of precision error as it relates to latitude and longitude.")]
#endif
#if Framework20 && !PocketPC
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
#endif
        public DilutionOfPrecision MaximumHorizontalDilutionOfPrecision
		{
			get
			{
				return pMaximumHorizontalDilutionOfPrecision;
			}
			set
			{
				if(pMaximumHorizontalDilutionOfPrecision.Equals(value)) return;
				pMaximumHorizontalDilutionOfPrecision = value;
				OnMaximumHorizontalDilutionOfPrecisionChanged(pMaximumHorizontalDilutionOfPrecision);
			}
		}
			
		/// <summary>
		/// Controls whether or not altitude measurements are ignored based on the current
		/// level of precision.
        /// </summary>
#if !PocketPC
        [Category("Precision")]
		[DefaultValue(typeof(DilutionOfPrecision), "50.0")]
		[Description("Controls the maximum amount of precision error as it relates to altitude.")]
#endif
#if Framework20 && !PocketPC
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
#endif
        public DilutionOfPrecision MaximumVerticalDilutionOfPrecision
		{
			get
			{
				return pMaximumVerticalDilutionOfPrecision;
			}
			set
			{
				if(pMaximumVerticalDilutionOfPrecision.Equals(value)) return;
				pMaximumVerticalDilutionOfPrecision = value;
				OnMaximumVerticalDilutionOfPrecisionChanged(pMaximumVerticalDilutionOfPrecision);
			}
		}			

		/// <summary>Returns the current satellite-derived time.</summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns the current satellite-derived time.")]
#endif
		public DateTime UtcDateTime
		{
			get
			{
				return pUtcDateTime;
			}
		}

		/// <summary>Returns the current satellite-derived time adjusted to the local time zone.</summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns the current satellite-derived time, adjusted to the local time zone.")]
#endif
		public DateTime DateTime
		{
			get
			{
				return pUtcDateTime.ToLocalTime();
			}
		}

		/// <summary>Returns the current location on Earth's surface.</summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns the current location on Earth's surface.")]
#endif
		public Position Position
		{
			get
			{
				return pPosition;
			}
		}

        /// <summary>Returns the current location after applying position-averaging filters.</summary>
#if DesignTime || !PocketPC
        [Category("Real-Time Data")]
        [Description("Returns the current location after applying position-averaging filters.")]
#endif
        public Position PositionAveraged
        {
            get
            {
                return pPositionAveragingItems.Average;
            }
        }

		/// <summary>Returns a collection describing all detected GPS satellites.</summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns a collection describing all detected GPS satellites.")]
#endif
		public SatelliteCollection Satellites
		{
			get
			{
				return pSatellites;
			}
		}

		/// <summary>Returns the current rate of travel.</summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns the current rate of travel.")]
#endif
		public Speed Speed
		{
			get
			{
				return pSpeed;
			}
		}

        /// <summary>Returns the current rate of change in altitude.</summary>
#if DesignTime || !PocketPC
        [Category("Real-Time Data")]
        [Description("Returns the current rate of change in altitude.")]
#endif
        public Speed VerticalSpeed
        {
            get
            {
                return pVerticalSpeed;
            }
        }

		/// <summary>Returns the name of the GPS protocol being interpreted.</summary>
#if DesignTime || !PocketPC
		[Category("Data")]
		[Description("Returns the name of the GPS protocol being interpreted.")]
#endif
		public override string Protocol
		{
			get
			{
				return ProtocolName;
			}
		}

		/// <summary>Returns the <strong>NmeaStream</strong> used to interpret raw GPS data.</summary>
#if DesignTime || !PocketPC
		[Category("Data")]
		[Description("Returns the NmeaStream used to interpret raw GPS data.")]
#endif
		public override Stream ParsingStream
		{
            get
            {			
                return pNmeaStream;
            }
        }

        /// <summary>Returns the current rate of change in speed per second.</summary>
        /// <remarks>This property uses satellite-derived time and speed measurements
        /// to calculate the rate of acceleration.  The value returned is the amount of 
        /// change per second.  Therefore, a value of two meters per second for this
        /// property means two meters per second per second of acceleration.</remarks>
#if DesignTime || !PocketPC
        [Category("Real-Time Data")]
        [Description("Returns the current rate of change in speed per second.")]
#endif
        public Speed Acceleration
        {
            get
            {
                return pAcceleration;
            }
        }

		/// <summary>Returns the current distance above sea level.</summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns the current distance above sea level.")]
#endif
		public Distance Altitude
		{
			get
			{
				return pAltitude;
			}
		}

		/// <summary>Returns the current distance above the ellipsoid.</summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns the current distance above the ellipsoid.")]
#endif
		public Distance AltitudeAboveEllipsoid
		{
			get
			{
				return pAltitudeAboveEllipsoid;
			}
		}

		/// <summary>
		/// Returns the difference between the Earth's surface and the surface of the
		/// ellipsoid.
		/// </summary>
#if DesignTime || !PocketPC
		[Category("Real-Time Data")]
		[Description("Returns the difference between the Earth's surface and the surface of the ellipsoid.")]
#endif
		public Distance GeoidalSeparation
		{
			get
			{
				return pGeoidalSeparation;
			}
		}

		/// <summary>
		/// Returns the current level of precision as it relates to latitude and
		/// longitude.
		/// </summary>
#if DesignTime || !PocketPC
		[Category("Precision")]
		[Description("Returns the current level of precision as it relates to latitude and longitude.")]
#endif
		public DilutionOfPrecision HorizontalDilutionOfPrecision
		{
			get
			{
				return pHorizontalDilutionOfPrecision;
			}
		}

		/// <summary>Returns the current level of precision as it relates to altitude.</summary>
#if DesignTime || !PocketPC
		[Category("Precision")]
		[Description("Returns the current level of precision as it relates to altitude.")]
#endif
		public DilutionOfPrecision VerticalDilutionOfPrecision
		{
			get
			{
				return pVerticalDilutionOfPrecision;
			}
		}

		/// <summary>
		/// Returns the current level of precision as it relates to latitude, longitude and
		/// altitude.
		/// </summary>
#if DesignTime || !PocketPC
		[Category("Precision")]
		[Description("Returns the current level of precision as it relates to latitude, longitude and altitude.")]
#endif
		public DilutionOfPrecision MeanDilutionOfPrecision
		{
			get
			{
				return pMeanDilutionOfPrecision;
			}
		}

		/// <summary>Indicates if a block of data can be recognized by the interpreter.</summary>
		/// <returns>
		/// A <strong>Boolean</strong>, <strong>True</strong> if data within the packet can
		/// be recognized as NMEA-0183 data.
		/// </returns>
		/// <remarks>
		/// This method is used to test a block of raw data in an attempt to discover its
		/// format. This method is used for rare situations where an application is dealing with
		/// data in an unknown format which needs to be recognized.
		/// </remarks>
		/// <param name="packet">A <strong>Packet</strong> object containing raw data to be analyzed.</param>
		public override bool CanParse(Packet packet)
		{
			return NmeaStream.CanParse(packet);
		}

		/// <summary>Returns information about a differential GPS fix is one is active.</summary>
#if DesignTime || !PocketPC
		[Category("Precision")]
		[Description("Returns information about a differential GPS fix is one is active.")]
#endif
		public DifferentialGps DifferentialGps
		{
			get
			{
				return pDifferentialGps;
			}
		}

		/// <summary>Interprets the specified block of one or more NMEA sentences.</summary>
		/// <param name="sentences">
		/// A <strong>String</strong> containing one or more NMEA sentences to
		/// interpret.
		/// </param>
		public void Parse(string sentences)
		{
			// Locate the first NMEA sentence
			int SentenceStart = sentences.IndexOf("$");
            while (SentenceStart != -1)
            {
    			// Locate the end of the sentence
	    		int SentenceEnd = sentences.IndexOf("*", SentenceStart, sentences.Length - SentenceStart);
                if (SentenceEnd != -1)
                {
                    // Extract the sentence
                    NmeaSentence Sentence = new NmeaSentence(sentences.Substring(SentenceStart, SentenceEnd + 3 - SentenceStart));
                    // And parse it
                    //DEBUG ParseSentence(Sentence);
                }
                // Do the next one
                SentenceStart = sentences.IndexOf("$", SentenceEnd + 3);
            }
		}

        /// <summary>
        /// Parses any recognizable data in the specified packet.
        /// </summary>
        /// <param name="packet">A <strong>Packet</strong> containing one or more NMEA sentences to interpret.</param>
        /// <returns>A <strong>Packet</strong> containing data which could not be parsed.</returns>
        /// <remarks>This method is typically used to process a raw buffer of data received from
        /// a GPS device.  If data is recognized as NMEA, any bytes before the first recognized
        /// sentence are discarded.  Any unrecognizable bytes after recognized sentences are returned
        /// in a new packet.</remarks>
        public override void Parse(Packet packet)
        {
			Parse(packet.ToString());
        }

        /// <summary>Interprets the specified NMEA-0183 sentence.</summary>
		/// <returns>
		/// A <strong>Boolean</strong>, <strong>True</strong> if the sentence was
		/// successfully interpreted.
		/// </returns>
		/// <remarks>
		/// This method is used for situations where data must be fed manually to the
		/// interpreter. Typically, this method is used when slowly playing back a log file -- a
		/// new line of data is read from a file, then fed into the interpreter.
		/// </remarks>
		/// <example>
		///     This example demonstrates the <strong>ParseSentence</strong> method by parsing a
		///     log file one line at a time. 
		///     <code lang="VB" title="[New Example]">
		/// Dim Interpreter As New NmeaInterpreter()
		/// Dim LogFileStream As New FileStream("C:\MyLogFile.txt")
		/// Dim LogFileReader As New StreamReader(LogFileStream)
		///  
		/// ' Process the entire file line by line
		/// Dim NextLine As String = LogFileReader.ReadLine()
		/// Do While Not NextLine Is Nothing
		///     ' Create an NMEA Sentence
		///     Dim NextSentence As New NmeaSentence(NextLine)
		///     ' Feed it to the interpreter
		///     Interpreter.ParseSentence(NextSentence)
		///     ' Delay for one second
		///     Thread.Sleep(1000)
		///     ' Now read the next line
		///     NextLine = LogFileReader.ReadLine()
		/// Loop
		///  
		/// ' Clean up
		/// LogFileReader.Close()
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// NmeaInterpreter Interpreter = new NmeaInterpreter();
		/// FileStream LogFileStream = new FileStream("C:\MyLogFile.txt");
		/// StreamReader LogFileReader = new StreamReader(LogFileStream);
		///  
		/// // Process the entire file line by line
		/// string NextLine = LogFileReader.ReadLine();
		/// while(NextLine != null)
		/// {
		///     // Create an NMEA Sentence
		///     NmeaSentence NextSentence = new NmeaSentence(NextLine);
		///     // Feed it to the interpreter
		///     Interpreter.ParseSentence(NextSentence);
		///     // Delay for one second
		///     Thread.Sleep(1000);
		///     // Now read the next line
		///     NextLine = LogFileReader.ReadLine();
		/// }
		///  
		/// // Clean up
		/// LogFileReader.Close();
		///     </code>
		/// </example>
        public bool ParseSentence(NmeaSentence sentence)
        {
            return ParseSentence(sentence, true);
        }

		/// <summary>Interprets the specified NMEA-0183 sentence.</summary>
		/// <returns>
		/// A <strong>Boolean</strong>, <strong>True</strong> if the sentence was
		/// successfully interpreted.
		/// </returns>
		/// <remarks>
		/// This method is used for situations where data must be fed manually to the
		/// interpreter. Typically, this method is used when slowly playing back a log file -- a
		/// new line of data is read from a file, then fed into the interpreter.
		/// </remarks>
		/// <example>
		///     This example demonstrates the <strong>ParseSentence</strong> method by parsing a
		///     log file one line at a time. 
		///     <code lang="VB" title="[New Example]">
		/// Dim Interpreter As New NmeaInterpreter()
		/// Dim LogFileStream As New FileStream("C:\MyLogFile.txt")
		/// Dim LogFileReader As New StreamReader(LogFileStream)
		///  
		/// ' Process the entire file line by line
		/// Dim NextLine As String = LogFileReader.ReadLine()
		/// Do While Not NextLine Is Nothing
		///     ' Create an NMEA Sentence
		///     Dim NextSentence As New NmeaSentence(NextLine)
		///     ' Feed it to the interpreter
		///     Interpreter.ParseSentence(NextSentence)
		///     ' Delay for one second
		///     Thread.Sleep(1000)
		///     ' Now read the next line
		///     NextLine = LogFileReader.ReadLine()
		/// Loop
		///  
		/// ' Clean up
		/// LogFileReader.Close()
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// NmeaInterpreter Interpreter = new NmeaInterpreter();
		/// FileStream LogFileStream = new FileStream("C:\MyLogFile.txt");
		/// StreamReader LogFileReader = new StreamReader(LogFileStream);
		///  
		/// // Process the entire file line by line
		/// string NextLine = LogFileReader.ReadLine();
		/// while(NextLine != null)
		/// {
		///     // Create an NMEA Sentence
		///     NmeaSentence NextSentence = new NmeaSentence(NextLine);
		///     // Feed it to the interpreter
		///     Interpreter.ParseSentence(NextSentence);
		///     // Delay for one second
		///     Thread.Sleep(1000);
		///     // Now read the next line
		///     NextLine = LogFileReader.ReadLine();
		/// }
		///  
		/// // Clean up
		/// LogFileReader.Close();
		///     </code>
		/// </example>
        public bool ParseSentence(NmeaSentence sentence, bool allowExceptions)
        {
            // Try to convert the item into a more specific object type
            switch (sentence.Command)
            {
                case "$GPRMC":
                    GprmcSentence Gprmc = sentence as GprmcSentence;
                    if (Gprmc == null)
                        Gprmc = new GprmcSentence(sentence.Value);
                    return ParseSentence(Gprmc);
                case "$GPGGA":
                    GpggaSentence Gpgga = sentence as GpggaSentence;
                    if (Gpgga == null)
                        Gpgga = new GpggaSentence(sentence.Value);
                    return ParseSentence(Gpgga);
                case "$GPGSA":
                    GpgsaSentence Gpgsa = sentence as GpgsaSentence;
                    if (Gpgsa == null)
                        Gpgsa = new GpgsaSentence(sentence.Value);
                    return ParseSentence(Gpgsa);
                case "$GPGSV":
                    GpgsvSentence Gpgsv = sentence as GpgsvSentence;
                    if (Gpgsv == null)
                        Gpgsv = new GpgsvSentence(sentence.Value);
                    return ParseSentence(Gpgsv);
                case "$GPGLL":
                    GpgllSentence Gpgll = sentence as GpgllSentence;
                    if (Gpgll == null)
                        Gpgll = new GpgllSentence(sentence.Value);
                    return ParseSentence(Gpgll);
                case "$GPVTG":
                    GpvtgSentence Gpvtg = sentence as GpvtgSentence;
                    if (Gpvtg == null)
                        Gpvtg = new GpvtgSentence(sentence.Value);
                    return ParseSentence(Gpvtg);
                case "$GPGGK":
                    GpggkSentence Gpggk = sentence as GpggkSentence;
                    if (Gpggk == null)
                        Gpggk = new GpggkSentence(sentence.Value);
                    return ParseSentence(Gpggk);
                case "$GPGGQ":
                    GpggqSentence Gpggq = sentence as GpggqSentence;
                    if (Gpggq == null)
                        Gpggq = new GpggqSentence(sentence.Value);
                    return ParseSentence(Gpggq);
                case "$GPLLK":
                    GpllkSentence Gpllk = sentence as GpllkSentence;
                    if (Gpllk == null)
                        Gpllk = new GpllkSentence(sentence.Value);
                    return ParseSentence(Gpllk);
                case "$GPLLQ":
                    GpllqSentence Gpllq = sentence as GpllqSentence;
                    if (Gpllq == null)
                        Gpllq = new GpllqSentence(sentence.Value);
                    return ParseSentence(Gpllq);
                case "$HCHDG":
                    HchdgSentence Hchdg = sentence as HchdgSentence;
                    if (Hchdg == null)
                        Hchdg = new HchdgSentence(sentence.Value);
                    return ParseSentence(Hchdg);
                case "$PGRME":
                    PgrmeSentence Pgrme = sentence as PgrmeSentence;
                    if (Pgrme == null)
                        Pgrme = new PgrmeSentence(sentence.Value);
                    return ParseSentence(Pgrme);
                case "$PGRMZ":
                    PgrmzSentence Pgrmz = sentence as PgrmzSentence;
                    if (Pgrmz == null)
                        Pgrmz = new PgrmzSentence(sentence.Value);
                    return ParseSentence(Pgrmz);
                case "$PMGNST":
                    PmgnstSentence Pmgnst = sentence as PmgnstSentence;
                    if (Pmgnst == null)
                        Pmgnst = new PmgnstSentence(sentence.Value);
                    return ParseSentence(Pmgnst);
                case "$PGLOR":
                    PglorSentence Pglor = sentence as PglorSentence;
                    if (Pglor == null)
                        Pglor = new PglorSentence(sentence.Value);
                    return ParseSentence(Pglor);
                default:
                    NmeaSentenceException ex = new NmeaSentenceException(sentence, "The specified " + sentence.Command + " sentence is currently not supported.");
                    OnExceptionOccurred(ex);
                    if (allowExceptions)
                        throw ex;
                    else
                        return false;
            }
        }

        public bool ParseSentence(GprmcSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update fix status
            OnFixReceived(sentence.IsFixObtained);
            // Update the time
            OnUtcDateTimeReceived(sentence.UtcDateTime);
            // Update the current position
            OnPositionReceived(sentence.Position);
            // Update magnetic variation
            OnMagneticVariationReceived(sentence.MagneticVariation);
            // Update the bearing
            OnBearingReceived(sentence.Bearing);
            // Update the speed
            OnSpeedReceived(sentence.Speed);
            // Update the acceleration
            OnAccelerationChanged(sentence.UtcDateTime, sentence.Speed);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            // And indicate success
            return true;
        }

        public bool ParseSentence(GpllqSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update fix status
            OnFixReceived(sentence.IsFixObtained);
            // Update date/time
            OnUtcDateTimeReceived(sentence.UtcDateTime);
            // Update fix quality
            OnFixQualityReceived(sentence.FixQuality);
            // Update mean DOP
            OnMeanDilutionOfPrecisionReceived(sentence.MeanDilutionOfPrecision);
            // Update current position
            OnPositionReceived(sentence.Position);
            // Update current altitude
            OnAltitudeReceived(sentence.Altitude);
            // Update rate of climb
            OnVerticalSpeedChanged(sentence.UtcDateTime, sentence.Altitude);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(GpggaSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update differential GPS 
            pDifferentialGps.SetAge(sentence.DifferentialGps.LastUpdateUtc, TimeSpan.Zero);
            // Also set the station ID
            pDifferentialGps.SetStationId(sentence.DifferentialGps.StationId);
            // What kind of fix is in effect?            
            SatelliteCollection DgpsSatellites = pSatellites.DifferentialGpsSatellites;
            if (DgpsSatellites.Count > 0)
            {
                // DGPS.  Mark ALL DGPS satellites as fixed.
                SatelliteCollection UpdatedSatellites = new SatelliteCollection();
                foreach(Satellite item in DgpsSatellites)
                {
                    // Mark the satellite as fixed then add it
                    if(item.IsTracked)
                        UpdatedSatellites.Add(item.ToFixedSatellite());
                }

                // Marge this with the regular list
                pSatellites.MergeWith(UpdatedSatellites);

                // Set the updated satellite list
                pDifferentialGps.SetSatellites(UpdatedSatellites);
                pDifferentialGps.SetIsActive(true);
                switch (DgpsSatellites[0].Class)
                {
                    case SatelliteClass.Egnos:
                        pDifferentialGps.SetType(DifferentialGpsType.Egnos);
                        break;
                    case SatelliteClass.Waas:
                        pDifferentialGps.SetType(DifferentialGpsType.Waas);
                        break;
                    case SatelliteClass.Msas:
                        pDifferentialGps.SetType(DifferentialGpsType.Msas);
                        break;
                }
            }
            else if (pDifferentialGps.StationId != 0)
            {
                pDifferentialGps.SetIsActive(true);
                pDifferentialGps.SetType(DifferentialGpsType.GroundStation);
                pDifferentialGps.SetSatellites(null);
            }
            else
            {
                pDifferentialGps.SetIsActive(false);
                pDifferentialGps.SetType(DifferentialGpsType.None);
                pDifferentialGps.SetSatellites(null);
            }

            // Update fix status
            OnFixReceived(sentence.IsFixObtained);
            // Update the current time
            OnUtcDateTimeReceived(sentence.UtcDateTime);
            // Update the HDOP
            OnHorizontalDilutionOfPrecisionReceived(sentence.HorizontalDilutionOfPrecision);
            // Update the current position
            OnPositionReceived(sentence.Position);
            // Update the altitude
            OnAltitudeReceived(sentence.Altitude);
            // Update rate of climb
            OnVerticalSpeedChanged(sentence.UtcDateTime, sentence.Altitude);
            // Update the altitude above the ellipsoid
            OnAltitudeAboveEllipsoidReceived(sentence.AltitudeAboveEllipsoid);
            // Update fix quality
            OnFixQualityReceived(sentence.FixQuality);
            // Update geoidal separation
            OnGeoidalSeparationReceived(sentence.GeoidalSeparation);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            // And return success
            return true;
        }

        public bool ParseSentence(GpgsaSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update horizontal DOP
            OnHorizontalDilutionOfPrecisionReceived(sentence.HorizontalDilutionOfPrecision);
            OnMeanDilutionOfPrecisionReceived(sentence.MeanDilutionOfPrecision);
            OnVerticalDilutionOfPrecisionReceived(sentence.VerticalDilutionOfPrecision);
            // Update fix status
            OnFixReceived(sentence.IsFixObtained);
            // Update fixed satellites
            OnFixedSatellitesReceived(sentence.FixedSatellites);
            // Update fix method
            OnFixMethodReceived(sentence.FixMethod);
            // Update fix mode
            OnFixModeReceived(sentence.FixMode);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(GpgllSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update fix status
            OnFixReceived(sentence.IsFixObtained);
            // Update UTC date/time
            OnUtcDateTimeReceived(sentence.UtcDateTime);
            // Update position
            OnPositionReceived(sentence.Position);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(GpggkSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update fix status
            OnFixReceived(sentence.IsFixObtained);
            // Update GPS date/time
            OnUtcDateTimeReceived(sentence.UtcDateTime);
            // Update mean DOP
            OnMeanDilutionOfPrecisionReceived(sentence.MeanDilutionOfPrecision);
            // Update fix quality
            OnFixQualityReceived(sentence.FixQuality);
            // Update position
            OnPositionReceived(sentence.Position);
            // Update altitude above ellipsoid
            OnAltitudeAboveEllipsoidReceived(sentence.AltitudeAboveEllipsoid);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(GpvtgSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update magnetic variation
            OnMagneticVariationReceived(new Longitude(sentence.MagneticBearing.Subtract(sentence.TrueBearing.DecimalDegrees).DecimalDegrees));
            // Update bearing
            OnBearingReceived(sentence.TrueBearing);
            // Update speed
            OnSpeedReceived(sentence.Speed);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(GpgsvSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Parse satellite information
            OnSatellitesReceived(sentence.Satellites);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(GpggqSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update fix status
            OnFixReceived(sentence.IsFixObtained);
            // Update GPS date/time
            OnUtcDateTimeReceived(sentence.UtcDateTime);
            // Update fix quality
            OnFixQualityReceived(sentence.FixQuality);
            // Update mean DOP
            OnMeanDilutionOfPrecisionReceived(sentence.MeanDilutionOfPrecision);
            // Update position
            OnPositionReceived(sentence.Position);
            // Update altitude
            OnAltitudeReceived(sentence.Altitude);
            // Update rate of climb
            OnVerticalSpeedChanged(sentence.UtcDateTime, sentence.Altitude);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(GpllkSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update fix status
            OnFixReceived(sentence.IsFixObtained);
            // Update satellite time
            OnUtcDateTimeReceived(sentence.UtcDateTime);
            // Update fix quality
            OnFixQualityReceived(sentence.FixQuality);
            // Update mean DOP
            OnMeanDilutionOfPrecisionReceived(sentence.MeanDilutionOfPrecision);
            // Update current location
            OnPositionReceived(sentence.Position);
            // Update altitude
            OnAltitudeReceived(sentence.Altitude);
            // Update rate of climb
            OnVerticalSpeedChanged(sentence.UtcDateTime, sentence.Altitude);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(PgrmzSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update altitude
            OnAltitudeReceived(sentence.Altitude);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(PgrmeSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update position error
            OnHorizontalPositionErrorReceived(sentence.HorizontalPositionError);
            OnVerticalPositionErrorReceived(sentence.VerticalPositionError);
            OnMeanPositionErrorReceived(sentence.MeanPositionError);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(PmgnstSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            OnFixMethodReceived(sentence.FixMethod);
            OnFixReceived(sentence.IsFixed);
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

        public bool ParseSentence(HchdgSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            // Update magnetic variation
            OnMagneticVariationReceived(sentence.MagneticVariation);
            // Update bearing
            OnBearingReceived(sentence.MagneticBearing.Add(sentence.MagneticVariation.DecimalDegrees));
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }

#if PocketPC
        public bool ParseSentence(PglorSentence sentence)
        {
            // Is it invalid?
            if (!sentence.IsValid)
                return false;

            Reset();
            // Show that parsing was successful
            OnSentenceInterpreted(sentence);
            return true;
        }
#endif

        /// <summary>
		/// Resets all recorded GPS values to their defaults.
		/// </summary>
		/// <remarks>
		/// This method is typically used to remove any cached values such as the current
		/// position, bearing and altitude, when restarting the interpreter. If an interpreter is
		/// paused or stopped, its property values can become outdated. This method prevents stale
		/// data from being used accidentally.
		/// </remarks>
		/// <example>
		///     This example resets all properties to their defaults after shutting down the
		///     interpreter. 
		///     <code lang="VB" title="[New Example]">
		/// Dim Interpreter As New NmeaInterpreter()
		/// ' Start processing GPS data
		/// Interpreter.Start()
		/// ' Let it process data for 20 seconds
		/// Thread.Sleep(TimeSpan.FromSeconds(20))
		/// ' Stop processing
		/// Interpreter.Stop()
		/// ' Delay for long enough for the properties to become outdated
		/// Thread.Sleep(TimeSpan.FromSeconds(10))
		/// ' Reset all properties
		/// Interpreter.Reset()
		///     </code>
		/// 	<code lang="CS" title="[New Example]">
		/// NmeaInterpreter Interpreter = new NmeaInterpreter();
		/// // Start processing GPS data
		/// Interpreter.Start();
		/// // Let it process data for 20 seconds
		/// Thread.Sleep(TimeSpan.FromSeconds(20));
		/// // Stop processing
		/// Interpreter.Stop();
		/// // Delay for long enough for the properties to become outdated
		/// Thread.Sleep(TimeSpan.FromSeconds(10));
		/// // Reset all properties
		/// Interpreter.Reset();
		///     </code>
		/// </example>
		public override void Reset()
		{
			pAltitude = Distance.Empty;
			pAltitudeAboveEllipsoid = Distance.Empty;
			pBearing = Azimuth.Empty;
			pDifferentialGps = new DifferentialGps();
			pFixMethod = FixMethod.NoFix;
			pFixMode = FixMode.Manual;
			pFixQuality = FixQuality.NoFix;
            pFixLikelihood = FixLikelihood.Unlikely;
			pGeoidalSeparation = Distance.Empty;
			pHorizontalDilutionOfPrecision = DilutionOfPrecision.Maximum;
            pIsHorizontalDilutionOfPrecisionSupported = false;
			pMagneticVariation = Longitude.Empty;
			pMeanDilutionOfPrecision = DilutionOfPrecision.Maximum;
            pIsMeanDilutionOfPrecisionSupported = false;
			pUtcDateTime = DateTime.UtcNow;
			pPosition = Position.Empty;
			pSpeed = Speed.Empty;
			pVerticalDilutionOfPrecision = DilutionOfPrecision.Maximum;
            pIsVerticalDilutionOfPrecisionSupported = false;
            pPreviousUtcDateTime = DateTime.MinValue;
            pPreviousAltitude = Distance.Invalid;
            pVerticalSpeed = Speed.Empty;
            pPreviousSpeed = Speed.Infinity;
            pAcceleration = Speed.AtRest;
			pSatellites = new SatelliteCollection();
            pPositionAveragingItems = new PositionCollection();
		}

		#region Satellite Events

		private void OnSatelliteAzimuthChanged(object sender, SatelliteEventArgs e)
		{
			if(SatellitePositionChanged != null)
#if PocketPC && !Framework20
				SatellitePositionChanged(sender, e);
#else
			DynamicEvent.Invoke(SatellitePositionChanged, sender, e);
#endif
		}

		private void OnSatelliteElevationChanged(object sender, SatelliteEventArgs e)
		{
			if(SatellitePositionChanged != null)
#if PocketPC && !Framework20
				SatellitePositionChanged(sender, e);
#else
			DynamicEvent.Invoke(SatellitePositionChanged, sender, e);
#endif
		}

		private void OnSatelliteFixed(object sender, SatelliteEventArgs e)
		{
			if(SatelliteFixObtained != null)
#if PocketPC && !Framework20
				SatelliteFixObtained(sender, e);
#else
			DynamicEvent.Invoke(SatelliteFixObtained, sender, e);
#endif
		}

		private void OnSatelliteFixLost(object sender, SatelliteEventArgs e)
		{
			if(SatelliteFixLost != null)
#if PocketPC && !Framework20
				SatelliteFixLost(sender, e);
#else
			DynamicEvent.Invoke(SatelliteFixLost, sender, e);
#endif
		}

		private void OnSatelliteSignalStrengthChanged(object sender, SatelliteEventArgs e)
		{
			if(SatelliteSignalStrengthChanged != null)
#if PocketPC && !Framework20
				SatelliteSignalStrengthChanged(sender, e);
#else
			DynamicEvent.Invoke(SatelliteSignalStrengthChanged, sender, e);
#endif
		}

		#endregion

        /// <summary>
        /// Indicates whether the GPS device is automatically calculating altitude, latitude and longitude.
        /// </summary>
#if DesignTime || !PocketPC
        [Category("Fix Status")]
        [Description("Indicates whether the GPS device is automatically calculating altitude, latitude and longitude.")]
#endif
        public FixMode FixMode
        {
            get
            {
                return pFixMode;
            }
        }

        /// <summary>
        /// Indicates the current technique used to calculate the current location.
        /// </summary>
#if DesignTime || !PocketPC
        [Category("Fix Status")]
        [Description("Indicates the current technique used to calculate the current location.")]
#endif
        public FixQuality FixQuality
        {
            get
            {
                return pFixQuality;
            }
        }

        /// <summary>
        /// Indicates whether the GPS device has a fix, and whether altitude can also be calculated.
        /// </summary>
#if DesignTime || !PocketPC
        [Category("Fix Status")]
        [Description("Indicates whether the GPS device has a fix, and whether altitude can also be calculated.")]
#endif
        public FixMethod FixMethod
        {
            get
            {
                return pFixMethod;
            }
        }

		/// <summary>
		/// Indicates if enough satellites are available to calculate the current
		/// location.
		/// </summary>
#if DesignTime || !PocketPC
        [Category("Fix Status")]
		[Description("Indicates if enough satellites are available to calculate the current location.")]
#endif
		public bool IsFixed
		{
			get
			{
				return pIsFixed;
			}
		}

        /// <summary>
        /// Controls whether a fix is required in order to accept time, position, altitude, or bearing measurements.
        /// </summary>
#if DesignTime || !PocketPC
        [Category("Precision")]
        [Description("Controls whether a fix is required in order to accept time, position, altitude, or bearing measurements.")]
#endif
        public bool IsFixRequired
        {
            get
            {
                return pIsFixRequired;
            }
            set
            {
                pIsFixRequired = value;
            }
        }

        /// <summary>
        /// Indicates whether the device is reporting precision information in regards to altitude.
        /// </summary>
        /// <remarks>
        /// In extremely rare cases, Dilution of Precision values are not provided, which can have
        /// an adverse effect on precision control.  When this property is False, precision control
        /// features cannot be enforced.  This poor behavior has only been witnessed on a Garmin iQue 
        /// M4.
        /// </remarks>
#if DesignTime || !PocketPC
        [Category("Precision")]
        [Description("Indicates whether the device is reporting precision information in regards to altitude.")]
#endif
        public bool IsVerticalDilutionOfPrecisionSupported
        {
            get
            {
                return pIsVerticalDilutionOfPrecisionSupported;
            }
        }

        /// <summary>
        /// Indicates whether the device is reporting precision information in regards to latitude and longitude.
        /// </summary>
        /// <remarks>
        /// In extremely rare cases, Dilution of Precision values are not provided, which can have
        /// an adverse effect on precision control.  When this property is False, precision control
        /// features cannot be enforced.  This poor behavior has not yet been witness on any GPS device.
        /// </remarks>
#if DesignTime || !PocketPC
        [Category("Precision")]
        [Description("Indicates whether the device is reporting precision information in regards to latitude and longitude.")]
#endif
        public bool IsHorizontalDilutionOfPrecisionSupported
        {
            get
            {
                return pIsHorizontalDilutionOfPrecisionSupported;
            }
        }

        /// <summary>
        /// Indicates whether the device is reporting overall precision information.
        /// </summary>
        /// <remarks>
        /// In extremely rare cases, Dilution of Precision values are not provided, which can have
        /// an adverse effect on precision control.  When this property is False, precision control
        /// features cannot be enforced.  This poor behavior has only been witnessed on a Garmin iQue 
        /// M4.
        /// </remarks>
#if DesignTime || !PocketPC
        [Category("Precision")]
        [Description("Indicates whether the device is reporting overall precision information.")]
#endif
        public bool IsMeanDilutionOfPrecisionSupported
        {
            get
            {
                return pIsMeanDilutionOfPrecisionSupported;
            }
        }
		#endregion

        /// <summary>
        /// Creates a new NMEA interpreter using the specified file as its data source.
        /// </summary>
        /// <param name="fileName">A <strong>String</strong> containing an absolute path to a file containing raw NMEA-0183 data.</param>
        /// <returns></returns>
        public static NmeaInterpreter FromFile(string fileName)
        {
            return new NmeaInterpreter(
                new GeoFramework.Gps.Emulators.TextFileEmulator(fileName));                    
        }

        /// <summary>
        /// Creates a new NMEA interpreter using the specified Stream as its data source.
        /// </summary>
        /// <param name="stream">Any <strong>Stream</strong> object.</param>
        /// <returns></returns>
        public static NmeaInterpreter FromStream(Stream stream)
        {
            return new NmeaInterpreter(stream);
        }

        /// <summary>
        /// Creates a new NMEA interpreter using the specified serial connection settings.
        /// </summary>
        /// <param name="port"></param>
        /// <param name="baudRate"></param>
        /// <returns></returns>
        public static NmeaInterpreter FromSerialDevice(GeoFramework.IO.Serial.CommunicationPort port, 
            GeoFramework.IO.Serial.BaudRate baudRate)
        {
            return new NmeaInterpreter(
                    new GeoFramework.IO.Serial.SerialStream(port, baudRate));
        }

        /// <summary>
        /// Creates an NMEA interpreter using any available GPS device.
        /// </summary>
        /// <returns></returns>
        public static NmeaInterpreter FromAnyGpsDevice()
        {
            return new NmeaInterpreter(new GpsStream());
        }

		#region Parsing Thread

		/// <summary>
		/// Performs analysis on a device to see if data can be recognized as NMEA-0183
		/// data.
		/// </summary>
		/// <returns>
		/// A <strong>Boolean</strong>, <strong>True</strong> if NMEA-0183 data was
		/// successfully received from the device.
		/// </returns>
		/// <remarks>
		/// 	<para>This method does most of the work when it comes to automatic GPS device
		///     detection. This method is called for each available device on the system in an
		///     attempt to recognize them all. This method creates an <strong>NmeaStream</strong>
		///     with the device's hardware stream as its data source, then attempts to read a
		///     sentence. If a full sentence is received, this method returns
		///     <strong>True</strong>, and the device becomes available for use by the
		///     <strong>GpsStream</strong> class.</para>
		/// 	<para>This method will typically never have to be called manually because it is
		///     called automatically during GPS device detection. This method could be used to
		///     re-detect a single device or to confirm that a device is <em>still</em> sending
		///     valid data.</para>
		/// </remarks>
		/// <param name="device">A <strong>Device</strong> object to analyze for NMEA-0183 data.</param>
#if !PocketPC
        [SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode=true)]
#endif
		public override DeviceRecognitionResult IsDeviceRecognizable(Device device)
		{
            // If a device is already detected and we're only looking for one, stop now
            if (Devices.IsDeviceDetected && Devices.IsOnlyFirstDeviceDetected)
                return DeviceRecognitionResult.NotRecognized;

            // Set thread priority
            int BaudRateIndex = 0;
			NmeaStream Checker = null;

			try
			{
                // Set the thread priority depending on the platform
                switch (Platform.HostPlatformID)
                {
                    case HostPlatformID.PocketPC:
                        // Bump the priority up if the device has been detected before
                        if (device.IsPreviouslyDetected)
                            Thread.CurrentThread.Priority = ThreadPriority.AboveNormal;
                        else
                            Thread.CurrentThread.Priority = ThreadPriority.BelowNormal;
                        break;

                }

                // Attempt to get a connection to the device
                Stream UnderlyingStream = device.GetHardwareStream();
                if (UnderlyingStream == null)
				{
					// The stream won't event open, so demote ourselves so other threads can have a better chance
                    if(Platform.HostPlatformID == HostPlatformID.PocketPC)
                        Thread.CurrentThread.Priority = ThreadPriority.Lowest;

                    // Exit, indicating that the detection failed
                    return DeviceRecognitionResult.NotRecognized;
                }
				SerialDevice serialDevice = device as SerialDevice;
				SerialStream serialStream = UnderlyingStream as SerialStream;

				// Is this a serial device?
				if (serialDevice != null)
				{
					// Try to open it
					if (!serialStream.TryOpen(false))
					{
                        // The stream won't event open, so demote ourselves so other threads can have a better chance
                        if (Platform.HostPlatformID == HostPlatformID.PocketPC)
                            Thread.CurrentThread.Priority = ThreadPriority.Lowest;

                        // Exit, indicating that the detection failed
                        return DeviceRecognitionResult.NotRecognized;
                    }

                    // If this is a Bluetooth device or the WindowsMobile split port, max out the
                    // baud rate, dude!
                    if (serialDevice.IsBluetoothDevice)
                    {
                        // Safe to max it out, baby.
                        serialStream.BaudRate = BaudRate.Baud115200;
                    }
                    else if (serialDevice.IsPreviouslyDetected)
                    {
                        // Don't change the baud rate.  It's already where it needs to be
                    }
                    else
                    {
                        lock (Devices.SerialBaudRates.SyncRoot)
                        {
                            // Start at the first baud rate
                            serialStream.BaudRate = (BaudRate)Devices.SerialBaudRates[0];
                        }
                        Thread.Sleep(0);
                    }

					// Use an NMEA stream to identify data
					Checker = new NmeaStream(serialStream);
				}
				else
				{
					// Use an NMEA stream to identify data
					Checker = new NmeaStream(UnderlyingStream);
                }

				// Set the timeout for device detection
				//Checker.ReadTimeout = Devices.DetectionTimeoutPerDevice;

                // Improve the thread priority if the device was never previously detected
                if (Platform.HostPlatformID == HostPlatformID.PocketPC && !device.IsPreviouslyDetected)
                    Thread.CurrentThread.Priority = ThreadPriority.Normal;
                        
//#if PocketPC
//                // For HP HW651x devices, don't even try to wait for data
//                if(GeoFramework.IO.HewlettPackard.Ipaq.IsSupported)
//                {
//                    // Do nothing
//                }
//                else
//                {
//#endif
//                    // Try to read a byte
//                    //int TestByte = UnderlyingStream.ReadByte();

//                    //// See if any bytes at all arrive
//                    //if (!Checker.WaitForData())
//                    //{
//                    //    // The stream won't event open, so demote ourselves so other threads can have a better chance
//                    //    if (Platform.HostPlatformID == HostPlatformID.PocketPC)
//                    //        Thread.CurrentThread.Priority = ThreadPriority.Lowest;

//                    //    // Close the connection
//                    //    Checker.Close();

//                    //    // Exit, indicating that the detection failed
//                    //    return DeviceRecognitionResult.NotRecognized;
//                    //}
//#if PocketPC
//                }
//#endif

                //// Data has arrived!  Continue to improve the thread priority
                //if (Platform.HostPlatformID == HostPlatformID.PocketPC && !device.IsPreviouslyDetected)
                //    Thread.CurrentThread.Priority = ThreadPriority.AboveNormal;

                // Data arrived.  Is it recognizable?
                while (device.IsDetectionInProgress)
                {
                    // If a device is already detected and we're only looking for one, stop now
                    if(Devices.IsDeviceDetected && Devices.IsOnlyFirstDeviceDetected)
                        return DeviceRecognitionResult.NotRecognized;

                    // Try to read in a sentence
                    NmeaSentence Test = Checker.ReadSentence(false);
					
					// Was anything returned?
					if(Test != null)
					{
                        // The device did not respond at all, so demote ourselves so other threads can have a better chance
                        if (Platform.HostPlatformID == HostPlatformID.PocketPC)
                            Thread.CurrentThread.Priority = ThreadPriority.AboveNormal;

						// Yes!  Was it too much data?
						if(Test.Length < NmeaStream.MaximumAllowedSentenceLength)
						{
							// No.  Was it valid?
							if(Test.IsValid)
							{
								// Yes!  Device detected!!
                                if (Platform.HostPlatformID == HostPlatformID.PocketPC)
                                    Thread.CurrentThread.Priority = ThreadPriority.Highest;

								// Yes!  We recognized the data!
								if (serialDevice != null)
								{
									// Reconfig the device for the right baud rate
									serialDevice.BaudRate = serialStream.BaudRate;
								}

								// NOTE: The stream is NOT Closed!  It is left open
								// in case there is an immediate need for it.  The
								// Devices.RegisterGpsDevice() method will close the
								// stream if the protocol it handles is not needed.
								GC.SuppressFinalize(Checker);

                                // Return success!
								return new DeviceRecognitionResult(true, UnderlyingStream);
							}
							else
							{
								// No.  Read another sentence to be sure
								continue;
							}
						}
						else
						{
                            // If a device was detected and only one device is needed, stop now
                            if (Devices.IsOnlyFirstDeviceDetected && Devices.IsDeviceDetected)
                                return DeviceRecognitionResult.NotRecognized;

							// Yes.  It was garbage!
							// The sentence was null, which suggests an error reading from the
							// stream.  If this is a Serial device, we can try adjust the baud rate
							// No.  If this is a serial device, try another baud rate
							if (serialDevice != null)
							{
								// Move to the next baud rate
								BaudRateIndex++;

								// Don't let anyone change the SerialBaudRates now
								lock (Devices.SerialBaudRates.SyncRoot)
								{
									// Are we at the end of the line?
									if (BaudRateIndex >= Devices.SerialBaudRates.Count)
									{
										// Yes.  No more baud rates can be checked, so detection failed
										Checker.Close();
										return DeviceRecognitionResult.NotRecognized;
									}
									else
									{
										// Try another baud rate
										serialStream.BaudRate = (BaudRate)Devices.SerialBaudRates[BaudRateIndex];

                                        // Update the device 
                                        serialDevice.BaudRate = serialStream.BaudRate;

                                        // And notify of the new attempt
                                        Devices.OnDeviceDetectionAttempted(device);

										// And keep going
										continue;
									}
								}
							}
							else
							{
                                // The device did not respond at all, so demote ourselves so other threads can have a better chance
                                if (Platform.HostPlatformID == HostPlatformID.PocketPC)
                                    Thread.CurrentThread.Priority = ThreadPriority.Lowest;

								// The stream is non-serial (i.e. Bluetooth) and no results are available.
								// Detection has failed.
								Checker.Close();
								return DeviceRecognitionResult.NotRecognized;
							}
						}
					}
					else
					{
                        // No.  No data whatsoever

                        // If a device was detected and only one device is needed, stop now
                        if (Devices.IsOnlyFirstDeviceDetected && Devices.IsDeviceDetected)
                            return DeviceRecognitionResult.NotRecognized;

                        // The sentence was null, which suggests an error reading from the
                        // stream.  If this is a Serial device, we can try adjust the baud rate
                        // No.  If this is a serial device, try another baud rate
                        if (serialDevice != null)
                        {
                            // Move to the next baud rate
                            BaudRateIndex++;
                            // Don't let anyone change the SerialBaudRates now
                            lock (Devices.SerialBaudRates.SyncRoot)
                            {
                                // Are we at the end of the line?
                                if (BaudRateIndex >= Devices.SerialBaudRates.Count)
                                {
                                    // Yes.  No more baud rates can be checked, so detection failed
                                    Checker.Close();
                                    return DeviceRecognitionResult.NotRecognized;
                                }
                                else
                                {
                                    // Try another baud rate
                                    serialStream.BaudRate = (BaudRate)Devices.SerialBaudRates[BaudRateIndex];

                                    // Update the device 
                                    serialDevice.BaudRate = serialStream.BaudRate;

                                    // And notify of the new attempt
                                    Devices.OnDeviceDetectionAttempted(device);

                                    // And keep going
                                    continue;
                                }
                            }
                        }
                        else
                        {
                            // The device did not respond at all, so demote ourselves so other threads can have a better chance
                            if (Platform.HostPlatformID == HostPlatformID.PocketPC)
                                Thread.CurrentThread.Priority = ThreadPriority.Lowest;

                            // The stream is non-serial (i.e. Bluetooth) and no results are available.
                            // Detection has failed.
                            Checker.Close();
                            return DeviceRecognitionResult.NotRecognized;
                        }
					}
                }

                // Detection is no longer in progress, so demote ourselves so other threads can have a better chance
                if (Platform.HostPlatformID == HostPlatformID.PocketPC)
                    Thread.CurrentThread.Priority = ThreadPriority.Lowest;

				// Detection is no longer in progress.  Abort
				Checker.Close();
				return DeviceRecognitionResult.NotRecognized;
            }
#if !PocketPC || Framework20
            catch (ThreadAbortException)
            {
                // Close the stream and exit
                if (Checker != null)
                    Checker.Close();
                return DeviceRecognitionResult.NotRecognized;
            }
#endif
			catch
			{
				throw;
			}			
		}

		/// <summary>Occurs when the next available data should be interpreted.</summary>
		protected override void OnParseIncomingData()
		{
			// If we're closing, just exit
			NmeaSentence Sentence = null;
            try
            {
                // Read the next sentence
                Sentence = pNmeaStream.ReadSentence(true);

                // Is the sentence null?  If so, skip
                if (Sentence == null)
                    return;

                // Notify of the received packet
                OnPacketReceived(Sentence);

                if (Sentence.IsValid && Sentence.IsSupported)
                {
                    // Parse the sentence
                    if (!ParseSentence(Sentence, false))
                        OnExceptionOccurred(new NmeaSentenceException(Sentence, "The NMEA sentence is not supported."));
                }
            }          
            catch
            {
                throw;
            }
		}

		#endregion

		#region Event Invokers

		private void OnMaximumMeanDilutionOfPrecisionChanged(DilutionOfPrecision value)
		{
			if(MaximumMeanDilutionOfPrecisionChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				MaximumMeanDilutionOfPrecisionChanged(this, new DilutionOfPrecisionEventArgs(value));
#else
			DynamicEvent.Invoke(MaximumMeanDilutionOfPrecisionChanged, this, new DilutionOfPrecisionEventArgs(value));
#endif
		}

		private void OnMaximumHorizontalDilutionOfPrecisionChanged(DilutionOfPrecision value)
		{
			if(MaximumHorizontalDilutionOfPrecisionChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				MaximumHorizontalDilutionOfPrecisionChanged(this, new DilutionOfPrecisionEventArgs(value));
#else
			DynamicEvent.Invoke(MaximumHorizontalDilutionOfPrecisionChanged, this, new DilutionOfPrecisionEventArgs(value));
#endif
		}

		private void OnMaximumVerticalDilutionOfPrecisionChanged(DilutionOfPrecision value)
		{
			if(MaximumVerticalDilutionOfPrecisionChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				MaximumVerticalDilutionOfPrecisionChanged(this, new DilutionOfPrecisionEventArgs(value));
#else
			DynamicEvent.Invoke(MaximumVerticalDilutionOfPrecisionChanged, this, new DilutionOfPrecisionEventArgs(value));
#endif
		}

		private void OnAltitudeReceived(Distance altitude)
		{
            // Skip if a fix is required but none is present
            if (pIsFixRequired && !pIsFixed)
                return;

			// Is the vertical dilution of precision above the max?  If so, IGNORE this sample!
			if(pIsVerticalDilutionOfPrecisionSupported && pVerticalDilutionOfPrecision > pMaximumVerticalDilutionOfPrecision) 
                return;

            // Ignore altitudes of zero
            if (altitude.IsEmpty)
                return;

            // Notify of the received altitude
            if (AltitudeReceived != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				AltitudeReceived(this, new DistanceEventArgs(altitude));
#else
                DynamicEvent.Invoke(AltitudeReceived, this, new DistanceEventArgs(altitude));
#endif

			// If it hasn't actually changed, ignore it
			if(pAltitude.Equals(altitude)) 
                return;

            // Set the global altitude
            Devices.SetCurrentAltitude(this, altitude);

            // Set the new value
			pAltitude = altitude;

			// And raise an event
			if(AltitudeChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				AltitudeChanged(this, new DistanceEventArgs(altitude));
#else
                DynamicEvent.Invoke(AltitudeChanged, this, new DistanceEventArgs(altitude));
#endif
        }

        private void OnVerticalSpeedChanged(DateTime utcDateTime, Distance altitude)
        {
            // Skip if a fix is required but none is present
            if (pIsFixRequired && !pIsFixed)
                return;

            // Is the vertical dilution of precision above the max?  If so, IGNORE this sample!
            if (pIsVerticalDilutionOfPrecisionSupported && pVerticalDilutionOfPrecision > pMaximumVerticalDilutionOfPrecision)
                return;

            // Is this the first reading?
            if (pPreviousUtcDateTime.Equals(DateTime.MinValue) || pPreviousAltitude.Equals(Distance.Invalid))
            {
                pPreviousAltitude = altitude;
                pPreviousUtcDateTime = utcDateTime;
            }
            else
            {
                // Calculate the change in seconds
                TimeSpan TimeDifference = utcDateTime.Subtract(pPreviousUtcDateTime);

                // Calculate the change in meters
                Distance AltitudeDifference = altitude.Subtract(pPreviousAltitude).ToMeters();

                // Now calculate meters-per-second
                pVerticalSpeed = new Speed(AltitudeDifference.Value / TimeDifference.TotalSeconds, SpeedUnit.MetersPerSecond).ToLocalUnitType();
            }
        }

        private void OnAccelerationChanged(DateTime utcDateTime, Speed speed)
        {
            // Skip if a fix is required but none is present
            if (pIsFixRequired && !pIsFixed)
                return;

            // Is this the first reading?
            if (pPreviousUtcDateTime.Equals(DateTime.MinValue) || pPreviousSpeed.Equals(Speed.Infinity))
            {
                pPreviousSpeed = speed;
                pPreviousUtcDateTime = utcDateTime;
            }
            else
            {
                // Calculate the change in seconds
                TimeSpan TimeDifference = utcDateTime.Subtract(pPreviousUtcDateTime);

                // Calculate the change in meters
                Speed SpeedDifference = speed.Subtract(pPreviousSpeed).ToMetersPerSecond();

                // Now calculate meters-per-second-per-second acceleration
                pAcceleration = new Speed(SpeedDifference.Value / TimeDifference.TotalSeconds, SpeedUnit.MetersPerSecond);
            }
        }

		private void OnAltitudeAboveEllipsoidReceived(Distance altitudeAboveEllipsoid)
		{
            // Skip if a fix is required but none is present
            if (pIsFixRequired && !pIsFixed)
                return;

			// Is the vertical dilution of precision above the max?  If so, IGNORE this sample!
            if (pIsVerticalDilutionOfPrecisionSupported && pVerticalDilutionOfPrecision > pMaximumVerticalDilutionOfPrecision) 
                return;

            // Ignore altitude of zero
            if (altitudeAboveEllipsoid.IsEmpty)
                return;

            // Notify of the received altitude
            if (AltitudeAboveEllipsoidReceived != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				AltitudeAboveEllipsoidReceived(this, new DistanceEventArgs(altitudeAboveEllipsoid));
#else
                DynamicEvent.Invoke(AltitudeAboveEllipsoidReceived, this, new DistanceEventArgs(altitudeAboveEllipsoid));
#endif

			// If it hasn't actually changed, ignore it
			if(pAltitudeAboveEllipsoid.Equals(altitudeAboveEllipsoid)) 
                return;

			// Set the new value
			pAltitudeAboveEllipsoid = altitudeAboveEllipsoid;
			
            // And raise an event
			if(AltitudeAboveEllipsoidChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				AltitudeAboveEllipsoidChanged(this, new DistanceEventArgs(altitudeAboveEllipsoid));
#else
            DynamicEvent.Invoke(AltitudeAboveEllipsoidChanged, this, new DistanceEventArgs(altitudeAboveEllipsoid));
#endif
        }

		private void OnBearingReceived(Azimuth bearing)
		{
            // Skip if a fix is required but none is present
            if (pIsFixRequired && !pIsFixed)
                return;

			// Is the vertical dilution of precision above the max?  If so, IGNORE this sample!
			if(pIsHorizontalDilutionOfPrecisionSupported && pHorizontalDilutionOfPrecision > pMaximumHorizontalDilutionOfPrecision) 
                return;

            // Notify of the received bearing
            if (BearingReceived != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				BearingReceived(this, new AzimuthEventArgs(bearing));
#else
                DynamicEvent.Invoke(BearingReceived, this, new AzimuthEventArgs(bearing));
#endif
		
            // If it hasn't actually changed, ignore it
			if(pBearing.Equals(bearing)) 
                return;

            // Update the global bearing
            Devices.SetCurrentBearing(this, bearing);

            // Set the new value
			pBearing = bearing;
			
            // And raise an event
			if(BearingChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				BearingChanged(this, new AzimuthEventArgs(bearing));
#else
            DynamicEvent.Invoke(BearingChanged, this, new AzimuthEventArgs(bearing));
#endif
        }	

		private void OnFixReceived(bool isFixed)
		{
			// If it hasn't actually changed, ignore it
			if(pIsFixed == isFixed) 
                return;
			
            // Set the new value
			pIsFixed = isFixed;

            // Set the global fix indicator
            Devices.SetCurrentlyFixed(this, isFixed);

			// And raise an event
			if (isFixed)
			{
				if(FixObtained != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
					FixObtained(this, EventArgs.Empty);
#else
                DynamicEvent.Invoke(FixObtained, this, EventArgs.Empty);
#endif
			}
			else
			{
				if(FixLost != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
					FixLost(this, EventArgs.Empty);
#else
                DynamicEvent.Invoke(FixLost, this, EventArgs.Empty);
#endif
			}
		}

        private void OnFixedSatellitesReceived(SatelliteCollection fixedSatellites)
        {
            SatelliteCollection FixedSatellites = null;
            SatelliteCollection LostSatellites = null;

            lock (pSatellites.SyncRoot)
            {
                // Compare the current satellite collection to the "fixed" list
                foreach (Satellite satellite in pSatellites)
                {
                    // Is this satellite fixed?
                    if (satellite.IsFixed)
                    {
                        // Yes.  Is it also in the fixed satellite list?
                        bool IsFound = false;
                        foreach (Satellite fixedSatellite in fixedSatellites)
                        {
                            if (fixedSatellite.PseudorandomNumber == satellite.PseudorandomNumber)
                            {
                                // Yes!  The satellite is unchanged
                                IsFound = true;
                                break;
                            }
                        }

                        // If the satellite was not found, it has dropped out of the fix.
                        if (!IsFound)
                        {
                            // Create a new list of dropped satellites
                            if (LostSatellites == null)
                                LostSatellites = new SatelliteCollection();

                            // Add the satellite to the list
                            LostSatellites.Add(satellite.ToUnfixedSatellite());
                        }
                        else
                        {
                            // The fix is already present, so do nothing.
                        }
                    }
                    else
                    {
                        // The satellite is not fixed.  If it exists in the
                        // "FixedSatellites" collection, then it must be
                        // flagged as fixed
                        bool IsFound = false;
                        foreach (Satellite fixedSatellite in fixedSatellites)
                        {
                            if (fixedSatellite.PseudorandomNumber == satellite.PseudorandomNumber)
                            {
                                // Yes!  The satellite is now fixed
                                IsFound = true;
                                break;
                            }
                        }

                        // If the satellite was found, it has joined
                        // an existing fix.
                        if (IsFound)
                        {
                            // Create a new list of fixed satellites
                            if (FixedSatellites == null)
                                FixedSatellites = new SatelliteCollection();

                            // Add the satellite to the list
                            FixedSatellites.Add(satellite.ToFixedSatellite());
                        }
                        else
                        {
                            // The satellite was not found, so it remains
                            // unfixed.  Do nothing.
                        }
                    }
                }
                if (FixedSatellites != null)
                {
                    // Yes.  Merge the two satellite collections
                    pSatellites.MergeWith(FixedSatellites);
                }
                if (LostSatellites != null)
                {
                    // Yes.  Merge the two satellite collections
                    pSatellites.MergeWith(LostSatellites);
                }
            }
            Thread.Sleep(0);

            // If anything has changed, we'll need to merge satellite collections
            if (LostSatellites != null || FixedSatellites != null)
            {
                if (FixedSatellites != null && SatelliteFixObtained != null && ((ExecutionState.SafeForEvents & State) != 0))
                {
                    foreach (Satellite item in FixedSatellites)
                    {
#if PocketPC && !Framework20
						SatelliteFixObtained(this, new SatelliteEventArgs(item));
#else
                        DynamicEvent.Invoke(SatelliteFixObtained, this, new SatelliteEventArgs(item));
#endif
                    }
                }

                if (LostSatellites != null && SatelliteFixLost != null && ((ExecutionState.SafeForEvents & State) != 0))
                {
                    foreach (Satellite item in LostSatellites)
                    {
#if PocketPC && !Framework20
						SatelliteFixLost(this, new SatelliteEventArgs(item));
#else
                        DynamicEvent.Invoke(SatelliteFixLost, this, new SatelliteEventArgs(item));
#endif
                    }
                }

                // Update the global satellite list if anything changed
                Devices.SetCurrentSatellites(this, pSatellites);
            }
        }

		private void OnGeoidalSeparationReceived(Distance geoidalSeparation)
		{
            // Skip if a fix is required but none is present
            if (pIsFixRequired && !pIsFixed)
                return;

			// Is the vertical dilution of precision above the max?  If so, IGNORE this sample!
			if(pIsVerticalDilutionOfPrecisionSupported && pVerticalDilutionOfPrecision > pMaximumVerticalDilutionOfPrecision) 
                return;

            // If it hasn't actually changed, ignore it
			if(pGeoidalSeparation.Equals(geoidalSeparation)) 
                return;
			
            // Set the new value
			pGeoidalSeparation = geoidalSeparation;

			// And raise an event
			if(GeoidalSeparationChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				GeoidalSeparationChanged(this, new DistanceEventArgs(pGeoidalSeparation));
#else
            DynamicEvent.Invoke(GeoidalSeparationChanged, this, new DistanceEventArgs(pGeoidalSeparation));
#endif
        }	

		/// <summary>Occurs when an NMEA sentence contains invalid data or its checksum fails.</summary>
		protected override void OnExceptionOccurred(Exception exception)
		{
            // Call the base exception notifier
			base.OnExceptionOccurred (exception);

            // Try to cast the exception
            NmeaSentenceException nmeaException = exception as NmeaSentenceException;

            // Is this an NMEA exception?
			if(SentenceExceptionOccurred != null 
                && State == ExecutionState.Running
                && nmeaException != null)
#if PocketPC && !Framework20
				SentenceExceptionOccurred(this, 
					new NmeaSentenceExceptionEventArgs((NmeaSentenceException)exception));
#else
			DynamicEvent.Invoke(SentenceExceptionOccurred, this,
                new NmeaSentenceExceptionEventArgs(nmeaException));
#endif
		}

		/// <summary>Occurs when a new NMEA sentence has been received.</summary>
		protected override void OnPacketReceived(Packet packet)
		{
			base.OnPacketReceived (packet);
			if(SentenceReceived != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				System.Threading.ThreadPool.QueueUserWorkItem(new WaitCallback(BeginSentenceReceived), new NmeaSentenceEventArgs((NmeaSentence)packet));
#else
            DynamicEvent.BeginInvoke(SentenceReceived, this, new NmeaSentenceEventArgs((NmeaSentence)packet));
#endif
		}

#if PocketPC && !Framework20
        private void BeginSentenceReceived(object state)
        {
			try
			{
				if(SentenceReceived != null)
					SentenceReceived(this, (NmeaSentenceEventArgs)state);
			}
			catch
			{
			}
        }
#endif

        /// <summary>Occurs when an NMEA sentence has been recorded to the recording stream.</summary>
		protected override void OnPacketRecorded(Packet packet)
		{
			base.OnPacketRecorded (packet);
			if(SentenceRecorded != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				System.Threading.ThreadPool.QueueUserWorkItem(new WaitCallback(BeginSentenceRecorded), new NmeaSentenceEventArgs((NmeaSentence)packet));
#else
                DynamicEvent.BeginInvoke(SentenceRecorded, this, new NmeaSentenceEventArgs((NmeaSentence)packet));
#endif
		}

#if PocketPC && !Framework20
		private void BeginSentenceRecorded(object state)
		{
			if(SentenceRecorded != null)
				SentenceRecorded(this, (NmeaSentenceEventArgs)state);
		}
#endif

        private void OnSentenceInterpreted(NmeaSentence sentence)
		{
			if(SentenceInterpreted != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				SentenceInterpreted(this, new NmeaSentenceEventArgs(sentence));
#else
            DynamicEvent.Invoke(SentenceInterpreted, this, new NmeaSentenceEventArgs(sentence));
#endif
        }

        private void OnSpeedReceived(Speed speed)
		{
            // Skip if a fix is required but none is present
            if (pIsFixRequired && !pIsFixed)
                return;

			// Is the dilution of precision above the max?  If so, IGNORE this sample!
			if(pIsHorizontalDilutionOfPrecisionSupported && pHorizontalDilutionOfPrecision > pMaximumHorizontalDilutionOfPrecision) 
                return;

            // Notify of the received speed
            if (SpeedReceived != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				SpeedReceived(this, new SpeedEventArgs(pSpeed));
#else
                DynamicEvent.Invoke(SpeedReceived, this, new SpeedEventArgs(speed));
#endif
			
            // If it hasn't actually changed, ignore it
			if(pSpeed.Equals(speed)) 
                return;

            // Set the new value
			pSpeed = speed;

            // Set the global speed
            Devices.SetCurrentSpeed(this, speed);

            // And raise an event
			if(SpeedChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				SpeedChanged(this, new SpeedEventArgs(speed));
#else
                DynamicEvent.Invoke(SpeedChanged, this, new SpeedEventArgs(speed));
#endif
        }

		private void OnSatellitesReceived(SatelliteCollection satellites)
		{
			//bool SatellitesHaveChanged = false;
			SatelliteCollection PositionChangedSatellites = null;
			SatelliteCollection SignalChangedSatellites = null;

			// Add the satellite to the collection if it's not already there
            lock (pSatellites.SyncRoot)
            {
                lock (satellites.SyncRoot)
                {
                    foreach (Satellite satellite in satellites)
                    {
                        if (!pSatellites.Contains(satellite.PseudorandomNumber))
                        {
                            // Raise the satellite event
                            if (PositionChangedSatellites == null)
                                PositionChangedSatellites = new SatelliteCollection();
                            PositionChangedSatellites.Add(satellite);

                            if (SignalChangedSatellites == null)
                                SignalChangedSatellites = new SatelliteCollection();
                            SignalChangedSatellites.Add(satellite);
                        }
                        else
                        {
                            /* This method is called when a GPGSV sentence is parsed.  Now,
                             * GPGSV sentences describe everything except for whether the
                             * satellite is fixed.  So we must merge the satellite's information
                             * with the current satellite's "fixed" flag.
                             */

                            // Update the satellite information
                            Satellite existing = pSatellites.GetSatellite(satellite.PseudorandomNumber);
                            // Has anything changed?
                            if (!existing.Azimuth.Equals(satellite.Azimuth)
                                || !existing.Elevation.Equals(satellite.Elevation)
                                || !existing.SignalToNoiseRatio.Equals(satellite.SignalToNoiseRatio))
                            {
                                if (PositionChangedSatellites == null)
                                    PositionChangedSatellites = new SatelliteCollection();
                                PositionChangedSatellites.Add(
                                    new Satellite(existing.PseudorandomNumber,
                                    satellite.Azimuth, satellite.Elevation,
                                    satellite.SignalToNoiseRatio, existing.IsFixed));
                            }
                        }
                    }
                }
                // Merge any necessary info
                if (PositionChangedSatellites != null)
                {
                    // Merge changes with our "master" collection
                    pSatellites.MergeWith(PositionChangedSatellites);
                }
                if (SignalChangedSatellites != null)
                {
                    // Merge changes with our "master" collection
                    pSatellites.MergeWith(SignalChangedSatellites);
                }
            }
			Thread.Sleep(0);

            if (PositionChangedSatellites != null || SignalChangedSatellites != null)
            {
                // Raise events (OUTSIDE OF THE LOCK) on Satellites which
                // have changed
                if (PositionChangedSatellites != null && SatellitePositionChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
                {
                    // And raise events
                    foreach (Satellite satellite in PositionChangedSatellites)
                    {
#if PocketPC && !Framework20
						SatellitePositionChanged(this, new SatelliteEventArgs(satellite));
#else
                        DynamicEvent.Invoke(SatellitePositionChanged, this, new SatelliteEventArgs(satellite));
#endif
                    }
                }

                // Raise events (OUTSIDE OF THE LOCK) on Satellites which
                // have changed
                if (SignalChangedSatellites != null && SatelliteSignalStrengthChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
                {
                    // And raise events
                    foreach (Satellite satellite in SignalChangedSatellites)
                    {
#if PocketPC && !Framework20					 
						SatelliteSignalStrengthChanged(this, new SatelliteEventArgs(satellite));
#else
                        DynamicEvent.Invoke(SatelliteSignalStrengthChanged, this, new SatelliteEventArgs(satellite));
#endif
                    }
                }

                // Update the global satellite list
                Devices.SetCurrentSatellites(this, pSatellites);
            }

            // Update fix liklihood
            OnFixLikelihoodReceived(pSatellites.FixLikelihood);
		}

		private void OnUtcDateTimeReceived(DateTime utcDateTime)
		{
            // If a fix is required, but none is present, skip!
            if (pIsFixRequired && !pIsFixed)
                return;

			// If it hasn't actually changed, ignore it
			if(pUtcDateTime.Equals(utcDateTime)) 
                return;
			
            // Set the new value
			pUtcDateTime = utcDateTime;

            // Set the global time
            Devices.SetCurrentUtcDateTime(this, utcDateTime);
            
            // And raise an event
			if(UtcDateTimeChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
			UtcDateTimeChanged(this, new DateTimeEventArgs(utcDateTime));
#else
            DynamicEvent.Invoke(UtcDateTimeChanged, this, new DateTimeEventArgs(utcDateTime));
#endif
        }

        private void OnFixQualityReceived(FixQuality fixQuality)
		{
			// If it hasn't actually changed, ignore it
			if(pFixQuality.Equals(fixQuality))
                return;

			// Set the new value
			pFixQuality = fixQuality;

            // If it's a differential fix, flag it as active
            pDifferentialGps.SetIsActive(pFixQuality == FixQuality.DifferentialGpsFix);                

			// And raise an event
			if(FixQualityChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				FixQualityChanged(this, new FixQualityEventArgs(fixQuality));
#else
            DynamicEvent.Invoke(FixQualityChanged, this, new FixQualityEventArgs(fixQuality));
#endif
        }

        private void OnMagneticVariationReceived(Longitude magneticVariation)
		{
			// If it hasn't actually changed, ignore it
			if(pMagneticVariation.Equals(magneticVariation)) return;
			// Set the new value
			pMagneticVariation = magneticVariation;
			// And raise an event
			if(MagneticVariationAvailable != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				MagneticVariationAvailable(this, new LongitudeEventArgs(magneticVariation));
#else
            DynamicEvent.Invoke(MagneticVariationAvailable, this, new LongitudeEventArgs(magneticVariation));
#endif
        }

        private void OnFixModeReceived(FixMode fixMode)
		{
			// If it hasn't actually changed, ignore it
			if(pFixMode.Equals(fixMode)) return;
			// Set the new value
			pFixMode = fixMode;
			// And raise an event
			if(FixModeChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				FixModeChanged(this, new FixModeEventArgs(pFixMode));
#else
            DynamicEvent.Invoke(FixModeChanged, this, new FixModeEventArgs(pFixMode));
#endif
        }

        private void OnFixMethodReceived(FixMethod fixMethod)
		{
			// If it hasn't actually changed, ignore it
			if(pFixMethod.Equals(fixMethod)) return;
			// Set the new value
			pFixMethod = fixMethod;
			// And raise an event
			if(FixMethodChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				FixMethodChanged(this, new FixMethodEventArgs(pFixMethod));
#else
            DynamicEvent.Invoke(FixMethodChanged, this, new FixMethodEventArgs(pFixMethod));
#endif
        }

        private void OnFixLikelihoodReceived(FixLikelihood FixLikelihood)
        {
            // If it hasn't actually changed, ignore it
            if (pFixLikelihood.Equals(FixLikelihood)) return;
            // Set the new value
            pFixLikelihood = FixLikelihood;
            // And raise an event
            if (FixLikelihoodChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				FixLikelihoodChanged(this, new FixLikelihoodEventArgs(pFixLikelihood));
#else
                DynamicEvent.Invoke(FixLikelihoodChanged, this, new FixLikelihoodEventArgs(pFixLikelihood));
#endif
        }

        private void OnPositionReceived(Position position)
		{
            // Skip if a fix is required but none is present
            if (pIsFixRequired && !pIsFixed)
                return;

			// Is the dilution of precision above the max?  If so, IGNORE this sample!
			if(pIsHorizontalDilutionOfPrecisionSupported && pHorizontalDilutionOfPrecision > pMaximumHorizontalDilutionOfPrecision) 
                return;

            // Ignore positions of 0,0 or null
            if (position.IsEmpty || position == null)
                return;

			// Notify of the received position
			if(PositionReceived != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				PositionReceived(this, new PositionEventArgs(position, pUtcDateTime));
#else
                DynamicEvent.Invoke(PositionReceived, this, new PositionEventArgs(position, pUtcDateTime));
#endif

            // If it hasn't actually changed, ignore it
			if(pPosition.Equals(position)) 
                return;

            // Set the global position
            Devices.SetCurrentPosition(this, position);

			// Set the new value
			pPosition = position;

            // Also apply position-averaging features
            pPositionAveragingItems.Add(pPosition);

            // Remove the first item if there are too many
            if(pPositionAveragingItems.Count > 4)
                pPositionAveragingItems.RemoveAt(0);

			// And raise an event
			if(PositionChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				PositionChanged(this, new PositionEventArgs(position, UtcDateTime));
#else
                DynamicEvent.Invoke(PositionChanged, this, new PositionEventArgs(position, UtcDateTime));
#endif

        }

        private void OnHorizontalDilutionOfPrecisionReceived(DilutionOfPrecision horizontalDilutionOfPrecision)
        {
            // If the value is non-zero, it's supported!
            if (!pIsHorizontalDilutionOfPrecisionSupported && horizontalDilutionOfPrecision.Value > 0)
                pIsHorizontalDilutionOfPrecisionSupported = true;

            // A value of 0.0 should be treated as 50.0
            if (horizontalDilutionOfPrecision.Value == 0)
                horizontalDilutionOfPrecision = 50.0;

            // If it hasn't actually changed, ignore it
            if (pHorizontalDilutionOfPrecision.Equals(horizontalDilutionOfPrecision))
                return;

            // Set the new value
            pHorizontalDilutionOfPrecision = horizontalDilutionOfPrecision;

            // Calculate horizontal position error
            if (pDifferentialGps.IsActive)
                // Error is cut in half by a differential GPS fix
                pHorizontalPositionError = pHorizontalDilutionOfPrecision.GetEstimatedError().Multiply(0.5);
            else
                // Calculate error using the standard 6-meter average
                pHorizontalPositionError = pHorizontalDilutionOfPrecision.GetEstimatedError();

            // And raise an event
            if (HorizontalDilutionOfPrecisionChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				HorizontalDilutionOfPrecisionChanged(this, new DilutionOfPrecisionEventArgs(pHorizontalDilutionOfPrecision));
#else
                DynamicEvent.Invoke(HorizontalDilutionOfPrecisionChanged, this, new DilutionOfPrecisionEventArgs(pHorizontalDilutionOfPrecision));
#endif

            // Calculate estimated position error
            if (pDifferentialGps.IsActive)
                OnHorizontalPositionErrorReceived(new Distance(4 * horizontalDilutionOfPrecision.Value, DistanceUnit.Meters));
            else
                OnHorizontalPositionErrorReceived(new Distance(6 * horizontalDilutionOfPrecision.Value, DistanceUnit.Meters));

        }

        private void OnVerticalDilutionOfPrecisionReceived(DilutionOfPrecision verticalDilutionOfPrecision)
        {
            // If the value is non-zero, it's supported!
            if (!pIsVerticalDilutionOfPrecisionSupported && verticalDilutionOfPrecision.Value > 0)
                pIsVerticalDilutionOfPrecisionSupported = true;

            // A value of 0.0 should be treated as 50.0
            if (verticalDilutionOfPrecision.Value == 0)
                verticalDilutionOfPrecision = 50.0;

            // If it hasn't actually changed, ignore it
            if (pVerticalDilutionOfPrecision.Equals(verticalDilutionOfPrecision)) return;

            // Set the new value
            pVerticalDilutionOfPrecision = verticalDilutionOfPrecision;

            // Calculate horizontal position error
            if (pDifferentialGps.IsActive)
                // Error is cut in half by a differential GPS fix
                pVerticalPositionError = pVerticalDilutionOfPrecision.GetEstimatedError().Multiply(0.5);
            else
                // Calculate error using the standard 6-meter average
                pVerticalPositionError = pVerticalDilutionOfPrecision.GetEstimatedError();

            // And raise an event
            if (VerticalDilutionOfPrecisionChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				VerticalDilutionOfPrecisionChanged(this, new DilutionOfPrecisionEventArgs(pVerticalDilutionOfPrecision));
#else
                DynamicEvent.Invoke(VerticalDilutionOfPrecisionChanged, this, new DilutionOfPrecisionEventArgs(pVerticalDilutionOfPrecision));
#endif

            // Calculate estimated position error
            if (pDifferentialGps.IsActive)
                OnVerticalPositionErrorReceived(new Distance(4 * verticalDilutionOfPrecision.Value, DistanceUnit.Meters));
            else
                OnVerticalPositionErrorReceived(new Distance(6 * verticalDilutionOfPrecision.Value, DistanceUnit.Meters));
        }

        private void OnMeanDilutionOfPrecisionReceived(DilutionOfPrecision meanDilutionOfPrecision)
		{
            // If the value is non-zero, it's supported!
            if (!pIsMeanDilutionOfPrecisionSupported && meanDilutionOfPrecision.Value > 0)
                pIsMeanDilutionOfPrecisionSupported = true;

            // A value of 0.0 should be treated as 50.0
            if (meanDilutionOfPrecision.Value == 0)
                meanDilutionOfPrecision = 50.0;

			// If it hasn't actually changed, ignore it
			if(pMeanDilutionOfPrecision.Equals(meanDilutionOfPrecision)) return;
			
            // Set the new value
			pMeanDilutionOfPrecision = meanDilutionOfPrecision;

            // Calculate horizontal position error
            if (pDifferentialGps.IsActive)
                // Error is cut in half by a differential GPS fix
                pMeanPositionError = pMeanDilutionOfPrecision.GetEstimatedError().Multiply(0.5);
            else
                // Calculate error using the standard 6-meter average
                pMeanPositionError = pMeanDilutionOfPrecision.GetEstimatedError();

            // And raise an event
			if(MeanDilutionOfPrecisionChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				MeanDilutionOfPrecisionChanged(this, new DilutionOfPrecisionEventArgs(pMeanDilutionOfPrecision));
#else
            DynamicEvent.Invoke(MeanDilutionOfPrecisionChanged, this, new DilutionOfPrecisionEventArgs(pMeanDilutionOfPrecision));
#endif

            // Calculate estimated position error
            if (pDifferentialGps.IsActive)
                OnMeanPositionErrorReceived(new Distance(4 * meanDilutionOfPrecision.Value, DistanceUnit.Meters));
            else
                OnMeanPositionErrorReceived(new Distance(6 * meanDilutionOfPrecision.Value, DistanceUnit.Meters));
        }

        private void OnHorizontalPositionErrorReceived(Distance horizontalPositionError)
		{
			// If it hasn't actually changed, ignore it
			if(pHorizontalPositionError.Equals(horizontalPositionError)) return;

            // Set the new value
			pHorizontalPositionError = horizontalPositionError;

			// And raise an event
			if(HorizontalPositionErrorChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				HorizontalPositionErrorChanged(this, new DistanceEventArgs(pHorizontalPositionError));
#else
            DynamicEvent.Invoke(HorizontalPositionErrorChanged, this, new DistanceEventArgs(pHorizontalPositionError));
#endif
        }

        private void OnVerticalPositionErrorReceived(Distance verticalPositionError)
		{
			// If it hasn't actually changed, ignore it
			if(pVerticalPositionError.Equals(verticalPositionError)) return;
			// Set the new value
			pVerticalPositionError = verticalPositionError;
			// And raise an event
			if(VerticalPositionErrorChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				VerticalPositionErrorChanged(this, new DistanceEventArgs(pVerticalPositionError));
#else
            DynamicEvent.Invoke(VerticalPositionErrorChanged, this, new DistanceEventArgs(pVerticalPositionError));
#endif
        }

        private void OnMeanPositionErrorReceived(Distance meanPositionError)
		{
			// If it hasn't actually changed, ignore it
			if(pMeanPositionError.Equals(meanPositionError)) return;
			// Set the new value
			pMeanPositionError = meanPositionError;
			// And raise an event
			if(MeanPositionErrorChanged != null && ((ExecutionState.SafeForEvents & State) != 0))
#if PocketPC && !Framework20
				MeanPositionErrorChanged(this, new DistanceEventArgs(pMeanPositionError));
#else
            DynamicEvent.Invoke(MeanPositionErrorChanged, this, new DistanceEventArgs(pMeanPositionError));
#endif
        }

        /// <summary>
        /// Returns the estimated amount of error in regards to the current altitude.
        /// </summary>
#if DesignTime || !PocketPC
        [Category("Precision")]
        [Description("Returns the estimated amount of error in regards to the current altitude.")]
#endif
        public Distance VerticalPositionError
        {
            get
            {
                // If WAAS or EGNOS is active, approximate error
                return pVerticalPositionError;
            }
        }

        /// <summary>
        /// Returns the estimated amount of error in regards to latitude, longitude, and altitude combined.
        /// </summary>
#if DesignTime || !PocketPC
        [Category("Precision")]
        [Description("Returns the estimated amount of error in regards to latitude, longitude, and altitude combined.")]
#endif
        public Distance MeanPositionError
        {
            get
            {
                return pMeanPositionError;
            }
        }

        /// <summary>
        /// Returns the estimated amount of error in regards to latitude and longitude.
        /// </summary>
#if DesignTime || !PocketPC
        [Category("Precision")]
        [Description("Returns the estimated amount of error in regards to latitude and longitude.")]
#endif
        public Distance HorizontalPositionError
        {
            get
            {
                return pHorizontalPositionError;
            }
        }

		#endregion
    }
}
