
using System;
using System.Text;
using System.Globalization;
using GeoFramework.IO;

namespace GeoFramework.Gps.Nmea
{
	/// <summary>
	/// Represents a single NMEA-0183 sentence.
	/// </summary>
	public class NmeaSentence : Packet
	{
		private string pValue;
		private string[] pWords;
		private string pChecksum;

		/// <summary>Returns the culture used to interpret all NMEA sentences.</summary>
		public static readonly CultureInfo NmeaCultureInfo = new CultureInfo("en-US");
        public static bool IsPartialSentenceParsingAllowed;

		protected NmeaSentence() 
			: base()
		{
		}

        public NmeaSentence(string sentence) 
			: base()
        {   
            SetSentence(sentence);
        }

		public NmeaSentence(byte[] bytes) 
		{
			SetSentence(ASCIIEncoding.ASCII.GetString(bytes, 0, bytes.Length));
		}

		protected void SetSentence(string sentence)
		{
			if(sentence.EndsWith("\r\n"))
			{
				pValue = sentence;
			}
			else
			{
				pValue = sentence + "\r\n";
			}
			//pCommand = GetCommand(pValue);
			pWords = GetWords(sentence);
			pChecksum = GetChecksum(pValue);
		}

		protected void SetChecksum()
		{
			string OldChecksum = pChecksum;
			pChecksum = CalculateChecksum();
			if(OldChecksum.Length == 0)
			{
				if(pValue.IndexOf("*") == -1)
					pValue = pValue.Replace("\r\n", "") + "*";
				pValue = pValue + pChecksum + "\r\n";
				pValue = pValue.Replace("\r\n\r\n", "\r\n");
			}	
			else
				pValue = pValue.Replace("*" + OldChecksum, "*" + pChecksum);			
		}

#if !Framework20 || PocketPC
		internal static int Truncate(double value)
		{
			return value > 0 
				? (int)(value - (value - Math.Floor(value))) 
				: (int)(value - (value - Math.Ceiling(value)));
		}
#endif


		public override byte[] GetBytes()
		{
			return System.Text.ASCIIEncoding.ASCII.GetBytes(pValue);
		}


        private const double roundingConstant = .0001 / 60.0;
        private const double inverseRoundingConstant = 60.0 / .0001;

        private static double RoundToNmea(double decimalDegrees)
        {
            // NMEA long/lats are accurate to .0001 minutes = .0001/60 hours

            return Math.Floor(Math.Round(decimalDegrees * inverseRoundingConstant)) * roundingConstant;
        }

        public static Position RoundToNmea(Position p)
        {
            return new Position(new Latitude(RoundToNmea(p.Latitude.DecimalDegrees)), new Longitude(RoundToNmea(p.Longitude.DecimalDegrees)));
        }

//
//		public static NmeaSentence GetSentence(string sentence)
//		{
//
//		}
//
//		public static NmeaSentence Parse(byte[] bytes)
//		{
//			return new NmeaSentence(bytes);
//		}

		/// <summary>Returns the two-letter checksum used to validate the sentence.</summary>
		public string Checksum
		{
			get
			{
				return pChecksum;
			}
		}

		/// <summary>Returns the first word of the sentence.</summary>
		public string Command
		{
			get
			{
				return pWords[0];
			}
		}

		/// <summary>Returns the sentence an an array of strings delimited by commas.</summary>
		public string[] Words
		{
			get
			{
				return pWords;
			}
		}

		/// <summary>Returns the number of comma-delimited strings within the sentence.</summary>
		public int WordCount
		{
			get
			{
				return pWords.Length;
			}
		}

		/// <summary>Returns the entire NMEA sentence.</summary>
		public string Value
		{
			get
			{
				return pValue;
			}
		}

		/// <summary>Returns the number of characters in the sentence.</summary>
		public int Length
		{
			get
			{
				return pValue.Length;
			}
		}

		// Returns the first word of an NMEA sentence
		/// <summary>Returns the first word of an NMEA sentence.</summary>
		public static string GetCommand(string sentence)
		{
			int start = sentence.IndexOf("$");
			if(start == -1)
				return "";
			int end = sentence.IndexOf(",", start);
			if(end == -1)
				return "";
			return sentence.Substring(start, end - start);			
		}

		/// <summary>Splits the specified string into an array of words.</summary>
		public static string[] GetWords(string sentence)
		{
            // Is there a CRLF?
            if (sentence.IndexOf("\r\n") != -1)
            {
                // Yes. Is there a checksum?
                if (sentence.IndexOf("*") != -1)
                    // Yes. Remove the CRLF and the checksum, then split by commas
                    return sentence.Replace("\r\n", "").Substring(0, sentence.IndexOf("*")).Split(',');
                else
                    // No. Remove the CRLF, then split by commas
                    return sentence.Replace("\r\n", "").Split(',');
            }
            else
            {
                // No. Is there a checksum?
                if (sentence.IndexOf("*") != -1)
                    // Yes. Remove the checksum, then split by commas
                    return sentence.Substring(0, sentence.IndexOf("*")).Split(',');
                else
                    // No. Split by commas
                    return sentence.Split(',');
            }
		}

		public string CalculateChecksum()
		{
			// Anytrhing to do?
			if(pValue.Length == 0)
				return "00";
			// Yes.
			int Position = 0;
			byte Total;
			byte[] Bytes = ASCIIEncoding.ASCII.GetBytes(pValue);
			// Start at the first chracter after the $ or the start if there's no dollar sign
			int StartPosition = pValue.IndexOf("$") + 1;
			if(StartPosition >= pValue.Length)
				return "00";
			// End at the asterisk or the end of the string if no asterisk is present
			int EndPosition = pValue.IndexOf("*");
			if (EndPosition == -1)
			{
				if(pValue.IndexOf("\r\n") != -1)
					EndPosition = pValue.Length - 2;
				else
					EndPosition = pValue.Length;			
			}
			// Calculate the checksum
			Total = Bytes[StartPosition];
			Position = StartPosition + 1;
			while (Position < EndPosition)
			{
				Total ^= Bytes[Position];
				Position++;
			}

			// Return the two's complement
			return Total.ToString("X", CultureInfo.InvariantCulture).PadLeft(2, '0');
		}

		public static string CalculateChecksum(string sentence)
		{
			NmeaSentence Result = new NmeaSentence(sentence);
			return Result.CalculateChecksum();
		}

		// Returns the checksum for an NMEA sentence
		/// <summary>Calculates the two-letter checksum for the specified sentence.</summary>
		public static string GetChecksum(string sentence)
		{
			int AsteriskPosition = sentence.IndexOf("*");
			// Is there a checksum?
			if(AsteriskPosition == -1)
				return "";
			else
			{
				string Checksum = sentence.Substring(AsteriskPosition + 1);
				if(Checksum.Length < 2)
					return "";
				else if(Checksum.Length > 2)
					return Checksum.Substring(0, 2);
				else
					return Checksum;
			}
		}

		/// <summary>Returns whether the sentence's data matches its checksum.</summary>
		public override bool IsValid
		{
			get
			{
				return pChecksum == CalculateChecksum();
			}
		}

		/// <summary>
		/// Indicates whether the sentence is currently supported by GPS.NET.
		/// </summary>
		/// <remarks>
		/// Sentence classes inheriting from this class should override this 
		/// property and set it to True to instruct GPS.NET to proceed with 
		/// parsing the sentence.
		/// </remarks>
		public virtual bool IsSupported
		{
			get
			{
				return false;
			}
		}

		/// <summary>Returns the entire NMEA sentence.</summary>
		public override string ToString(string format, IFormatProvider formatProvider)
		{
			return pValue.Replace("\r\n", "");
		}

		#region Conversions

		public static explicit operator string(NmeaSentence sentence)
		{
			return sentence.ToString();
		}

		public static explicit operator NmeaSentence(string sentence)
		{
			return new NmeaSentence(sentence);
		}

		#endregion
	}

	/// <summary>
	/// Represents information about raw data received from a GPS device.
	/// </summary>
	/// <remarks>This object is used primarily by the <see cref="Receiver.SentenceReceived">SentenceReceived</see>
	/// and <see cref="Receiver.UnrecognizedSentenceReceived">UnrecognizedSentenceReceived</see> events of
	/// the <see cref="Receiver">Receiver</see> class to provide notification when raw data has been received
	/// from the GPS device.</remarks>
	/// <example>This example demonstrates how to use this class when raising an event.
	/// <code lang="VB">
	/// ' Declare a new event
	/// Dim MySentenceEvent As SentenceEventHandler
	/// ' Create an NMEA $GPGLL sentence
	/// Dim MySentence As String = "$GPGLL,4122,N,14628,W,104243.00,A,D*7E"
	/// 
	/// Sub Main()
	///   ' Raise our custom event, also indicating that the sentence was completely recognized
	///   RaiseEvent MySentenceEvent(Me, New SentenceEventArgs(MySentence, SentenceType.Recognized))
	/// End Sub
	/// </code>
	/// <code lang="C#">
	/// // Declare a new event
	/// SentenceEventHandler MySentenceEvent;
	/// // Create an NMEA $GPGLL sentence
	/// string MySentence = "$GPGLL,4122,N,14628,W,104243.00,A,D*7E"
	/// 
	/// void Main()
	/// {
	///   // Raise our custom event, also indicating that the sentence was completely recognized
	///   MySentenceEvent(this, New SentenceEventArgs(MySentence, SentenceType.Recognized));
	/// }
	/// </code>
	/// </example>
	/// <seealso cref="Receiver">Receiver Class</seealso>
	/// <seealso cref="Receiver.SentenceReceived">SentenceReceived Event (Receiver Class)</seealso>
	/// <seealso cref="Receiver.UnrecognizedSentenceReceived">UnrecognizedSentenceReceived Event (Receiver Class)</seealso>
	public sealed class NmeaSentenceEventArgs : PacketEventArgs
	{
		public new static readonly NmeaSentenceEventArgs Empty = new NmeaSentenceEventArgs(null);

		/// <summary>
		/// Creates a new instance with the specified block of raw GPS data and a flag indicating if the sentence was fully parsed.
		/// </summary>
		/// <param name="sentence">A <strong>String</strong> containing raw GPS data.</param>
		/// <param name="sentenceType">A value from the <see cref="SentenceType">SentenceType</see> enumeration indicating if the sentence was fully parsed or not.</param>
		/// <seealso cref="Type">Type Property</seealso>
		/// <seealso cref="Sentence">Sentence Property</seealso>
		public NmeaSentenceEventArgs(NmeaSentence sentence) : base(sentence)
		{
		}

		/// <summary>
		/// Stores one line of raw GPS data.
		/// </summary>
		/// <value>A String containing one line of raw NMEA or Garmin text data.</value>
		/// <remarks>When using the NMEA-0183 or Garmin text protocols, this property stores text for one individual line.  For the Garmin binary protocol, a signature such as "(Garmin waypoint data)" is stored instead of actual binary data.</remarks>
		/// <seealso cref="Type">Type Property</seealso>
		public NmeaSentence Sentence
		{
			get
			{
				return (NmeaSentence)base.Packet;
			}
		}
	}

	/// <summary>
	/// Facilitates events and asynchronous calls which use a sentence-related
	/// <strong>String</strong> as a parameter.
	/// </summary>
	/// <example>
	/// This example declares a custom event using this delegate.
	/// <code lang="VB">
	/// Dim MySentenceEvent As SentenceEventHandler
	/// </code>
	/// <code lang="C#">
	/// SentenceEventHandler MySentenceEvent;
	/// </code>
	/// </example>
	/// <param name="e">A <see cref="SentenceEventArgs">SentenceEventArgs</see> object containing raw, unmodified GPS data.</param>
	/// <param name="sender">The object which initiated the call to the delegate.</param>
	public delegate void NmeaSentenceEventHandler(object sender, NmeaSentenceEventArgs e);

}
