using System;
using System.Text;
using System.Collections;
using System.IO;
using System.Globalization;
using System.Threading;
using GeoFramework.IO;
#if !PocketPC || DesignTime
using System.ComponentModel;
#endif

namespace GeoFramework.Gps.Nmea
{
	/// <summary>
	/// Provides features for reading and writing NMEA-0183 to a stream.
	/// </summary>
	/// <remarks>This class is used to interpret individual values extracted from NMEA data, as
	/// well as to transmit NMEA data on certain receivers.  The methods in this class, such as
	/// ReadPosition and ReadAltitude, will parse all received data until the requested information
	/// is found.  Methods are designed to block until the specified value is received, or until
	/// the underlying stream is closed.</remarks>
#if !PocketPC
	[TypeConverter(typeof(ExpandableObjectConverter))]
#endif
	public sealed class NmeaStream : Stream
	{
		#region Private Variables
		private Stream pBaseStream;
		private TimeSpan pReadTimeout = _DefaultReadTimeout;
		private TimeSpan pWriteTimeout = _DefaultWriteTimeout;
        private bool pIsStreamClosed;

#if Smartphone
		private static TimeSpan _DefaultReadTimeout = TimeSpan.FromSeconds(3);
		private static TimeSpan _DefaultWriteTimeout = TimeSpan.FromSeconds(3);
#elif PocketPC
        private static TimeSpan _DefaultReadTimeout = TimeSpan.FromSeconds(4);
        private static TimeSpan _DefaultWriteTimeout = TimeSpan.FromSeconds(4);
#else
        private static TimeSpan _DefaultReadTimeout = TimeSpan.FromSeconds(5);
        private static TimeSpan _DefaultWriteTimeout = TimeSpan.FromSeconds(5);
#endif

        public static TimeSpan DefaultReadTimeout
        {
            get { return _DefaultReadTimeout; }
            set 
            {
                if (value.TotalSeconds <= 0)
#if PocketPC
					throw new ArgumentOutOfRangeException("The read timeout for an NMEA stream must be greater than zero.  A value of 3-5 seconds is recommended.");
#else
                    throw new ArgumentOutOfRangeException("DefaultReadTimeout", "The read timeout for an NMEA stream must be greater than zero.  A value of 3-5 seconds is recommended.");
#endif

                _DefaultReadTimeout = value;
            }
        }

        public static TimeSpan DefaultWriteTimeout
        {
            get { return _DefaultWriteTimeout; }
            set
            {
                if (value.TotalSeconds <= 0)
#if PocketPC
					throw new ArgumentOutOfRangeException("The write timeout for an NMEA stream must be greater than zero.  A value of 3-5 seconds is recommended.");
#else
                    throw new ArgumentOutOfRangeException("DefaultWriteTimeout", "The write timeout for an NMEA stream must be greater than zero.  A value of 3-5 seconds is recommended.");
#endif

                _DefaultWriteTimeout = value;
            }
        }

		#endregion

		#region Static Members

		private static int pMaximumAllowedSentenceLength = 128;

		/// <summary>
		/// Controls the maximum allowed characters in a single NMEA sentence.
		/// </summary>
		/// <remarks>
		/// This property is used to determine when an NMEA sentence is invalid due to too many characters.  This can occur if,
		/// for example, the GPS device is set to the wrong baud rate.  In such situations, the device will transmit a large amount of
		/// garbage characters.  By setting this property to a lower level, GPS.NET can more quickly discover communication problems with 
		/// the device, which can speed up Automatic GPS Device Detection.  This property defaults to a value of 128.
		/// </remarks>
		public static int MaximumAllowedSentenceLength
		{
			get
			{
				return pMaximumAllowedSentenceLength;
			}
			set
			{
				if(value <= 0)
#if PocketPC
					throw new ArgumentOutOfRangeException("The maximum allowed sentence length must be a value greater than zero.");
#else
					throw new ArgumentOutOfRangeException("MaximumAllowedSentenceLength", value, "The maximum allowed sentence length must be a value greater than zero.");
#endif

                pMaximumAllowedSentenceLength = value;
			}
		}

		/// <summary>
		/// Returns wether a packet of raw data can be interpreted as an NMEA sentence.
		/// </summary>
		/// <param name="packet"></param>
		/// <returns></returns>
        public static bool CanParse(Packet packet)
        {
            // Convert to a string
            string sentence = packet.ToString("G", CultureInfo.InvariantCulture);
            // If there's no dollar sign, return false
            int DollarSignIndex = sentence.IndexOf("$");
            if(DollarSignIndex == -1) return false;
            // If there's no comma after the dollar sign, return false
            if (sentence.IndexOf(",", DollarSignIndex) == -1) return false;
            // If there's no asterisk after the dollar sign, return false
            if (sentence.IndexOf("*", DollarSignIndex) == -1) return false;
            // This is an NMEA sentence
            return true;
        }

        public static bool CanParse(NmeaSentence sentence)
        {
            return true;
        }

		#endregion

		#region Constructors

		/// <summary>
		/// Creates a new instance using the first available GPS device.
		/// </summary>
		public NmeaStream()
			: this(new GpsStream(NmeaInterpreter.ProtocolName))
		{
		}

		public NmeaStream(Stream input)
		{
			pBaseStream = input;

#if PocketPC && !Smartphone
			// On the 6515, timeouts must be higher to accomodate
			// the 10-second warm-up period for the integrated
			// GPS device.
			if(Environment.OSVersion.Platform == PlatformID.WinCE)
			{
				if(GeoFramework.IO.HewlettPackard.Ipaq.IsSupported)
				{
					// Change the timeout to 10 seconds by default
					pReadTimeout = TimeSpan.FromSeconds(10);
				}
			}
#endif
		}

		#endregion

		#region Overrides

#if !PocketPC
		/// <summary>Returns the stream providing raw data.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
#endif
		public Stream BaseStream
		{
			get
			{
				return pBaseStream;
			}
		}

		/// <summary>Begins an asynchronous read operation.</summary>
		public override IAsyncResult BeginRead(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			return pBaseStream.BeginRead (buffer, offset, count, callback, state);
		}

		/// <summary>Begins an asynchronous write operation.</summary>
		public override IAsyncResult BeginWrite(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			return pBaseStream.BeginWrite (buffer, offset, count, callback, state);
		}

		/// <summary>Returns whether read operations are allowed.</summary>
		public override bool CanRead
		{
			get
			{
				return !pIsStreamClosed && pBaseStream != null && pBaseStream.CanRead;
			}
		}

		/// <summary>Returns whether seek operations are allowed.</summary>
		public override bool CanSeek
		{
			get
			{
                return pBaseStream != null && pBaseStream.CanSeek && !pIsStreamClosed;
			}
		}

		/// <summary>Returns whether write operations are allowed.</summary>
		public override bool CanWrite
		{
			get
			{
				return pBaseStream != null && pBaseStream.CanWrite && !pIsStreamClosed;
			}
		}

		/// <summary>Closes the current stream as well as the underlying raw data stream.</summary>
		public override void Close()
		{
			if (pIsStreamClosed)
				return;

			// Flag us as closed
			pIsStreamClosed = true;

			if (pBaseStream != null)
				pBaseStream.Close();            
		}

//#if !PocketPC
//        /// <summary>Obsolete. Returns a new instance of the specified type.</summary>
//        public override System.Runtime.Remoting.ObjRef CreateObjRef(Type requestedType)
//        {
//            return pBaseStream.CreateObjRef(requestedType);
//        }
//#endif

		/// <summary>Completes an asynchronous read operation.</summary>
		public override int EndRead(IAsyncResult asyncResult)
		{
			return pBaseStream.EndRead (asyncResult);
		}

		/// <summary>Completes an asynchronous write operation.</summary>
		public override void EndWrite(IAsyncResult asyncResult)
		{
			pBaseStream.EndWrite (asyncResult);
		}

		/// <summary>Sends any pending data to the underlying stream.</summary>
		public override void Flush()
		{
            if (pIsStreamClosed)
                throw new ObjectDisposedException(GetType().ToString(), "A flush operation cannot begin because the stream has been closed.");
            
            pBaseStream.Flush();
		}

		/// <summary>Returns the number of bytes available on the underlying stream.</summary>
		public override long Length
		{
			get
			{
                if (pIsStreamClosed)
                    throw new ObjectDisposedException(GetType().ToString(), "The stream length cannot be returned because the stream has been closed.");
                
                return pBaseStream.Length;
			}
		}

		/// <summary>Returns the current index of the underlying stream.</summary>
		public override long Position
		{
			get
			{
                if (pIsStreamClosed)
                    throw new ObjectDisposedException(GetType().ToString(), "The stream position cannot be read because the stream has been closed.");

				return pBaseStream.Position;
			}
			set
			{
                if (pIsStreamClosed)
                    throw new ObjectDisposedException(GetType().ToString(), "The stream position cannot be set because the stream has been closed.");
                
                pBaseStream.Position = value;
			}
		}

		/// <summary>Reads one or more bytes from the underlying stream.</summary>
		public override int Read(byte[] buffer, int offset, int count)
		{
            if (pIsStreamClosed)
                throw new ObjectDisposedException(GetType().ToString(), "A read operation cannot be executed because the stream has been closed.");

			return pBaseStream.Read(buffer, offset, count);
		}

		/// <summary>Reads a single byte from the underlying stream.</summary>
		public override int ReadByte()
		{
            if (pIsStreamClosed)
                throw new ObjectDisposedException(GetType().ToString(), "A read operation cannot be executed because the stream has been closed.");
            
            return pBaseStream.ReadByte();
		}

		/// <summary>Changes the current location of the underlying stream.</summary>
		public override long Seek(long offset, SeekOrigin origin)
		{
            if (pIsStreamClosed)
                throw new ObjectDisposedException(GetType().ToString(), "A seek operation cannot be executed because the stream has been closed.");

			return pBaseStream.Seek(offset, origin);
		}

		/// <summary>Changes the length of the underlying stream.</summary>
		public override void SetLength(long value)
		{
            if (pIsStreamClosed)
                throw new ObjectDisposedException(GetType().ToString(), "A set length operation cannot be executed because the stream has been closed.");
            
            pBaseStream.SetLength(value);
		}

		public override void Write(byte[] buffer, int offset, int count)
		{
            if (pIsStreamClosed)
                throw new ObjectDisposedException(GetType().ToString(), "A white operation cannot be executed because the stream has been closed.");
            
            pBaseStream.Write(buffer, offset, count);
		}

		/// <summary>Writes a single byte to the underlying stream.</summary>
		public override void WriteByte(byte value)
		{
            if (pIsStreamClosed)
                throw new ObjectDisposedException(GetType().ToString(), "A white operation cannot be executed because the stream has been closed.");

            pBaseStream.WriteByte(value);
		}

		#endregion

		#region Public Members

		/// <summary>Returns the name of the current instance.</summary>
		public override string ToString()
		{
			return "NMEA-0183 Data Stream (" + pBaseStream.ToString() + ")";
		}

		/// <summary>Controls the maximum amount of time to wait for an NMEA sentence.</summary>
        public
#if Framework20
        new
#endif
		TimeSpan ReadTimeout
        {
			get
			{
				return pReadTimeout;
			}
			set
			{
				pReadTimeout = value;
			}
		}

        /// <summary>Controls the maximum amount of time to wait for an NMEA sentence to be sent to the device.</summary>
        public
#if Framework20
 new
#endif
 TimeSpan WriteTimeout
        {
            get
            {
                return pWriteTimeout;
            }
            set
            {
                pWriteTimeout = value;
            }
        }

		/// <summary>Reads a single line of text from the underlying stream.</summary>
		/// <returns></returns>
		public string ReadLine()
		{
			return ReadLine(true);
		}

		/// <summary>Reads a single line of text from the underlying stream.</summary>
		/// <param name="allowExceptions"></param>
		/// <returns></returns>
		public string ReadLine(bool allowExceptions)
		{
			try
			{
				if (pIsStreamClosed)
					throw new ObjectDisposedException(GetType().ToString(), "A wait for data operation cannot begin because the stream has been closed.");
				// We have data!  Now read it in.
				StringBuilder Result = new StringBuilder(pMaximumAllowedSentenceLength);
				int BadByteCount = 0;
				DateTime TimeoutDateTime = DateTime.MinValue;
				// And loop while we're able to read
				while(!pIsStreamClosed)
				{
					// Read in a byte
					int Data = pBaseStream.ReadByte();

					// Is the data valid?
					if(Data > 0)
					{
						// Yes. Add it to the list
						Result.Append((char)Data);
						// If it's a line feed or dollar sign, we're done
						if(Data == 10)
							return Result.ToString();
						// If the length is OK, keep going
						if(Result.Length < pMaximumAllowedSentenceLength)
							continue;
						// Maybe there's something right in here still, but without
						// a CRLF.   i.e.  $GPABC,,,,*70$GPRMC
						string FinalResult = Result.ToString();
						if(FinalResult.IndexOf("$") != -1
							&& FinalResult.IndexOf("*") != -1
							&& FinalResult.IndexOf(",") != -1)
						{
							try
							{
								// Return the extracted sentence
								int start = FinalResult.IndexOf("$");
								int end = FinalResult.IndexOf("*", start) + 2;
								string Extracted = FinalResult.Substring(start, end - start);
								return Extracted;
							}
							catch
							{
								// The buffer is full of bad data
								if(allowExceptions)
									throw new FormatException("Invalid characters were received which are not found in NMEA sentences.  The baud rate may be incorrect.  Try changing the baud rate to 4800 and try again.");
								else
									return Result.ToString();
							}
						}
						else
						{
							// The buffer is full of bad data
							if(allowExceptions)
								throw new FormatException("Invalid characters were received which are not found in NMEA sentences.  The baud rate may be incorrect.  Try changing the baud rate to 4800 and try again.");
							else
								return Result.ToString();
						}						
					}
					else if (Data == 0)
					{
#if PocketPC
                        if(GeoFramework.IO.HewlettPackard.Ipaq.IsSupported)
                        {
                            // On the 6515, null values are allowed for a while before valid data is sent.  Epic fail.
                            continue;
                        }
#endif
						// Do the next one
                        BadByteCount++;
                        // If the count is over the buffer, flag failure
                        if (BadByteCount > 128)
                        {
                            if (allowExceptions)
                                throw new FormatException("The device has returned nothing but null characters.  The device is not a GPS device, or the baud rate is incorrect.  Try changing the baud rate.");
                            else
                                return null;
                        }
                        continue;
					}

					/* If we get here, we encountered either a null (0)
					 * or -1, which means no data was available.  Increase
					 * the "bad byte" count.
					 */

                    // Is this a file?  If so, we're at an EOF
                    if (pBaseStream is FileStream)
                        return null;

                    // No. Increase the bad byte count
					BadByteCount++;

					// If the count is over the buffer, flag failure
					if (BadByteCount > 512)
					{
                        if (allowExceptions)
                            throw new FormatException("The buffer has become full trying to read a single line of data.  The baud rate may be incorrect.  Try changing the baud rate, or use the SetLength method to increase the size of the buffer.");
                        else
                            return null;
					}

					// If we have no timeout set, set it now
					if(TimeoutDateTime.Equals(DateTime.MinValue))
						TimeoutDateTime = DateTime.Now.Add(pReadTimeout);

					// If the read timeout has passed, throw a timeout
					if(DateTime.Now > TimeoutDateTime)
					{
						if(allowExceptions)
							// No. flag an exception
							throw new TimeoutException("The GPS device did not respond in a timely manner with any data.");
						else
							return null;
					}

#if PocketPC  && !Smartphone
					// On the HP 6515, give the spawned process some time to execute
					if(GeoFramework.IO.HewlettPackard.Ipaq.IsSupported)
					{
						System.Windows.Forms.Application.DoEvents();
					}
#endif

					// If we get here, we have a bad byte, but not too
					// many to get in a fuss over.  Sleep for a moment
					Thread.Sleep(10);
					Thread.Sleep(0);
				}
                // Return what we have
                return null;
			}
#if !PocketPC
			catch(ThreadAbortException)
			{
				// Just exit
				return null;
			}
#endif
			catch
			{
				if(allowExceptions)
					throw;
				else
					return null;
			}
		}

        /// <summary>
        /// Blocks execution until until a GPS satellite fix has been acquired, or a timeout occurs.
        /// </summary>
        /// <returns>A <strong>Boolean</strong> value, <strong>True</strong> if a fix has been obtained.</returns>
        /// <remarks>When a connection is first established to a GPS device, some time is typically required for the
        /// device to obtain a fix.  This method will keep reading NMEA data until a fix is obtained. A value of <strong>False</strong>
        /// is returned if sixty seconds have elapsed before a fix has been acquired.</remarks>
        public bool WaitForFix()
        {
            return WaitForFix(TimeSpan.FromSeconds(60));
        }

        /// <summary>
        /// Blocks execution until until a GPS satellite fix has been acquired, or a timeout occurs.
        /// </summary>
        /// <returns>A <strong>Boolean</strong> value, <strong>True</strong> if a fix has been obtained.</returns>
        /// <remarks>When a connection is first established to a GPS device, some time is typically required for the
        /// device to obtain a fix.  This method will keep reading NMEA data until a fix is obtained. A value of <strong>False</strong>
        /// is returned if sixty seconds have elapsed before a fix has been acquired.</remarks>
        public bool WaitForFix(TimeSpan timeout)
        {
            // Calculate the time to stop reading
            DateTime TimeoutTime = DateTime.Now.Add(timeout);

            // Loop until a fix is obtained or the timeout occurs
            while (DateTime.Now < TimeoutTime)
            {
                bool IsFixed = ReadIsFixObtained();
                if (IsFixed)
                    return true;
            }

            // No fix was found
            return false;
        }

        /// <summary>
        /// Blocks execution until any data arrives from the underlying stream.
        /// </summary>
        /// <returns>A <strong>Boolean</strong> value, <strong>True</strong> if data has arrived from the device.  A value
        /// of <strong>False</strong> is returned if the number of seconds specified in <strong>ReadTimeout</strong> has elapsed
        /// with no data read.</returns>
		public bool WaitForData()
		{
			return WaitForData(DateTime.Now, pReadTimeout);
		}

        /// <summary>
        /// Blocks execution until any data arrives from the underlying stream.
        /// </summary>
        /// <returns>A <strong>Boolean</strong> value, <strong>True</strong> if data has arrived from the device.  A value
        /// of <strong>False</strong> is returned if the number of seconds specified has elapsed
        /// with no data read.</returns>
        public bool WaitForData(TimeSpan timeout)
		{
			return WaitForData(DateTime.Now, timeout);
		}

        /// <summary>
        /// Blocks execution until any data arrives from the underlying stream.
        /// </summary>
        /// <returns>A <strong>Boolean</strong> value, <strong>True</strong> if data has arrived from the device.  A value
        /// of <strong>False</strong> is returned if the number of seconds specified has elapsed
        /// with no data read.</returns>
        public bool WaitForData(DateTime timeLastDataReceived, TimeSpan timeout)
		{
            try
            {
                // If this stream is closed, throw an exception
                if (pIsStreamClosed)
                    throw new ObjectDisposedException(GetType().ToString(), "A wait for data operation cannot begin because the stream has been closed.");
                // If the stream is closed, return false without waiting
                if (!CanRead)
                    return false;
                // Check the stream length	
				if(pBaseStream is System.Net.Sockets.NetworkStream)
				{
					System.Net.Sockets.NetworkStream NetStream = pBaseStream as System.Net.Sockets.NetworkStream;
					if(NetStream.DataAvailable)
						return true;
					// Calculate the termination time
					DateTime TerminationTime = timeLastDataReceived.Add(timeout);
					// Wait for data
					while (!pIsStreamClosed && CanRead && !NetStream.DataAvailable && (DateTime.Now < TerminationTime))
					{
						Thread.Sleep(100);
						Thread.Sleep(0);
					}
					return NetStream.DataAvailable;					
				}
				else
				{
					if (pBaseStream.Position < pBaseStream.Length)
						return true;                    

					// Calculate the termination time
					DateTime TerminationTime = timeLastDataReceived.Add(timeout);
					// Wait for data
					while (!pIsStreamClosed && pBaseStream.CanRead && (pBaseStream.Position == pBaseStream.Length) && (DateTime.Now < TerminationTime))
					{
						Thread.Sleep(100);
						Thread.Sleep(0);
					}
					return pBaseStream.Position < pBaseStream.Length;
				}
            }
            catch
            {
                throw;
            }
		}

		public void Write(string value)
		{
			byte[] Bytes = ASCIIEncoding.ASCII.GetBytes(value);
			pBaseStream.Write(Bytes, 0, Bytes.Length);
		}

		/// <summary>
		/// Writes a blank line to the stream.
		/// </summary>
		public void WriteLine()
		{
			Write("\r\n");
		}

		/// <summary>
		/// Writes the specified string to the stream, including a carriage return and line feed.
		/// </summary>
		/// <param name="value"></param>
		public void WriteLine(string value)
		{
			// Append a CRLF
			Write(value + "\r\n");
		}

		/// <summary>Writes a single NMEA sentence to the underlying stream.</summary>
		public void WriteSentence(NmeaSentence sentence)
		{
			WriteLine(sentence.Value);
		}

		/// <summary>Reads a single NMEA sentence from the stream.</summary>
		public NmeaSentence ReadSentence()
		{
			return ReadSentence(true);
		}

		/// <summary>Reads a single NMEA sentence from the stream.</summary>
		public NmeaSentence ReadSentence(bool allowExceptions)
		{
			string sentence = null;
			try
			{	
				// Get the next line
				sentence = ReadLine(allowExceptions);
				// Are we at the end of the stream?
				if(sentence == null)
					return null;
                // Remove any preceeding characters
                if (sentence.IndexOf("$") > 0)
                    sentence = sentence.Substring(sentence.IndexOf("$"));
				// Extract the first word from the sentence
				switch (NmeaSentence.GetCommand(sentence))
				{
					case "$GPGGA": 
						return new GpggaSentence(sentence);
					case "$GPGGK": 
						return new GpggkSentence(sentence);
					case "$GPGGQ": 
						return new GpggqSentence(sentence);
					case "$GPGLL": 
						return new GpgllSentence(sentence);
					case "$GPGSA": 
						return new GpgsaSentence(sentence);
					case "$GPGSV": 
						return new GpgsvSentence(sentence);
					case "$GPLLK": 
						return new GpllkSentence(sentence);
					case "$GPLLQ": 
						return new GpllqSentence(sentence);
					case "$GPRMC": 
						return new GprmcSentence(sentence);
					case "$GPVTG": 
						return new GpvtgSentence(sentence);
					case "$HCHDG": 
						return new HchdgSentence(sentence);
					case "$PGRME": 
						return new PgrmeSentence(sentence);
					case "$PGRMZ": 
						return new PgrmzSentence(sentence);
					case "$PMGNST": 
						return new PmgnstSentence(sentence);
#if PocketPC
                    case "$PGLOR":
                        return new PglorSentence(sentence);
#endif
					default:
						return new NmeaSentence(sentence);
				}
				// convert to a proper sentence object
				//return NmeaSentence.GetSentence(sentence);
			}
#if !PocketPC
			catch(ThreadAbortException)
			{
				return null;
			}
#endif
			catch 
			{
				if(allowExceptions)
					throw;
				else
					return null;
			}
//			finally
//			{
//				//Console.WriteLine("Test");
//			}
		}

		/// <summary>Reads the current location from the underlying Stream.</summary>
		public Position ReadPosition()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.Position;
				}
				else if(Sentence is GpggkSentence)
				{
					GpggkSentence Result = (GpggkSentence)Sentence;
					return Result.Position;
				}	
				else if(Sentence is GpggqSentence)
				{
					GpggqSentence Result = (GpggqSentence)Sentence;
					return Result.Position;
				}	
				else if(Sentence is GpgllSentence)
				{
					GpgllSentence Result = (GpgllSentence)Sentence;
					return Result.Position;
				}	
				else if(Sentence is GpllkSentence)
				{
					GpllkSentence Result = (GpllkSentence)Sentence;
					return Result.Position;
				}	
				else if(Sentence is GpllqSentence)
				{
					GpllqSentence Result = (GpllqSentence)Sentence;
					return Result.Position;
				}				
				else if(Sentence is GprmcSentence)
				{
					GprmcSentence Result = (GprmcSentence)Sentence;
					return Result.Position;
				}		
			}
			return GeoFramework.Position.Empty;
		}

		/// <summary>
		/// Reads the current direction of travel as it relates to True North from the
		/// underlying stream.
		/// </summary>
		/// <returns>An <strong>Azimuth</strong> indicating the current direction of travel.</returns>
		public Azimuth ReadTrueBearing()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GprmcSentence)
				{
					GprmcSentence Result = (GprmcSentence)Sentence;
					return Result.Bearing;
				}
				else if(Sentence is GpvtgSentence)
				{
					GpvtgSentence Result = (GpvtgSentence)Sentence;
					return Result.TrueBearing;
				}			
			}
			return Azimuth.Empty;
		}

		/// <returns>An <strong>Azimuth</strong> indicating the current direction of travel.</returns>
		/// <remarks>This method waits for a $HCHDG sentence to be received from the device.  Some care should
		/// be taken when using this method because it is not supported by most GPS devices.</remarks>
		/// <summary>
		/// Reads the current direction of travel as it relates to Magnetic North from the
		/// underlying stream.
		/// </summary>
		public Azimuth ReadMagneticBearing()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is HchdgSentence)
				{
					HchdgSentence Result = (HchdgSentence)Sentence;
					return Result.MagneticBearing;
				}			
			}
			return Azimuth.Empty;
		}

		/// <summary>
		/// Reads the difference between True North and Magnetic North from the underlying
		/// stream.
		/// </summary>
		public Longitude ReadMagneticVariation()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GprmcSentence)
				{
					GprmcSentence Result = (GprmcSentence)Sentence;
					return Result.MagneticVariation;
				}
			}
			return Longitude.Empty;
		}

		/// <summary>Reads whether a satellite fix is active from the underlying stream.</summary>
		public bool ReadIsFixObtained()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpgllSentence)
				{
					GpgllSentence Result = (GpgllSentence)Sentence;
					return Result.IsFixObtained;
				}
				else if(Sentence is GprmcSentence)
				{
					GprmcSentence Result = (GprmcSentence)Sentence;
					return Result.IsFixObtained;
				}
			}
			return false;
		}

		/// <summary>Reads GPS satellite information from the underlying stream.</summary>
		public SatelliteCollection ReadSatellites()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpgsvSentence)
				{
					GpgsvSentence Result = (GpgsvSentence)Sentence;
					return Result.Satellites;
				}
			}
			return new SatelliteCollection(0);
		}

		/// <summary>Reads the current rate of travel from the underlying stream.</summary>
		public Speed ReadSpeed()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GprmcSentence)
				{
					GprmcSentence Result = (GprmcSentence)Sentence;
					return Result.Speed;
				}
				else if(Sentence is GpvtgSentence)
				{
					GpvtgSentence Result = (GpvtgSentence)Sentence;
					return Result.Speed;
				}
			}
			return Speed.Empty;
		}

		/// <summary>Reads the current satellite-derived time from the underlying stream.</summary>
		public DateTime ReadUtcDateTime()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.UtcDateTime;
				}
				else if(Sentence is GpggkSentence)
				{
					GpggkSentence Result = (GpggkSentence)Sentence;
					return Result.UtcDateTime;
				}
				else if(Sentence is GpggqSentence)
				{
					GpggqSentence Result = (GpggqSentence)Sentence;
					return Result.UtcDateTime;
				}
				else if(Sentence is GpgllSentence)
				{
					GpgllSentence Result = (GpgllSentence)Sentence;
					return Result.UtcDateTime;
				}
				else if(Sentence is GpllkSentence)
				{
					GpllkSentence Result = (GpllkSentence)Sentence;
					return Result.UtcDateTime;
				}				
				else if(Sentence is GpllqSentence)
				{
					GpllqSentence Result = (GpllqSentence)Sentence;
					return Result.UtcDateTime;
				}								
				else if(Sentence is GprmcSentence)
				{
					GprmcSentence Result = (GprmcSentence)Sentence;
					return Result.UtcDateTime;
				}
			}
			return DateTime.UtcNow;
		}

		/// <summary>Reads the current distance above sea level from the underlying stream.</summary>
		public Distance ReadAltitude()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.Altitude;
				}
				else if(Sentence is GpggqSentence)
				{
					GpggqSentence Result = (GpggqSentence)Sentence;
					return Result.Altitude;
				}
				else if(Sentence is GpllkSentence)
				{
					GpllkSentence Result = (GpllkSentence)Sentence;
					return Result.Altitude;
				}
				else if(Sentence is GpllqSentence)
				{
					GpllqSentence Result = (GpllqSentence)Sentence;
					return Result.Altitude;
				}				
				else if(Sentence is PgrmzSentence)
				{
					PgrmzSentence Result = (PgrmzSentence)Sentence;
					return Result.Altitude;
				}				
			}
			return Distance.Empty;
		}

		/// <summary>
		/// Reads the current distance above the current ellipsoid from the underlying
		/// stream.
		/// </summary>
		public Distance ReadAltitudeAboveEllipsoid()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.AltitudeAboveEllipsoid;
				}
				else if(Sentence is GpggkSentence)
				{
					GpggkSentence Result = (GpggkSentence)Sentence;
					return Result.AltitudeAboveEllipsoid;
				}
			}
			return Distance.Empty;
		}

		/// <summary>Reads the current amount of precision error from the underlying stream.</summary>
		public Distance ReadHorizontalPositionError()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is PgrmeSentence)
				{
					PgrmeSentence Result = (PgrmeSentence)Sentence;
					return Result.HorizontalPositionError;
				}
			}
			return Distance.Empty;
		}

		/// <summary>
		/// Garmin devices only. Reads the amount of error as it relates to altitude from
		/// the underlying stream.
		/// </summary>
		/// <returns>A <strong>Distance</strong> object describing the amount of error.</returns>
		/// <remarks>Some Garmin devices will return the amount of positional error in the form of a 
		/// measurable distance.  For example, a value of "21 meters" means that the current position may be
		/// inaccurate by 21 meters.</remarks>
		public Distance ReadVerticalPositionError()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is PgrmeSentence)
				{
					PgrmeSentence Result = (PgrmeSentence)Sentence;
					return Result.VerticalPositionError;
				}
			}
			return Distance.Empty;
		}

		/// <summary>Reads the amount of positional error from the underlying stream.</summary>
		public Distance ReadMeanPositionError()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is PgrmeSentence)
				{
					PgrmeSentence Result = (PgrmeSentence)Sentence;
					return Result.MeanPositionError;
				}
			}
			return Distance.Empty;
		}

		/// <summary>
		/// Reads the current level of precision as it relates to latitude and longitude from
		/// the underlying stream.
		/// </summary>
		public DilutionOfPrecision ReadHorizontalDilutionOfPrecision()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.HorizontalDilutionOfPrecision;
				}
				else if(Sentence is GpgsaSentence)
				{
					GpgsaSentence Result = (GpgsaSentence)Sentence;
					return Result.HorizontalDilutionOfPrecision;
				}
			}
			return DilutionOfPrecision.Maximum;
		}

		/// <summary>
		/// Reads the current level of precision as it relates to altitude from the
		/// underlying stream.
		/// </summary>
		public DilutionOfPrecision ReadVerticalDilutionOfPrecision()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpgsaSentence)
				{
					GpgsaSentence Result = (GpgsaSentence)Sentence;
					return Result.VerticalDilutionOfPrecision;
				}
			}
			return DilutionOfPrecision.Maximum;
		}

		/// <summary>
		/// Reads the current level of precision as it relates to latitude, longitude and
		/// altitude from the underlying stream.
		/// </summary>
		public DilutionOfPrecision ReadMeanDilutionOfPrecision()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpgsaSentence)
				{
					GpgsaSentence Result = (GpgsaSentence)Sentence;
					return Result.MeanDilutionOfPrecision;
				}
				else if(Sentence is GpggkSentence)
				{
					GpggkSentence Result = (GpggkSentence)Sentence;
					return Result.MeanDilutionOfPrecision;
				}
				else if(Sentence is GpggqSentence)
				{
					GpggqSentence Result = (GpggqSentence)Sentence;
					return Result.MeanDilutionOfPrecision;
				}
				else if(Sentence is GpllkSentence)
				{
					GpllkSentence Result = (GpllkSentence)Sentence;
					return Result.MeanDilutionOfPrecision;
				}
				else if(Sentence is GpllqSentence)
				{
					GpllqSentence Result = (GpllqSentence)Sentence;
					return Result.MeanDilutionOfPrecision;
				}				
			}
			return DilutionOfPrecision.Maximum;
		}

		/// <summary>Reads differential GPS information from the underlying stream.</summary>
		public DifferentialGps ReadDifferentialGps()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.DifferentialGps;
				}
			}
			return null;
		}

		/// <summary>
		/// Reads satellite information involved in the current fix from the underlying
		/// stream.
		/// </summary>
		public SatelliteCollection ReadFixedSatellites()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpgsaSentence)
				{
					GpgsaSentence Result = (GpgsaSentence)Sentence;
					return Result.FixedSatellites;
				}
			}
			return null;
		}

		/// <summary>
		/// Reads the distance between the Earth's surface and the surface of the ellipsoid
		/// from the underlying stream.
		/// </summary>
		public Distance ReadGeoidalSeparation()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.GeoidalSeparation;
				}
			}
			return Distance.Empty;
		}

		/// <summary>Reads the current fix method from the underlying stream.</summary>
		public FixMethod ReadFixMethod()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpgsaSentence)
				{
					GpgsaSentence Result = (GpgsaSentence)Sentence;
					return Result.FixMethod;
				}
			}
			return FixMethod.NoFix;
		}

		/// <summary>Reads the current fix mode from the underlying stream.</summary>
		public FixMode ReadFixMode()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpgsaSentence)
				{
					GpgsaSentence Result = (GpgsaSentence)Sentence;
					return Result.FixMode;
				}
			}
			return FixMode.Manual;
		}

		/// <summary>Reads the current fix quality from the underlying stream.</summary>
		public FixQuality ReadFixQuality()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.FixQuality;
				}
				else if(Sentence is GpggkSentence)
				{
					GpggkSentence Result = (GpggkSentence)Sentence;
					return Result.FixQuality;
				}				
				else if(Sentence is GpggqSentence)
				{
					GpggqSentence Result = (GpggqSentence)Sentence;
					return Result.FixQuality;
				}
				else if(Sentence is GpllkSentence)
				{
					GpllkSentence Result = (GpllkSentence)Sentence;
					return Result.FixQuality;
				}
				else if(Sentence is GpllqSentence)
				{
					GpllqSentence Result = (GpllqSentence)Sentence;
					return Result.FixQuality;
				}				
			}
			return FixQuality.NoFix;
		}

		/// <summary>
		/// Reads the number of satellites involved in the current fix from the underlying
		/// stream.
		/// </summary>
		public int ReadFixedSatelliteCount()
		{
			// Keep reading until we get a sentence containing position data
			while(CanRead)
			{
				// Read another line
				NmeaSentence Sentence = ReadSentence();
				// What kind of sentence is it?
				if(Sentence is GpggaSentence)
				{
					GpggaSentence Result = (GpggaSentence)Sentence;
					return Result.FixedSatelliteCount;
				}
				else if(Sentence is GpggkSentence)
				{
					GpggkSentence Result = (GpggkSentence)Sentence;
					return Result.FixedSatelliteCount;
				}
				else if(Sentence is GpggqSentence)
				{
					GpggqSentence Result = (GpggqSentence)Sentence;
					return Result.FixedSatelliteCount;
				}
				else if(Sentence is GpllkSentence)
				{
					GpllkSentence Result = (GpllkSentence)Sentence;
					return Result.FixedSatelliteCount;
				}
				else if(Sentence is GpllqSentence)
				{
					GpllqSentence Result = (GpllqSentence)Sentence;
					return Result.FixedSatelliteCount;
				}				
			}
			return 0;
		}

		#endregion
     }
}
