
using System;
using System.Text;
using GeoFramework;

namespace GeoFramework.Gps.Nmea
{
	/// <summary>
	/// Garmin devices only.  Represents information about the current altitude.
	/// </summary>
	/// <remarks>This sentence (along with all sentences starting with "$PGRM") is transmitted
	/// by Garmin devices only.  Altitude is given in either feet or meters.</remarks>
	public class PgrmzSentence : NmeaSentence
	{
		#region Private Variables

		private Distance pAltitude;
		private bool pIsGpsAltitude;

		#endregion

		#region Constructors

		public PgrmzSentence(string sentence) : base(sentence)
		{
			// Is the checksum valid?
			if (!IsValid && !IsPartialSentenceParsingAllowed)
			{
				pAltitude = Distance.Empty;
				pIsGpsAltitude = true;
				// No.  Do no parsing
				return;
			}

			//$PGRMZ,93,f,3*21
			// where:
			//	  93   Altitude in feet
			//			 f		Altitude units
			try
			{
				if(WordCount > 2 && Words[1].Length != 0 && Words[2].Length != 0)
				{
					pAltitude = new Distance(
						double.Parse(Words[1], NmeaCultureInfo),
						Words[2] == "M" ? DistanceUnit.Meters : DistanceUnit.Feet);
				}
				else
				{
					pAltitude = Distance.Empty;
				}
				
				if(WordCount > 3)					
				{
					switch(Words[3])
					{
                        //case "2":
                        //    pIsGpsAltitude = false;
                        //    break;
						case "3":
							pIsGpsAltitude = true;
							break;
					}
				}
				else
				{
					// This defaults to TRUE.  False is the special-case scenario here.
					pIsGpsAltitude = true;
				}
			}
			catch (Exception ex)
			{
				throw new NmeaSentenceException(this, "A $PGRMZ sentence contained unrecognizable data.", ex);
			}
		}

		public PgrmzSentence(byte[] bytes) : this(ASCIIEncoding.ASCII.GetString(bytes, 0, bytes.Length))
		{}

		//[CLSCompliant(false)]
		public PgrmzSentence(Distance altitude)
		{
			// Set the variables
			pAltitude = altitude;
			// Build the sentence from the position information
			StringBuilder Sentence = new StringBuilder();
			Sentence.Append("$PGRMZ,");
			// Convert UTC data time to a string
			Sentence.Append(Altitude.ToFeet().ToString("v.v"));
			Sentence.Append(",f,");
			// Append an undocumented number
			if(IsGpsAltitude)
				Sentence.Append("3");
			else
				Sentence.Append("2");
			Sentence.Append(",");
			// Set the sentence
			SetSentence(Sentence.ToString());
			SetChecksum();
		}

		#endregion

		#region Public Members

		public Distance Altitude
		{
			get
			{
				return pAltitude;
			}
		}

		public bool IsGpsAltitude
		{
			get
			{
				return pIsGpsAltitude;
			}
		}

		/// <summary>
		/// Indicates that this sentence is supported by GPS.NET.
		/// </summary>
		public override bool IsSupported
		{
			get
			{
				return true;
			}
		}

		#endregion

		public static PgrmzSentence Parse(string sentence)
		{
			return new PgrmzSentence(sentence);
		}
	}
}
