using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Globalization;
using System.Threading;
using System.Windows.Forms;
#if !PocketPC || DesignTime || Framework20
using System.ComponentModel;
#endif
#if PocketPC
using GeoFramework.Licensing;
#endif

namespace GeoFramework.Controls
{
	/// <summary>Represents a user control used to display altitude.</summary>
	/// <remarks>
	/// Altimeters are used to visually represent some form of altitude, typically the
	/// user's current altitude above sea level. Altitude is measured using three needles which
	/// represent (from longest to shortest) hundreds, thousands and tens-of-thousands. The
	/// display of the Altimeter is controlled via the <strong>Value</strong> property.
	/// </remarks>
#if !PocketPC || DesignTime
	[ToolboxBitmap("Altimeter.bmp")]
	[DefaultProperty("Value")]
#endif
#if Framework20
#if !PocketPC
    [ToolboxItem(true)]
#endif
    //[DesignTimeVisible(true)]
#endif
	[CLSCompliant(false)]
    public sealed class Altimeter : PolarControl
    {
#if PocketPC
        private Font pAltitudeLabelFont = new Font("Tahoma", 8.0f, FontStyle.Bold);
#if Framework20
		private Pen pMajorTickPen = new Pen(Color.Black, 2.0f);
#else
		private Pen pMajorTickPen = new Pen(Color.Black);
#endif
#else
        private Font pAltitudeLabelFont = new Font("Tahoma", 12.0f, FontStyle.Regular);
		private Thread InterpolationThread;
		//private ManualResetEvent InterpolationThreadWaitHandle = new ManualResetEvent(false);
		private ManualResetEvent AnimationWaitHandle = new ManualResetEvent(false);
		private Interpolator ValueInterpolator = new Interpolator(15, InterpolationMethod.CubicEaseInOut);
        private int InterpolationIndex;
		private bool IsInterpolationActive;
        private SolidBrush pNeedleShadowBrush = new SolidBrush(Color.FromArgb(128, 0, 0, 0));
        private Size pNeedleShadowSize = new Size(5, 5);
		private Pen pMajorTickPen = new Pen(Color.Black, 2.0f);
#endif
		private Pen pCenterPen = new Pen(Color.Gray);
		private Pen pMinorTickPen = new Pen(Color.Black);
		private Pen pTensOfThousandsPen = new Pen(Color.Black);
		private Pen pThousandsPen = new Pen(Color.Black);
		private Pen pHundredsPen = new Pen(Color.Black);
		private SolidBrush pAltitudeLabelBrush = new SolidBrush(Color.Black);
		private Distance pAltitude = Distance.Empty;
        private Font pValueFont = new Font("Tahoma", 9.0f, FontStyle.Regular);
        private SolidBrush pTensOfThousandsBrush = new SolidBrush(Color.Red);
		private SolidBrush pThousandsBrush = new SolidBrush(Color.Red);
		private SolidBrush pHundredsBrush = new SolidBrush(Color.Red);
		private SolidBrush pValueBrush = new SolidBrush(Color.Black);
		private string pValueFormat = "v uu";
        private bool pIsUsingRealTimeData = false;

#if (PocketPC && Framework20)
		private const int MaximumGracefulShutdownTime = 2000;
#elif !PocketPC
		private const int MaximumGracefulShutdownTime = 500;
#endif

		private static PolarCoordinate[] TensOfThousandsNeedle = new PolarCoordinate[]
			{
				new PolarCoordinate(6, 345, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(25, 358, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(30, 0, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(25, 2, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(6, 165, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(6, 195, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(6, 345, Azimuth.North, PolarCoordinateOrientation.Clockwise)
			};
        private static PolarCoordinate[] HundredsNeedle = new PolarCoordinate[]
			{
				new PolarCoordinate(10, 345, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(85, 358, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(90, 0, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(85, 2, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(10, 165, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(10, 195, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(10, 345, Azimuth.North, PolarCoordinateOrientation.Clockwise)
			};
        private static PolarCoordinate[] ThousandsNeedle = new PolarCoordinate[]
			{
				new PolarCoordinate(8, 345, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(40, 350, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(65, 0, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(40, 10, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(8, 165, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(8, 195, Azimuth.North, PolarCoordinateOrientation.Clockwise),
				new PolarCoordinate(8, 345, Azimuth.North, PolarCoordinateOrientation.Clockwise)
			};

		private const double ConversionFactor = 3.6;

		//[CLSCompliant(false)]
        public event DistanceEventHandler ValueChanged;

		/// <summary>Creates a new instance.</summary>
		public Altimeter()
			: base("GeoFramework Multithreaded Altimeter Control (http://www.geoframeworks.com)")
		{
#if !PocketPC
			// Start the interpolation thread
			InterpolationThread = new Thread(new ThreadStart(InterpolationLoop));
			InterpolationThread.IsBackground = true;
			InterpolationThread.Name = "GeoFramework Altimeter Needle Animation Thread (http://www.geoframeworks.com)";
			IsInterpolationActive = true;
			InterpolationThread.Start();

#endif
			Origin = Azimuth.North;
			Orientation = PolarCoordinateOrientation.Clockwise;

//#if PocketPC && !Framework20 && !DesignTime
//            GeoFramework.IO.Devices.CurrentAltitudeChanged += new DistanceEventHandler(Devices_CurrentAltitudeChanged);
//#endif
        }
        

        protected override void Dispose(bool disposing)
        {
            if (pIsUsingRealTimeData)
            {
                try
                {
                    // Only work with events if we're not in design mode
                    if (LicenseManager.CurrentContext.UsageMode == LicenseUsageMode.Runtime
                        && pIsUsingRealTimeData)
                    {
                        // Disconnect from the altitude changed event
                        GeoFramework.IO.Devices.CurrentAltitudeChanged -= new DistanceEventHandler(Devices_CurrentAltitudeChanged);
                    }
                }
                catch
                {
                }
            }
            
#if !PocketPC
            // Get the interpolation thread out of a loop 
            IsInterpolationActive = false;

            if (InterpolationThread != null)
            {

                if (AnimationWaitHandle != null)
                {
                    try
                    {
                        AnimationWaitHandle.Set();
                    }
                    catch
                    {
                    }
                }

                if (!InterpolationThread.Join(MaximumGracefulShutdownTime))
                {
                    try
                    {
                        InterpolationThread.Abort();
                    }
                    catch
                    {
                    }
                }

            }

			if (AnimationWaitHandle != null)
			{
				try
				{
					AnimationWaitHandle.Close();
				}
				catch
				{
				}
				finally
				{
					AnimationWaitHandle = null;
				}
			}
#endif

            if (pCenterPen != null)
            {
                try
                {
                    pCenterPen.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pCenterPen = null;
                }
            }
            if (pMinorTickPen != null)
            {
                try
                {
                    pMinorTickPen.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pMinorTickPen = null;
                }
            }
            if (pMajorTickPen != null)
            {
                try
                {
                    pMajorTickPen.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pMajorTickPen = null;
                }
            }
            if (pAltitudeLabelBrush != null)
            {
                try
                {
                    pAltitudeLabelBrush.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pAltitudeLabelBrush = null;
                }
            }
            if (pAltitudeLabelFont != null)
            {
                try
                {
                    pAltitudeLabelFont.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pAltitudeLabelFont = null;
                }
            }
#if !PocketPC
                if (pNeedleShadowBrush != null)
                {
                    pNeedleShadowBrush.Dispose();
                    pNeedleShadowBrush = null;
                }
#endif
            if (pTensOfThousandsBrush != null)
            {
                try
                {
                    pTensOfThousandsBrush.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pTensOfThousandsBrush = null;
                }
            }
            if (pThousandsBrush != null)
            {
                try
                {
                    pThousandsBrush.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pThousandsBrush = null;
                }
            }
            if (pHundredsBrush != null)
            {
                try
                {
                    pHundredsBrush.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pHundredsBrush = null;
                }
            }
            if (pTensOfThousandsPen != null)
            {
                try
                {
                    pTensOfThousandsPen.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pTensOfThousandsPen = null;
                }
            }
            if (pThousandsPen != null)
            {
                try
                {
                    pThousandsPen.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pThousandsPen = null;
                }
            }
            if (pHundredsPen != null)
            {
                try
                {
                    pHundredsPen.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pHundredsPen = null;
                }
            }
            if (pValueFont != null)
            {
                try
                {
                    pValueFont.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pValueFont = null;
                }
            }
            if (pValueBrush != null)
            {
                try
                {
                    pValueBrush.Dispose();
                }
                catch
                {
                }
                finally
                {
                    pValueBrush = null;
                }
            }

            // Move on down the line
            try
            {
                base.Dispose(disposing);
            }
            catch
            {
            }
        }

      

		/// <summary>
		/// Indicates if the control should automatically display the current
		/// altitude.
		/// </summary>
		/// <value>
		/// A <strong>Boolean</strong>, <strong>True</strong> if the control automatically
		/// displays the current altitude.
		/// </value>
		/// <remarks>
		/// When this property is enabled, the control will examine the
		/// <strong>CurrentAltitude</strong> property of the
		/// <strong>GeoFramework.Gps.Devices</strong> class and update itself when the property
		/// changes. When disabled, the <strong>Value</strong> property must be set manually to
		/// change the control.
		/// </remarks>
#if !PocketPC || DesignTime
		[Category("Behavior")]
		[DefaultValue(typeof(bool), "False")]
		[Description("Controls whether the Value property is set manually, or automatically read from any available GPS device.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif        
		public bool IsUsingRealTimeData
		{
			get
			{
				return pIsUsingRealTimeData;
			}
			set
			{
				// Has anything changed?  If not, exit
				if(pIsUsingRealTimeData == value) 
                    return;

                // Set the new value
				pIsUsingRealTimeData = value;

				// Set the control to the last known altitude (if any)
                if (pIsUsingRealTimeData)
                {
                    try
                    {
                        // Hook into the AltitudeChanged event\
                        // When in design mode, do not hook into events
                        if (LicenseManager.CurrentContext.UsageMode == LicenseUsageMode.Runtime)
                        {
                            // Hook into the altitude changed event
                            GeoFramework.IO.Devices.CurrentAltitudeChanged += new DistanceEventHandler(Devices_CurrentAltitudeChanged);
                        }

                        // Set the value of the control
#if !PocketPC
						if(DesignMode)
							Value = Distance.Empty;
						else
#endif
                            Value = GeoFramework.IO.Devices.CurrentAltitude;

                    }
                    catch
                    {
                        Value = Distance.Empty;
                    }
                }
                else
                {
                    // When in design mode, do not hook into events
                    if (LicenseManager.CurrentContext.UsageMode == LicenseUsageMode.Runtime)
                    {
                        // Hook into the altitude changed event
                        GeoFramework.IO.Devices.CurrentAltitudeChanged -= new DistanceEventHandler(Devices_CurrentAltitudeChanged);
                    }

                    // Set teh control to no value
                    Value = Distance.Empty;
                }

                // Tell the control to repaint
				InvokePaintOffScreen();
			}
		}


		/// <summary>Controls the font used for displaying altitude text.</summary>
		/// <remarks>
		/// This property controls the font used to display altitude text on the control. To
		/// control the font of numbers drawn around the edge of the control, see the
		/// <strong>AltitudeLabelFont</strong> property.
		/// </remarks>
		/// <seealso cref="AltitudeLabelFont">AltitudeLabelFont Property</seealso>
#if !PocketPC || DesignTime
		[Category("Altitude Label")]
		[DefaultValue(typeof(Font), "Tahoma, 9pt")]
		[Description("Controls the font of the label displaying the current altitude.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
		public Font ValueFont
		{
			get
			{
				return pValueFont;
			}
			set
			{
				if(pValueFont.Equals(value)) return;
				pValueFont = value;
				InvokePaintOffScreen();
			}
		}

		/// <summary>Controls the color of altitude text.</summary>
		/// <remarks>
		/// This property controls the color of the altitude text on the control. To change
		/// the color of numbers drawn around the edge of the control, see the
		/// <strong>AltitudeLabelColor</strong> property.
		/// </remarks>
		/// <seealso cref="AltitudeLabelColor">AltitudeLabelColor Property</seealso>
#if !PocketPC || DesignTime
		[Category("Altitude Label")]
		[DefaultValue(typeof(Color), "Black")]
		[Description("Controls the color of the label displaying the current altitude.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
		public Color ValueColor 
		{
			get
			{
				return pValueBrush.Color;
			}
			set
			{
				if(pValueBrush.Color.Equals(value)) return;
				pValueBrush.Color = value;
				InvokePaintOffScreen();
			}
		}

#if !PocketPC
		[Category("Behavior")]
		[DefaultValue(typeof(InterpolationMethod), "CubicEaseInOut")]
		[Description("Controls how the control smoothly transitions from one value to another.")]
		//[CLSCompliant(false)]
		public InterpolationMethod ValueInterpolationMethod
		{
			get
			{
				return ValueInterpolator.InterpolationMethod;
			}
			set
			{
				ValueInterpolator.InterpolationMethod = value;
			}
		}
#endif

		/// <summary>Controls the format of altitude values.</summary>
		/// <value>
		/// A <strong>String</strong> which is compatible with the <strong>ToString</strong>
		/// method of the <strong>Distance</strong> class.
		/// </value>
		/// <remarks>
		/// This property controls how text is output on the control. By default, the format
		/// is "<strong>v uu</strong>" where <strong>v</strong> represents the numeric portion of
		/// the altitude, and <strong>uu</strong> represents units.
		/// </remarks>
#if !PocketPC || DesignTime
		[Category("Altitude Label")]
		[DefaultValue(typeof(string), "v uu")]
		[Description("Controls the format of the label displaying the current altitude.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
		public string ValueFormat
		{
			get
			{
				return pValueFormat;
			}
			set
			{
				if(pValueFormat.Equals(value)) return;
				pValueFormat = value;
				InvokePaintOffScreen();
			}
		}


		/// <summary>Controls the altitude to display on the control.</summary>
		/// <value>A <strong>Distance</strong> structure measuring the altitude to display.</value>
		/// <remarks>
		/// Changing this property causes the needles on the control to move so that they
		/// represent the value.
		/// </remarks>
#if !PocketPC || DesignTime
		[Category("Appearance")]
		[DefaultValue(typeof(Distance), "0 m")]
		[Description("Controls the amount of distance above sea level being displayed in the control.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
		//[CLSCompliant(false)]
		public Distance Value
		{
			get
			{
				return pAltitude;
			}
			set
			{
				if(pAltitude.Equals(value)) 
                    return;
				pAltitude = value;
				// Force either feet or meters
				if(pAltitude.IsMetric)
					pAltitude = pAltitude.ToMeters();
				else
					pAltitude = pAltitude.ToFeet();
#if PocketPC
                InvokePaintOffScreen();
#else
				if(IsDisposed)
					return;

				lock(ValueInterpolator)
				{
					// Are we changing direction?
					if(pAltitude.Value >= ValueInterpolator.Minimum
						&& pAltitude.Value > ValueInterpolator[InterpolationIndex])
					{
						// No.  Just set the new maximum
						ValueInterpolator.Maximum = pAltitude.Value;
					}					
					else if(pAltitude.Value < ValueInterpolator.Minimum)
					{
						// We're changing directions, so stop then accellerate again
						ValueInterpolator.Minimum = ValueInterpolator[InterpolationIndex];
						ValueInterpolator.Maximum = pAltitude.Value;
						InterpolationIndex = 0;
					}
					else if(pAltitude.Value > ValueInterpolator.Minimum
						&& pAltitude.Value < ValueInterpolator[InterpolationIndex])
					{
						// We're changing directions, so stop then accellerate again
						ValueInterpolator.Minimum = ValueInterpolator[InterpolationIndex];
						ValueInterpolator.Maximum = pAltitude.Value;
						InterpolationIndex = 0;
					}
					else if(pAltitude.Value > ValueInterpolator.Maximum)
					{
						// No.  Just set the new maximum
						ValueInterpolator.Maximum = pAltitude.Value;
					}
				}
				// And activate the interpolation thread
//				if((InterpolationThread.ThreadState & ThreadState.Suspended) != 0)
//					InterpolationThread.Resume();
				AnimationWaitHandle.Set();
#endif

                // Signal that the value has changed
                OnValueChanged(new DistanceEventArgs(pAltitude));
			}
		}

        private void OnValueChanged(DistanceEventArgs e)
        {
            if (ValueChanged != null)
#if PocketPC && !Framework20
				ValueChanged(this, e);
#else
                DynamicEvent.BeginInvoke(ValueChanged, this, e);
#endif
        }

		/// <summary>Controls the color of numbers around the edge of the control.</summary>
		/// <remarks>
		/// This property controls the color of numbers around the edge of the control. To
		/// control the color of altitude text, see the <strong>ValueColor</strong>
		/// property.
		/// </remarks>
		/// <seealso cref="ValueColor">ValueColor Property</seealso>
#if !PocketPC || DesignTime
		[Category("Altitude Label")]
		[DefaultValue(typeof(Color), "Black")]
		[Description("Controls the color of altitude digits drawn around the control.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
		public Color AltitudeLabelColor 
		{
			get
			{
				return pAltitudeLabelBrush.Color;
			}
			set
			{
				if(pAltitudeLabelBrush.Color.Equals(value)) return;
				pAltitudeLabelBrush.Color = value;
				InvokePaintOffScreen();
			}
		}

		/// <summary>Controls the font of numbers drawn around the edge of the control.</summary>
		/// <remarks>
		/// This property controls the font used to draw numbers around the edge of the
		/// control. To control the font used to draw altitude text, see the
		/// <strong>ValueFont</strong> property.
		/// </remarks>
		/// <seealso cref="ValueFont">ValueFont Property</seealso>
#if !PocketPC || DesignTime
		[Category("Altitude Label")]
#if PocketPC
		[DefaultValue(typeof(Font), "Tahoma, 8pt, style=Bold")]		
#else
		[DefaultValue(typeof(Font), "Tahoma, 12pt")]
#endif
		[Description("Controls the font of altitude digits drawn around the control.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
		public Font AltitudeLabelFont 
		{
			get
			{
				return pAltitudeLabelFont;
			}
			set
			{
				if(pAltitudeLabelFont.Equals(value)) return;
				pAltitudeLabelFont = value;
				InvokePaintOffScreen();
			}
		}

//#if Framework20 && !PocketPC
//        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
//        public override Azimuth Origin
//        {
//            get
//            {
//                return base.Origin;
//            }
//            set
//            {
//                base.Origin = value;
//            }
//        }
//
//        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
//        public override Angle Rotation
//        {
//            get
//            {
//                return base.Rotation;
//            }
//            set
//            {
//                base.Rotation = value;
//            }
//        }
//#endif
        

		/// <summary>
		/// Controls the color of smaller tick marks drawn around the edge of the
		/// control.
		/// </summary>
		/// <remarks>
		/// Minor tick marks are drawn between major tick marks around the control. These
		/// tick marks can be made invisible by changing the color to
		/// <strong>Transparent</strong>.
		/// </remarks>
#if !PocketPC || DesignTime
		[Category("Tick Marks")]
		[DefaultValue(typeof(Color), "Black")]
		[Description("Controls the color of smaller tick marks drawn around the control.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
        public Color MinorTickColor
		{
			get
			{
				return pMinorTickPen.Color;
			}
			set
			{
				pMinorTickPen.Color = value;
				InvokePaintOffScreen();
			}
		}

		/// <remarks>
		/// There are ten major tick marks in an altimeter, drawn next to numbers on the
		/// control. These tick marks can be made invisible by changing the color to
		/// <strong>Transparent</strong>.
		/// </remarks>
#if !PocketPC || DesignTime
		[Category("Tick Marks")]
		[DefaultValue(typeof(Color), "Black")]
		[Description("Controls the color of larger tick marks drawn around the control.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
		public Color MajorTickColor
		{
			get
			{
				return pMajorTickPen.Color;
			}
			set
			{
				pMajorTickPen.Color = value;
				InvokePaintOffScreen();
			}
		}
		/// <summary>Controls the interior color of the tens-of-thousands needle.</summary>
		/// <remarks>
		/// The tens-of-thousands needle is the smallest needle of the control. The interior
		/// can be made invisible by setting this property to <strong>Transparent</strong>.
		/// </remarks>
#if !PocketPC || DesignTime
		[Category("Needle Colors")]
		[DefaultValue(typeof(Color), "Red")]
		[Description("Controls the color of the interior of the tens-of-thousands needle.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
		public Color TensOfThousandsNeedleFillColor
		{
			get
			{
				return pTensOfThousandsBrush.Color;
			}
			set
			{
				if(pTensOfThousandsBrush.Color.Equals(value)) return;
				pTensOfThousandsBrush.Color = value;
				InvokePaintOffScreen();
			}
		}

#if !PocketPC || DesignTime
		[Category("Needle Colors")]
		[DefaultValue(typeof(Color), "Red")]
		[Description("Controls the color of the interior of the thousands needle.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
        public Color ThousandsNeedleFillColor
		{
			get
			{
				return pThousandsBrush.Color;
			}
			set
			{
				if(pThousandsBrush.Color.Equals(value)) return;
				pThousandsBrush.Color = value;
				InvokePaintOffScreen();
			}
		}

#if !PocketPC || DesignTime
		[Category("Needle Colors")]
		[DefaultValue(typeof(Color), "Red")]
		[Description("Controls the color of the interior of the hundreds needle.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
        public Color HundredsNeedleFillColor
		{
			get
			{
				return pHundredsBrush.Color;
			}
			set
			{
				if(pHundredsBrush.Color.Equals(value)) return;
				pHundredsBrush.Color = value;
				InvokePaintOffScreen();
			}
		}

#if !PocketPC || DesignTime
		[Category("Needle Colors")]
		[DefaultValue(typeof(Color), "Black")]
		[Description("Controls the color of the edge of the tens-of-thousands needle.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
        public Color TensOfThousandsNeedleOutlineColor
		{
			get
			{
				return pTensOfThousandsPen.Color;
			}
			set
			{
				if(pTensOfThousandsPen.Color.Equals(value)) return;
				pTensOfThousandsPen.Color = value;
				InvokePaintOffScreen();
			}
		}

#if !PocketPC || DesignTime
		[Category("Needle Colors")]
		[DefaultValue(typeof(Color), "Black")]
		[Description("Controls the color of the edge of the thousands needle.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
        public Color ThousandsNeedleOutlineColor
		{
			get
			{
				return pThousandsPen.Color;
			}
			set
			{
				if(pThousandsPen.Color.Equals(value)) return;
				pThousandsPen.Color = value;
				InvokePaintOffScreen();
			}
		}

		
#if !PocketPC || DesignTime
		[Category("Needle Colors")]
		[DefaultValue(typeof(Color), "Black")]
		[Description("Controls the color of the edge of the hundreds needle.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
#endif
        public Color HundredsNeedleOutlineColor
		{
			get
			{
				return pHundredsPen.Color;
			}
			set
			{
				if(pHundredsPen.Color.Equals(value)) return;
				pHundredsPen.Color = value;
				InvokePaintOffScreen();
			}
		}

#if !PocketPC
		[Category("Appearance")]
		[DefaultValue(typeof(Color), "128, 0, 0, 0")]
		[Description("Controls the color of the shadow cast by the altitude needles.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
        public Color NeedleShadowColor
		{
			get
			{
				return pNeedleShadowBrush.Color;
			}
			set
			{
				if(pNeedleShadowBrush.Color.Equals(value)) return;
				pNeedleShadowBrush.Color = value;
				InvokePaintOffScreen();
			}
		}
#endif

#if !PocketPC || Framework20
        ///// <summary>
        ///// Occurs when the GDI handle for the control is initialized.
        ///// </summary>
        //protected override void OnHandleCreated(EventArgs e)
        //{
        //    // Subscribe to events
        //    try
        //    {
        //        base.OnHandleCreated(e);
        //    }
        //    catch
        //    {
        //    }
        //}

        ///// <summary>
        ///// Occurs when the GDI handle for the control is destroyed.
        ///// </summary>
        //protected override void OnHandleDestroyed(EventArgs e)
        //{
        //    try
        //    {
        //        // When in design mode, do not hook into events
        //        if (LicenseManager.CurrentContext.UsageMode == LicenseUsageMode.Runtime)
        //        {
        //            // Hook into the altitude changed event
        //            GeoFramework.IO.Devices.CurrentAltitudeChanged -= new DistanceEventHandler(Devices_CurrentAltitudeChanged);
        //        }
        //    }
        //    catch
        //    {
        //    }
        //    finally
        //    {
        //        base.OnHandleDestroyed(e);
        //    }
        //}
#endif

#if !PocketPC
		[Category("Appearance")]
		[DefaultValue(typeof(Size), "5, 5")]
		[Description("Controls the size of the shadow cast by the altitude needles.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[Browsable(true)]
		[EditorBrowsable(EditorBrowsableState.Always)]
        public Size NeedleShadowSize
		{
			get
			{
				return pNeedleShadowSize;
			}
			set
			{
				pNeedleShadowSize = value;
				InvokePaintOffScreen();
			}
		}
#endif
		//[CLSCompliant(false)]
        protected override void OnPaintOffScreen(CancelablePaintEventArgs e)
		{
            PolarGraphics f = base.CreatePolarGraphics(e.Graphics);
			

			// What altitude are we drawing?
#if PocketPC
            Distance AltitudeToRender = pAltitude;
#else
			Distance AltitudeToRender = new Distance(ValueInterpolator[InterpolationIndex], pAltitude.Units);
#endif

				// There are 100 tick marks in 360 degrees.   3.6 per tick mark
				double ConversionFactor = 3.6;
				// Draw tick marks
				if(pMinorTickPen != null)
				{
					for(double alt = 0; alt < 100; alt += 1)
					{
						// Convert the speed to an angle
						Angle angle = new Angle(alt * ConversionFactor);
						// Get the coordinate of the line's start
						PolarCoordinate start = new PolarCoordinate(95, angle, Azimuth.North, PolarCoordinateOrientation.Clockwise);
						PolarCoordinate end = new PolarCoordinate(100, angle, Azimuth.North, PolarCoordinateOrientation.Clockwise);
						// And draw a line
						f.DrawLine(pMinorTickPen, start, end);
					}
				}
				// Draw tick marks
				if(pMajorTickPen != null)
				{
					for(double alt = 0; alt < 100; alt += 10)
					{
						// Convert the speed to an angle
						Angle angle = new Angle(alt * ConversionFactor);
						// Get the coordinate of the line's start
						PolarCoordinate start = new PolarCoordinate(94, angle, Azimuth.North, PolarCoordinateOrientation.Clockwise);
						PolarCoordinate end = new PolarCoordinate(100, angle, Azimuth.North, PolarCoordinateOrientation.Clockwise);
						// And draw a line
						f.DrawLine(pMajorTickPen, start, end);
						// And also a string
						string s = Convert.ToString(alt * 0.1, CultureInfo.CurrentCulture);
						f.DrawCenteredString(s, pAltitudeLabelFont, pAltitudeLabelBrush,
							new PolarCoordinate(85, angle, Azimuth.North, PolarCoordinateOrientation.Clockwise));
					}
				}
				// Calculate all needle values
				double pTensOfThousandsValue = AltitudeToRender.Value / 1000.0;
				double pThousandsValue = AltitudeToRender.Value / 100.0;
				double pValue = AltitudeToRender.Value / 10.0;			

				// Now draw the tens-of-thousands needle
				// Rotate the needle to the right place
				PolarCoordinate[] Needle1 = new PolarCoordinate[TensOfThousandsNeedle.Length];
				for(int index = 0; index < Needle1.Length; index++)
					Needle1[index] = TensOfThousandsNeedle[index].Rotate(pTensOfThousandsValue * ConversionFactor);

				// Now draw the tens-of-thousands needle
				// Rotate the needle to the right place
				PolarCoordinate[] Needle2 = new PolarCoordinate[ThousandsNeedle.Length];
				for(int index = 0; index < Needle2.Length; index++)
				{
					Needle2[index] = ThousandsNeedle[index].Rotate(pThousandsValue * ConversionFactor);
				}

				// Now draw the tens-of-Hundreds needle
				// Rotate the needle to the right place
				PolarCoordinate[] Needle3 = new PolarCoordinate[HundredsNeedle.Length];
				for(int index = 0; index < Needle3.Length; index++)
				{
					Needle3[index] = HundredsNeedle[index].Rotate(pValue * ConversionFactor);
				}

				string AltitudeString = AltitudeToRender.ToString(pValueFormat, CultureInfo.CurrentCulture);
				//SizeF FontSize = f.Graphics.MeasureString(AltitudeString, pValueFont);	
		
#if PocketPC
            f.DrawCenteredString(AltitudeString, pValueFont,
                pValueBrush, new PolarCoordinate(45.0f, 0.0, Azimuth.North, PolarCoordinateOrientation.Clockwise));
#else
				f.DrawRotatedString(AltitudeString, pValueFont,
					pValueBrush, new PolarCoordinate(45.0f, Angle.Empty, Azimuth.North, PolarCoordinateOrientation.Clockwise));
#endif
				// Draw an ellipse at the center
				f.DrawEllipse(pCenterPen, PolarCoordinate.Empty, 10);

#if !PocketPC
				f.Graphics.TranslateTransform(pNeedleShadowSize.Width, pNeedleShadowSize.Height, MatrixOrder.Append);

				f.FillPolygon(pNeedleShadowBrush, Needle1);
				f.FillPolygon(pNeedleShadowBrush, Needle2);
				f.FillPolygon(pNeedleShadowBrush, Needle3);

				f.Graphics.ResetTransform();
#endif

				f.FillPolygon(pTensOfThousandsBrush, Needle1);
				f.DrawPolygon(pTensOfThousandsPen, Needle1);
				f.FillPolygon(pThousandsBrush, Needle2);
				f.DrawPolygon(pThousandsPen, Needle2);
				f.FillPolygon(pHundredsBrush, Needle3);
				f.DrawPolygon(pHundredsPen, Needle3);

				// Draw an ellipse at the center
				f.FillEllipse(pHundredsBrush, PolarCoordinate.Empty, 7);
				f.DrawEllipse(pHundredsPen, PolarCoordinate.Empty, 7);
			
		}

		private void Devices_CurrentAltitudeChanged(object sender, DistanceEventArgs e)
		{
			if(pIsUsingRealTimeData)
				Value = e.Distance;
		}

#if !PocketPC
		protected override void OnTargetFrameRateChanged(int framesPerSecond)
		{
			base.OnTargetFrameRateChanged(framesPerSecond);
			// Recalculate our things
			ValueInterpolator.Count = framesPerSecond;
			// Adjust the index if it's outside of bounds
			if(InterpolationIndex > ValueInterpolator.Count - 1)
				InterpolationIndex = ValueInterpolator.Count - 1;
		}
//
//		protected override void OnHandleCreated(EventArgs e)
//		{
//            try
//            {
//            }
//            catch
//            {
//                throw;
//            }
//            finally
//            {
//                base.OnHandleCreated(e);
//            }
//        }

//        protected override void OnHandleDestroyed(EventArgs e)
//        {
//            try
//            {
//#if !PocketPC
//                // Get the interpolation thread out of a loop 
//                IsInterpolationActive = false;
//
//				if(InterpolationThread != null)
//				{
//					if((InterpolationThread.ThreadState & ThreadState.Suspended) != 0)
//						InterpolationThread.Resume();
//					while(!InterpolationThread.Join(1000))
//					{
//						if((InterpolationThread.ThreadState & ThreadState.AbortRequested) == 0)
//							InterpolationThread.Abort();
//					}
//				}
//#endif
//            }
//            catch
//            {
//                throw;
//            }
//            finally
//            {
//                base.OnHandleDestroyed(e);
//            }
//        }

		private void InterpolationLoop()
		{
			// Flag that we're alive
			//InterpolationThreadWaitHandle.Set();
			// Are we at the end?
			while(IsInterpolationActive)
			{
				try
				{
					// Wait for interpolation to actually be needed
					AnimationWaitHandle.WaitOne();
					//InterpolationThread.Suspend();
					// If we're shutting down, just exit
					if (!IsInterpolationActive)
						break;
					// Keep updating interpolation until we're done
					while (IsInterpolationActive && InterpolationIndex < ValueInterpolator.Count)
					{
						// Render the next value
						InvokePaintOffScreen();
						InterpolationIndex++;
						// Wait for the next frame
						Thread.Sleep(1000 / ValueInterpolator.Count);
					}
					// Reset interpolation
					ValueInterpolator.Minimum = ValueInterpolator.Maximum;
					InterpolationIndex = 0;
					AnimationWaitHandle.Reset();
				}
				catch(ThreadAbortException)
				{
					// Just exit!
					break;
				}
				catch
				{					
				}
			}
			// Flag that we're alive
			//InterpolationThreadWaitHandle.Set();
		}
#endif
	}
}
