using System;
using System.Globalization;
using System.Threading;
using GeoFramework;
using System.Xml;
using System.Xml.Serialization;
using System.Security.Permissions;
#if !PocketPC || DesignTime
using System.Runtime.Serialization;
using System.ComponentModel;
#endif

namespace GeoFramework.Gps
{
	/// <summary>Indicates the government project responsible for a GPS satellite.</summary>
    public enum SatelliteClass
    {
		/// <summary>The satellite could not be identified.</summary>
        Unknown,
		/// <summary>The satellite is part of the NAVSTAR project.</summary>
        Navstar,
		/// <summary>The satellite is part of the Wide Area Augmentation System (WAAS).</summary>
        Waas,
		/// <summary>
		/// The satellite is part of the European Geostationary Navigation Overlay Service
		/// (EGNOS).
		/// </summary>
        Egnos,
		/// <summary>
		/// The satellite is pard of Japan's MTSAT Satellite Augmentation System
		/// (MSAS).
		/// </summary>
        Msas
    }

	/// <summary>Indicates the launch block of a group of NAVSTAR GPS satellites.</summary>
    public enum SatelliteBlock
    {
		/// <summary>The block is unknown.</summary>
        Unknown,
		/// <summary>Represents Block I.</summary>
        I,
		/// <summary>Represents Block II.</summary>
        II,
		/// <summary>Represents Block IIA</summary>
        IIA,
		/// <summary>Represents Block IIR.</summary>
        IIR
    }

	/// <summary>Indicates the main atomic clock in service aboard a GPS satellite.</summary>
    public enum SatelliteAtomicClockType
    {
		/// <summary>The clock type is unknown.</summary>
        Unknown,
		/// <summary>The satellite's Cesium clock is currently in service.</summary>
        Cesium,
		/// <summary>The satellite's Rubidium clock is currently in service.</summary>
        Rubidium
    }

	/// <summary>Represents information about a GPS satellite in orbit above Earth.</summary>
	/// <remarks>
	/// 	<para>GPS devices are able to isolate information about GPS satellites in orbit. Each
	///  satellite's <see cref="Satellite.PseudorandomNumber">unique identifier</see>,
	///  <see cref="Satellite.SignalToNoiseRatio">radio signal strength</see>,
	///  <see cref="Satellite.Azimuth">azimuth</see> and
	///  <see cref="Satellite.Elevation">elevation</see> are available once its
	///  radio signal is detected.</para>
	/// 	<para>Properties in this class are updated automatically as new information is
	///  received from the GPS device.</para>
	/// 	<para><img src="Satellite.jpg"/></para>
	/// </remarks>
	/// <seealso cref="SatelliteCollection">SatelliteCollection Class</seealso>
#if !PocketPC || DesignTime
    [Serializable()]
    [ImmutableObject(true)]
    [TypeConverter(typeof(ExpandableObjectConverter))]
#endif
    public sealed class Satellite : IFormattable, ICloneable
#if !PocketPC || (PocketPC && Framework20)
		, IXmlSerializable
#endif
#if !PocketPC || DesignTime
		, ISerializable
#endif
	{
		#region Private Variables

		private Azimuth pAzimuth = Azimuth.Empty;
		private Elevation pElevation = Elevation.Empty;
		private int pPseudorandomNumber;
		private SignalToNoiseRatio pSignalToNoiseRatio = SignalToNoiseRatio.Empty;
		private bool pIsFixed;
		private DateTime pLastSignalReceived = DateTime.Now;

		#endregion

		#region  Constructors 

        /// <summary>
        /// Creates a new instance with empty values.
        /// </summary>
        public Satellite()
        {
        }
        
        /// <summary>
        /// Creates a new instance using the specified unique identifier.
        /// </summary>
        /// <param name="pseudorandomNumber"></param>
        public Satellite(int pseudorandomNumber)
		{
			pPseudorandomNumber = pseudorandomNumber;
		}

        /// <summary>
        /// Creates a new instance using the specified unique identifier, location around the horizon, and elevation up from the horizon.
        /// </summary>
        /// <param name="pseudorandomNumber"></param>
        /// <param name="azimuth"></param>
        /// <param name="elevation"></param>
		//[CLSCompliant(false)]
        public Satellite(int pseudorandomNumber, Azimuth azimuth, Elevation elevation) 
            : this(pseudorandomNumber)
        {
			pAzimuth = azimuth;
			pElevation = elevation;
		}

		//[CLSCompliant(false)]
        public Satellite(int pseudorandomNumber, Azimuth azimuth, Elevation elevation, SignalToNoiseRatio signalToNoiseRatio) 
            : this(pseudorandomNumber, azimuth, elevation)
        {
			pSignalToNoiseRatio = signalToNoiseRatio;
		}

		//[CLSCompliant(false)]
        public Satellite(int pseudorandomNumber, Azimuth azimuth, Elevation elevation, SignalToNoiseRatio signalToNoiseRatio, bool isFixed) 
            : this(pseudorandomNumber, azimuth, elevation, signalToNoiseRatio)
        {
			pIsFixed = isFixed;
		}

#if !PocketPC || DesignTime
        /// <summary>
        /// Creates a new instance by de-serializing previously-serialized information.
        /// </summary>
        /// <param name="info">A <strong>SerializationInfo</strong> object containing serialized data.</param>
        /// <param name="context">A <strong>StreamingContext</strong> object describing the context of serialization.</param>
        /// <remarks>This constructor will de-serialize a previously-serialized object.</remarks>
        private Satellite(SerializationInfo info, StreamingContext context)
        {
            pPseudorandomNumber = info.GetInt32("PseudorandomNumber");
            pAzimuth = new Azimuth(info.GetDouble("Azimuth"));
            pElevation = new Elevation(info.GetDouble("Elevation"));
            pSignalToNoiseRatio = new SignalToNoiseRatio(info.GetInt32("SignalToNoiseRatio"));
            pIsFixed = info.GetBoolean("IsFixed");
        }
#endif

		#endregion

        //public Satellite(int pseudorandomNumber,
        //                string name,
        //                SatelliteClass satelliteClass, 
        //                SatelliteAtomicClockType clockType,
        //                SatelliteBlock block,
        //                DateTime dateLaunched,
        //                DateTime dateCommissioned,
        //                DateTime dateDecommissioned)
        //{
        //    pPseudorandomNumber = pseudorandomNumber;
            
        //}
                        

        //#region Static Members

        //public static Satellite Navstar22 = new Satellite(22, 

        

        //#endregion

		#region  Public Members 

		/// <summary>
		/// Returns the horizontal direction towards the satellite from the current
		/// location.
		/// </summary>
		//[CLSCompliant(false)]
		public Azimuth Azimuth
		{
			get
			{
				return pAzimuth;
			}
		}

		/// <summary>
		/// Returns the vertical direction towards the satellite from the current
		/// location.
		/// </summary>
		//[CLSCompliant(false)]
		public Elevation Elevation
		{
			get
			{
				return pElevation;
			}
		}

		/// <summary>Returns the date and time that the satellite's signal was detected.</summary>
		public DateTime LastSignalReceived
		{
			get
			{
				return pLastSignalReceived;
			}
		}

		/// <summary>Compares the current instance with the specified object by value.</summary>
		public bool Equals(Satellite value)
		{
            return ValueEquals(value);
		}

        /// <summary>
        /// Returns whether the current instance has the same values as the specified Satellite object.
        /// </summary>
        /// <param name="value">A <strong>Satellite</strong> object to examine.</param>
        /// <returns>A <strong>Boolean</strong>, <strong>True</strong> if the <strong>IsFixed</strong>,
        /// <strong>PseudorandomNumber</strong>, <strong>Azimuth</strong>, <strong>Elevation</strong>,
        /// and <strong>SignalToNoiseRatio</strong> property values match the specified object.</returns>
        public bool ValueEquals(Satellite value)
        {
            if (value == null)
                return false;

            return pIsFixed == value.IsFixed
                && pPseudorandomNumber == value.PseudorandomNumber
                && pAzimuth.Equals(value.Azimuth)
                && pElevation.Equals(value.Elevation)
                && pSignalToNoiseRatio.Equals(value.SignalToNoiseRatio);
        }

#if !PocketPC || Framework20
        /// <summary>
        /// Returns whether the current object is the same instance as the specified object.
        /// </summary>
        /// <param name="value">A <strong>Satellite</strong> to compare with.</param>
        /// <returns><A <strong>Boolean</strong>, <strong>True></strong> if the current instance is the same
        /// as the specified object.</returns>
        public bool ReferenceEquals(Satellite value)
        {
            return Object.ReferenceEquals(this, value);
        }
#endif

		/// <summary>Returns a unique code for the current instance used for hash tabled.</summary>
		public override int GetHashCode()
		{
			return PseudorandomNumber.GetHashCode();
		}

		/// <summary>
		/// Returns whether the satellite's signal is being used to calculate the current
		/// location.
		/// </summary>
		public bool IsFixed
		{
			get
			{
				return pIsFixed;
			}
		}

        /// <summary>
        /// Returns a copy of the current instance with IsFixed set to True.
        /// </summary>
        /// <returns></returns>
        public Satellite ToFixedSatellite()
        {
            if (pIsFixed)
                return this;
            else
                return new Satellite(pPseudorandomNumber, pAzimuth,
                pElevation, pSignalToNoiseRatio, true);
        }

        /// <summary>
        /// Returns a copy of the current instance with IsFixed set to False.
        /// </summary>
        /// <returns></returns>
        public Satellite ToUnfixedSatellite()
        {
            if (!pIsFixed)
                return this;
            else
                return new Satellite(pPseudorandomNumber,pAzimuth,
                    pElevation, pSignalToNoiseRatio, false);
        }

        //internal static Satellite FromPseudorandomNumber(int pseudorandomNumber)
        //{
        //    // Try to find a known satellite
        //    foreach(Satellite satellite in GeoFramework.IO.Devices.CurrentSatellites)
        //    {
        //        if(satellite.PseudorandomNumber == pseudorandomNumber)
        //            return satellite;
        //    }
        //    // It was not found, so create a new one
        //    Satellite NewSatellite = new Satellite(pseudorandomNumber);
        //    GeoFramework.IO.Devices.CurrentSatellites.Add(NewSatellite);
        //    // And return it
        //    return NewSatellite;
        //}

		/// <summary>Returns whether the satellite's signal is currently being detected.</summary>
		public bool IsTracked
		{
			get
			{
				return !pSignalToNoiseRatio.IsEmpty;
			}
		}

		/// <summary>Returns the amount of time elapsed since the signal was last received.</summary>
		public TimeSpan SignalAge
		{
			get
			{
				return DateTime.Now.Subtract(pLastSignalReceived);
			}
		}

		/// <summary>Returns whether the satellite's signal has recently been received.</summary>
		public bool IsActive
		{
			get
			{
#if PocketPC
				return pIsFixed || SignalAge.TotalSeconds < 20.0;
#else
				return pIsFixed || SignalAge.TotalSeconds < 10.0;
#endif
            }
		}

        /// <summary>
        /// Indicates whether the satellite is providing additional corrective
        /// signals to increase precision.
        /// </summary>
        /// <remarks>This property will return a value of <strong>True</strong>
        /// if the GPS satellite has been identified as a WAAS, EGNOS or MSAS satellite.
        /// These satellites are geostationary (unlike typical NAVSTAR GPS satellites)
        /// and re-broadcast correction signals from ground stations.  When this property
        /// is true, the GPS device has improved precision.</remarks>
        public bool IsDifferentialGpsSatellite
        {
            get
            {
                return this.Class == SatelliteClass.Waas
                        || this.Class == SatelliteClass.Egnos
                        || this.Class == SatelliteClass.Msas; 
            }
        }

		/// <summary>Returns the unique identifier of the satellite.</summary>
		public int PseudorandomNumber
		{
			get
			{
				return pPseudorandomNumber;
			}            
        }

		/// <summary>
		/// Returns the strength of the satellite's radio signal as it is being
		/// received.
		/// </summary>
		public SignalToNoiseRatio SignalToNoiseRatio
		{
			get
			{
				return pSignalToNoiseRatio;
			}
//			set
//			{
//				// If the signal is non-zero, set the last known date/time
//				if(!value.IsEmpty)
//					pLastSignalReceived = DateTime.Now;
//				// Has the value actually changed?
//				if(pSignalToNoiseRatio.Equals(value)) return;
//				pSignalToNoiseRatio = value;
//				OnSignalToNoiseRatioChanged(pSignalToNoiseRatio);
//			}
		}

//		/// <summary>
//		/// Occurs when the horizontal direction to the satellite from the current location
//		/// has changed.
//		/// </summary>
//		protected virtual void OnAzimuthChanged(Azimuth azimuth)
//		{
//            if (AzimuthChanged != null)
//#if PocketPC && !Framework20
//				AzimuthChanged(this, new AzimuthEventArgs(azimuth));
//#else
//				DynamicEvent.Invoke(AzimuthChanged, this, new AzimuthEventArgs(azimuth));
//#endif
//		}
//
//		/// <summary>
//		/// Occurs when the vertical direction to the satellite from the current location has
//		/// changed.
//		/// </summary>
//		protected virtual void OnElevationChanged(Elevation elevation)
//		{
//            if (ElevationChanged != null)
//#if PocketPC && !Framework20
//				ElevationChanged(this, new ElevationEventArgs(elevation));
//#else
//				DynamicEvent.Invoke(ElevationChanged, this, new ElevationEventArgs(elevation));
//#endif
//		}
//
//		/// <summary>Occurs when a satellite's radio signal becomes more or less obscured.</summary>
//		protected virtual void OnSignalToNoiseRatioChanged(SignalToNoiseRatio signalToNoiseRatio)
//		{
//            if (SignalStrengthChanged != null)
//#if PocketPC && !Framework20
//				SignalStrengthChanged(this, new SignalToNoiseRatioEventArgs(pSignalToNoiseRatio));
//#else
//				DynamicEvent.Invoke(SignalStrengthChanged, this, new SignalToNoiseRatioEventArgs(pSignalToNoiseRatio));
//#endif
//		}
//
//		/// <summary>
//		/// Occurs when the satellite's signal becomes involved in calculating the current
//		/// location.
//		/// </summary>
//		protected virtual void OnFixed()
//		{
//            if (Fixed != null)
//#if PocketPC && !Framework20
//				Fixed(this, EventArgs.Empty);
//#else
//				DynamicEvent.Invoke(Fixed, this, EventArgs.Empty);
//#endif
//		}
//
//		/// <summary>
//		/// Occurs when the satellite's signal is too obscured to be involved in calculating
//		/// the current location.
//		/// </summary>
//		protected virtual void OnFixLost()
//		{
//            if (FixLost != null)
//#if PocketPC && !Framework20
//				FixLost(this, EventArgs.Empty);
//#else
//				DynamicEvent.Invoke(FixLost, this, EventArgs.Empty);
//#endif
//		}

		#endregion

		/// <summary>Returns the government project responsible for launching the satellite.</summary>
        public SatelliteClass Class
        {
            get
            {
                switch(pPseudorandomNumber)
                {
                    case 122:
                    case 134:
                    case 125:
                    case 135:
                    case 138:
                    case 35:
                    case 47:
                    case 48:
                    case 51:
                        return SatelliteClass.Waas;
                    case 120:
                    case 124:
                    case 126:
                    case 131:
                    case 33:
                    case 37:
                    case 39:
                    case 44:
                        return SatelliteClass.Egnos;
                    case 129:
                    case 137:
                    case 42:
                    case 50:
                        return SatelliteClass.Msas;
                    case 1:
                    case 2:
                    case 3:
                    case 4:
                    case 5:
                    case 6:
                    case 7:
                    case 8:
                    case 9:
                    case 10:
                    case 11:
                    case 13:
                    case 14:
                    case 15:
                    case 16:
                    case 18:
                    case 19:
                    case 20:
                    case 21:
                    case 22:
                    case 23:
                    case 24:
                    case 25:
                    case 26:
                    case 27:
                    case 28:
                    case 29:
                    case 30:
                    case 31:
                        return SatelliteClass.Navstar;
                    default:
                        return SatelliteClass.Unknown;
                }
            }
        }

		/// <summary>Returns the atomic clock currently in service.</summary>
        public SatelliteAtomicClockType AtomicClockType
        {
            get
            {
                switch(pPseudorandomNumber)
                {
                    case 1:
                    case 3:
                    case 5:
                    case 8:
                    case 9:
                    case 10:
                    case 15:
                    case 24:
                    case 25:
                        return SatelliteAtomicClockType.Cesium;
                    case 2:
                    case 4:
                    case 6:
                    case 7:
                    case 11:
                    case 13:
                    case 14:
                    case 16:
                    case 18:
                    case 19:
                    case 20:
                    case 21:
                    case 22:
                    case 23:
                    case 26:
                    case 27:
                    case 28:
                    case 29:
                    case 30:
                    case 31:
                        return SatelliteAtomicClockType.Rubidium;
                    default:
                        return SatelliteAtomicClockType.Unknown;
                }
            }
        }

		/// <summary>Returns the launch block of the satellite.</summary>
        public SatelliteBlock Block
        {
            get
            {
                switch(pPseudorandomNumber)
                {
                    case 3:
                    case 4:
                    case 7:
                    case 6:
                    case 8:
                    case 5:
                    case 9:
                    case 10:
                    case 11:
                    case 12:
                    case 13:
                        return SatelliteBlock.I;
                    case 2:
                    case 14:
                    case 15:
                    case 16:
                    case 17:
                    case 18:
                    case 19:
                    case 20:
                    case 21:
                        return SatelliteBlock.II;
                    case 22:
                    case 23:
                    case 24:
                    case 25:
                    case 26:
                    case 27:
                    case 28:
                    case 29:
                    case 30:
                    case 31:
                    case 32:
                        return SatelliteBlock.IIA;
                    default:
                        return SatelliteBlock.Unknown;
                }
            }
        }

		/// <summary>Returns the date the satellite was placed into orbit.</summary>
        public DateTime DateLaunched
        {
            get
            {
                switch (pPseudorandomNumber)
                {
                    case 1:
                        return new DateTime(1992, 11, 22);
                    case 2:
                        return new DateTime(1989, 6, 10);
                    case 3:
                        return new DateTime(1985, 10, 9);
                    case 4:
                        return new DateTime(1993, 10, 26);
                    case 5:
                        return new DateTime(1980, 2, 9);
                    case 6:
                        return new DateTime(1978, 10, 6);
                    case 7:
                        return new DateTime(1978, 5, 13);
                    case 8:
                        return new DateTime(1978, 12, 11);
                    case 9:
                        return new DateTime(1980, 4, 26);
                    case 11:
                        return new DateTime(1983, 7, 14);
                    case 12:
                        return new DateTime(1984, 9, 8);
                    case 13:
                        return new DateTime(1984, 6, 13);
                    case 14:
                        return new DateTime(1989, 2, 14);
                    case 15:
                        return new DateTime(1990, 10, 10);
                    case 16:
                        return new DateTime(1989, 8, 17);
                    case 17:
                        return new DateTime(1989, 12, 11);
                    case 18:
                        return new DateTime(1990, 1, 24);
                    case 19:
                        return new DateTime(1989, 10, 21);
                    case 20:
                        return new DateTime(1990, 3, 25);
                    case 21:
                        return new DateTime(1990, 8, 2);
                    case 22:
                        return new DateTime(1993, 2, 3);
                    case 23:
                        return new DateTime(1990, 11, 26);
                    case 24:
                        return new DateTime(1991, 7, 3);
                    case 25:
                        return new DateTime(1992, 2, 23);
                    case 26:
                        return new DateTime(1992, 7, 7);
                    case 27:
                        return new DateTime(1992, 9, 9);
                    case 28:
                        return new DateTime(1992, 4, 10);
                    case 29:
                        return new DateTime(1992, 12, 18);

                    case 31:
                        return new DateTime(1993, 3, 30);

                    case 37:
                        return new DateTime(1993, 5, 13);
                    default:
                        return new DateTime(0);
                }
            }
        }

		/// <summary>Returns the date the satellite was placed into service.</summary>
        public DateTime DateCommissioned
        {
            get
            {
                switch (pPseudorandomNumber)
                {
                    case 1:
                        return new DateTime(1992, 12, 11);
                    case 2:
                        return new DateTime(1989, 6, 10);
                    case 3:
                        return new DateTime(1985, 10, 30);
                    case 4:
                        return new DateTime(1978, 3, 29);
                    case 5:
                        return new DateTime(1980, 2, 27);
                    case 6:
                        return new DateTime(1978, 11, 9);
                    case 7:
                        return new DateTime(1978, 7, 14);
                    case 8:
                        return new DateTime(1979, 1, 8);
                    case 9:
                        return new DateTime(1980, 5, 16);
                    case 11:
                        return new DateTime(1983, 8, 10);
                    case 12:
                        return new DateTime(1984, 10, 3);
                    case 13:
                        return new DateTime(1984, 7, 19);
                    case 14:
                        return new DateTime(1989, 4, 14);
                    case 15:
                        return new DateTime(1990, 10, 20);
                    case 16:
                        return new DateTime(1989, 8, 17);
                    case 17:
                        return new DateTime(1990, 1, 11);
                    case 18:
                        return new DateTime(1990, 2, 14);
                    case 19:
                        return new DateTime(1989, 10, 21);
                    case 20:
                        return new DateTime(1990, 4, 19);
                    case 21:
                        return new DateTime(1990, 8, 31);
                    case 22:
                        return new DateTime(1993, 2, 3);
                    case 23:
                        return new DateTime(1990, 12, 10);
                    case 24:
                        return new DateTime(1991, 8, 30);
                    case 25:
                        return new DateTime(1992, 3, 24);
                    case 26:
                        return new DateTime(1992, 7, 23);
                    case 27:
                        return new DateTime(1992, 9, 30);
                    case 28:
                        return new DateTime(1992, 4, 10);
                    case 29:
                        return new DateTime(1995, 1, 5);

                    case 31:
                        return new DateTime(1993, 4, 13);

                    case 37:
                        return new DateTime(1993, 6, 12);
                    default:
                        return new DateTime(0);
                }
            }
        }

		/// <summary>Returns the date the satellite was removed from service.</summary>
        public DateTime DateDecommissioned
        {
            get
            {
                switch (pPseudorandomNumber)
                {
                    case 2:
                        return new DateTime(2004, 5, 12);
                    case 14:
                        return new DateTime(2000, 4, 14);
                    default:
                        return new DateTime(0);
                }
            }
        }

		/// <summary>Returns the friendly name of the satellite.</summary>
        public string Name
        {
            get
            {
                switch (pPseudorandomNumber)
                {
                    case 1:
                        return "NAVSTAR 32";
                    case 2:
                        return "NAVSTAR 13";
                    case 3:
                        return "NAVSTAR 33";
                    case 4:
                        return "NAVSTAR 34";
                    case 5:
                        return "NAVSTAR 35";
                    case 6:
                        return "NAVSTAR 36";
                    case 7:
                        return "NAVSTAR 37";
                    case 8:
                        return "NAVSTAR 4";
                    case 9:
                        return "NAVSTAR 39";
                    case 10:
                        return "GPS 2A-17";
                    case 11:
                        return "NAVSTAR 8";
                    case 12:
                        return "NAVSTAR 10";
                    case 13:
                        return "NAVSTAR 9";
                    case 14:
                    case 15:
                    case 16:
                    case 17:
                    case 18:
                    case 19:
                    case 20:
                    case 21:
                    case 22:
                    case 23:
                    case 24:
                    case 25:
                    case 26:
                    case 27:
                    case 28:
                    case 29:
                    case 31:
                    case 32:
                        return "NAVSTAR " + pPseudorandomNumber;
                    case 30:
                        return "GPS 2-27";
                    case 122:
                    case 35:
                        return "INMARSAT 3 F4 (AOR-W) (WAAS East)";
                    case 134:
                    case 47:
                        return "INMARSAT 3 F3 (POR) (WAAS West)";
                    case 120:
                    case 33:
                        return "INMARSAT 3 F2 (AOR-E) (EGNOS Western Africa)";
                    case 131:
                    case 44:
                        return "INMARSAT 3 F3 (IOR) (EGNOS Indian Ocean)";
                    case 124:
                    case 37:
                        return "Artmeis (Africa Congo)";
                    case 126:
                    case 39:
                        return "INMARSAT 3 F5 (IOR-W) (Africa Congo)";
                    case 129:
                    case 42:
                        return "MTSAT-1R (MSAS)";
                    case 137:
                    case 50:
                        return "MTSAT-2 (MSAS)";
                    case 138:
                    case 51:
                        return "Anik F1R (WAAS)";
                    case 135:
                    case 48:
                        return "PanAm Galaxy XV (WAAS)";
                    default:
                        return "Invalid";

                }
            }
        }

		#region IFormattable Members

		public override string ToString()
		{
			return ToString("g", CultureInfo.CurrentCulture);
		}

		public string ToString(string format)
		{
			return ToString(format, CultureInfo.CurrentCulture);
		}

		public string ToString(string format, IFormatProvider formatProvider)
		{
			CultureInfo culture = (CultureInfo)formatProvider;
            return Name + " (" + PseudorandomNumber.ToString(format, culture) + "): "
                        + Azimuth.ToString(format, culture) + culture.TextInfo.ListSeparator + " "
                        + Elevation.ToString(format, culture) + culture.TextInfo.ListSeparator + " "
                        + SignalToNoiseRatio.ToString(format, culture);
		}

		#endregion

		#region ICloneable Members

		/// <summary>Returns a copy of the current instance.</summary>
		public Satellite Clone()
		{
			return new Satellite(pPseudorandomNumber, pAzimuth, pElevation, pSignalToNoiseRatio, pIsFixed);
		}

		object ICloneable.Clone()
		{
			return this.Clone();
		}

		#endregion

        #region IXmlSerializable Members

#if !PocketPC || (PocketPC && Framework20)

        System.Xml.Schema.XmlSchema IXmlSerializable.GetSchema()
        {
            return null;
        }

        void IXmlSerializable.ReadXml(XmlReader reader)
        {
            // Read the PRN
            pPseudorandomNumber = int.Parse(reader.GetAttribute("PseudorandomNumber"));

            // Read the start element
            reader.ReadStartElement();

            // Get the depth
            int depth = reader.Depth;

            // Now read in the remaining values
            while (reader.Depth >= depth)
            {
                // Keep reading until we reach an element
                switch (reader.NodeType)
                {                    
                    case XmlNodeType.Element:
                        switch (reader.LocalName)
                        {

                            case "Azimuth":
#if Framework20
                                pAzimuth = new Azimuth(reader.ReadElementContentAsDouble());
#else
								pAzimuth = new Azimuth(reader.ReadElementString());
#endif
                                break;
                            case "Elevation":
#if Framework20
                                pElevation = new Elevation(reader.ReadElementContentAsDouble());
#else
								pElevation = new Elevation(reader.ReadElementString());
#endif
								break;
                            case "SignalToNoiseRatio":
#if Framework20
                                pSignalToNoiseRatio = new SignalToNoiseRatio(reader.ReadElementContentAsInt());
#else
								pSignalToNoiseRatio = new SignalToNoiseRatio(reader.ReadElementString());
#endif
								break;
                            case "IsFixed":
#if Framework20
                                pIsFixed = bool.Parse(reader.ReadElementContentAsString());
#else
								pIsFixed = bool.Parse(reader.ReadElementString());
#endif
								break;
                            default:
                                // Read the next item
                                reader.Read();
                                break;
                        }
                        break;
                    default:
                        // Read the next item
                        reader.Read();
                        break;
                }
            }

            reader.ReadEndElement();
        }

        void IXmlSerializable.WriteXml(XmlWriter writer)
        {
            writer.WriteStartElement("Satellite");

            writer.WriteAttributeString("PseudorandomNumber", pPseudorandomNumber.ToString());

            writer.WriteElementString("Azimuth", pAzimuth.DecimalDegrees.ToString("G15"));
            writer.WriteElementString("Elevation", pElevation.DecimalDegrees.ToString("G15"));
            writer.WriteElementString("SignalToNoiseRatio", pSignalToNoiseRatio.Value.ToString(CultureInfo.InvariantCulture));
            writer.WriteElementString("IsFixed", pIsFixed.ToString(CultureInfo.InvariantCulture));

            // These other values are miscellaneous and are not needed to de-serialize the satellite
            writer.WriteElementString("Block", Block.ToString());
            writer.WriteElementString("Class", Class.ToString());
            writer.WriteElementString("AtomicClockType", AtomicClockType.ToString());
            writer.WriteElementString("DateCommissioned", DateCommissioned.ToString(CultureInfo.InvariantCulture));
            writer.WriteElementString("DateDecommissioned", DateDecommissioned.ToString(CultureInfo.InvariantCulture));
            writer.WriteElementString("DateLaunched", DateLaunched.ToString(CultureInfo.InvariantCulture));
            writer.WriteElementString("Name", Name.ToString(CultureInfo.InvariantCulture));
            writer.WriteElementString("IsActive", IsActive.ToString(CultureInfo.InvariantCulture));
            writer.WriteElementString("IsTracked", IsTracked.ToString(CultureInfo.InvariantCulture));
            writer.WriteElementString("IsDifferentialGpsSatellite", IsDifferentialGpsSatellite.ToString(CultureInfo.InvariantCulture));
            writer.WriteElementString("Block", Block.ToString());

            writer.WriteEndElement();
        }
#endif

		#endregion


#if !PocketPC || DesignTime
        #region ISerializable Members

        [SecurityPermission(SecurityAction.LinkDemand, SerializationFormatter=true)]
        void ISerializable.GetObjectData(SerializationInfo info, StreamingContext context)
        {
            info.AddValue("PseudorandomNumber", pPseudorandomNumber);
            info.AddValue("Azimuth", pAzimuth.DecimalDegrees);
            info.AddValue("Elevation", pElevation.DecimalDegrees);
            info.AddValue("SignalToNoiseRatio", pSignalToNoiseRatio.Value);
            info.AddValue("IsFixed", pIsFixed);
        }

        #endregion
#endif
    }

	/// <summary>
	/// Represents information about a satellite when a satellite-related event is raised.
	/// </summary>
	/// <remarks>This object is used primarily by the <see cref="Satellite">Satellite</see>
	/// class to provide notification when information such as <see cref="Satellite.Azimuth">azimuth</see>,
	/// <see cref="Satellite.Elevation">elevation</see>, or <see cref="Satellite.SignalToNoiseRatio">radio signal strength</see> 
	/// has changed.</remarks>
	/// <example>This example demonstrates how to use this class when raising an event.
	/// <code lang="VB">
	/// ' Start a new receiver
	/// Dim MyReceiver As New Receiver()
	/// ' Declare a new event
	/// Dim MySatelliteEvent As EventHandler
	/// ' Get a handle on the first satellite in the receiver's collection
	/// Dim MySatellite As Satellite = MyReceiver.Satellites(0)
	/// 
	/// Sub Main()
	///   ' Raise our custom event
	///   RaiseEvent MySatelliteEvent(Me, New SatelliteEventArgs(MySatellite))
	/// End Sub
	/// </code>
	/// <code lang="C#">
	/// // Start a new receiver
	/// Receiver MyReceiver = new Receiver();
	/// // Declare a new event
	/// EventHandler MySatelliteEvent;
	/// // Create an Satellite of 90
	/// Satellite MySatellite = new Satellite(90);
	/// 
	/// void Main()
	/// {
	///   // Raise our custom event
	///   MySatelliteEvent(this, New SatelliteEventArgs(MySatellite));
	/// }
	/// </code>
	/// </example>
	/// <seealso cref="Satellite">SatelliteCollection Class</seealso>
	/// <seealso cref="Satellite.Azimuth">Azimuth Property (Satellite Class)</seealso>
	/// <seealso cref="Satellite.Elevation">Elevation Property (Satellite Class)</seealso>
	/// <seealso cref="Satellite.SignalToNoiseRatio">SignalToNoiseRatio Property (Satellite Class)</seealso>
	public sealed class SatelliteEventArgs : EventArgs
	{
		private Satellite oSatellite;

		/// <summary>
		/// Creates a new instance.
		/// </summary>
		/// <param name="Satellite"></param>

		public SatelliteEventArgs(Satellite satellite)
		{
			oSatellite = satellite;
		}

		/// <summary>
		/// Indicates which satellite is the target of the event.
		/// </summary>
		/// <value>A <strong>Satellite</strong> object containing modified information.</value>

		public Satellite Satellite //'Implements ISatelliteEventArgs.Satellite
		{
			get
			{
				return oSatellite;
			}
		}
	}

	/// <summary>
	/// Facilitates events and asynchronous calls which use an <strong>Satellite</strong>
	/// object as a parameter.
	/// </summary>
	/// <example>
	/// This example declares a custom event using this delegate.
	/// <code lang="VB">
	/// Dim MySatelliteEvent As EventHandler
	/// </code>
	/// <code lang="C#">
	/// EventHandler MySatelliteEvent;
	/// </code>
	/// </example>
	/// <param name="e">A <see cref="SatelliteEventArgs">SatelliteEventArgs</see> object containing a <see cref="Satellite">Satellite</see> object involved with the event.</param>
	/// <param name="sender">The object which initiated the call to the delegate.</param>
	public delegate void SatelliteEventHandler(object sender, SatelliteEventArgs e);
}